/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef DBUG_ARBITER_MPI_COLLECTIVE_H_
#define DBUG_ARBITER_MPI_COLLECTIVE_H_ 1

#include <set>
#include <string>

#include "dbug/common/clock.h"
#include "dbug/common/macros.h"

namespace dbug {
namespace arbiter {

class Request;

namespace mpi {

class Collective {
 public:
  explicit Collective(int communicator);
  virtual ~Collective();

  // accessors
  std::set<int> arrived() const;
  Clock clock() const;
  int communicator() const;
  std::set<int> departed() const;
  int num_arrived() const;
  int num_departed() const;
  std::set<Request *> waiting() const;

  // mutators
  void add_to_arrived(int key);
  void add_to_departed(int key);
  void add_to_waiting(Request *key);
  void join_with_clock(const Clock &clock);
  void remove_from_waiting(Request *key);

  virtual std::string ToString() const;

 protected:
  std::set<int> arrived_;
  Clock clock_;
  int communicator_;
  std::set<int> departed_;
  std::set<Request *> waiting_;

 private:
  DISALLOW_COPY_AND_ASSIGN(Collective);
};

class AllToAll : public Collective {
 public:
  enum Type {
    ALLGATHER,
    ALLGATHERV,
    ALLREDUCE,
    ALLTOALL,
    ALLTOALLV,
    ALLTOALLW,
    BARRIER,
    REDUCE_SCATTER,
    REDUCE_SCATTER_BLOCK
  };

  AllToAll(int communicator, int size, int type);

  // accessors
  std::set<int> ranks() const;
  int size() const;
  Type type() const;

  // mutators
  void add_to_ranks(int rank);

  virtual std::string ToString() const;

 protected:
  std::set<int> ranks_;
  int size_;
  Type type_;

 private:
  DISALLOW_COPY_AND_ASSIGN(AllToAll);
};

class AllToOne : public Collective {
 public:
  enum Type {
    GATHER,
    GATHERV,
    REDUCE
  };

  AllToOne(int communicator, int size, int sink, int type);

  // accessors
  int sink() const;
  int size() const;
  Type type() const;

  virtual std::string ToString() const;

 protected:
  int sink_;
  int size_;
  Type type_;

 private:
  DISALLOW_COPY_AND_ASSIGN(AllToOne);
};

class CommCreate : public Collective {
 public:
  CommCreate(int communicator, std::set<int> group, int tag);

  // accessors
  std::set<int> group() const;
  int group_size() const;
  std::set<int> ranks() const;
  int tag() const;

  // mutators
  void add_to_ranks(int rank);

  virtual std::string ToString() const;

 protected:
  std::set<int> group_;
  std::set<int> ranks_;
  int tag_;

 private:
  DISALLOW_COPY_AND_ASSIGN(CommCreate);
};

class OneToAll : public Collective {
 public:
  enum Type {
    BCAST,
    SCATTER,
    SCATTERV
  };

  OneToAll(int communicator, int root, int size, int type);

  // accessors
  int root() const;
  int size() const;
  Type type() const;

  virtual std::string ToString() const;

 protected:
  int root_;
  int size_;
  Type type_;

 private:
  DISALLOW_COPY_AND_ASSIGN(OneToAll);
};

class Scan : public Collective {
 public:
  enum Type {
    EXCLUSIVE,
    INCLUSIVE
  };

  Scan(int communicator, int size, int type);

  // accessors
  int num_posted() const;
  std::set<int> posted() const;
  int size() const;
  Type type() const;

  // mutators
  void add_to_posted(int key);

  virtual std::string ToString() const;

 protected:
  std::set<int> posted_;
  int size_;
  Type type_;

 private:
  DISALLOW_COPY_AND_ASSIGN(Scan);
};

}  // namespace mpi
}  // namespace arbiter
}  // namespace dbug

#endif  // DBUG_ARBITER_MPI_COLLECTIVE_H_
