/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef DBUG_ARBITER_MPI_H_
#define DBUG_ARBITER_MPI_H_ 1

#include <list>
#include <map>
#include <set>
#include <sstream>
#include <string>

#include "dbug/common/clock.h"
#include "dbug/common/macros.h"

namespace dbug {
namespace arbiter {

class Request;

namespace mpi {

class Event;
class MatchedEvent;
class AllToAll;
class AllToOne;
class CommCreate;
class OneToAll;
class Scan;

class MPI {
 public:
  struct CommIDKey {
    bool operator<(const CommIDKey &other) const;
    CommCreate *create;
    int tag;
    std::string subgroup;
  };

  typedef std::list<AllToAll *> AllToAlls;
  typedef std::map<int, AllToAlls> AllToAllTypeMap;
  typedef std::map<int, AllToAllTypeMap> AllToAllCommMap;
  typedef std::list<AllToOne *> AllToOnes;
  typedef std::map<int, AllToOnes> AllToOneTypeMap;
  typedef std::map<int, AllToOneTypeMap> AllToOneCommMap;
  typedef std::list<CommCreate *> CommCreates;
  typedef std::map<int, CommCreates> CommCreateCommMap;
  typedef std::list<Event *> Events;
  typedef std::map<int, Events> EventOriginMap;
  typedef std::map<int, EventOriginMap> EventCommMap;
  typedef std::list<OneToAll *> OneToAlls;
  typedef std::map<int, OneToAlls> OneToAllTypeMap;
  typedef std::map<int, OneToAllTypeMap> OneToAllCommMap;
  typedef std::list<Scan *> Scans;
  typedef std::map<int, Scans> ScanTypeMap;
  typedef std::map<int, ScanTypeMap> ScanCommMap;

  MPI();

  // accessors
  AllToAllCommMap alltoalls() const;
  AllToAllTypeMap alltoalls(int comm) const;
  AllToAlls alltoalls(int comm, int type) const;
  AllToAll *alltoall_lookup(Request *request) const;
  AllToOneCommMap alltoones() const;
  AllToOneTypeMap alltoones(int comm) const;
  AllToOnes alltoones(int comm, int type) const;
  AllToOne *alltoone_lookup(Request *request) const;
  CommCreateCommMap comm_creates() const;
  CommCreates comm_creates(int comm) const;
  CommCreate *comm_create_lookup(Request *request) const;
  int comm_id(CommCreate *create, int tag, std::string subgroup) const;
  int any_source() const;
  int any_tag() const;
  EventCommMap events() const;
  EventOriginMap events(int comm) const;
  Events events(int comm, int origin) const;
  Event *event_lookup(Request *request) const;
  Event *matched_event_lookup(std::string id) const;
  OneToAllCommMap onetoalls() const;
  OneToAllTypeMap onetoalls(int comm) const;
  OneToAlls onetoalls(int comm, int type) const;
  OneToAll *onetoall_lookup(Request *request) const;
  int rank(int rank) const;
  bool request_exists(std::string request_id) const;
  Clock request_clock(std::string request_id) const;
  ScanCommMap scans() const;
  ScanTypeMap scans(int comm) const;
  Scans scans(int comm, int type) const;
  Scan *scan_lookup(Request *request) const;

  // mutators
  void add_alltoall(AllToAll *alltoall);
  void remove_alltoall(AllToAll *alltoall);
  void set_alltoall_lookup(Request *key, AllToAll *value);
  void remove_alltoall_lookup(Request *key);
  void add_alltoone(AllToOne *alltoone);
  void remove_alltoone(AllToOne *alltoone);
  void set_alltoone_lookup(Request *key, AllToOne *value);
  void remove_alltoone_lookup(Request *key);
  void set_any_source(int any_source);
  void set_any_tag(int any_tag);
  void add_comm_create(CommCreate *create);
  void remove_comm_create(CommCreate *create);
  void set_comm_create_lookup(Request *key, CommCreate *value);
  void remove_comm_create_lookup(Request *key);
  void create_comm_id(CommCreate *create, int tag, std::string subgroup);
  void remove_comm_id(CommCreate *create, int tag);
  void add_event(Event *event);
  void remove_event(Event *event);
  void add_onetoall(OneToAll *onetoall);
  void remove_onetoall(OneToAll *onetoall);
  void set_onetoall_lookup(Request *key, OneToAll *value);
  void remove_onetoall_lookup(Request *key);
  void set_rank(int rank, int dbug_rank);
  void add_request(std::string request_id, Clock clock);
  void remove_request(std::string request_id);
  void add_scan(Scan *scan);
  void remove_scan(Scan *scan);
  void set_scan_lookup(Request *key, Scan *value);
  void remove_scan_lookup(Request *key);

 private:
  AllToAllCommMap alltoalls_;
  std::map<Request *, AllToAll *> alltoalls_lookup_;
  AllToOneCommMap alltoones_;
  std::map<Request *, AllToOne *> alltoones_lookup_;
  int any_source_;
  int any_tag_;
  CommCreateCommMap comm_creates_;
  std::map<Request *, CommCreate *> comm_creates_lookup_;
  int comm_id_counter_;
  std::map<CommIDKey, int> comm_ids_;
  EventCommMap events_;
  std::map<Request *, Event *> events_lookup_;
  std::map<std::string, Event *> matched_events_;
  OneToAllCommMap onetoalls_;
  std::map<Request *, OneToAll *> onetoalls_lookup_;
  std::map<int, int> rank_;
  std::map<std::string, Clock> requests_;
  ScanCommMap scans_;
  std::map<Request *, Scan *> scans_lookup_;

  DISALLOW_COPY_AND_ASSIGN(MPI);
};

}  // namespace mpi
}  // namespace arbiter
}  // namespace dbug

#endif  // DBUG_ARBITER_MPI_H_
