/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

// TODO(jsimsa): Track position for individual file descriptors in the
// File class.

#ifndef DBUG_ARBITER_POSIX_RESOURCE_H_
#define DBUG_ARBITER_POSIX_RESOURCE_H_ 1

#include <list>
#include <map>
#include <set>
#include <string>
#include <utility>

#include "dbug/arbiter/request.h"
#include "dbug/common/clock.h"
#include "dbug/common/macros.h"

namespace dbug {
namespace arbiter {
namespace posix {

class Resource {
 public:
  Resource(std::string id, Clock clock);
  virtual ~Resource();

  // accessors
  std::string id() const;
  Clock clock() const;

  // mutators
  void join_with_clock(const Clock &clock);
  void set_clock(const Clock &clock);

  virtual std::list<std::string> ToString() const;

 protected:
  Clock clock_;
  std::string id_;

 private:
  DISALLOW_COPY_AND_ASSIGN(Resource);
};

class Barrier : public Resource {
 public:
  Barrier(std::string id, int count, Clock clock);
  virtual ~Barrier();

  // accessors
  int count() const;
  int num_waiters() const;
  Request::Set waiters() const;

  // mutators
  void add_waiter(Request *request);
  void remove_waiter(Request *request);
  void set_count(int count);

  virtual std::list<std::string> ToString() const;

 private:
  int count_;
  Request::Set waiters_;

  DISALLOW_COPY_AND_ASSIGN(Barrier);
};

class CVar : public Resource {
 public:
  CVar(std::string id, Clock clock);
  virtual ~CVar();

  // accessors
  bool signaled(Request *request);
  bool signaled_empty() const;
  std::list<Request *> waiters() const;
  bool waiters_empty() const;
  Request *waiters_front() const;

  // mutators
  void waiters_clear();
  void waiters_pop_front();
  void waiters_push_back(Request * element);
  void add_signaled(Request *request);
  void remove_signaled(Request *request);

  virtual std::list<std::string> ToString() const;

 private:
  Request::Set signaled_;
  std::list<Request *> waiters_;

  DISALLOW_COPY_AND_ASSIGN(CVar);
};

class EPoll : public Resource {
 public:
  class Event {
   public:
    Event(int events, System::Int64 data);

    // accessors
    System::Int64 data() const;
    int events() const;

    // mutators
    void set_data(System::Int64 data);
    void set_events(int events);

    std::string ToString() const;

   private:
    System::Int64 data_;
    int events_;

    DISALLOW_COPY_AND_ASSIGN(Event);
  };

  EPoll(std::string id, Clock clock);
  virtual ~EPoll();

  // accessors
  std::map<std::string, Event *> events() const;
  Event *get_event(std::string rid) const;

  // mutators
  void add_event(std::string rid, Event *event);
  void remove_event(std::string rid);

  virtual std::list<std::string> ToString() const;

 private:
  std::map<std::string, Event *> events_;

  DISALLOW_COPY_AND_ASSIGN(EPoll);
};

class File : public Resource {
 public:
  File(std::string id, Clock clock);
  virtual ~File();

  // accessors
  int flags() const;
  int nrefs() const;

  // mutators
  void decrement_nrefs();
  void increment_nrefs();
  void set_flags(int flags);

  bool IsBlocking() const;
  bool IsReadable() const;
  bool IsWritable() const;
  virtual std::list<std::string> ToString() const;

 private:
  int flags_;
  int nrefs_;

  DISALLOW_COPY_AND_ASSIGN(File);
};

class Mutex : public Resource {
 public:
  enum State {
    LOCKED,
    UNLOCKED
  };

  static std::string StateToString(State state);

  Mutex(std::string id, bool recursive, Clock clock);
  virtual ~Mutex();

  // accessors
  bool recursive() const;
  State state() const;
  int owner() const;
  int counter() const;

  // mutators
  void set_state(State state);
  void set_owner(int owner);
  void increment_counter();
  void decrement_counter();

  virtual std::list<std::string> ToString() const;

 private:
  int counter_;
  int owner_;
  bool recursive_;
  State state_;

  DISALLOW_COPY_AND_ASSIGN(Mutex);
};

class Pipe : public Resource {
 public:
  static int GenerateID();

  Pipe(std::string id, Clock clock);
  virtual ~Pipe();

  // accessors
  int flags() const;
  int input() const;
  int nbytes() const;
  int output() const;
  int size() const;

  // mutators
  void decrement_nbytes(int delta);
  void increment_nbytes(int delta);
  void reset_nbytes();
  void set_flags(int flags);
  void set_input(int fd);
  void set_output(int fd);
  void set_size(int size);

  bool IsBlocking() const;
  bool IsReadable() const;
  bool IsWritable() const;
  virtual std::list<std::string> ToString() const;

 private:
  static int id_counter_;
  int flags_;
  int input_;
  int nbytes_;
  int output_;
  int size_;

  DISALLOW_COPY_AND_ASSIGN(Pipe);
};

class RWLock : public Resource {
 public:
  enum State {
    FREE,
    SHARED,
    EXCLUSIVE
  };

  static std::string StateToString(State state);

  RWLock(std::string id, Clock cloc);
  virtual ~RWLock();

  // accessors
  bool readers_empty() const;
  State state() const;
  int writer() const;

  // mutators
  void readers_add(int element);
  void readers_clear();
  bool readers_contains(int element);
  void readers_remove(int element);
  void set_state(State state);
  void set_writer(int writer);

  virtual std::list<std::string> ToString() const;

 private:
  std::multiset<int> readers_;
  State state_;
  int writer_;

  DISALLOW_COPY_AND_ASSIGN(RWLock);
};

class Semaphore : public Resource {
 public:
  Semaphore(std::string id, bool named, int value, Clock clock);
  virtual ~Semaphore();

  // accessors
  bool marked() const;
  bool named() const;
  int nrefs() const;
  int value() const;

  // mutators
  void decrement_nrefs();
  void decrement_value();
  void increment_nrefs();
  void increment_value();
  void set_marked(bool marked);

  virtual std::list<std::string> ToString() const;

 private:
  bool marked_;
  bool named_;
  int nrefs_;
  int value_;

  DISALLOW_COPY_AND_ASSIGN(Semaphore);
};

class Socket : public Resource {
 public:
  Socket(std::string id, int domain, int type, int protocol, Clock clock);
  virtual ~Socket() {}

  // accessors
  std::string address() const;
  Socket *connection() const;
  int domain() const;
  int flags() const;
  bool listening() const;
  int nbytes() const;
  std::string node_id() const;
  std::list<Socket *> pending_connections() const;
  std::list<Socket *> pending_messages() const;
  int port() const;
  int protocol() const;
  int size() const;
  int type() const;

  // mutators
  void add_pending_connection(Socket *socket);
  void add_pending_message(Socket *socket);
  void decrement_nbytes(int delta);
  void increment_nbytes(int delta);
  Socket *remove_pending_connection();
  Socket *remove_pending_message();
  void reset_nbytes();
  void set_address(std::string address);
  void set_connection(Socket *connection);
  void set_flags(int flags);
  void set_listening(bool listening);
  void set_node_id(std::string node_id);
  void set_port(int port);
  void set_size(int size);

  Socket *GetReadableRemote() const;
  bool IsBlocking() const;
  bool IsReadable() const;
  bool IsWritable() const;
  virtual std::list<std::string> ToString() const;

 private:
  std::string address_;
  Socket *connection_;
  int domain_;
  int flags_;
  bool listening_;
  int nbytes_;
  std::string node_id_;
  std::list<Socket *> pending_connections_;
  std::list<Socket *> pending_messages_;
  int port_;
  int protocol_;
  int size_;
  int type_;

  DISALLOW_COPY_AND_ASSIGN(Socket);
};

class Spin : public Resource {
 public:
  enum State {
    UNLOCKED,
    LOCKED
  };

  static std::string StateToString(State state);

  Spin(std::string id, Clock clock);
  virtual ~Spin();

  // accessors
  int owner() const;
  State state() const;

  // mutators
  void set_owner(int owner);
  void set_state(State state);

  virtual std::list<std::string> ToString() const;

 private:
  int owner_;
  State state_;

  DISALLOW_COPY_AND_ASSIGN(Spin);
};

}  // namespace posix
}  // namespace arbiter
}  // namespace dbug

#endif  // DBUG_ARBITER_POSIX_RESOURCE_H_
