/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef DBUG_ARBITER_SERVER_H_
#define DBUG_ARBITER_SERVER_H_ 1

#include <thrift/server/TServer.h>
#include <thrift/server/TThreadedServer.h>

#include <boost/scoped_array.hpp>
#include <boost/scoped_ptr.hpp>
#include <list>
#include <map>
#include <set>
#include <string>
#include <vector>

#include "dbug/arbiter/mpi.h"
#include "dbug/arbiter/posix.h"
#include "dbug/arbiter/request.h"
#include "dbug/common/clock.h"
#include "dbug/common/logging.h"
#include "dbug/common/macros.h"
#include "dbug/common/system.h"
#include "src/thrift/explorer/explorer_types.h"

using apache::thrift::server::TThreadedServer;

namespace dbug {
namespace arbiter {

void *arbiter_thread(void *args);
void *input_thread(void *args);

class Choose;
class Process;
class Request;
class Thread;

class Server {
 public:
  class EventHandler : public ::apache::thrift::server::TServerEventHandler {
   public:
    explicit EventHandler(Server *server);
    virtual void preServe();

   private:
    Server *server_;

    DISALLOW_COPY_AND_ASSIGN(EventHandler);
  };

  class Lock {
   public:
    explicit Lock(Server *server);
    ~Lock();

   private:
    Server *server_;

    DISALLOW_COPY_AND_ASSIGN(Lock);
  };

  class Statistics {
   public:
    Statistics();
    ~Statistics();

    std::list<std::string> ToString() const;

   private:
    friend class Server;
    std::map<std::string, int> function_calls_;
    int nprocesses_;
    int nrequests_;
    int nthreads_;

    DISALLOW_COPY_AND_ASSIGN(Statistics);
  };

  enum Mode {
    INTERACTIVE,
    ROUND_ROBIN,
    RANDOM
  };

  Server(std::string explorer_hostname, int explorer_port, int port);
  ~Server();

  // accessors
  bool has_nondet_events() const;
  Mode mode() const;
  mpi::MPI *mpi() const;
  int nthreads() const;
  int nenabled() const;
  int port() const;
  posix::POSIX *posix() const;
  Process *process_lookup(int pid) const;
  std::list<Request *> requests() const;
  std::set<int> sleep_set() const;
  std::vector<dbug::thrift::Move> strategy() const;
  bool terminate() const;
  Thread *thread_lookup(int itd) const;

  // mutators
  void add_move(dbug::thrift::Move move);
  void add_process(Process *process);
  void add_request(Request *request);
  void add_to_sleep_set(int tid);
  void clear_sleep_set();
  int consume_nondet_event();
  void decrement_nenabled();
  void decrement_nthreads();
  void increment_nenabled();
  void increment_nthreads();
  void record_nondet_event(int event);
  void remove_process(Process *process);
  void remove_request(Request *request);
  void set_end_nthreads(int nthreads);
  void set_explorer_hostname(std::string explorer_hostname);
  void set_explorer_port(int explorer_port);
  void set_mode(Mode mode);
  void set_port(int port);
  void set_random_seed(unsigned int random_seed);
  void set_start_nthreads(int nthreads);
  void set_terminate(bool terminate);
  void set_thrift_server(TThreadedServer *thrift_server);

  void Arbitrate();
  bool CreateThread(const int parent_tid,
                    const bool create_process,
                    int *tid,
                    int *pid);
  void InputLoop();
  void LoadStrategyFromFile(const char *filename);
  void NewProcessNotification();
  void PrintRequestGroups(Logging::Verbosity verbosity);
  void PrintRequests(Logging::Verbosity verbosity);
  void PrintResources(Logging::Verbosity verbosity);
  void PrintProcesses(Logging::Verbosity verbosity);
  void RegisterArbiter();
  void RequestCanceled();
  void RequestExecuted();
  void Stop();
  void TerminateThread(Thread *thread, bool cleanup);
  void UnregisterArbiter();
  void WaitForCancelation();
  void WaitForExecution();
  void WaitingRequest(System::CVar *cond);

 private:
  // Checks is the arbiter can proceed to executing a request.
  //
  // INTERACTIVE: A request can be executed if the user inputted a
  // valid ID through the standard input.
  //
  // ROUND_ROBIN and RANDOM: The system needs to reach a quiescent state.
  bool CanExecuteRequest();
  // Checks if all running processes of a process tree still exist. If a
  // non-existing process is detected, all of its threads are terminated
  // to prevent the arbiter from waiting indefinitely.
  void CheckProcessTree(const Process *process);
  // Checks if all running processes of the system still exist.
  void CheckSystem();
  void GenerateRequestGroups();
  Process *InternalProcessLookup(int pid, Process *process) const;
  Thread *InternalThreadLookup(int tid, const Process *process) const;
  // Checks if a process has reached a quiescent state. A process is in
  // a quiescent state if it is not RUNNING or CONTINUED or every
  // RUNNING or DETACHED thread it owns has a pending request.
  bool IsProcessQuiescent(const Process *process);
  // Checks if a process tree has reached a quiescent state. A process
  // tree is in a quiescent state if every RUNNING or DETACHED thread
  // owned by a RUNNING or CONTINUED process of the tree has a pending
  // request.
  bool IsProcessTreeQuiescent(const Process *process);
  // Checks if the system has reached a quiescent state. The system is
  // in a quiescent state if every RUNNING or DETACHED thread owned by a
  // RUNNING or CONTINUED process has a pending request.
  bool IsSystemQuiescent();
  // Checks if a thread has reached a quiescent state. A thread is in a
  // quiescent state if it is not RUNNING or DETACHED or it has a
  // pending request.
  bool IsThreadQuiescent(const Thread *thread);
  void Log();
  void PrintProcessTree(Logging::Verbosity verbosity, const Process *process);
  void ResetSchedulingState();
  Request::GroupSet::iterator SelectRequestGroup();
  // Terminates all threads of a process tree.
  void TerminateProcessTree(const Process *process);
  // Terminates all threads in the test.
  void TerminateSystem();
  bool TestReentrant();
  void TimedWait();
  void Wait();

  std::vector<dbug::thrift::Node> branch_;
  int choice_;
  System::CVar cvar_;
  System::CVar cancelation_cvar_;
  int end_nthreads_;
  System::CVar execution_cvar_;
  std::string explorer_hostname_;
  int explorer_port_;
  Request::GroupSet groups_;
  int last_choice_;
  Mode mode_;
  boost::scoped_ptr<mpi::MPI> mpi_;
  System::Mutex mutex_;
  int nenabled_;
  int nnondet_events_;
  std::vector<int> nondet_events_;
  int nscheduled_;
  int nthreads_;
  int port_;
  boost::scoped_ptr<posix::POSIX> posix_;
  std::list<Process *> process_forest_;
  boost::scoped_ptr<System::RandomData> random_buffer_;
  unsigned int random_seed_;
  boost::scoped_array<char> random_state_;
  std::list<Request *> requests_;
  std::set<int> sleep_set_;
  int start_nthreads_;
  Statistics statistics_;
  std::vector<dbug::thrift::Move> strategy_;
  System::File *strategy_file_;
  bool terminate_;
  TThreadedServer *thrift_server_;

  DISALLOW_COPY_AND_ASSIGN(Server);
};

}  // namespace arbiter
}  // namespace dbug

#endif  // DBUG_ARBITER_SERVER_H_
