/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef DBUG_COMMON_SYSTEM_H_
#define DBUG_COMMON_SYSTEM_H_ 1

#include <arpa/inet.h>
#include <assert.h>
#include <dlfcn.h>
#include <errno.h>
#include <fcntl.h>
#include <ftw.h>
#include <net/if.h>
#include <netdb.h>
#include <netinet/in.h>
#include <pthread.h>
#include <semaphore.h>
#include <signal.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <time.h>
#include <unistd.h>

#include <map>
#include <string>
#include <vector>

#include "dbug/common/macros.h"

namespace dbug {

class System {
 public:
  typedef pthread_barrier_t Barrier;
  typedef pthread_barrierattr_t BarrierAttr;
  typedef clock_t ClockType;
  typedef pthread_cond_t CVar;
  typedef pthread_condattr_t CVarAttr;
  typedef Dl_info DLInfo;
  typedef fd_set FDSet;
  typedef FILE File;
  typedef struct stat FileStat;
  typedef struct hostent HostEntry;
  typedef struct ifreq IFRequest;
  typedef struct ifconf IFConfig;
  typedef struct in_addr IPv4Address;
  typedef struct in6_addr IPv6Address;
  typedef int8_t Int8;
  typedef int16_t Int16;
  typedef int32_t Int32;
  typedef int64_t Int64;
  typedef pthread_mutex_t Mutex;
  typedef pthread_mutexattr_t MutexAttr;
  typedef pid_t Process;
  typedef struct random_data RandomData;
  typedef pthread_rwlock_t RWLock;
  typedef pthread_rwlockattr_t RWLockAttr;
  typedef struct sched_param SchedParam;
  typedef sem_t Semaphore;
  typedef struct sigaction SignalAction;
  typedef sighandler_t SignalHandler;
  typedef sigset_t SignalSet;
  typedef size_t SizeType;
  typedef struct sockaddr SocketAddress;
  typedef struct sockaddr_in SocketIPv4Address;
  typedef struct sockaddr_in6 SocketIPv6Address;
  typedef pthread_t Thread;
  typedef pthread_attr_t ThreadAttr;
  typedef pthread_key_t ThreadKey;
  typedef time_t TimeType;
  typedef struct timespec TimeSpec;
  typedef struct timeval TimeValue;
  typedef struct timezone TimeZone;
  typedef uint8_t UInt8;
  typedef uint16_t UInt16;
  typedef uint32_t UInt32;
  typedef uint64_t UInt64;
#ifdef __WAIT_STATUS
  typedef __WAIT_STATUS WaitStatus;
#else
  typedef int * WaitStatus;
#endif

  static void Initialize();
  static void Finalize();

  // Process
  static Process Fork();
  static Process Wait(Process process, int *status, int options);
  static int Kill(Process process, int signal);
  static int AtExit(void (*function)(void));
  static void Exit(int status) DBUG_NORETURN;
  static Process GetProcessID();
  static Process GetParentProcessID();
  static Process GetProcessGroupID();
  static int Sleep(int seconds);
  static int Execute(const char *path, char *const argv[], char *const envp[]);
  static int ExecuteShellCommand(const char *command);
  static int Yield();

  // Signal
  static int SignalSetEmpty(SignalSet *set);
  static int SignalSetFill(SignalSet *set);
  static int SignalSetAdd(SignalSet *set, int signum);
  static int SignalSetRemove(SignalSet *set, int signum);
  static int SignalSetContains(const SignalSet *set, int signum);
  static int ControlSignalAction(int signum,
                                 const SignalAction *act,
                                 SignalAction *oldact);

  // String
  static char *Basename(std::string path);
  static int CompareStrings(const char *s1, const char *s2);
  static char *FindString(const char *haystack, const char *needle);
  static char *StringCopy(char *dest, const char *src, SizeType n);
  static SizeType StringLength(const char *string);
  static int StringToInt(const char *string);

  // I/O
  static int Select(int nfds, FDSet *readfds, FDSet *writefds,
                    FDSet *exceptfds, TimeValue *timeout);
  static void FDSetAdd(int fd, FDSet *set);
  static int FDSetContains(int fd, FDSet *set);
  static void FDSetEmpty(FDSet *set);
  static void FDSetRemove(int fd, FDSet *set);
  static int Open(const char *path, int oflag, ...);
  static int DuplicateFD(int old_fd, int new_fd);
  static bool FileExists(const char *path);
  static File *FileOpen(const char *path, const char *mode);
  static int FileControl(int fd, int cmd, ...);
  static std::string FileControlCommandToString(int cmd);
  static int FileClose(File *file);
  static int FileFlush(File *file);
  static int FilePrint(File *file, const char *format, ...);
  static int FileScan(File *file, const char *format, ...);
  static void FileGetString(char *buffer, int size, File *file);
  static int FileEOF(File *file);
  static int ErrorOutput(const char *path, ...);
  static int Output(const char *path, ...);
  static int Input(const char *path, ...);
  static int CreateDirectory(const char *path);
  static int Rename(const char *old_path, const char *new_path);
  static int Remove(const char *path);
  static int Stat(const char *path, FileStat *buffer);
  static int IOControl(int fd, int request, void *data);
  static int GetSocketName(int sockfd,
                           SocketAddress *addr,
                           socklen_t *addrlen);
  static int GetSocketOption(int sockfd, int level, int optname,
                             void *optval, socklen_t *optlen);
  static int SetSocketOption(int sockfd, int level, int optname,
                             const void *optval, socklen_t optlen);
  static UInt32 HostToNetworkLong(UInt32 host);
  static UInt16 HostToNetworkShort(UInt16 host);
  static UInt32 NetworkToHostLong(UInt32 net);
  static UInt16 NetworkToHostShort(UInt16 net);
  static int IPv4StringToNetwork(const char *cp, IPv4Address *inp);
  static int IPv4GetNetwork(const char *cp);
  static int IPv4GetHost(const char *cp);
  static std::string IPv4NetworkToString(IPv4Address in);
  static std::string IPv6NetworkToString(IPv6Address in);
  static IPv4Address IPv4MakeAddress(int net, int host);
  static int IPv4GetLocalNetworkOf(IPv4Address in);
  static int IPv4GetNetworkOf(IPv4Address in);

  // Thread management
  static int ThreadCreate(Thread *thread, const ThreadAttr *attr,
                          void *(*function)(void *), void *arg);  // NOLINT
  static int ThreadCancel(Thread thread);
  static int ThreadDetach(Thread thread);
  static int ThreadJoin(Thread thread, void **retval);
  static void ThreadExit(void *retval) DBUG_NORETURN;
  static Thread ThreadSelf();
  static int ThreadSetCancelState(int state, int *oldstate);
  static int ThreadSetCancelType(int type, int *oldtype);

  // Thread attribute
  static int ThreadAttrInit(ThreadAttr *attr);
  static int ThreadAttrDestroy(ThreadAttr *attr);
  static int ThreadAttrSetInheritSched(ThreadAttr *attr, int inheritsched);
  static int ThreadAttrGetInheritSched(ThreadAttr *attr, int *inheritsched);
  static int ThreadAttrSetSchedParam(ThreadAttr *attr, const SchedParam *param);
  static int ThreadAttrGetSchedParam(ThreadAttr *attr, SchedParam *param);
  static int ThreadAttrSetSchedPolicy(ThreadAttr *attr, int policy);
  static int ThreadAttrGetSchedPolicy(ThreadAttr *attr, int *policy);
  static int ThreadAttrSetScope(ThreadAttr *attr, int scope);
  static int ThreadAttrGetScope(ThreadAttr *attr, int *scope);
  static int ThreadAttrSetStackSize(ThreadAttr *attr, SizeType stacksize);
  static int ThreadAttrGetStackSize(ThreadAttr *attr, SizeType *stacksize);
  static int ThreadAttrSetStackAddr(ThreadAttr *attr, void *stackaddr);
  static int ThreadAttrGetStackAddr(ThreadAttr *attr, void **stackaddr);
  static int ThreadAttrSetStack(ThreadAttr *attr,
                                void *stackaddr,
                                SizeType stacksize);
  static int ThreadAttrGetStack(ThreadAttr *attr,
                                void **stackaddr,
                                SizeType *stacksize);
  static int ThreadAttrSetGuardSize(ThreadAttr *attr, SizeType guardsize);
  static int ThreadAttrGetGuardSize(ThreadAttr *attr, SizeType *guardsize);
  static int ThreadAttrSetDetachState(ThreadAttr *attr, int detachstate);
  static int ThreadAttrGetSetachState(ThreadAttr *attr, int *detachstate);

  // Barrier
  static int BarrierDestroy(Barrier *barrier);
  static int BarrierInit(Barrier *DBUG_RESTRICT barrier,
                         const BarrierAttr *DBUG_RESTRICT attr,
                         unsigned count);
  static int BarrierWait(Barrier *barrier);

  // Condition variable
  static int CVarDestroy(CVar *cond);
  static int CVarInit(CVar *DBUG_RESTRICT cond,
                      const CVarAttr *DBUG_RESTRICT attr);
  static int CVarTimedWait(CVar *DBUG_RESTRICT cond,
                           Mutex *DBUG_RESTRICT mutex,
                           const TimeSpec *DBUG_RESTRICT abstime);
  static int CVarWait(CVar *DBUG_RESTRICT cond,
                      Mutex *DBUG_RESTRICT mutex);
  static int CVarBroadcast(CVar *cond);
  static int CVarSignal(CVar *cond);

  // Thread-specific data
  static int KeyCreate(ThreadKey *key, void (*destructor)(void*));  // NOLINT
  static int KeyDelete(ThreadKey key);
  static void *KeyGetSpecific(ThreadKey key);
  static int KeySetSpecific(ThreadKey key, const void *value);

  // Mutex
  static int MutexDestroy(Mutex *mutex);
  static int MutexInit(Mutex *DBUG_RESTRICT mutex,
                       const MutexAttr *DBUG_RESTRICT attr);
  static int MutexLock(Mutex *mutex);
  static int MutexTryLock(Mutex *mutex);
  static int MutexUnlock(Mutex *mutex);

  // Read-write lock
  static int RWLockDestroy(RWLock *rwlock);
  static int RWLockInit(RWLock *DBUG_RESTRICT rwlock,
                        const RWLockAttr *DBUG_RESTRICT attr);
  static int RWLockReadLock(RWLock *rwlock);
  static int RWLockTryReadLock(RWLock *rwlock);
  static int RWLockWriteLock(RWLock *rwlock);
  static int RWLockTryWriteLock(RWLock *rwlock);
  static int RWLockUnlock(RWLock *rwlock);

  // Semaphore
  static int SemaphoreInit(Semaphore *sem, int pshared, unsigned int value);
  static int SemaphoreDestroy(Semaphore *sem);
  static int SemaphoreWait(Semaphore *sem);
  static int SemaphoreTryWait(Semaphore *sem);
  static int SemaphoreTimedWait(Semaphore *sem, const TimeSpec *abs_timeout);
  static int SemaphorePost(Semaphore *sem);
  static int SemaphoreGetValue(Semaphore *sem, int *sval);
  static Semaphore *SemaphoreOpen(const char *path, int oflag, ...);
  static int SemaphoreClose(Semaphore *sem);
  static int SemaphoreUnlink(const char *path);

  // Dynamic memory
  static void *Malloc(SizeType num);
  static void *Realloc(void *p, SizeType num);
  static void *Calloc(SizeType num, SizeType size);
  static void Free(void *p);

  // Time
  static ClockType Clock();
  static int GetTimeOfDay(TimeValue *time_value, TimeZone *time_zone);
  static TimeType Time(TimeType *tloc);

  // Randomness
  static char *InitState(unsigned int seed, char *state, SizeType statelen);
  static int Random();
  static int Random(int n);
  static void SeedRandom(unsigned int seed);
  static char *SetState(char *state);
  static int ReentrantInitState(unsigned int seed,
                                char *state,
                                SizeType statelen,
                                RandomData *buffer);
  static int ReentrantRandom(RandomData *buffer, int *result);
  static int ReentrantRandom(RandomData *buffer, int *result, int n);
  static int ReentrantSetState(char *state, RandomData *buffer);
  static int ReentrantSeedRandom(unsigned int seed, RandomData *buffer);

  // Miscellaneous
  static int Backtrace(void **buffer, int size);
  static char **BacktraceSymbols(void **buffer, int size);
  static void CopyBytes(const void *src, void *dest, SizeType n);
  static int DLAddress(void *addr, DLInfo *info);
  static char **GetEnvironment();
  static void SetEnvironment(char **environment);
  static char *GetEnvironmentVariable(const char *name);
  static int SetEnvironmentVariable(const char *name,
                                    const char *value,
                                    int overwrite);
  static int UnsetEnvironmentVariable(const char *name);

 private:
  static int RemoveDirectory(const char *path, bool recursive);
  static int RemoveFile(const char *path);
  static int RemoveHelper(const char *path, const FileStat *buffer,
                          int typeflag, struct FTW *ftw_buffer);

  DISALLOW_ALL_DEFAULT(System);
};

}  // namespace dbug

#endif  // DBUG_COMMON_SYSTEM_H_
