/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef DBUG_COMMON_UTIL_H_
#define DBUG_COMMON_UTIL_H_ 1

#include <bfd.h>

#include <list>
#include <map>
#include <set>
#include <sstream>
#include <string>
#include <vector>

#include "dbug/common/clock.h"
#include "dbug/common/command.h"
#include "dbug/common/logging.h"
#include "dbug/common/macros.h"
#include "dbug/common/system.h"
#include "src/thrift/arbiter/arbiter_types.h"

using dbug::thrift::Location;

namespace dbug {

// Contains utility functions for common operations used across
// different dbug components. This class contains only static members
// and cannot be instantiated. To initialize it, call the Initialize()
// function and to finalize it, call the Finalize() function.
class Util {
 public:
  // Generates a string representation of a list.
  template <class T>
  static std::string ListToString(const std::list<T> &list) {
    if (list.empty()) { return "[]"; }
    typename std::list<T>::const_iterator lit = list.begin();
    std::stringstream stream;
    stream << "[" << *(lit++);
    for (; lit != list.end(); ++lit) { stream << " " << *lit; }
    stream << "]";
    return stream.str();
  }
  // Generates a string representation of a list of pointers.
  template <class T>
  static std::string ListToString(const std::list<T *> &list) {
    if (list.empty()) { return "[]"; }
    typename std::list<T *>::const_iterator lit = list.begin();
    std::stringstream stream;
    stream << "[" << (*(lit++))->ToString();
    for (; lit != list.end(); ++lit) { stream << " " << (*lit)->ToString(); }
    stream << "]";
    return stream.str();
  }
  // Generates a string representation of a map.
  template <class T, class U>
  static std::string MapToString(const std::map<T, U> &map) {
    if (map.empty()) { return "{}"; }
    typename std::map<T, U>::const_iterator mit = map.begin();
    std::stringstream stream;
    stream << "{ " << mit->first << " = " << mit->second;
    ++mit;
    for (mit = map.begin(); mit != map.end(); ++mit) {
      stream << ", " << mit->first << " = " << mit->second;
    }
    stream << " }";
    return stream.str();
  }
  // Generates a string representation of a map of pointers.
  template <class T, class U>
  static std::string MapToString(const std::map<T, U *> &map) {
    if (map.empty()) { return "{}"; }
    typename std::map<T, U *>::const_iterator mit = map.begin();
    std::stringstream stream;
    stream << "{ " << mit->first << " = " << (mit->second)->ToString();
    ++mit;
    for (mit = map.begin(); mit != map.end(); ++mit) {
      stream << ", " << mit->first << " = " << (mit->second)->ToString();
    }
    stream << " }";
    return stream.str();
  }
  // Generates a string representation of a set.
  template <class T>
  static std::string SetToString(const std::set<T> &set) {
    if (set.empty()) { return "{}"; }
    typename std::set<T>::const_iterator sit = set.begin();
    std::stringstream stream;
    stream << "{" << *(sit++);
    for (; sit != set.end(); ++sit) { stream << " " << *sit; }
    stream << "}";
    return stream.str();
  }
  // Generates a string representation of a set of pointers.
  template <class T>
  static std::string SetToString(const std::set<T *> &set) {
    if (set.empty()) { return "{}"; }
    typename std::set<T *>::const_iterator sit = set.begin();
    std::stringstream stream;
    stream << "{" << (*(sit++))->ToString();
    for (; sit != set.end(); ++sit) { stream << " " << (*sit)->ToString(); }
    stream << "}";
    return stream.str();
  }
  // Generates a string representation of a vector.
  template <class T>
  static std::string VectorToString(const std::vector<T> &vector) {
    if (vector.empty()) { return "[]"; }
    typename std::vector<T>::const_iterator vit = vector.begin();
    std::stringstream stream;
    stream << "[" << *(vit++);
    for (; vit != vector.end(); ++vit) { stream << " " << *vit; }
    stream << "]";
    return stream.str();
  }
  // Generates a string representation of a vector of pointers.
  template <class T>
  static std::string VectorToString(const std::vector<T *> &vector) {
    if (vector.empty()) { return "[]"; }
    typename std::vector<T *>::const_iterator vit = vector.begin();
    std::stringstream stream;
    stream << "[" << (*(vit++))->ToString();
    for (; vit != vector.end(); ++vit) { stream << " " << (*vit)->ToString(); }
    stream << "]";
    return stream.str();
  }

  // Static "constructor".
  static void Initialize();
  // Static "destructor".
  static void Finalize();

  // Given a resource type <type>, a process identifier <pid>, and an
  // opaque identifier <id>, generates a unique resource idenfitier.
  static std::string CreateResourceID(std::string type,
                                      int pid,
                                      System::UInt64 id);
  // Creates a sequence of (function name, file name, line) tuples
  // describing the addresses on the current call stack.
  static std::vector<Location> GetBacktrace();
  // Given a socket descriptor, generates the MAC address of the
  // interface this socket uses.
  static std::string GetMACAddress(int socket);
  // Generates a new unique identifier.
  static int NewIdentifier();
  // Given a file name, opens a log file.
  static System::File *OpenLogFile(const char *name);
  // Given a variable-length argument list <arglist> and argument 0
  // <arg>, creates a vector of arguments <argv> and environment
  // variables <envp>.
  static void ParseArgumentList(char ***argv,
                                char ***envp,
                                char *arg,
                                va_list arglist);
  // Outputs the current backtrace using the target verbosity level
  // <level> to the logging channel <channel>.
  static void PrintBacktrace(Logging::Verbosity verbosity,
                             Logging::Channel channel);
  // Generates a textual description of a signal number.
  static std::string SignalToString(int signal);
  // Sets up a vector of environment variables that can be used to
  // continue using dbug after calls that modify environment variables
  // such as to execve().
  static void SetupEnvironment(char ***envp, int pid, int tid);
  // Execute the command <command> using the list of environment
  // variables <environment>.
  static void StartCommand(const Command *command,
                           std::map<std::string, std::string> environment);

 private:
  // Stores a context used for looking up symbol information.
  struct Context {
    bfd *descriptor;
    asymbol **symbol_table;
  };

  struct FindAddressArguments {
    bfd_vma address;
    const char *file_name;
    bool found;
    const char *function_name;
    unsigned line;
    asymbol **symbol_table;
  };

  static void FindAddressInSection(bfd *descriptor,
                                   asection *section,
                                   void *data);
  // Given a binary file descriptor <descriptor>, creates a symbol
  // table and returns a pointer to it, or fails returning NULL.
  static asymbol **GetSymbolTable(bfd *descriptor);
  // Given a file name <name>, creates a context that can be used for
  // looking up symbol information and returns a pointer to it, or
  // fails returning a NULL.
  static Context *CreateContext(const char *name);
  // Deletes a given context <context>, deallocating dynamically
  // allocated memory used for storing the context.
  static void DeleteContext(Context *context);

  // A mapping from file names to contexts that can be used for
  // looking up symbol information.
  static std::map<std::string, Context *> *context_map_;
  // Mutex to protect concurrent accesses to <context_map_>.
  static System::Mutex context_map_mutex_;
  // Counter used for generating deterministic resource identifiers.
  static int next_id_;
  // Mutex to protext concurrent accesses to <next_id_>.
  static System::Mutex id_mutex_;

  DISALLOW_ALL_DEFAULT(Util);
};

}  // namespace dbug

#endif  // DBUG_COMMON_UTIL_H_
