/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef DBUG_EXPLORER_NODE_H_
#define DBUG_EXPLORER_NODE_H_ 1

#include <list>
#include <map>
#include <string>
#include <vector>

#include "dbug/common/big-float.h"
#include "dbug/common/logging.h"
#include "dbug/common/macros.h"

namespace dbug {
namespace explorer {

class Edge;

class Node {
 public:
  // Comparison function used to sort exploration seeds in the
  // decreasing order of their depth.
  static bool DecreasingOrder(const Node *n1, const Node *n2);

  // Comparison function used to sort exploration seeds in the
  // increasing order of their depth.
  static bool IncreasingOrder(const Node *n1, const Node *n2);

  Node(int id, Node *parent);
  ~Node();

  // accessors
  std::map<int, Node *> back_pointers() const;
  int depth() const;
  BigFloat estimate() const;
  std::map<int, Edge *> edges() const;
  bool explored() const;
  int id() const;
  std::vector<int> nondet_events() const;
  int num_explored_branches() const;
  Node * parent() const;
  BigFloat probability() const;
  bool seeded() const;
  bool seeds_computed() const;
  bool visited() const;

  // mutators
  void add_edge(int id, Edge *edge);
  void clear_edges();
  void set_back_pointers(const std::map<int, Node *> &back_pointers);
  void set_estimate(BigFloat estimate);
  void set_explored(bool explored);
  void set_nondet_events(std::vector<int> nondet_events);
  void set_num_explored_branches(int num_explored_branches);
  void set_parent(Node *parent);
  void set_probability(BigFloat probability);
  void set_seeded(bool seeded);
  void set_seeds_computed(bool seeds_computed);
  void set_visited(bool visited);

  // Although deallocation could be easily implemented using
  // destructors and recursion, for large trees the recursion could
  // run over the default stack size. So a non-recursive explicit
  // deallocation is used instead.
  void DeallocateSubtree();
  Edge *GetEdge(int id) const;
  Edge *GetThreadEdge(int tid) const;
  Node *FollowEdge(int id) const;
  void Print(Logging::Verbosity verbosity) const;
  std::list<std::string> ToString() const;

 private:
  std::map<int, Node *> back_pointers_;
  int depth_;
  BigFloat estimate_;
  std::map<int, Edge *> edges_;
  bool explored_;
  int id_;
  std::vector<int> nondet_events_;
  int num_explored_branches_;
  Node *parent_;
  BigFloat probability_;
  bool seeded_;
  bool seeds_computed_;
  bool visited_;

  DISALLOW_COPY_AND_ASSIGN(Node);
};

}  // namespace explorer
}  // namespace dbug

#endif  // DBUG_EXPLORER_NODE_H_
