/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef DBUG_INTERPOSITION_HANDLER_H_
#define DBUG_INTERPOSITION_HANDLER_H_ 1

#include <atomic>
#include <map>
#include <set>
#include <string>
#include <vector>

#include "dbug/common/macros.h"
#include "dbug/common/system.h"
#include "dbug/interposition/arbiter.h"
#include "dbug/interposition/types.h"

namespace dbug {
namespace interposition {

class Handler {
 public:
  class Interposition {
   public:
    explicit Interposition(const char *function) {
      Handler::Disable(static_cast<const void *>(function));
    }
    ~Interposition() { Handler::Enable(); }

   private:
    DISALLOW_COPY_AND_ASSIGN(Interposition);
  };

  struct MPIMessage {
    int source;
    int tag;
    int comm;
  };

  // A constant used for distinguishing between function calls coming
  // from the application and function calls coming from dbug.
  static const void *kEnabled;

  static int Choose(int n);
  static void Detach();
  static void Disable(const void *value);
  static void DisableCancellation();
  static void DisableFork();
  static void DisableInterposition(const void *value);
  static void Enable();
  static void EnableCancellation();
  static void EnableFork();
  static void EnableInterposition();
  static void ExportResourceIdentifiers();
  static void Finalize();
  static void IgnoreMutex(pthread_mutex_t *mutex);
  static void ImportResourceIdentifiers();
  static void Initialize();
  static bool Interpose();
  static void Off();
  static void On();
  static int RecordNonDeterminism(int event);
  static void ThreadActive(pthread_t tid);
  static void ThreadWaiting();

  static void *IBSendHelper(void *args);
  static void *IMRecvHelper(void *args);
  static void *IRecvHelper(void *args);
  static void *IRSendHelper(void *args);
  static void *ISendHelper(void *args);
  static void *ISSendHelper(void *args);
  static void *SendHelper(void *args);
  static void *ThreadCreateHelper(void *args);

#include "dbug/interposition/gen/handler.h"

 private:
  static void ContinuationHandler(int sig);
  static void InstallHandler(int sig, System::SignalHandler handler);
  static void Reset();
  static void StopHandler(int sig);
  static void TerminationHandler(int sig);

  static Arbiter *arbiter_;
  static std::map<pthread_barrier_t *, std::string> barrier_map_;
  static System::Mutex barrier_map_mutex_;
  static int cancel_state_;
  static System::ThreadKey cancel_state_key_;
  static System::ClockType clock_;
  static std::map<pthread_cond_t *, std::string> cvar_map_;
  static System::Mutex cvar_map_mutex_;
  static int dbug_comm_internal_;
  static int dbug_comm_world_;
  static int dbug_group_world_;
  static std::map<int, std::string> epoll_map_;
  static System::Mutex epoll_map_mutex_;
  static std::map<int, std::string> file_map_;
  static System::Mutex file_map_mutex_;
  static std::atomic_bool finalized_;
  static System::RWLock fork_rwlock_;
  static System::TimeValue gettimeofday_;
  static std::set<pthread_mutex_t *> ignore_mutex_set_;
  static System::Mutex ignore_mutex_set_mutex_;
  static std::atomic_bool initialized_;
  static System::ThreadKey interpose_key_;
  static std::map<int, int> mpi_comms_map_;
  static System::Mutex mpi_comms_map_mutex_;
  static int mpi_group_world_;
  static std::map<void *, MPIMessage> mpi_message_map_;
  static System::Mutex mpi_message_map_mutex_;
  static int mpi_rank_;
  static std::set<System::Thread> mpi_threads_;
  static System::Mutex mpi_threads_mutex_;
  static std::map<pthread_mutex_t *, std::string> mutex_map_;
  static System::Mutex mutex_map_mutex_;
  static System::SignalAction new_actions_[NSIG];
  static std::map<sem_t *, std::string> nsem_map_;
  static System::Mutex nsem_map_mutex_;
  static System::SignalAction old_actions_[NSIG];
  static std::map<int, std::string> pipe_map_;
  static System::Mutex pipe_map_mutex_;
  static std::map<pthread_rwlock_t *, std::string> rwlock_map_;
  static System::Mutex rwlock_map_mutex_;
  static std::map<int, std::string> socket_map_;
  static System::Mutex socket_map_mutex_;
  static std::map<pthread_spinlock_t *, std::string> spin_map_;
  static System::Mutex spin_map_mutex_;
  static std::map<sem_t *, int> usem_map_;
  static System::Mutex usem_map_mutex_;
  static System::TimeType time_;

  DISALLOW_ALL_DEFAULT(Handler);
};

}  // namespace interposition
}  // namespace dbug

#endif  // DBUG_INTERPOSITION_HANDLER_H_
