/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "dbug/arbiter/handler.h"

#include <thrift/protocol/TBinaryProtocol.h>
#include <thrift/server/TThreadedServer.h>
#include <thrift/transport/TServerSocket.h>
#include <thrift/transport/TTransportUtils.h>

#include <boost/scoped_ptr.hpp>
#include <boost/shared_ptr.hpp>
#include <list>
#include <set>
#include <sstream>
#include <string>

#include "dbug/arbiter/logging.h"
#include "dbug/arbiter/process.h"
#include "dbug/arbiter/request.h"
#include "dbug/arbiter/posix-resource.h"
#include "dbug/arbiter/thread.h"
#include "src/thrift/arbiter/arbiter_types.h"

using dbug::thrift::Location;
using dbug::thrift::RegisterThreadReturn;

namespace dbug {
namespace arbiter {

using posix::File;
using posix::Pipe;
using posix::Socket;

void *handler_thread(void *args) {
  Server *arbiter = reinterpret_cast<Server *>(args);
  boost::shared_ptr<Handler> handler(new Handler(arbiter));
  boost::shared_ptr<apache::thrift::TProcessor>
      processor(new dbug::thrift::arbiterProcessor(handler));
  boost::shared_ptr<apache::thrift::transport::TServerTransport>
      transport(new apache::thrift::transport::TServerSocket(arbiter->port()));
  boost::shared_ptr<apache::thrift::transport::TTransportFactory>
      t_factory(new apache::thrift::transport::TBufferedTransportFactory());
  boost::shared_ptr<apache::thrift::protocol::TProtocolFactory>
      p_factory(new apache::thrift::protocol::TBinaryProtocolFactory());
  boost::scoped_ptr<apache::thrift::server::TThreadedServer>
      server(new apache::thrift::server::TThreadedServer(
          processor, transport, t_factory, p_factory));
  boost::shared_ptr<apache::thrift::server::TServerEventHandler>
      event_handler(new Server::EventHandler(arbiter));
  arbiter->set_thrift_server(server.get());
  server->setServerEventHandler(event_handler);
  server->serve();
  return NULL;
}

Handler::Handler(Server *server) : DefaultHandler(server) {}

void Handler::AdvanceClock(const int tid, const int index) {
  Server::Lock lock(server_);
  Thread *thread = server_->thread_lookup(tid);
  if (thread == NULL) {
    WARNING("Thread %u does not exists.", tid);
  } else {
    thread->advance_clock(index);
  }
}

int Handler::RecordNonDeterminism(const int event) {
  Server::Lock lock(server_);
  int ret;
  if (server_->has_nondet_events()) {
    ret = server_->consume_nondet_event();
  } else {
    ret = event;
  }
  server_->record_nondet_event(ret);
  return ret;
}

void Handler::RegisterThread(RegisterThreadReturn &result,  // NOLINT
                             const int parent_tid,
                             const bool create_process) {
  Server::Lock lock(server_);
  if (server_->CreateThread(parent_tid,
                            create_process,
                            &result.pid,
                            &result.tid)) {
    MESSAGE("Thread %u registered.", result.tid);
  }
}

void Handler::SetFileFlags(const int tid,
                           const std::string &rid,
                           const int flags) {
  Server::Lock lock(server_);
  File *file = server_->posix()->file_lookup(rid);
  if (!file) {
    WARNING("%s does not exists.", rid.c_str());
  } else {
    file->set_flags(flags);
    MESSAGE("Thread %d sets flags of %s to %u.", tid, rid.c_str(), flags);
  }
}

void Handler::SetPipeFlags(const int tid,
                           const std::string &rid,
                           const int flags) {
  Server::Lock lock(server_);
  Pipe *pipe = server_->posix()->pipe_lookup(rid);
  if (!pipe) {
    WARNING("%s does not exists.", rid.c_str());
  } else {
    pipe->set_flags(flags);
    MESSAGE("Thread %d sets flags of %s to %u.", tid, rid.c_str(), flags);
  }
}

void Handler::SetPipeInput(const int tid,
                           const std::string &rid,
                           const int fd) {
  Server::Lock lock(server_);
  Pipe *pipe = server_->posix()->pipe_lookup(rid);
  if (pipe == NULL) {
    WARNING("%s does not exists.", rid.c_str());
  } else {
    pipe->set_input(fd);
    MESSAGE("Thread %d sets input of %s to %d.", tid, rid.c_str(), fd);
  }
}

void Handler::SetPipeOutput(const int tid,
                            const std::string &rid,
                            const int fd) {
  Server::Lock lock(server_);
  Pipe *pipe = server_->posix()->pipe_lookup(rid);
  if (pipe == NULL) {
    WARNING("%s does not exists.", rid.c_str());
  } else {
    pipe->set_output(fd);
    MESSAGE("Thread %d sets output of %s to %d.", tid, rid.c_str(), fd);
  }
}

void Handler::SetPipeSize(const int tid,
                          const std::string &rid,
                          const int size) {
  Server::Lock lock(server_);
  Pipe *pipe = server_->posix()->pipe_lookup(rid);
  if (!pipe) {
    WARNING("%s does not exists.", rid.c_str());
  } else {
    pipe->set_size(size);
    MESSAGE("Thread %d sets size of %s to %u.", tid, rid.c_str(), size);
  }
}

void Handler::SetProcessInfo(const int tid,
                             const int pid,
                             const int ppid,
                             const int pgid,
                             const int state,
                             const int return_status,
                             const int signal) {
  Server::Lock lock(server_);
  Thread *thread = server_->thread_lookup(tid);
  if (thread == NULL) {
    WARNING("Thread %u does not exists.", tid);
  } else {
    if (pid != 0) { thread->pgrp()->set_pid(pid); }
    if (ppid != 0) { thread->pgrp()->set_ppid(ppid); }
    if (pgid != 0) { thread->pgrp()->set_pgid(pgid); }
    if ((Process::State) state != thread->pgrp()->state()) {
      thread->pgrp()->set_collected(false);
    }
    if ((Process::State) state ==  Process::TERMINATED) {
      thread->pgrp()->set_signal(0);
    }
    thread->pgrp()->set_state((Process::State) state);
    if (return_status != 0) {
      thread->pgrp()->set_return_status(return_status);
    }
    if (signal != 0) { thread->pgrp()->set_signal(signal); }
    MESSAGE("Thread %u updated its process information.", tid);
  }
}

void Handler::SetProcessPGID(const int tid, const int pid, const int pgid) {
  Server::Lock lock(server_);
  Process *process = server_->process_lookup(pid);
  if (process == NULL) {
    WARNING("Process %u does not exists.", pid);
  } else {
    process->set_pgid(pgid);
    MESSAGE("Thread %d updated process group ID of process %u.", tid, pid);
  }
}

void Handler::SetProcessReturnStatus(const int tid,
                                     const int pid,
                                     const int return_status) {
  Server::Lock lock(server_);
  Process *process = server_->process_lookup(pid);
  if (process == NULL) {
    WARNING("Process %u does not exists.", pid);
  } else {
    process->set_return_status(return_status);
    MESSAGE("Thread %d Updated return status of process %u.", tid, pid);
  }
}

void Handler::SetSocketFlags(const int tid,
                             const std::string &rid,
                             const int flags) {
  Server::Lock lock(server_);
  Socket *socket = server_->posix()->socket_lookup(rid);
  if (!socket) {
    WARNING("%s does not exists.", rid.c_str());
  } else {
    socket->set_flags(flags);
    MESSAGE("Thread %d sets flags of %s to %u.", tid, rid.c_str(), flags);
  }
}

void Handler::SetSocketNodeID(const int tid,
                              const std::string &rid,
                              const std::string &id) {
  Server::Lock lock(server_);
  Socket *socket = server_->posix()->socket_lookup(rid);
  if (!socket) {
    WARNING("%s does not exists.", rid.c_str());
  } else {
    socket->set_node_id(id);
    MESSAGE("Thread %d sets node ID of %s to %s.",
            tid, rid.c_str(), id.c_str());
  }
}

void Handler::SetSocketSize(const int tid,
                            const std::string &rid,
                            const int size) {
  Server::Lock lock(server_);
  Socket *socket = server_->posix()->socket_lookup(rid);
  if (!socket) {
    WARNING("%s does not exists.", rid.c_str());
  } else {
    socket->set_size(size);
    MESSAGE("Thread %d sets size of %s to %u.", tid, rid.c_str(), size);
  }
}

void Handler::SetThreadState(const int origin,
                             const int target,
                             const int state) {
  Server::Lock lock(server_);
  Thread *thread;
  thread = server_->thread_lookup(target);
  if (thread == NULL) {
    WARNING("Thread %u does not exists.", target);
  } else {
    thread->set_state((Thread::State) state);
    MESSAGE("Thread %u updated state of thread %u to %s.",
            origin, target,
            Thread::StateToString(
                static_cast<Thread::State>(thread->state())).c_str());
  }
}

void Handler::ShutdownServer() {
  Server::Lock lock(server_);
  server_->set_terminate(true);
  MESSAGE("Server shutdown initiated.");
}

void Handler::UnregisterThread(const int tid) {
  Server::Lock lock(server_);
  Thread *thread = server_->thread_lookup(tid);
  if (thread == NULL) {
    WARNING("Thread %u does not exists.", tid);
  } else {
    server_->TerminateThread(thread, true);
    MESSAGE("Thread %u unregistered.", thread->id());
  }
}

void Handler::UpdatePipeBytes(const int tid,
                              const std::string &rid,
                              const int nbytes) {
  Server::Lock lock(server_);
  Pipe *pipe = server_->posix()->pipe_lookup(rid);
  if (!pipe) {
    WARNING("%s does not exists.", rid.c_str());
  } else {
    if (nbytes > 0) {
      pipe->increment_nbytes(nbytes);
      MESSAGE("Thread %d wrote %d bytes to %s, which now holds %d bytes.",
              tid, nbytes, rid.c_str(), pipe->nbytes());
    }
    if (nbytes < 0) {
      pipe->decrement_nbytes(-nbytes);
      MESSAGE("Thread %d read %d bytes from %s, which now holds %d bytes.",
              tid, -nbytes, rid.c_str(), pipe->nbytes());
    }
  }
}

void Handler::UpdateSocketBytes(const int tid,
                                const std::string &rid,
                                const int nbytes) {
  Server::Lock lock(server_);
  Socket *socket = server_->posix()->socket_lookup(rid);
  if (!socket) {
    WARNING("%s does not exists.", rid.c_str());
  } else {
    if (nbytes > 0) {
      socket->increment_nbytes(nbytes);
      MESSAGE("Thread %d wrote %d bytes to %s, which now holds %d bytes.",
              tid, nbytes, rid.c_str(), socket->nbytes());
    }
    if (nbytes < 0) {
      Socket *remote_socket = socket->connection();
      if (remote_socket == NULL) {
        WARNING("Reading from an unconnected socket.");
      } else {
        remote_socket->decrement_nbytes(-nbytes);
        MESSAGE("Thread %d read %d bytes from %s, which now holds %d bytes.",
                tid, -nbytes,
                remote_socket->id().c_str(), remote_socket->nbytes());
      }
    }
  }
}

}  // namespace arbiter
}  // namespace dbug
