/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <boost/scoped_ptr.hpp>
#include <boost/program_options/options_description.hpp>
#include <boost/program_options/parsers.hpp>
#include <boost/program_options/value_semantic.hpp>
#include <boost/program_options/variables_map.hpp>
#include <sstream>
#include <string>

#include "dbug/arbiter/handler.h"
#include "dbug/arbiter/logging.h"
#include "dbug/arbiter/server.h"
#include "dbug/common/configuration.h"
#include "dbug/common/init.h"
#include "dbug/common/system.h"
#include "dbug/common/util.h"

using boost::program_options::notify;
using boost::program_options::options_description;
using boost::program_options::parse_command_line;
using boost::program_options::store;
using boost::program_options::value;
using boost::program_options::variables_map;

using dbug::Configuration;
using dbug::Logging;
using dbug::System;
using dbug::Util;
using dbug::arbiter::Server;

int main(int argc, char **argv) {
  // Initialize common dbug components.
  dbug::Initialize();

  // Initialize the arbiter server.
  boost::scoped_ptr<Server>
      server(new Server(Configuration::explorer_hostname(),
                        Configuration::explorer_port(),
                        Configuration::arbiter_port()));

  // Declare supported command-line options.
  int arbiter_port;
  int end_nthreads;
  std::string explorer_hostname;
  int explorer_port;
  unsigned int random_seed;
  int scheduler_mode;
  int start_nthreads;
  std::string strategy;
  options_description desc("Allowed options");
  desc.add_options()
      ("help", "print help message")
      ("arbiter_port",
       value<int>(&arbiter_port)->
       default_value(Configuration::arbiter_port()),
       "set arbiter port")
      ("end_num_threads",
       value<int>(&end_nthreads)->default_value(0),
       "set final number of threads")
      ("explorer_hostname",
       value<std::string>(&explorer_hostname),
       "set explorer hostname")
      ("explorer_port",
       value<int>(&explorer_port)->
       default_value(Configuration::explorer_port()),
       "set explorer port")
      ("random_seed",
       value<unsigned int>(&random_seed)->default_value(1),
       "set random seed for the random scheduler")
      ("scheduler_mode",
       value<int>(&scheduler_mode)->default_value(Server::Mode::ROUND_ROBIN),
       "set scheduler mode (0 = interactive, 1 = round-robin, 2 = random)")
      ("start_num_threads",
       value<int>(&start_nthreads)->default_value(1),
       "set initial number of threads")
      ("strategy",
       value<std::string>(&strategy),
       "set initial scheduler strategy")
      ("use_log_file", "output logging information into a file");


  // Parse command-line options.
  variables_map vm;
  store(parse_command_line(argc, argv, desc), vm);
  notify(vm);

  if (vm.count("help")) {
    std::stringstream stream;
    stream << desc;
    System::Output("%s\n", stream.str().c_str());
    System::Exit(1);
  }

  // Configure the arbiter server.
  if (vm.count("arbiter_port")) {
    server->set_port(arbiter_port);
  }
  if (vm.count("end_num_threads")) {
    server->set_end_nthreads(end_nthreads);
  }
  if (vm.count("explorer_hostname")) {
    server->set_explorer_hostname(explorer_hostname);
  }
  if (vm.count("explorer_port")) {
    server->set_explorer_port(explorer_port);
  }
  if (vm.count("random_seed")) {
    server->set_random_seed(random_seed);
  }
  if (vm.count("scheduler_mode")) {
    server->set_mode((Server::Mode) scheduler_mode);
  }
  if (vm.count("start_num_threads")) {
    server->set_start_nthreads(start_nthreads);
  }
  if (vm.count("strategy")) {
    server->LoadStrategyFromFile(strategy.c_str());
  }
  System::File *log_file = NULL;
  if (vm.count("use_log_file")) {
    log_file = Util::OpenLogFile("dbug-arbiter.log");
  }

  // Create a logging channel for the arbiter server.
  Logging::Verbosity verbosity =
      (server->mode() == Server::INTERACTIVE &&
       Configuration::arbiter_verbosity() < Logging::MESSAGE) ?
      Logging::MESSAGE : Configuration::arbiter_verbosity();
  Logging::add_channel(Logging::ARBITER, log_file, verbosity);

  // Start the input thread and the handler thread.
  System::Thread handler, input;
  System::ThreadCreate(&handler, NULL, dbug::arbiter::handler_thread,
                       reinterpret_cast<void *>(server.get()));
  if (server->mode() == Server::INTERACTIVE) {
    System::ThreadCreate(&input, NULL, dbug::arbiter::input_thread,
                         reinterpret_cast<void *>(server.get()));
  }

  server->Arbitrate();
  server->Stop();

  if (server->mode() == Server::INTERACTIVE) {
    System::ThreadCancel(input);
    System::ThreadJoin(input, NULL);
  }
  System::ThreadJoin(handler, NULL);

  // Deallocate the arbiter server.
  server.reset();

  // Finalize common dbug components.
  dbug::Finalize();

  return 0;
}
