/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <sstream>
#include <string>

#include "dbug/arbiter/logging.h"
#include "dbug/arbiter/process.h"
#include "dbug/arbiter/posix-resource.h"
#include "dbug/arbiter/request-prototypes.h"

namespace dbug {
namespace arbiter {

using posix::Barrier;

void Choose::Execute() {
  ASSERT(!nondet_choices_.empty(), "Missing nondeterministic choices.");
  return_value_->return_value = nondet_choices_.front();
  MESSAGE("Thread %d called choose(%d).", thread_->id(), n_);
}

void Choose::Test(Set requests, GroupSet *groups) {
  for (int i = 0; i < n_; i++) {
    Request::Group group;
    std::vector<int> choices;
    choices.push_back(i);
    group.add_request(this, choices);
    std::stringstream stream;
    stream << "choose-" << thread_->id();
    group.add_to_write_set(stream.str().c_str());
    Enable(&group);
    groups->insert(group);
  }
  Request::Test(requests, groups);
}

std::list<std::string> Choose::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: executing non-deterministic choice " << n_;
  result.push_back(stream.str());
  return result;
}

void NonReentrantFunction::Execute() {
  Process *process = thread_->pgrp();
  ASSERT(process, "Thread missing its process.");
  process->decrement_nonreentrant_function_count(function_);
  MESSAGE("Thread %d called non-reentrant function.", thread_->id());
}

void NonReentrantFunction::PreSchedule() {
  Process *process = thread_->pgrp();
  ASSERT(process, "Thread missing its process.");
  process->increment_nonreentrant_function_count(function_);
}

void NonReentrantFunction::Test(Set requests, GroupSet *groups) {
  Request::Group group;
  group.add_request(this);
  Process *process = thread_->pgrp();
  if (process) {
    std::stringstream stream;
    stream << "process-" << process->id() << "-reentrancy-" << function_;
    group.add_to_write_set(stream.str().c_str());
    if (process->nonreentrant_function_count(function_) == 1) {
      Enable(&group);
    }
  }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> NonReentrantFunction::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: executing a non-reentrant function";
  result.push_back(stream.str());
  return result;
}

void InternalBarrier::Execute() {
  std::stringstream stream;
  stream << "internal-barrier-" << id_;
  std::string id = stream.str();
  Barrier *barrier = server_->posix()->barrier_lookup(id);
  ASSERT(barrier, "Barrier does not exist.");
  thread_->join_with_clock(barrier->clock());
  barrier->remove_waiter(this);
  if (barrier->num_waiters() == 0) {
    server_->posix()->remove_barrier(barrier);
    delete barrier;
  }
  MESSAGE("Thread %d returned from InternalBarrier(%d, %d).",
          thread_->id(), id_, count_);
}

void InternalBarrier::PreSchedule() {
  std::stringstream stream;
  stream << "internal-barrier-" << id_;
  std::string id = stream.str();
  Barrier *barrier = server_->posix()->barrier_lookup(id);
  if (!barrier) {
    barrier = new Barrier(id, count_, thread_->clock());
    server_->posix()->add_barrier(barrier);
  }
  barrier->add_waiter(this);
  barrier->join_with_clock(thread_->clock());
}

void InternalBarrier::Test(Set requests, GroupSet *groups) {
  std::stringstream stream;
  stream << "internal-barrier-" << id_;
  std::string id = stream.str();
  Barrier *barrier = server_->posix()->barrier_lookup(id);
  Request::Group group;
  group.add_to_write_set(id);
  group.set_serialize(false);
  for (Request *request : barrier->waiters()) {
    ASSERT(requests.erase(request) == 1, "Request does not exist.");
    std::vector<int> nondet_choices;
    group.add_request(request, nondet_choices);
  }
  if (barrier->num_waiters() == count_) { Enable(&group); }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> InternalBarrier::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: internal barrier";
  result.push_back(stream.str());
  return result;
}

void Sleep::Execute() {
  // server_->add_to_sleep_set(thread_->id());
  MESSAGE("Thread %d called sleep().", thread_->id());
}

void Sleep::Test(Set requests, GroupSet *groups) {
  // A sleep / yield request is disabled if no other thread has
  // executed a non-sleep / non-sleep request since the last time this
  // thread executed a sleep / yield request.
  std::set<int> sleep_set = server_->sleep_set();
  Request::Group group;
  group.add_request(this);
  if (sleep_set.find(thread_->id()) == sleep_set.end()) { Enable(&group); }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> Sleep::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: sleeping";
  result.push_back(stream.str());
  return result;
}

void Yield::Execute() {
  // server_->add_to_sleep_set(thread_->id());
  MESSAGE("Thread %d called yield().", thread_->id());
}

void Yield::Test(Set requests, GroupSet *groups) {
  // A sleep / yield request is disabled if no other thread has
  // executed a non-sleep / non-sleep request since the last time this
  // thread executed a sleep / yield request.
  std::set<int> sleep_set = server_->sleep_set();
  Request::Group group;
  group.add_request(this);
  if (sleep_set.find(thread_->id()) == sleep_set.end()) { Enable(&group); }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> Yield::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: yielding";
  result.push_back(stream.str());
  return result;
}

}  // namespace arbiter
}  // namespace dbug
