/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "dbug/arbiter/mpi-collective.h"

#include <sstream>
#include <string>

#include "dbug/arbiter/request.h"
#include "dbug/common/clock.h"
#include "dbug/common/util.h"

namespace dbug {
namespace arbiter {
namespace mpi {

// Collective class

Collective::Collective(int communicator) : communicator_(communicator) {}

Collective::~Collective() {}

// accessors

std::set<int> Collective::arrived() const { return arrived_; }

Clock Collective::clock() const { return clock_; }

int Collective::communicator() const { return communicator_; }

std::set<int> Collective::departed() const { return departed_; }

int Collective::num_arrived() const { return arrived_.size(); }

int Collective::num_departed() const { return departed_.size(); }

std::set<Request *> Collective::waiting() const { return waiting_; }

// mutators

void Collective::add_to_arrived(int key) { arrived_.insert(key); }

void Collective::add_to_departed(int key) { departed_.insert(key); }

void Collective::add_to_waiting(Request *key) { waiting_.insert(key); }

void Collective::join_with_clock(const Clock &clock) { clock_.JoinWith(clock); }

void Collective::remove_from_waiting(Request *key) { waiting_.erase(key); }

// public methods

std::string Collective::ToString() const {
  std::stringstream stream;
  stream << "communicator = " << communicator_;
  return stream.str();
}

// AllToAll class

AllToAll::AllToAll(int communicator, int size, int type)
    : Collective(communicator), size_(size), type_((Type) type) {}

// accessors

std::set<int> AllToAll::ranks() const { return ranks_; }

int AllToAll::size() const { return size_; }

AllToAll::Type AllToAll::type() const { return type_; }

// mutators

void AllToAll::add_to_ranks(int rank) { ranks_.insert(rank); }

// public methods

std::string AllToAll::ToString() const {
  std::stringstream stream;
  stream << Collective::ToString();
  stream << " ranks = " << Util::SetToString<int>(ranks_);
  stream << " size = " << size_;
  stream << " type = " << type_;
  return stream.str();
}

// AllToOne class

AllToOne::AllToOne(int communicator, int sink, int size, int type)
    : Collective(communicator), sink_(sink), size_(size), type_((Type) type) {}

// accessors

int AllToOne::sink() const { return sink_; }

int AllToOne::size() const { return size_; }

AllToOne::Type AllToOne::type() const { return type_; }

// public methods

std::string AllToOne::ToString() const {
  std::stringstream stream;
  stream << Collective::ToString();
  stream << " sink = " << sink_;
  stream << " size = " << size_;
  stream << " type = " << type_;
  return stream.str();
}

// CommCreate class

CommCreate::CommCreate(int communicator, std::set<int> group, int tag)
    : Collective(communicator), group_(group), tag_(tag) {}

// accessors

std::set<int> CommCreate::group() const { return group_; }

int CommCreate::group_size() const { return group_.size(); }

std::set<int> CommCreate::ranks() const { return ranks_; }

int CommCreate::tag() const { return tag_; }

// mutators

void CommCreate::add_to_ranks(int rank) { ranks_.insert(rank); }

// public methods

std::string CommCreate::ToString() const {
  std::stringstream stream;
  stream << Collective::ToString();
  stream << " group = " << Util::SetToString<int>(group_);
  stream << " tag = " << tag_;
  stream << " ranks = " << Util::SetToString<int>(ranks_);
  return stream.str();
}

// OneToAll class

OneToAll::OneToAll(int communicator, int root, int size, int type)
    : Collective(communicator), root_(root), size_(size), type_((Type) type) {}

// accessors

int OneToAll::root() const { return root_; }

int OneToAll::size() const { return size_; }

OneToAll::Type OneToAll::type() const { return type_; }

// public methods

std::string OneToAll::ToString() const {
  std::stringstream stream;
  stream << Collective::ToString();
  stream << " root = " << root_;
  stream << " size = " << size_;
  stream << " type = " << type_;
  return stream.str();
}

// Scan class

Scan::Scan(int communicator, int size, int type)
    : Collective(communicator), size_(size), type_((Type) type) {}

// accessors

int Scan::num_posted() const { return posted_.size(); }

std::set<int> Scan::posted() const { return posted_; }

int Scan::size() const { return size_; }

Scan::Type Scan::type() const { return type_; }

// mutators

void Scan::add_to_posted(int key) { posted_.insert(key); }

// public methods

std::string Scan::ToString() const {
  std::stringstream stream;
  stream << Collective::ToString();
  stream << " posted = " << Util::SetToString<int>(posted_);
  stream << " size = " << size_;
  stream << " type = " << type_;
  return stream.str();
}

}  // namespace mpi
}  // namespace arbiter
}  // namespace dbug
