/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <set>
#include <sstream>
#include <string>

#include "dbug/arbiter/logging.h"
#include "dbug/arbiter/process.h"
#include "dbug/arbiter/request-prototypes.h"
#include "dbug/arbiter/thread.h"

namespace dbug {
namespace arbiter {

void ProcessSignal::Execute() {
  // TODO(jsimsa): Record happens-before relation between the
  // signaling and the signaled thread.
  Process *process = server_->process_lookup(pid_);
  if (process == NULL) {
    WARNING("Process %d does not exists.", pid_);
  } else {
    process->set_signal(signal_);
    process->set_collected(false);
    switch (signal_) {
      case SIGCONT:
        process->set_state(Process::CONTINUED);
        break;
      case SIGTSTP:
      case SIGTTIN:
      case SIGTTOU:
        process->set_state(Process::STOPPED);
        break;
      default:
        WARNING("Process %d received signal %d.", pid_, signal_);
        process->set_return_status(128 + signal_);
        process->set_state(Process::SIGNALED);
        // If this process has any running threads, we need to
        // unregister them and cancel their requests.
        int nterminated = 0;
        for (Thread *thread : process->threads()) {
          if (thread->id() != thread_->id() &&
              (thread->state() == Thread::RUNNING ||
               thread->state() == Thread::DETACHED)) {
            nterminated++;
            server_->TerminateThread(thread, true);
          }
        }
        if (nterminated != 0) {
          WARNING("Signaling a process with %d running threads.", nterminated);
        }
        // Terminate the calling thread.
        server_->TerminateThread(thread_, false);
        break;
    }
    MESSAGE("Process %d received signal %d.", pid_, signal_);
  }
}

void ProcessSignal::Test(Set requests, GroupSet *groups) {
  Request::Group group;
  group.add_request(this);
  std::stringstream stream;
  stream << "process-" << pid_;
  group.add_to_write_set(stream.str());
  const Process *process = server_->process_lookup(pid_);
  if (process != NULL) {
    for (const Thread *thread : process->threads()) {
      std::stringstream stream;
      stream << "thread-" << thread->id();
      group.add_to_write_set(stream.str());
      stream.str(std::string());
    }
  }
  Enable(&group);
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> ProcessSignal::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: receiving signal " << signal_ << " in process " << pid_;
  result.push_back(stream.str());
  return result;
}

void ProcessTerminate::Execute() {
  Process *process = server_->process_lookup(pid_);
  if (process == NULL) {
    WARNING("Process %d does not exists.", pid_);
  } else {
    process->set_signal(0);
    process->set_state(Process::TERMINATED);
    // If this process has any running threads, we need to unregister
    // them and cancel their requests.
    int nterminated = 0;
    for (Thread *thread : process->threads()) {
      if (thread->id() != thread_->id() &&
          (thread->state() == Thread::RUNNING ||
           thread->state() == Thread::DETACHED)) {
        nterminated++;
        server_->TerminateThread(thread, true);
      }
    }
    if (nterminated != 0) {
      WARNING("Terminating a process with %d running thread(s).", nterminated);
    }
    // Terminate the calling thread.
    server_->TerminateThread(thread_, false);
    MESSAGE("Process %d terminated.", pid_);
  }
}

void ProcessTerminate::Test(Set requests, GroupSet *groups) {
  Request::Group group;
  group.add_request(this);
  std::stringstream stream;
  stream << "process-" << pid_;
  group.add_to_write_set(stream.str());
  const Process *process = server_->process_lookup(pid_);
  if (process != NULL) {
    for (const Thread *thread : process->threads()) {
      std::stringstream stream;
      stream << "thread-" << thread->id();
      group.add_to_write_set(stream.str());
      stream.str(std::string());
    }
  }
  Enable(&group);
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> ProcessTerminate::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: terminating process " << pid_;
  result.push_back(stream.str());
  return result;
}

void ComputeRWSets(Thread *thread, int pid, Request::Group *group) {
  std::stringstream stream;
  Process *process = thread->pgrp();
  if (pid > 0) {
    stream << "process-" << pid;
    group->add_to_write_set(stream.str());
  }
  if (pid == -1) {
    for (const Process *child : process->children()) {
      stream << "process-" << child->pid();
      group->add_to_write_set(stream.str());
      stream.str(std::string());
    }
  }
  if (pid == 0) {
    for (const Process *child : process->children()) {
      if (child->pgid() == process->pgid()) {
        stream << "process-" << child->pid();
        group->add_to_write_set(stream.str());
        stream.str(std::string());
      }
    }
  }
  if (pid < -1) {
    for (const Process *child : process->children()) {
      if (child->pgid() == -pid) {
        stream << "process-" << child->pid();
        group->add_to_write_set(stream.str());
        stream.str(std::string());
      }
    }
  }
}

bool IsCandidate(const Process *child,
                 const Process *parent,
                 int pid) {
  if (pid == -1) {
    // If pid == -1, wait for any children process.
    return true;
  } else if (pid > 0) {
    // If pid > 0, wait for the children with process id == pid.
    return child->pid() == pid;
  } else if (pid == 0) {
    // if pid == 0, wait for any children in the same process group.
    return child->pgid() == parent->pgid();
  } else {
    // If pid < -1, wait for any children with process group id == -pid.
    return child->pgid() == -pid;
  }
}

bool IsWaitable(const Process *process, int options) {
  if (process->collected()) { return false; }
  if (process->state() == Process::TERMINATED ||
      process->state() == Process::SIGNALED) {
    return true;
  }
  if ((options & WUNTRACED) == WUNTRACED &&
      process->state() == Process::STOPPED) {
    return true;
  }
  if ((options & WCONTINUED) == WCONTINUED &&
      process->state() == Process::CONTINUED) {
    return true;
  }
  return false;
}

void ProcessWait::Execute() {
  dbug::thrift::ProcessWaitReturn return_value;
  return_value.value = -1;
  return_value.state = -1;
  return_value.return_status = 0;
  return_value.signal = 0;
  return_value.errnum = -0;
  std::set<Process *> candidates;
  Process *process = thread_->pgrp();
  for (Process *child : process->children()) {
    if (IsCandidate(child, process, pid_)) { candidates.insert(child); }
  }
  bool waitable = false;
  if (candidates.size() == 0) {
    return_value.errnum = ECHILD;
    goto exit;
  }
  // Test if there exists a child that can be collected.
  for (Process *child : candidates) {
    if (IsWaitable(child, options_)) {
      ASSERT(!child->collected(), "A candidate child has been collected.");
      return_value.value = child->pid();
      return_value.state = child->state();
      return_value.return_status = child->return_status();
      return_value.signal = child->signal();
      return_value.errnum = 0;
      if (child->state() == Process::CONTINUED) {
        // TODO(jsimsa): Check that the thread waited on this
        // process to be continued.
        child->set_state(Process::RUNNING);
      } else {
        child->set_collected(true);
      }
      // Join clock with clocks of the waited for process.
      for (const Thread *thread : child->threads()) {
        thread_->join_with_clock(thread->clock());
      }
      goto exit;
    } else {
      waitable |= !child->collected();
    }
  }
  if ((options_ & WNOHANG) == WNOHANG) {
    return_value.value = 0;
    return_value.errnum = 0;
    goto exit;
  }
  ASSERT(!waitable, "A waitable child exists.");
  return_value.errnum = ECHILD;
exit:  // NOLINT
  return_value_->return_value = return_value;
  MESSAGE("Thread %d woke up from waiting on pid %d.", thread_->id(), pid_);
}

void ProcessWait::Test(Set requests, GroupSet *groups) {
  std::set<const Process *> candidates;
  Process *process = thread_->pgrp();
  for (const Process *child : process->children()) {
    if (IsCandidate(child, process, pid_)) { candidates.insert(child); }
  }
  bool enabled = false;
  if (candidates.size() == 0) {
    enabled = true;
    DEBUG("waitpid(%d) enabled, no viable candidates.", pid_);
  }
  bool waitable = false;
  // Test if there exists a child that can be collected.
  for (const Process *child : candidates) {
    if (IsWaitable(child, options_)) {
      enabled = true;
      DEBUG("waitpid(%d) enabled, a child process can be collected.", pid_);
    } else {
      waitable |= !child->collected();
    }
  }
  // If control reaches here, no child can be collected.
  if ((options_ & WNOHANG) == WNOHANG) {
    enabled = true;
    DEBUG("waitpid(%d) enabled, WNOHANG.", pid_);
  }
  // If control reaches here, there are no waitable children, return ECHILD.
  if (!waitable) {
    enabled = true;
    DEBUG("waitpid(%d) enabled, no viable candidates.", pid_);
  }
  Request::Group group;
  group.add_request(this);
  ComputeRWSets(thread_, pid_, &group);
  if (enabled) { Enable(&group); }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> ProcessWait::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: waiting for process " << pid_;
  result.push_back(stream.str());
  return result;
}

}  // namespace arbiter
}  // namespace dbug
