/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "dbug/arbiter/process.h"

#include <sstream>
#include <string>

#include "dbug/arbiter/logging.h"
#include "dbug/arbiter/thread.h"

namespace dbug {
namespace arbiter {

std::string Process::StateToString(State state) {
  switch (state) {
    case RUNNING:
      return "running";
    case STOPPED:
      return "stopped";
    case TERMINATED:
      return "terminated";
    case SIGNALED:
      return "signaled";
    case CONTINUED:
      return "continued";
    default:
      return "unknown";
  }
}

Process::Process()
    : collected_(false), id_(++id_counter_), parent_(NULL),
      pid_(0), ppid_(0), pgid_(0),
      return_status_(0), signal_(0), state_(RUNNING) {}

Process::~Process() {
  for (const Thread *thread : threads_) { delete thread; }
  threads_.clear();
  for (const Process *process : children_) { delete process; }
  children_.clear();
}

// accessors

std::set<Process *> Process::children() const { return children_; }

bool Process::collected() const { return collected_; }

int Process::id() const { return id_; }

int Process::nonreentrant_function_count(std::string fname) const {
  if (nonreentrant_map_.find(fname) != nonreentrant_map_.end()) {
    return nonreentrant_map_.at(fname);
  }
  return 0;
}

Process *Process::parent() const { return parent_; }

int Process::pid() const { return pid_; }

int Process::ppid() const { return ppid_; }

int Process::pgid() const { return pgid_; }

int Process::return_status() const { return return_status_; }

int Process::signal() const { return signal_; }

Process::State Process::state() const { return state_; }

std::set<Thread *> Process::threads() const { return threads_; }

// mutators

void Process::add_child(Process *child) { children_.insert(child); }

void Process::add_thread(Thread *thread) { threads_.insert(thread); }

void Process::decrement_nonreentrant_function_count(std::string fname) {
  ASSERT(nonreentrant_map_.find(fname) != nonreentrant_map_.end(),
         "Function does not have an entry.");
  if (nonreentrant_map_[fname] > 1) {
    nonreentrant_map_[fname] = nonreentrant_map_[fname] - 1;
  } else {
    nonreentrant_map_.erase(fname);
  }
}

void Process::increment_nonreentrant_function_count(std::string fname) {
  if (nonreentrant_map_.find(fname) != nonreentrant_map_.end()) {
    nonreentrant_map_[fname] = nonreentrant_map_[fname] + 1;
  } else {
    nonreentrant_map_[fname] = 1;
  }
}

void Process::set_collected(bool collected) { collected_ = collected; }

void Process::set_parent(Process *parent) { parent_ = parent; }

void Process::set_pid(int pid) { pid_ = pid; }

void Process::set_ppid(int ppid) { ppid_ = ppid; }

void Process::set_pgid(int pgid) { pgid_ = pgid; }

void Process::set_return_status(int return_status) {
  return_status_ = return_status;
}

void Process::set_signal(int signal) { signal_ = signal; }

void Process::set_state(Process::State state) { state_ = state; }

// public methods

std::list<std::string> Process::ToString() const {
  std::list<std::string> result;
  std::stringstream stream;
  stream << " Process:";
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  id: " << id_;
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  process id: " << pid_;
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  parent process id: " << ppid_;
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  process group id: " << pgid_;
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  stats: " << StateToString(state_);
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  return status: " << return_status_;
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  signal: " << signal_;
  result.push_back(stream.str());
  stream.str(std::string());
  for (const Process *child : children_) {
    stream << "  child: " << child->pid();
    result.push_back(stream.str());
    stream.str(std::string());
  }
  std::set<Thread *>::const_iterator it;
  for (const Thread *thread : threads_) {
    result.splice(result.end(), thread->ToString());
  }
  return result;
}

// private static fields

int Process::id_counter_ = 0;

}  // namespace arbiter
}  // namespace dbug
