/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "dbug/arbiter/logging.h"
#include "dbug/arbiter/posix-resource.h"
#include "dbug/arbiter/request-prototypes.h"
#include "dbug/arbiter/server.h"

namespace dbug {
namespace arbiter {

using posix::Barrier;
using posix::CVar;
using posix::Mutex;
using posix::RWLock;
using posix::Spin;

void BarrierCreate::Execute() {
  switch (state_) {
    case BUSY:
      return_value_->return_value = EBUSY;
      WARNING("Device or resource busy.");
      break;
    case INVALID:
      return_value_->return_value = EINVAL;
      WARNING("Invalid argument.");
      break;
    case VALID:
      ASSERT(!server_->posix()->barrier_lookup(id_), "Barrier already exists.");
      server_->posix()->add_barrier(
          new Barrier(id_, count_, thread_->clock()));
      return_value_->return_value = 0;
      MESSAGE("Thread %d creates %s.", thread_->id(), id_.c_str());
      break;
    default:
      WARNING("No Execute() transition applies.");
  }
}

void BarrierCreate::Test(Set requests, GroupSet *groups) {
  if (server_->posix()->barrier_lookup(id_)) {
    state_ = BUSY;
  } else if (count_ == 0) {
    state_ = INVALID;
  } else {
    state_ = VALID;
  }
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(id_);
  Enable(&group);
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> BarrierCreate::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: BarrierCreate"
         << "  barrier: " << id_
         << "  count: " << count_;
  result.push_back(stream.str());
  return result;
}

void BarrierDelete::Execute() {
  switch (state_) {
    case BUSY:
      return_value_->return_value = EBUSY;
      WARNING("Device or resource busy.");
      break;
    case INVALID:
      return_value_->return_value = EINVAL;
      WARNING("Invalid argument.");
      break;
    case VALID: {
      Barrier *barrier = server_->posix()->barrier_lookup(id_);
      ASSERT(barrier, "Barrier does not exist.");
      barrier->join_with_clock(thread_->clock());
      thread_->join_with_clock(barrier->clock());
      server_->posix()->remove_barrier(barrier);
      delete barrier;
      return_value_->return_value = 0;
      MESSAGE("Thread %d deletes %s.", thread_->id(), id_.c_str());
      break;
    }
    default:
      WARNING("No Execute() transition applies.");
  }
}

void BarrierDelete::Test(Set requests, GroupSet *groups) {
  Barrier *barrier = server_->posix()->barrier_lookup(id_);
  if (!barrier) {
    state_ = INVALID;
  } else if (barrier->num_waiters() != 0) {
    state_ = BUSY;
  } else {
    state_ = VALID;
  }
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(id_);
  Enable(&group);
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> BarrierDelete::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: BarrierDelete"
         << "  barrier: " << id_;
  result.push_back(stream.str());
  return result;
}

void BarrierWait::Execute() {
  switch (state_) {
    case INVALID:
      return_value_->return_value = EINVAL;
      WARNING("Invalid argument.");
      break;
    case VALID: {
      Barrier *barrier = server_->posix()->barrier_lookup(id_);
      ASSERT(barrier, "Barrier does not exist.");
      thread_->join_with_clock(barrier->clock());
      if (nondet_choices_.size() == 0) {
        return_value_->return_value = 0;
      } else {
        return_value_->return_value = PTHREAD_BARRIER_SERIAL_THREAD;
      }
      barrier->remove_waiter(this);
      MESSAGE("Thread %d wakes up from waiting on %s.",
              thread_->id(), id_.c_str());
      break;
    }
    default:
      WARNING("No Execute() transition applies.");
  }
}

void BarrierWait::PreSchedule() {
  Barrier *barrier = server_->posix()->barrier_lookup(id_);
  if (barrier) {
    barrier->add_waiter(this);
    barrier->join_with_clock(thread_->clock());
  }
}

void BarrierWait::Test(Set requests, GroupSet *groups) {
  Barrier *barrier = server_->posix()->barrier_lookup(id_);
  state_ = barrier ? VALID : INVALID;
  if (state_ == INVALID) {
    Request::Group group;
    group.add_request(this);
    group.add_to_write_set(id_);
    Enable(&group);
    groups->insert(group);
  } else {
    if (barrier->num_waiters() != barrier->count()) {
      Request::Group group;
      group.add_to_write_set(id_);
      for (Request *request : barrier->waiters()) {
        request->set_state(VALID);
        ASSERT(requests.erase(request) == 1, "Request does not exist.");
        group.add_request(request);
      }
      groups->insert(group);
    } else {
      // TODO(jsimsa): Create barrier->count() groups, each returning
      // PTHREAD_BARRIER_SERIAL_THREAD to a different thread.
      Request::Group group;
      group.add_to_write_set(id_);
      bool first = true;
      for (Request *request : barrier->waiters()) {
        request->set_state(VALID);
        ASSERT(requests.erase(request) == 1, "Request does not exist.");
        std::vector<int> nondet_choices;
        if (first) {
          nondet_choices.push_back(0);
          first = false;
        }
        group.add_request(request, nondet_choices);
      }
      Enable(&group);
      groups->insert(group);
    }
  }
  Request::Test(requests, groups);
}

std::list<std::string> BarrierWait::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: BarrierWait"
         << "  barrier: " << id_;
  result.push_back(stream.str());
  return result;
}

void CVarBroadcast::Execute() {
  switch (state_) {
    case INVALID:
      return_value_->return_value = EINVAL;
      WARNING("Invalid argument.");
      break;
    case VALID: {
      CVar *cvar = server_->posix()->cvar_lookup(id_);
      ASSERT(cvar, "Condition variable does not exist.");
      cvar->join_with_clock(thread_->clock());
      thread_->join_with_clock(cvar->clock());
      for (Request *request : cvar->waiters()) { cvar->add_signaled(request); }
      cvar->waiters_clear();
      return_value_->return_value = 0;
      MESSAGE("Thread %d broadcasts on a %s.", thread_->id(), id_.c_str());
      break;
    }
    default:
      WARNING("No Execute() transition applies.");
  }
}

void CVarBroadcast::PreSchedule() {
  CVar *cvar = server_->posix()->cvar_lookup(id_);
  if (!cvar) {
    // If the resource does not exist, we create it to account for the
    // possibility of static initialization.
    cvar = new CVar(id_, thread_->clock());
    server_->posix()->add_cvar(cvar);
  }
}

void CVarBroadcast::Test(Set requests, GroupSet *groups) {
  CVar *cvar = server_->posix()->cvar_lookup(id_);
  state_ = cvar ? VALID : INVALID;
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(id_);
  Enable(&group);
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> CVarBroadcast::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: CVarBroadcast"
         << "  cvar: " << id_;
  result.push_back(stream.str());
  return result;
}

void CVarCreate::Execute() {
  switch (state_) {
    case BUSY:
      return_value_->return_value = EBUSY;
      WARNING("Device or resource busy.");
      break;
    case VALID: {
      ASSERT(!server_->posix()->cvar_lookup(id_),
             "Condition variable already exists.");
      server_->posix()->add_cvar(new CVar(id_, thread_->clock()));
      return_value_->return_value = 0;
      MESSAGE("Thread %d creates %s.", thread_->id(), id_.c_str());
      break;
    }
    default:
      WARNING("No Execute() transition applies.");
  }
}

void CVarCreate::Test(Set requests, GroupSet *groups) {
  CVar *cvar = server_->posix()->cvar_lookup(id_);
  state_ = cvar ? BUSY : VALID;
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(id_);
  Enable(&group);
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> CVarCreate::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: CVarCreate"
         << "  cvar: " << id_;
  result.push_back(stream.str());
  return result;
}

void CVarDelete::Execute() {
  switch (state_) {
    case BUSY:
      return_value_->return_value = EBUSY;
      WARNING("Device or resource busy.");
      break;
    case INVALID:
      return_value_->return_value = EINVAL;
      WARNING("Invalied argument.");
      break;
    case VALID: {
      CVar *cvar = server_->posix()->cvar_lookup(id_);
      ASSERT(cvar, "Condition variable does not exist.");
      cvar->join_with_clock(thread_->clock());
      thread_->join_with_clock(cvar->clock());
      server_->posix()->remove_cvar(cvar);
      delete cvar;
      return_value_->return_value = 0;
      MESSAGE("Thread %d deletes %s.", thread_->id(), id_.c_str());
      break;
    }
    default:
      WARNING("No Execute() transition applies.");
  }
}

void CVarDelete::Test(Set requests, GroupSet *groups) {
  CVar *cvar = server_->posix()->cvar_lookup(id_);
  if (!cvar) {
    state_ = INVALID;
  } else if (!cvar->waiters_empty() || !cvar->signaled_empty()) {
    state_ = BUSY;
  } else {
    state_ = VALID;
  }
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(id_);
  Enable(&group);
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> CVarDelete::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: CVarDelete"
         << "  cvar: " << id_;
  result.push_back(stream.str());
  return result;
}

void CVarSignal::Execute() {
  switch (state_) {
    case INVALID:
      return_value_->return_value = EINVAL;
      WARNING("Invalid argument.");
      break;
    case VALID: {
      CVar *cvar = server_->posix()->cvar_lookup(id_);
      ASSERT(cvar, "Condition variable does not exist.");
      cvar->join_with_clock(thread_->clock());
      thread_->join_with_clock(cvar->clock());
      if (!cvar->waiters_empty()) {
        Request *request = cvar->waiters_front();
        cvar->add_signaled(request);
        cvar->waiters_pop_front();
      }
      return_value_->return_value = 0;
      MESSAGE("Thread %d signals %s.", thread_->id(), id_.c_str());
      break;
    }
    default:
      WARNING("No Execute() transition applies.");
  }
}

void CVarSignal::PreSchedule() {
  CVar *cvar = server_->posix()->cvar_lookup(id_);
  if (!cvar) {
    // If the resource does not exist, we create it to account for the
    // possibility of static initialization.
    cvar = new CVar(id_, thread_->clock());
    server_->posix()->add_cvar(cvar);
  }
}

void CVarSignal::Test(Set requests, GroupSet *groups) {
  // TODO(jsimsa): If there are any threads waiting on the condition
  // variable, create one request group for each waiting thread,
  // signaling that particular thread.
  CVar *cvar = server_->posix()->cvar_lookup(id_);
  state_ = cvar ? VALID : INVALID;
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(id_);
  Enable(&group);
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> CVarSignal::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: CVarSignal"
         << "  cvar: " << id_;
  result.push_back(stream.str());
  return result;
}

void CVarWait::Execute() {
  switch (state_) {
    case INVALID:
      return_value_->return_value = EINVAL;
      WARNING("Invalid argument.");
      break;
    case PERMISSION:
      return_value_->return_value = EPERM;
      WARNING("Operation not permitted.");
      break;
    case VALID: {
      CVar *cvar = server_->posix()->cvar_lookup(cvar_id_);
      ASSERT(cvar, "Condition variable does not exist.");
      Mutex *mutex = server_->posix()->mutex_lookup(mutex_id_);
      ASSERT(mutex, "Mutex does not exist.");
      ASSERT(mutex->state() == Mutex::UNLOCKED, "Mutex not available.");
      ASSERT(cvar->signaled(this),
             "Waking up a request that has not been signaled.");
      cvar->join_with_clock(thread_->clock());
      thread_->join_with_clock(cvar->clock());
      mutex->join_with_clock(thread_->clock());
      thread_->join_with_clock(mutex->clock());
      mutex->set_state(Mutex::LOCKED);
      mutex->set_owner(thread_->id());
      cvar->remove_signaled(this);
      return_value_->return_value = 0;
      MESSAGE("Thread %d wakes up from waiting on %s using %s.",
              thread_->id(), cvar_id_.c_str(), mutex_id_.c_str());
      break;
    }
    default:
      WARNING("No Execute() transition applies.");
  }
}

void CVarWait::PreSchedule() {
  CVar *cvar = server_->posix()->cvar_lookup(cvar_id_);
  if (!cvar) {
    // If the resource does not exist, we create it to account for the
    // possibility of static initialization.
    cvar = new CVar(cvar_id_, thread_->clock());
    server_->posix()->add_cvar(cvar);
  }
  Mutex *mutex = server_->posix()->mutex_lookup(mutex_id_);
  if (mutex) {
    if (mutex->state() == Mutex::LOCKED &&
        mutex->owner() == thread_->id()) {
      mutex->set_state(Mutex::UNLOCKED);
      mutex->set_owner(0);
      cvar->waiters_push_back(this);
    } else {
      state_ = PERMISSION;
    }
  }
}

void CVarWait::Test(Set requests, GroupSet *groups) {
  CVar *cvar = server_->posix()->cvar_lookup(cvar_id_);
  Mutex *mutex = server_->posix()->mutex_lookup(mutex_id_);
  if (state_ != PERMISSION) {
    state_ = cvar && mutex ? VALID : INVALID;
  }
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(cvar_id_);
  group.add_to_write_set(mutex_id_);
  if (state_ != VALID ||
      (mutex->state() == Mutex::UNLOCKED &&
       cvar->signaled(this))) { Enable(&group); }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> CVarWait::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: CVarWait"
         << "  cvar: " << cvar_id_
         << "  mutex: " << mutex_id_;
  result.push_back(stream.str());
  return result;
}

void MutexCreate::Execute() {
  switch (state_) {
    case BUSY:
      return_value_->return_value = EBUSY;
      WARNING("Device or resource busy.");
      break;
    case VALID: {
      ASSERT(!server_->posix()->mutex_lookup(id_), "Mutex already exists.");
      server_->posix()->add_mutex(new Mutex(id_, recursive_, thread_->clock()));
      return_value_->return_value = 0;
      MESSAGE("Thread %d creates %s.", thread_->id(), id_.c_str());
      break;
    }
    default:
      WARNING("No Execute() transition applies.");
  }
}

void MutexCreate::Test(Set requests, GroupSet *groups) {
  Mutex *mutex = server_->posix()->mutex_lookup(id_);
  state_ = mutex ? BUSY : VALID;
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(id_);
  Enable(&group);
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> MutexCreate::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: MutexCreate"
         << "  mutex: " << id_;
  result.push_back(stream.str());
  return result;
}

void MutexDelete::Execute() {
  switch (state_) {
    case BUSY:
      return_value_->return_value = EBUSY;
      WARNING("Device or resource busy.");
      break;
    case INVALID:
      return_value_->return_value = EINVAL;
      WARNING("Invalid argument.");
      break;
    case VALID: {
      Mutex *mutex = server_->posix()->mutex_lookup(id_);
      ASSERT(mutex, "Mutex does not exist.");
      mutex->join_with_clock(thread_->clock());
      thread_->join_with_clock(mutex->clock());
      server_->posix()->remove_mutex(mutex);
      delete mutex;
      return_value_->return_value = 0;
      MESSAGE("Thread %d deletes %s.", thread_->id(), id_.c_str());
      break;
    }
    default:
      WARNING("No Execute() transition applies.");
  }
}

void MutexDelete::Test(Set requests, GroupSet *groups) {
  Mutex *mutex = server_->posix()->mutex_lookup(id_);
  if (!mutex) {
    state_ = INVALID;
  } else if (mutex->state() != Mutex::UNLOCKED) {
    state_ = BUSY;
  } else {
    state_ = VALID;
  }
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(id_);
  Enable(&group);
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> MutexDelete::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: MutexDelete"
         << "  mutex: " << id_;
  result.push_back(stream.str());
  return result;
}

void MutexLock::Execute() {
  switch (state_) {
    case INVALID:
      return_value_->return_value = EINVAL;
      WARNING("Invalid argument.");
      break;
    case DEADLOCK:
      return_value_->return_value = EDEADLK;
      WARNING("Resource deadlock would occur.");
      break;
    case VALID: {
      Mutex *mutex = server_->posix()->mutex_lookup(id_);
      ASSERT(mutex, "Mutex does not exist.");
      mutex->join_with_clock(thread_->clock());
      thread_->join_with_clock(mutex->clock());
      if (mutex->recursive()) {
        if (mutex->state() == Mutex::UNLOCKED) {
          ASSERT(mutex->owner() == 0, "Inconsistent mutex state.");
          mutex->set_state(Mutex::LOCKED);
          mutex->set_owner(thread_->id());
        } else {
          ASSERT(mutex->owner() == thread_->id(),
                 "Mutex is owned by someone else.");
        }
        mutex->increment_counter();
      } else {
        ASSERT(mutex->state() == Mutex::UNLOCKED, "Mutex is not available.");
        ASSERT(mutex->owner() == 0, "Inconsistent mutex state.");
        mutex->set_state(Mutex::LOCKED);
        mutex->set_owner(thread_->id());
      }
      return_value_->return_value = 0;
      MESSAGE("Thread %d locks %s.", thread_->id(), id_.c_str());
      break;
    }
    default:
      WARNING("No Execute() transition applies.");
  }
}

void MutexLock::PreSchedule() {
  Mutex *mutex = server_->posix()->mutex_lookup(id_);
  if (!mutex) {
    // If the resource does not exist, we create it to account for the
    // possibility of static initialization.
    mutex = new Mutex(id_, false, thread_->clock());
    server_->posix()->add_mutex(mutex);
  }
}

void MutexLock::Test(Set requests, GroupSet *groups) {
  Mutex *mutex = server_->posix()->mutex_lookup(id_);
  if (!mutex) {
    state_ = INVALID;
  } else if (mutex->state() == Mutex::LOCKED &&
             mutex->owner() == thread_->id() &&
             !mutex->recursive()) {
    state_ = DEADLOCK;
  } else {
    state_ = VALID;
  }
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(id_);
  if (state_ != VALID ||
      mutex->state() == Mutex::UNLOCKED ||
      (mutex->recursive() &&
       mutex->owner() == thread_->id())) { Enable(&group); }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> MutexLock::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: MutexLock"
         << "  mutex: " << id_;
  result.push_back(stream.str());
  return result;
}

void MutexTryLock::Execute() {
  switch (state_) {
    case INVALID:
      return_value_->return_value = EINVAL;
      WARNING("Invalid argument.");
      break;
    case DEADLOCK:
      return_value_->return_value = EDEADLK;
      WARNING("Resource deadlock would occur.");
      break;
    case VALID: {
      Mutex *mutex = server_->posix()->mutex_lookup(id_);
      ASSERT(mutex, "Mutex does not exist.");
      mutex->join_with_clock(thread_->clock());
      thread_->join_with_clock(mutex->clock());
      if (mutex->recursive()) {
        if (mutex->state() == Mutex::UNLOCKED) {
          ASSERT(mutex->owner() == 0, "Inconsistent mutex state.");
          mutex->set_state(Mutex::LOCKED);
          mutex->set_owner(thread_->id());
          mutex->increment_counter();
          return_value_->return_value = 0;
        } else {
          if (mutex->owner() == thread_->id()) {
            mutex->increment_counter();
            return_value_->return_value = 0;
          } else {
            return_value_->return_value = EBUSY;
          }
        }
      } else {
        if (mutex->state() == Mutex::UNLOCKED) {
          ASSERT(mutex->owner() == 0, "Inconsistent mutex state.");
          mutex->set_state(Mutex::LOCKED);
          mutex->set_owner(thread_->id());
          return_value_->return_value = 0;
        } else {
          return_value_->return_value = EBUSY;
        }
      }
      break;
    }
    default:
      WARNING("No Execute() transition applies.");
  }
  if (return_value_->return_value == 0) {
    MESSAGE("Thread %d locks %s.", thread_->id(), id_.c_str());
  }
}

void MutexTryLock::PreSchedule() {
  Mutex *mutex = server_->posix()->mutex_lookup(id_);
  if (!mutex) {
    // If the resource does not exist, we create it to account for the
    // possibility of static initialization.
    mutex = new Mutex(id_, false, thread_->clock());
    server_->posix()->add_mutex(mutex);
  }
}

void MutexTryLock::Test(Set requests, GroupSet *groups) {
  Mutex *mutex = server_->posix()->mutex_lookup(id_);
  if (!mutex) {
    state_ = INVALID;
  } else if (mutex->state() == Mutex::LOCKED &&
             mutex->owner() == thread_->id() &&
             !mutex->recursive()) {
    state_ = DEADLOCK;
  } else {
    state_ = VALID;
  }
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(id_);
  Enable(&group);
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> MutexTryLock::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: MutexTryLock"
         << "  mutex: " << id_;
  result.push_back(stream.str());
  return result;
}

void MutexUnlock::Execute() {
  switch (state_) {
    case INVALID:
      return_value_->return_value = EINVAL;
      WARNING("Invalid argument.");
      break;
    case PERMISSION:
      return_value_->return_value = EPERM;
      WARNING("Operation not permitted.");
      break;
    case VALID: {
      Mutex *mutex = server_->posix()->mutex_lookup(id_);
      ASSERT(mutex, "Mutex does not exist.");
      ASSERT(mutex->state() == Mutex::LOCKED &&
             mutex->owner() == thread_->id(),
             "Mutex is not owned by the caller.");
      mutex->join_with_clock(thread_->clock());
      thread_->join_with_clock(mutex->clock());
      if (mutex->recursive()) {
        if (mutex->counter() == 1) {
          mutex->set_state(Mutex::UNLOCKED);
          mutex->set_owner(0);
        }
        mutex->decrement_counter();
      } else {
        mutex->set_state(Mutex::UNLOCKED);
        mutex->set_owner(0);
      }
      return_value_->return_value = 0;
      MESSAGE("Thread %d unlocks %s.", thread_->id(), id_.c_str());
      break;
    }
    default:
      WARNING("No Execute() transition applies.");
  }
}

void MutexUnlock::Test(Set requests, GroupSet *groups) {
  Mutex *mutex = server_->posix()->mutex_lookup(id_);
  if (!mutex) {
    state_ = INVALID;
  } else if (mutex->state() == Mutex::UNLOCKED ||
             mutex->owner() != thread_->id()) {
    state_ = PERMISSION;
  } else {
    state_ = VALID;
  }
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(id_);
  Enable(&group);
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> MutexUnlock::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: MutexUnlock"
         << "  mutex: " << id_;
  result.push_back(stream.str());
  return result;
}

void RWLockCreate::Execute() {
  switch (state_) {
    case BUSY:
      return_value_->return_value = EBUSY;
      WARNING("Device or resource busy.");
      break;
    case VALID: {
      ASSERT(!server_->posix()->rwlock_lookup(id_),
             "Read-write lock already exists.");
      server_->posix()->add_rwlock(new RWLock(id_, thread_->clock()));
      return_value_->return_value = 0;
      MESSAGE("Thread %d creates %s.", thread_->id(), id_.c_str());
      break;
    }
    default:
      WARNING("No Execute() transition applies.");
  }
}

void RWLockCreate::Test(Set requests, GroupSet *groups) {
  RWLock *rwlock = server_->posix()->rwlock_lookup(id_);
  state_ = rwlock ? BUSY : VALID;
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(id_);
  Enable(&group);
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> RWLockCreate::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: RWLockCreate"
         << "  rwlock: " << id_;
  result.push_back(stream.str());
  return result;
}

void RWLockDelete::Execute() {
  switch (state_) {
    case BUSY:
      return_value_->return_value = EBUSY;
      WARNING("Device or resource busy.");
      break;
    case INVALID:
      return_value_->return_value = EINVAL;
      WARNING("Invalid argument.");
      break;
    case VALID: {
      RWLock *rwlock = server_->posix()->rwlock_lookup(id_);
      ASSERT(rwlock, "Read-write lock does not exist.");
      rwlock->join_with_clock(thread_->clock());
      thread_->join_with_clock(rwlock->clock());
      server_->posix()->remove_rwlock(rwlock);
      delete rwlock;
      return_value_->return_value = 0;
      MESSAGE("Thread %d deletes %s.", thread_->id(), id_.c_str());
      break;
    }
    default:
      WARNING("No Execute() transition applies.");
  }
}


void RWLockDelete::Test(Set requests, GroupSet *groups) {
  RWLock *rwlock = server_->posix()->rwlock_lookup(id_);
  if (!rwlock) {
    state_ = INVALID;
  } else if (rwlock->state() != RWLock::FREE) {
    state_ = BUSY;
  } else {
    state_ = VALID;
  }
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(id_);
  Enable(&group);
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> RWLockDelete::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: RWLockDelete";
  result.push_back(stream.str());
  return result;
}

void RWLockReadLock::Execute() {
  switch (state_) {
    case INVALID:
      return_value_->return_value = EINVAL;
      WARNING("Invalid argument.");
      break;
    case DEADLOCK:
      return_value_->return_value = EDEADLK;
      WARNING("Resource deadlock would occur.");
      break;
    case VALID: {
      RWLock *rwlock = server_->posix()->rwlock_lookup(id_);
      ASSERT(rwlock, "Read-write lock does not exist.");
      ASSERT(rwlock->state() != RWLock::EXCLUSIVE,
             "Read-write lock is exclusively owned.");
      rwlock->join_with_clock(thread_->clock());
      thread_->join_with_clock(rwlock->clock());
      if (rwlock->state() == RWLock::FREE) {
        ASSERT(rwlock->readers_empty(), "Inconsistent read-write lock state.");
        ASSERT(rwlock->writer() == 0, "Inconsistent read-write lock state.");
        rwlock->set_state(RWLock::SHARED);
      }
      rwlock->readers_add(thread_->id());
      return_value_->return_value = 0;
      MESSAGE("Thread %d read-locks %s.", thread_->id(), id_.c_str());
      break;
    }
    default:
      WARNING("No Execute() transition applies.");
  }
}

void RWLockReadLock::PreSchedule() {
  RWLock *rwlock = server_->posix()->rwlock_lookup(id_);
  if (!rwlock) {
    // If the resource does not exist, we create it to account for the
    // possibility of static initialization.
    rwlock = new RWLock(id_, thread_->clock());
    server_->posix()->add_rwlock(rwlock);
  }
}

void RWLockReadLock::Test(Set requests, GroupSet *groups) {
  RWLock *rwlock = server_->posix()->rwlock_lookup(id_);
  if (!rwlock) {
    state_ = INVALID;
  } else if (rwlock->state() == RWLock::EXCLUSIVE &&
             rwlock->writer() == thread_->id()) {
    state_ = DEADLOCK;
  } else {
    state_ = VALID;
  }
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(id_);
  if (state_ != VALID ||
      rwlock->state() != RWLock::EXCLUSIVE) { Enable(&group); }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> RWLockReadLock::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: RWLockReadLock"
         << "  rwlock: " << id_;
  result.push_back(stream.str());
  return result;
}

void RWLockTryReadLock::Execute() {
  switch (state_) {
    case INVALID:
      return_value_->return_value = EINVAL;
      WARNING("Invalid argument.");
      break;
    case VALID: {
      RWLock *rwlock = server_->posix()->rwlock_lookup(id_);
      ASSERT(rwlock, "Read-write lock does not exist.");
      rwlock->join_with_clock(thread_->clock());
      thread_->join_with_clock(rwlock->clock());
      if (rwlock->state() == RWLock::EXCLUSIVE) {
        return_value_->return_value = EBUSY;
      } else {
        if (rwlock->state() == RWLock::FREE) {
          ASSERT(rwlock->readers_empty(),
                 "Inconsistent read-write lock state.");
          ASSERT(rwlock->writer() == 0, "Inconsistent read-write lock state.");
          rwlock->set_state(RWLock::SHARED);
        }
        rwlock->readers_add(thread_->id());
        return_value_->return_value = 0;
        MESSAGE("Thread %d read-locks %s.", thread_->id(), id_.c_str());
      }
      break;
    }
    default:
      WARNING("No Execute() transition applies.");
  }
}

void RWLockTryReadLock::PreSchedule() {
  RWLock *rwlock = server_->posix()->rwlock_lookup(id_);
  if (!rwlock) {
    // If the resource does not exist, we create it to account for the
    // possibility of static initialization.
    rwlock = new RWLock(id_, thread_->clock());
    server_->posix()->add_rwlock(rwlock);
  }
}

void RWLockTryReadLock::Test(Set requests, GroupSet *groups) {
  RWLock *rwlock = server_->posix()->rwlock_lookup(id_);
  state_ = rwlock ? VALID : INVALID;
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(id_);
  Enable(&group);
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> RWLockTryReadLock::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: RWLockTryReadLock"
         << "  rwlock: " << id_;
  result.push_back(stream.str());
  return result;
}

void RWLockTryWriteLock::Execute() {
  switch (state_) {
    case INVALID:
      return_value_->return_value = EINVAL;
      WARNING("Invalid argument.");
      break;
    case VALID: {
      RWLock *rwlock = server_->posix()->rwlock_lookup(id_);
      ASSERT(rwlock, "Read-write lock does not exist.");
      rwlock->join_with_clock(thread_->clock());
      thread_->join_with_clock(rwlock->clock());
      if (rwlock->state() != RWLock::FREE) {
        return_value_->return_value = EBUSY;
      } else {
        ASSERT(rwlock->readers_empty(), "Inconsistent read-write lock state.");
        ASSERT(rwlock->writer() == 0, "Inconsistent read-write lock state.");
        rwlock->set_state(RWLock::EXCLUSIVE);
        rwlock->set_writer(thread_->id());
        return_value_->return_value = 0;
        MESSAGE("Thread %d write-locks %s.", thread_->id(), id_.c_str());
      }
      break;
    }
    default:
      WARNING("No Execute() transition applies.");
  }
}

void RWLockTryWriteLock::PreSchedule() {
  RWLock *rwlock = server_->posix()->rwlock_lookup(id_);
  if (!rwlock) {
    // If the resource does not exist, we create it to account for the
    // possibility of static initialization.
    rwlock = new RWLock(id_, thread_->clock());
    server_->posix()->add_rwlock(rwlock);
  }
}

void RWLockTryWriteLock::Test(Set requests, GroupSet *groups) {
  RWLock *rwlock = server_->posix()->rwlock_lookup(id_);
  state_ = rwlock ? VALID : INVALID;
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(id_);
  Enable(&group);
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> RWLockTryWriteLock::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: RWLockTryWriteLock"
         << "  rwlock: " << id_;
  result.push_back(stream.str());
  return result;
}

void RWLockUnlock::Execute() {
  switch (state_) {
    case INVALID:
      return_value_->return_value = EINVAL;
      WARNING("Invalid argument.");
      break;
    case PERMISSION:
      return_value_->return_value = EPERM;
      WARNING("Operation not permitted.");
      break;
    case VALID: {
      RWLock *rwlock = server_->posix()->rwlock_lookup(id_);
      ASSERT(rwlock, "Read-write lock does not exist.");
      ASSERT(rwlock->state() != RWLock::FREE,
             "Read-write lock is not locked.");
      ASSERT((rwlock->state() == RWLock::SHARED &&
              rwlock->readers_contains(thread_->id())) ||
             (rwlock->state() == RWLock::EXCLUSIVE &&
              rwlock->writer() == thread_->id()),
             "Read-write lock is not owned by the caller.");
      rwlock->join_with_clock(thread_->clock());
      thread_->join_with_clock(rwlock->clock());
      if (rwlock->state() == RWLock::SHARED) {
        rwlock->readers_remove(thread_->id());
        if (rwlock->readers_empty()) {
          rwlock->set_state(RWLock::FREE);
        }
      } else {
        rwlock->set_state(RWLock::FREE);
        rwlock->set_writer(0);
      }
      return_value_->return_value = 0;
      MESSAGE("Thread %d unlocks %s.", thread_->id(), id_.c_str());
      break;
    }
    default:
      WARNING("No Execute() transition applies.");
  }
}

void RWLockUnlock::Test(Set requests, GroupSet *groups) {
  RWLock *rwlock = server_->posix()->rwlock_lookup(id_);
  if (!rwlock) {
    state_ = INVALID;
  } else if (rwlock->state() == RWLock::FREE ||
             (rwlock->state() == RWLock::SHARED &&
              !rwlock->readers_contains(thread_->id())) ||
             (rwlock->state() == RWLock::EXCLUSIVE &&
              rwlock->writer() != thread_->id())) {
    state_ = PERMISSION;
  } else {
    state_ = VALID;
  }
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(id_);
  Enable(&group);
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> RWLockUnlock::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: RWLockUnlock";
  result.push_back(stream.str());
  return result;
}

void RWLockWriteLock::Execute() {
  switch (state_) {
    case INVALID:
      return_value_->return_value = EINVAL;
      WARNING("Invalid argument.");
      break;
    case DEADLOCK:
      return_value_->return_value = EDEADLK;
      WARNING("Resource deadlock would occur.");
      break;
    case VALID: {
      RWLock *rwlock = server_->posix()->rwlock_lookup(id_);
      ASSERT(rwlock, "Read-write lock does not exist.");
      ASSERT(rwlock->state() == RWLock::FREE,
             "Read-write lock is not available.");
      ASSERT(rwlock->readers_empty(), "Inconsistent read-write lock state.");
      ASSERT(rwlock->writer() == 0, "Inconsistent read-write lock state.");
      rwlock->join_with_clock(thread_->clock());
      thread_->join_with_clock(rwlock->clock());
      rwlock->set_state(RWLock::EXCLUSIVE);
      rwlock->set_writer(thread_->id());
      return_value_->return_value = 0;
      MESSAGE("Thread %d write-locks %s.", thread_->id(), id_.c_str());
      break;
    }
    default:
      WARNING("No Execute() transition applies.");
  }
}

void RWLockWriteLock::PreSchedule() {
  RWLock *rwlock = server_->posix()->rwlock_lookup(id_);
  if (!rwlock) {
    // If the resource does not exist, we create it to account for the
    // possibility of static initialization.
    rwlock = new RWLock(id_, thread_->clock());
    server_->posix()->add_rwlock(rwlock);
  }
}

void RWLockWriteLock::Test(Set requests, GroupSet *groups) {
  RWLock *rwlock = server_->posix()->rwlock_lookup(id_);
  if (!rwlock) {
    state_ = INVALID;
  } else if ((rwlock->state() == RWLock::SHARED &&
              rwlock->readers_contains(thread_->id())) ||
             (rwlock->state() == RWLock::EXCLUSIVE &&
              rwlock->writer() == thread_->id())) {
    state_ = DEADLOCK;
  } else {
    state_ = VALID;
  }
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(id_);
  if (state_ != VALID ||
      rwlock->state() == RWLock::FREE) { Enable(&group); }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> RWLockWriteLock::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: RWLockWriteLock"
         << "  rwlock: " << id_;
  result.push_back(stream.str());
  return result;
}

void SpinCreate::Execute() {
  switch (state_) {
    case BUSY:
      return_value_->return_value = EBUSY;
      WARNING("Device or resource busy.");
      break;
    case VALID: {
      ASSERT(!server_->posix()->spin_lookup(id_), "Spinlock already exists.");
      server_->posix()->add_spin(new Spin(id_, thread_->clock()));
      return_value_->return_value = 0;
      MESSAGE("Thread %d creates %s.", thread_->id(), id_.c_str());
      break;
    }
    default:
      WARNING("No Execute() transition applies.");
  }
}

void SpinCreate::Test(Set requests, GroupSet *groups) {
  Spin *spin = server_->posix()->spin_lookup(id_);
  state_ = spin ? BUSY : VALID;
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(id_);
  Enable(&group);
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> SpinCreate::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: SpinCreate"
         << "  spinlock: " << id_;
  result.push_back(stream.str());
  return result;
}

void SpinDelete::Execute() {
  switch (state_) {
    case BUSY:
      return_value_->return_value = EBUSY;
      WARNING("Device or resource busy.");
      break;
    case INVALID:
      return_value_->return_value = EINVAL;
      WARNING("Invalid argument.");
      break;
    case VALID: {
      Spin *spin = server_->posix()->spin_lookup(id_);
      ASSERT(spin, "Spinlock does not exist.");
      spin->join_with_clock(thread_->clock());
      thread_->join_with_clock(spin->clock());
      server_->posix()->remove_spin(spin);
      delete spin;
      return_value_->return_value = 0;
      MESSAGE("Thread %d deletes %s.", thread_->id(), id_.c_str());
      break;
    }
    default:
      WARNING("No Execute() transition applies.");
  }
}

void SpinDelete::Test(Set requests, GroupSet *groups) {
  Spin *spin = server_->posix()->spin_lookup(id_);
  if (!spin) {
    state_ = INVALID;
  } else if (spin->state() != Spin::UNLOCKED) {
    state_ = BUSY;
  } else {
    state_ = VALID;
  }
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(id_);
  Enable(&group);
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> SpinDelete::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: SpinDelete"
         << "  spinlock: " << id_;
  result.push_back(stream.str());
  return result;
}

void SpinLock::Execute() {
  switch (state_) {
    case INVALID:
      return_value_->return_value = EINVAL;
      WARNING("Invalid argument.");
      break;
    case DEADLOCK:
      return_value_->return_value = EDEADLK;
      WARNING("Resource deadlock would occur.");
      break;
    case VALID: {
      Spin *spin = server_->posix()->spin_lookup(id_);
      ASSERT(spin, "Spinlock does not exist.");
      ASSERT(spin->state() == Spin::UNLOCKED, "Spinlock is not available.");
      ASSERT(spin->owner() == 0, "Inconsistent spinlock state.");
      spin->join_with_clock(thread_->clock());
      thread_->join_with_clock(spin->clock());
      spin->set_state(Spin::LOCKED);
      spin->set_owner(thread_->id());
      return_value_->return_value = 0;
      MESSAGE("Thread %d locks %s.", thread_->id(), id_.c_str());
      break;
    }
    default:
      WARNING("No Execute() transition applies.");
  }
}

void SpinLock::Test(Set requests, GroupSet *groups) {
  Spin *spin = server_->posix()->spin_lookup(id_);
  if (!spin) {
    state_ = INVALID;
  } else if (spin->state() == Spin::LOCKED &&
             spin->owner() == thread_->id()) {
    state_ = DEADLOCK;
  } else {
    state_ = VALID;
  }
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(id_);
  if (state_ != VALID ||
      spin->state() == Spin::UNLOCKED) { Enable(&group); }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> SpinLock::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: SpinLock"
         << "  spinlock: " << id_;
  result.push_back(stream.str());
  return result;
}

void SpinTryLock::Execute() {
  switch (state_) {
    case INVALID:
      return_value_->return_value = EINVAL;
      WARNING("Invalid argument.");
      break;
    case DEADLOCK:
      return_value_->return_value = EDEADLK;
      WARNING("Resource deadlock would occur.");
      break;
    case VALID: {
      Spin *spin = server_->posix()->spin_lookup(id_);
      ASSERT(spin, "Spinlock does not exist.");
      spin->join_with_clock(thread_->clock());
      thread_->join_with_clock(spin->clock());
      if (spin->state() == Spin::UNLOCKED) {
        ASSERT(spin->owner() == 0, "Inconsistent spinlock state.");
        spin->set_state(Spin::LOCKED);
        spin->set_owner(thread_->id());
        return_value_->return_value = 0;
        MESSAGE("Thread %d locks %s.", thread_->id(), id_.c_str());
      } else {
        ASSERT(spin->owner() != thread_->id(),
               "A previously undetected deadlock.");
        return_value_->return_value = EBUSY;
      }
      break;
    }
    default:
      WARNING("No Execute() transition applies.");
  }
}

void SpinTryLock::Test(Set requests, GroupSet *groups) {
  Spin *spin = server_->posix()->spin_lookup(id_);
  if (!spin) {
    state_ = INVALID;
  } else if (spin->state() == Spin::LOCKED &&
             spin->owner() == thread_->id()) {
    state_ = DEADLOCK;
  } else {
    state_ = VALID;
  }
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(id_);
  Enable(&group);
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> SpinTryLock::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: SpinTryLock"
         << "  spinlock: " << id_;
  result.push_back(stream.str());
  return result;
}

void SpinUnlock::Execute() {
  switch (state_) {
    case INVALID:
      return_value_->return_value = EINVAL;
      WARNING("Invalid argument.");
      break;
    case PERMISSION:
      return_value_->return_value = EPERM;
      WARNING("Operation not permitted.");
      break;
    case VALID: {
      Spin *spin = server_->posix()->spin_lookup(id_);
      ASSERT(spin, "Spinlock does not exist.");
      ASSERT(spin->state() == Spin::LOCKED &&
             spin->owner() == thread_->id(),
             "Spinlock is not owned by the caller.");
      spin->join_with_clock(thread_->clock());
      thread_->join_with_clock(spin->clock());
      spin->set_state(Spin::UNLOCKED);
      spin->set_owner(0);
      return_value_->return_value = 0;
      MESSAGE("Thread %d unlocks %s.", thread_->id(), id_.c_str());
      break;
    }
    default:
      WARNING("No Execute() transition applies.");
  }
}

void SpinUnlock::Test(Set requests, GroupSet *groups) {
  Spin *spin = server_->posix()->spin_lookup(id_);
  if (!spin) {
    state_ = INVALID;
  } else if (spin->state() == Spin::UNLOCKED ||
             spin->owner() != thread_->id()) {
    state_ = PERMISSION;
  } else {
    state_ = VALID;
  }
  Request::Group group;
  group.add_request(this);
  group.add_to_write_set(id_);
  Enable(&group);
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> SpinUnlock::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: SpinUnlock"
         << "  spinlock: " << id_;
  result.push_back(stream.str());
  return result;
}

void ThreadCreate::Execute() {
  if (server_->CreateThread(thread_->id(),
                            create_process_,
                            &return_value_->return_value.pid,
                            &return_value_->return_value.tid)) {
    MESSAGE("Thread %d created thread %d.",
            thread_->id(), return_value_->return_value.tid);
  } else {
    WARNING("Thread %d failed to create a thread.", thread_->id());
  }
}

void ThreadCreate::Test(Set requests, GroupSet *groups) {
  Request::Group group;
  group.add_request(this);
  Enable(&group);
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> ThreadCreate::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: creating a thread";
  result.push_back(stream.str());
  return result;
}

void ThreadJoin::Execute() {
  Thread *join_thread = server_->thread_lookup(join_tid_);
  if (!join_thread) {
    WARNING("Thread %d to be joined does not exist.", join_tid_);
  } else {
    if (join_thread->state() == Thread::JOINABLE) {
      thread_->join_with_clock(join_thread->clock());
      join_thread->set_state(Thread::TERMINATED);
    }
  }
  MESSAGE("Thread %d joined thread %d.", thread_->id(), join_tid_);
}

void ThreadJoin::Test(Set requests, GroupSet *groups) {
  Request::Group group;
  group.add_request(this);
  std::stringstream stream;
  stream << "thread-" << join_tid_;
  group.add_to_write_set(stream.str());
  Thread *join_thread = server_->thread_lookup(join_tid_);
  if (join_thread && join_thread->state() == Thread::JOINABLE) {
    Enable(&group);
  }
  groups->insert(group);
  Request::Test(requests, groups);
}

std::list<std::string> ThreadJoin::ToString() const {
  std::list<std::string> result = Request::ToString();
  std::stringstream stream;
  stream << "  action: joining thread " << join_tid_;
  result.push_back(stream.str());
  return result;
}

}  // namespace arbiter
}  // namespace dbug
