/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "dbug/arbiter/request.h"

#include <list>
#include <set>
#include <sstream>
#include <string>
#include <vector>

#include "dbug/arbiter/logging.h"
#include "dbug/arbiter/server.h"
#include "dbug/arbiter/thread.h"
#include "dbug/common/system.h"
#include "dbug/common/util.h"
#include "src/thrift/arbiter/arbiter_types.h"

using dbug::thrift::Location;

namespace dbug {
namespace arbiter {

bool Request::PointerComparison::operator()(const Request *r1,
                                            const Request *r2) {
  if (!r1 && !r2) { return false; }
  if (!r1 && r2) { return true; }
  if (r1 && !r2) { return false; }
  return r1->thread()->id() < r2->thread()->id();
}

// Request::Resolved class

bool Request::Resolved::operator<(const Resolved &rhs) const {
  Request *r1 = request;
  Request *r2 = rhs.request;
  if (r1->thread()->id() < r2->thread()->id()) { return true; }
  if (r1->thread()->id() > r2->thread()->id()) { return false; }
  std::vector<int> choices1 = choices;
  std::vector<int> choices2 = rhs.choices;
  unsigned i = 0;
  for (; i < choices1.size() && i < choices2.size(); ++i) {
    if (choices1[i] < choices2[i]) { return true; }
    if (choices1[i] > choices2[i]) { return false; }
  }
  if (i < choices1.size()) { return false; }
  if (i < choices2.size()) { return true; }
  return false;
}

// Request::Group class

bool Request::Group::operator<(const Group &rhs) const {
  if (enabled_ && !rhs.enabled()) { return true; }
  if (!enabled_ && rhs.enabled()) { return false; }
  ResolvedSet set1 = resolved_set_;
  ResolvedSet set2 = rhs.resolved_set_;
  ResolvedSet::const_iterator it1 = set1.begin();
  ResolvedSet::const_iterator it2 = set2.begin();
  for (; it1 != set1.end() && it2 != set2.end(); ++it1, ++it2) {
    Resolved resolved1 = *it1;
    Resolved resolved2 = *it2;
    if (resolved1 < resolved2) { return true; }
    if (resolved2 < resolved1) { return false; }
  }
  if (it1 != set1.end()) { return false; }
  if (it2 != set2.end()) { return true; }
  return false;
}

Request::Group::Group() : enabled_(false), serialize_(true) {}

// accessors

bool Request::Group::enabled() const { return enabled_; }

std::set<std::string> Request::Group::read_set() const { return read_set_; }

Request::ResolvedSet Request::Group::resolved_set() const {
  return resolved_set_;
}

bool Request::Group::serialize() const { return serialize_; }

std::set<std::string> Request::Group::write_set() const { return write_set_; }

// mutators

void Request::Group::add_request(Request *request) {
  Resolved resolved;
  resolved.request = request;
  resolved_set_.insert(resolved);
}

void Request::Group::add_request(Request *request, std::vector<int> choices) {
  Resolved resolved;
  resolved.request = request;
  resolved.choices = choices;
  resolved_set_.insert(resolved);
}

void Request::Group::add_to_read_set(std::string id) { read_set_.insert(id); }

void Request::Group::add_to_write_set(std::string id) { write_set_.insert(id); }

void Request::Group::set_enabled(bool enabled) { enabled_ = enabled; }

void Request::Group::set_serialize(bool serialize) { serialize_ = serialize; }

// public methods

std::list<std::string> Request::Group::ToString() const {
  std::list<std::string> result;
  std::stringstream stream;
  result.push_back(" Group:");
  stream << "  Enabled: " << (enabled_ ? "YES" : "NO");
  result.push_back(stream.str());
  stream.str(std::string());
  for (const Resolved &resolved : resolved_set_) {
    stream << "  Non-deterministic chocies: "
           << Util::VectorToString<int>(resolved.choices);
    result.push_back(stream.str());
    stream.str(std::string());
    std::list<std::string> subresult = resolved.request->ToString();
    result.splice(result.end(), subresult);
  }
  stream << "  Read set: " << Util::SetToString<std::string>(read_set_);
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "  Write set: " << Util::SetToString<std::string>(write_set_);
  result.push_back(stream.str());
  return result;
}

// Request class

bool Request::SortByTimestamps(Request *r1, Request *r2) {
  return (r1->timestamp_ < r2->timestamp_);
}

bool Request::SortByThreadIDs(Request *r1, Request *r2) {
  return (r1->thread_->id() < r2->thread_->id());
}

std::string Request::StateToString(State state) {
  switch (state) {
    case INITIAL:
      return "initial";
    case BUSY:
      return "busy";
    case DEADLOCK:
      return "deadlock";
    case INVALID:
      return "invalid";
    case PERMISSION:
      return "permission";
    case VALID:
      return "valid";
    default:
      return "unknown";
  }
}

bool Request::operator<(const Request *rhs) const {
  return thread_->id() < rhs->thread_->id();
}

Request::Request(Server *server,
                 Thread *thread,
                 const std::string &func,
                 const std::vector<Location> &backtrace) :
    backtrace_(backtrace), canceled_(false), function_(func), server_(server),
    state_(INITIAL), thread_(thread), timestamp_(clock()) {
  System::CVarInit(&cvar_, NULL);
  thread_->advance_clock(thread_->id());
}

Request::~Request() { System::CVarDestroy(&cvar_); }

// accessors

std::vector<Location> Request::backtrace() const { return backtrace_; }

bool Request::canceled() const { return canceled_; }

std::string Request::function() const { return function_; }

Thread *Request::thread() const { return thread_; }

int Request::timestamp() const { return timestamp_; }

// mutators

void Request::set_nondet_choices(std::vector<int> nondet_choices) {
  nondet_choices_ = nondet_choices;
}

void Request::set_state(State state) { state_ = state; }

// public methods

void Request::Cancel() {
  canceled_ = true;
  System::CVarSignal(&cvar_);
}

void Request::Enable(Group *group) {
  group->set_enabled(true);
  server_->increment_nenabled();
}

void Request::PreSchedule() {}

void Request::Remove() {
  server_->remove_request(this);
  std::set<int> sleep_set = server_->sleep_set();
  // If the request has executed and is a not a sleep request, empty
  // the sleep set.
  if (!canceled_ && sleep_set.find(thread_->id()) == sleep_set.end()) {
    server_->clear_sleep_set();
  }
}

void Request::Schedule() {
  // Add this request to the list of existing requests.
  server_->add_request(this);
  // Wait for the arbiter to execute or cancel the request.
  server_->WaitingRequest(&cvar_);
  if (!canceled_) {
    // Execute the request.
    Execute();
    Remove();
    server_->RequestExecuted();
  } else {
    // Cancel the request.
    Remove();
    server_->RequestCanceled();
  }
}

void Request::Test(Set requests, GroupSet *groups) {
  requests.erase(this);
  if (!requests.empty()) {
    Request *request = *(requests.begin());
    request->Test(requests, groups);
  }
}

std::list<std::string> Request::ToString() const {
  std::list<std::string> result;
  std::stringstream stream;
  result.push_back("  Request:");
  for (unsigned i = 0; i < backtrace_.size(); i++) {
    stream << "   backtrace: "
           << backtrace_[i].func_name
           << " at " << System::Basename(backtrace_[i].file_name.c_str())
           << ":" << backtrace_[i].line;
    result.push_back(stream.str());
    stream.str(std::string());
  }
  stream << "   clock: " << Util::VectorToString<int>(thread_->clock());
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "   function: " << function_;
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "   id: " << thread_->id();
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "   state: " << StateToString(state_);
  result.push_back(stream.str());
  stream.str(std::string());
  return result;
}

void Request::WakeUp() { System::CVarSignal(&cvar_); }

}  // namespace arbiter
}  // namespace dbug
