/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "dbug/arbiter/thread.h"

#include <sstream>
#include <string>

#include "dbug/arbiter/logging.h"
#include "dbug/common/util.h"

namespace dbug {
namespace arbiter {

std::string Thread::StateToString(State state) {
  switch (state) {
    case DETACHED:
      return "detached";
    case JOINABLE:
      return "joinable";
    case RUNNING:
      return "running";
    case TERMINATED:
      return "terminated";
    case WAITING:
      return "waiting";
    default:
      return "unknown";
  }
}

int Thread::id_counter() { return id_counter_; }

Thread::Thread()
    : id_(++id_counter_),
      old_state_(RUNNING),
      pgrp_(0),
      state_(RUNNING) {}

// accessors

Clock Thread::clock() const { return clock_; }

int Thread::id() const { return id_; }

Process *Thread::pgrp() const { return pgrp_; }

Thread::State Thread::state() const { return state_; }

// mutators

void Thread::advance_clock(int tid) {
  if (clock_.size() < (unsigned) tid) { clock_.resize(tid, 0); }
  clock_[tid - 1]++;
}

void Thread::join_with_clock(const Clock &clock) { clock_.JoinWith(clock); }

void Thread::set_clock(const Clock &clock) { clock_ = clock; }

void Thread::set_pgrp(Process *process) { pgrp_ = process; }

void Thread::set_state(Thread::State state) {
  if (state == PREVIOUS) {
    ASSERT(state_ == WAITING,
           "Only waiting threads can transition to a previous state.");
    state_ = old_state_;
  } else {
    ASSERT((state != WAITING) || (state_ == RUNNING || state_ == DETACHED),
           "Only running threads can transition to a waiting state.");
    old_state_ = state_;
    state_ = state;
  }
}

// public methods

std::list<std::string> Thread::ToString() const {
  std::list<std::string> result;
  std::stringstream stream;
  stream << "  Thread:";
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "    clock: " << Util::VectorToString<int>(clock_);
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "    id: " << id_;
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "    old state: " << StateToString(old_state_);
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "    state: " << StateToString(state_);
  result.push_back(stream.str());
  stream.str(std::string());
  return result;
}

// private static fields

int Thread::id_counter_ = 0;

}  // namespace arbiter
}  // namespace dbug
