/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <libxml/parser.h>
#include <libxml/tree.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>

#include <boost/algorithm/string/predicate.hpp>
#include <map>
#include <set>
#include <sstream>
#include <string>
#include <vector>

struct Enum {
  std::string name;
  std::vector<std::string> values;
};

struct Field {
  std::string name;
  std::string thrift_type;
  std::string c_type;
};

struct Struct {
  std::string name;
  std::vector<Field> fields;
};

struct ReturnValue {
  std::string thrift_type;
  std::string c_type;
};

struct Argument {
  std::string name;
  std::string thrift_type;
  std::string c_type;
};

struct RPC {
  std::string name;
  std::string preschedule;
  std::string schedule;
  std::vector<Argument> arguments;
  ReturnValue return_value;
};

const char *copyright =
    "// Authors: Jiri Simsa\n"
    "// Copyright (c) 2013, Carnegie Mellon University.\n"
    "// All rights reserved.\n"
    "//\n"
    "// Redistribution and use in source and binary forms, with or without\n"
    "// modification, are permitted provided that the following conditions\n"
    "// are met:\n"
    "// 1. Redistributions of source code must retain the above copyright\n"
    "//    notice, this list of conditions and the following disclaimer.\n"
    "// 2. Redistributions in binary form must reproduce the above copyright\n"
    "//    notice, this list of conditions and "
    "the following disclaimer in the\n"
    "//    documentation and/or other materials "
    "provided with the distribution.\n"
    "// 3. Neither the name of the University "
    "nor the names of its contributors\n"
    "//    may be used to endorse or promote "
    "products derived from this software\n"
    "//    without specific prior written permission.\n"
    "//\n"
    "// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS\n"
    "// ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT\n"
    "// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR\n"
    "// A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT\n"
    "// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,\n"
    "// INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,\n"
    "// BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS\n"
    "// OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED\n"
    "// AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT\n"
    "// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY\n"
    "// WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE\n"
    "// POSSIBILITY OF SUCH DAMAGE.\n";

void Die(const char *format, ...) {
  va_list ap;
  va_start(ap, format);
  vfprintf(stderr, format, ap);
  va_end(ap);
  exit(-1);
}

bool Exists(const char *path) { return (access(path, F_OK) != -1); }

void CreateDirectory(const char *path) {
  int ret = mkdir(path, S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH);
  if (ret == -1) { Die("Could not create directory %s.\n", path); }
}

void UpdateFile(std::string previous, std::string current) {
  // Overwrite the previous version only if it is different from the
  // current version to avoid recompilation of unchanged files.
  std::stringstream stream;
  stream << "diff -q " << previous.c_str() << " " << current.c_str();
  if (!Exists(previous.c_str()) || system(stream.str().c_str())) {
    int ret = rename(current.c_str(), previous.c_str());
    if (ret == -1) {
      Die("Could not move %s to %s.\n", current.c_str(), previous.c_str());
    } else {
      printf("Updated file %s.\n", previous.c_str());
    }
  } else {
    printf("File %s is up-to-date.\n", previous.c_str());
    int ret = unlink(current.c_str());
    if (ret == -1) { Die("Could not remove %s.\n", current.c_str()); }
  }
}

bool IsPrimitiveType(std::string type) {
  if (type == "int" ||
      type == "bool" ||
      boost::algorithm::ends_with(type, "::type")) {
    return true;
  }
  return false;
}

Enum ParseEnumNode(xmlNode *node) {
  Enum result;
  const xmlChar *name_xml = reinterpret_cast<const xmlChar *>("name");
  if (xmlHasProp(node, name_xml)) {
    xmlChar *prop = xmlGetProp(node, name_xml);
    result.name = reinterpret_cast<char *>(prop);
    xmlFree(prop);
  } else {
    Die("The name attribute of the <enum> element is missing.\n");
  }
  for (xmlNode *child = node->children; child; child = child->next) {
    if (child->type == XML_ELEMENT_NODE) {
      if (xmlStrcmp(child->name,
                    reinterpret_cast<const xmlChar *>("value")) == 0) {
        const xmlChar *name_xml = reinterpret_cast<const xmlChar *>("name");
        if (xmlHasProp(child, name_xml)) {
          xmlChar *prop = xmlGetProp(child, name_xml);
          std::string value(reinterpret_cast<char *>(prop));
          xmlFree(prop);
          result.values.push_back(value);
        } else {
          Die("The name attribute of the <value> element is missing.\n");
        }
      }
    }
  }
  return result;
}

Struct ParseStructNode(xmlNode *node) {
  Struct result;
  const xmlChar *name_xml = reinterpret_cast<const xmlChar *>("name");
  if (xmlHasProp(node, name_xml)) {
    xmlChar *prop = xmlGetProp(node, name_xml);
    result.name = reinterpret_cast<char *>(prop);
    xmlFree(prop);
  } else {
    Die("The name attribute of the <struct> element is missing.\n");
  }
  for (xmlNode *child = node->children; child; child = child->next) {
    if (child->type == XML_ELEMENT_NODE) {
      if (xmlStrcmp(child->name,
                    reinterpret_cast<const xmlChar *>("field")) == 0) {
        Field field;
        const xmlChar *name_xml = reinterpret_cast<const xmlChar *>("name");
        if (xmlHasProp(child, name_xml)) {
          xmlChar *prop = xmlGetProp(child, name_xml);
          field.name = reinterpret_cast<char *>(prop);
          xmlFree(prop);
        } else {
          Die("The name attribute of the <field> element is missing.\n");
        }
        const xmlChar *thrift_type_xml =
            reinterpret_cast<const xmlChar *>("thrift_type");
        if (xmlHasProp(child, thrift_type_xml)) {
          xmlChar *prop = xmlGetProp(child, thrift_type_xml);
          field.thrift_type = reinterpret_cast<char *>(prop);
          xmlFree(prop);
        } else {
          Die("The thrift_type attribute of the <field> element is missing.\n");
        }
        const xmlChar *c_type_xml = reinterpret_cast<const xmlChar *>("c_type");
        if (xmlHasProp(child, c_type_xml)) {
          xmlChar *prop = xmlGetProp(child, c_type_xml);
          field.c_type = reinterpret_cast<char *>(prop);
          xmlFree(prop);
        } else {
          Die("The c_type attribute of the <field> element is missing.\n");
        }
        result.fields.push_back(field);
      }
    }
  }
  return result;
}

RPC ParseRPCNode(xmlNode *node) {
  RPC result;
  const xmlChar *name_xml = reinterpret_cast<const xmlChar *>("name");
  if (xmlHasProp(node, name_xml)) {
    xmlChar *prop = xmlGetProp(node, name_xml);
    result.name = reinterpret_cast<char *>(prop);
    xmlFree(prop);
  } else {
    Die("The name attribute of the <rpc> element is missing.\n");
  }
  const xmlChar *schedule_xml = reinterpret_cast<const xmlChar *>("schedule");
  if (xmlHasProp(node, schedule_xml)) {
    xmlChar *prop = xmlGetProp(node, schedule_xml);
    result.schedule = reinterpret_cast<char *>(prop);
    xmlFree(prop);
    if (result.schedule != "true" &&
        result.schedule != "false") {
      Die("Unsupported value %s for the schedule attribute "
          "of the <rpc> element.\n", result.schedule.c_str());
    }
  } else {
    result.schedule = "false";
  }
  const xmlChar *pre_xml = reinterpret_cast<const xmlChar *>("preschedule");
  if (xmlHasProp(node, pre_xml)) {
    xmlChar *prop = xmlGetProp(node, pre_xml);
    result.preschedule = reinterpret_cast<char *>(prop);
    xmlFree(prop);
    if (result.preschedule != "true" &&
        result.preschedule != "false") {
      Die("Unsupported value %s for the preschedule attribute "
          "of the <rpc> element.\n", result.preschedule.c_str());
    }
  } else {
    result.preschedule = "false";
  }
  for (xmlNode *child = node->children; child; child = child->next) {
    if (child->type == XML_ELEMENT_NODE) {
      if (xmlStrcmp(child->name,
                    reinterpret_cast<const xmlChar *>("return_value")) == 0) {
        const xmlChar *thrift_type_xml =
            reinterpret_cast<const xmlChar *>("thrift_type");
        if (xmlHasProp(child, thrift_type_xml)) {
          xmlChar *prop = xmlGetProp(child, thrift_type_xml);
          result.return_value.thrift_type = reinterpret_cast<char *>(prop);
          xmlFree(prop);
        } else {
          Die("The thrift_type attribute of the "
              "<return_value> element is missing.\n");
        }
        const xmlChar *c_type_xml = reinterpret_cast<const xmlChar *>("c_type");
        if (xmlHasProp(child, c_type_xml)) {
          xmlChar *prop = xmlGetProp(child, c_type_xml);
          result.return_value.c_type = reinterpret_cast<char *>(prop);
          xmlFree(prop);
        } else {
          Die("The c_type attribute of the "
              "<return_value> element is missing.\n");
        }
      }
      if (xmlStrcmp(child->name,
                    reinterpret_cast<const xmlChar *>("argument")) == 0) {
        Argument argument;
        const xmlChar *name_xml = reinterpret_cast<const xmlChar *>("name");
        if (xmlHasProp(child, name_xml)) {
          xmlChar *prop = xmlGetProp(child, name_xml);
          argument.name = reinterpret_cast<char *>(prop);
          xmlFree(prop);
        } else {
          Die("The name attribute of the <argument> element is missing.\n");
        }
        const xmlChar *thrift_type_xml =
            reinterpret_cast<const xmlChar *>("thrift_type");
        if (xmlHasProp(child, thrift_type_xml)) {
          xmlChar *prop = xmlGetProp(child, thrift_type_xml);
          argument.thrift_type = reinterpret_cast<char *>(prop);
          xmlFree(prop);
        } else {
          Die("The thrift_type attribute of the "
              "<argument> element is missing.\n");
        }
        const xmlChar *c_type_xml = reinterpret_cast<const xmlChar *>("c_type");
        if (xmlHasProp(child, c_type_xml)) {
          xmlChar *prop = xmlGetProp(child, c_type_xml);
          argument.c_type = reinterpret_cast<char *>(prop);
          xmlFree(prop);
        } else {
          Die("The c_type attribute of the <argument> element is missing.\n");
        }
        result.arguments.push_back(argument);
      }
    }
  }
  return result;
}

void AddCopyright(FILE *file) { fprintf(file, "%s\n", copyright); }

void GenerateArbiterThrift(std::map<std::string, Enum> enums,
                           std::map<std::string, Struct> structs,
                           std::map<std::string, RPC> rpcs) {
  if (!Exists(SRCDIR "/src/thrift/gen")) {
    CreateDirectory(SRCDIR "/src/thrift/gen");
  }
  std::string name = SRCDIR "/src/thrift/gen/arbiter.thrift";
  std::string tmp_name = name + ".tmp";
  FILE *file = fopen(tmp_name.c_str(), "w");
  if (file == NULL) { Die("Could not create file %s.\n", tmp_name.c_str()); }

  AddCopyright(file);
  fprintf(file, "// This is an automatically generated file. ");
  fprintf(file, "Any modifications will be lost.\n");
  fprintf(file, "\n");
  fprintf(file, "namespace cpp dbug.thrift\n");
  fprintf(file, "\n");
  fprintf(file, "include \"../common.thrift\"\n");
  fprintf(file, "\n");

  for (const std::pair<std::string, Enum> &pair : enums) {
    const Enum &e = pair.second;
    std::stringstream stream;
    stream << "enum " << e.name << " {\n";
    for (unsigned i = 0; i < e.values.size(); i++) {
      std::string value = e.values[i];
      stream << "  " << value;
      if (i < e.values.size() - 1) { stream << ","; }
      stream << "\n";
    }
    stream << "}\n\n";
    fprintf(file, "%s", stream.str().c_str());
  }

  for (const std::pair<std::string, Struct> &pair : structs) {
    const Struct &s = pair.second;
    std::stringstream stream;
    stream << "struct " << s.name << " {\n";
    for (unsigned i = 0; i < s.fields.size(); i++) {
      Field field = s.fields[i];
      stream << "  " << i+1 << ":" << field.thrift_type << " " << field.name;
      if (i < s.fields.size() - 1) { stream << ","; }
      stream << "\n";
    }
    stream << "}\n\n";
    fprintf(file, "%s", stream.str().c_str());
  }

  fprintf(file, "service arbiter {\n");

  for (const std::pair<std::string, RPC> &pair : rpcs) {
    const RPC &rpc = pair.second;
    std::stringstream stream;
    stream << "  " << rpc.return_value.thrift_type << " " << rpc.name << "(";
    for (unsigned i = 0; i < rpc.arguments.size(); i++) {
      Argument argument = rpc.arguments[i];
      if (i > 0) { stream << ", "; }
      stream << i + 1 << ":" << argument.thrift_type << " " << argument.name;
    }
    stream << ")\n";
    fprintf(file, "%s", stream.str().c_str());
  }
  fprintf(file, "}\n");

  fclose(file);
  UpdateFile(name, tmp_name);
}

void GenerateArbiterInterfaceHeader(std::map<std::string, Struct> structs,
                                    std::map<std::string, RPC> rpcs) {
  if (!Exists(SRCDIR "/include/dbug/interposition/gen")) {
    CreateDirectory(SRCDIR "/include/dbug/interposition/gen");
  }
  std::string name =
      SRCDIR "/include/dbug/interposition/gen/arbiter-interface.h";
  std::string tmp_name = name + ".tmp";
  FILE *file = fopen(tmp_name.c_str(), "w");
  if (file == NULL) { Die("Could not create file %s.\n", tmp_name.c_str()); }

  AddCopyright(file);
  fprintf(file, "// This is an automatically generated file. ");
  fprintf(file, "Any modifications will be lost.\n");
  fprintf(file, "\n");
  fprintf(file, "#include <set>\n");
  fprintf(file, "#include <string>\n");
  fprintf(file, "#include <vector>\n");
  fprintf(file, "\n");
  fprintf(file, "#include \"dbug/common/macros.h\"\n");
  fprintf(file, "#include \"src/thrift/arbiter/arbiter_types.h\"\n");
  fprintf(file, "\n");
  fprintf(file, "namespace dbug {\n");
  fprintf(file, "namespace interposition {\n");
  fprintf(file, "\n");
  fprintf(file, "class ArbiterInterface {\n");
  fprintf(file, " public:\n");
  fprintf(file,
          "  virtual void CancelThreadRegistration(const int tid) = 0;\n");
  fprintf(file,
          "  virtual void CompleteThreadRegistration(const int pid, "
          "const int tid) = 0;\n");
  fprintf(file,
          "  virtual void PostThreadRegistration(const int tid, "
          "const std::string &func, "
          "const std::vector<dbug::thrift::Location> &backtrace, "
          "const bool create_process, int *new_pid, int *new_tid) = 0;\n");
  fprintf(file, "  virtual void RegisterThread() = 0;\n");
  fprintf(file, "  virtual void UnregisterAllThreads() = 0;\n");
  fprintf(file, "  virtual void UnregisterThread() = 0;\n");
  fprintf(file, "\n");
  for (const std::pair<std::string, RPC> &pair : rpcs) {
    const RPC &rpc = pair.second;
    std::stringstream stream;
    stream << "  virtual ";
    std::string return_type;
    if (rpc.schedule == "true") {
      if (rpc.return_value.c_type == "bool") {
        return_type = "void";
      } else {
        std::map<std::string, Struct>::iterator sit =
            structs.find(rpc.return_value.thrift_type);
        if (sit != structs.end()) {
          if (sit->second.fields[1].name == "return_value") {
            return_type = sit->second.fields[1].c_type;
          } else {
            Die("Could not find field 'return_value'.\n");
          }
        } else {
          Die("Could not find struct %s.\n",
              rpc.return_value.thrift_type.c_str());
        }
      }
    } else {
      return_type = rpc.return_value.c_type;
    }
    stream << return_type << " " << rpc.name << "(";
    for (unsigned i = 0; i < rpc.arguments.size(); i++) {
      Argument argument = rpc.arguments[i];
      if (i > 0) { stream << ", "; }
      stream << "const " << argument.c_type << " ";
      if (!IsPrimitiveType(argument.c_type)) { stream << "&"; }
      stream << argument.name;
    }
    stream << ") = 0;\n";
    fprintf(file, "%s", stream.str().c_str());
  }
  fprintf(file, "};\n");
  fprintf(file, "\n");
  fprintf(file, "}  // namespace interposition\n");
  fprintf(file, "}  // namespace dbug\n");

  fclose(file);
  UpdateFile(name, tmp_name);
}

void GenerateDefaultLocalArbiterHeader(std::map<std::string, Struct> structs,
                                       std::map<std::string, RPC> rpcs) {
  if (!Exists(SRCDIR "/include/dbug/interposition/gen")) {
    CreateDirectory(SRCDIR "/include/dbug/interposition/gen");
  }
  std::string name =
      SRCDIR "/include/dbug/interposition/gen/default-local-arbiter.h";
  std::string tmp_name = name + ".tmp";
  FILE *file = fopen(tmp_name.c_str(), "w");
  if (file == NULL) { Die("Could not create file %s.\n", tmp_name.c_str()); }

  AddCopyright(file);
  fprintf(file, "// This is an automatically generated file. ");
  fprintf(file, "Any modifications will be lost.\n");
  fprintf(file, "\n");
  fprintf(file, "#include <set>\n");
  fprintf(file, "#include <string>\n");
  fprintf(file, "#include <vector>\n");
  fprintf(file, "\n");
  fprintf(file, "#include \"dbug/arbiter/handler.h\"\n");
  fprintf(file, "#include \"dbug/arbiter/server.h\"\n");
  fprintf(file, "#include \"dbug/common/macros.h\"\n");
  fprintf(file, "#include \"dbug/interposition/arbiter.h\"\n");
  fprintf(file, "#include \"src/thrift/arbiter/arbiter_types.h\"\n");
  fprintf(file, "\n");
  fprintf(file, "namespace dbug {\n");
  fprintf(file, "namespace interposition {\n");
  fprintf(file, "\n");
  fprintf(file, "class DefaultLocalArbiter : public Arbiter {\n");
  fprintf(file, " public:\n");
  fprintf(file, "  DefaultLocalArbiter();\n");
  fprintf(file, "  virtual ~DefaultLocalArbiter();\n");
  fprintf(file, "\n");
  for (const std::pair<std::string, RPC> &pair : rpcs) {
    const RPC &rpc = pair.second;
    std::stringstream stream;
    stream << "  virtual ";
    std::string return_type;
    if (rpc.schedule == "true") {
      if (rpc.return_value.c_type == "bool") {
        return_type = "void";
      } else {
        std::map<std::string, Struct>::iterator sit =
            structs.find(rpc.return_value.thrift_type);
        if (sit != structs.end()) {
          if (sit->second.fields[1].name == "return_value") {
            return_type = sit->second.fields[1].c_type;
          } else {
            Die("Could not find field 'return_value'.\n");
          }
        } else {
          Die("Could not find struct %s.\n",
              rpc.return_value.thrift_type.c_str());
        }
      }
    } else {
      return_type = rpc.return_value.c_type;
    }
    stream << return_type << " " << rpc.name << "(";
    for (unsigned i = 0; i < rpc.arguments.size(); i++) {
      Argument argument = rpc.arguments[i];
      if (i > 0) { stream << ", "; }
      stream << "const " << argument.c_type << " ";
      if (!IsPrimitiveType(argument.c_type)) { stream << "&"; }
      stream << argument.name;
    }
    stream << ");\n";
    fprintf(file, "%s", stream.str().c_str());
  }
  fprintf(file, "\n");
  fprintf(file, " private:\n");
  fprintf(file, "  System::Thread arbiter_;\n");
  fprintf(file, "  dbug::arbiter::Server *server_;\n");
  fprintf(file, "  dbug::arbiter::Handler *handler_;\n");
  fprintf(file, "\n");
  fprintf(file, "  DISALLOW_COPY_AND_ASSIGN(DefaultLocalArbiter);\n");
  fprintf(file, "};\n");
  fprintf(file, "\n");
  fprintf(file, "}  // namespace interposition\n");
  fprintf(file, "}  // namespace dbug\n");

  fclose(file);
  UpdateFile(name, tmp_name);
}

void GenerateDefaultLocalArbiterSource(std::map<std::string, Struct> structs,
                                       std::map<std::string, RPC> rpcs) {
  if (!Exists(SRCDIR "/src/interposition/gen")) {
    CreateDirectory(SRCDIR "/src/interposition/gen");
  }
  std::string name =
      SRCDIR "/src/interposition/gen/default-local-arbiter.cc";
  std::string tmp_name = name + ".tmp";
  FILE *file = fopen(tmp_name.c_str(), "w");
  if (file == NULL) { Die("Could not create file %s.\n", tmp_name.c_str()); }

  AddCopyright(file);
  fprintf(file, "// This is an automatically generated file. ");
  fprintf(file, "Any modifications will be lost.\n");
  fprintf(file, "\n");
  fprintf(file, "#include \"dbug/interposition/default-local-arbiter.h\"\n");
  fprintf(file, "\n");
  fprintf(file, "#include <set>\n");
  fprintf(file, "#include <string>\n");
  fprintf(file, "#include <vector>\n");
  fprintf(file, "\n");
  fprintf(file, "#include \"dbug/arbiter/handler.h\"\n");
  fprintf(file, "#include \"dbug/arbiter/server.h\"\n");
  fprintf(file, "#include \"dbug/interposition/logging.h\"\n");
  fprintf(file, "#include \"src/thrift/arbiter/arbiter_types.h\"\n");
  fprintf(file, "\n");
  fprintf(file, "namespace dbug {\n");
  fprintf(file, "namespace interposition {\n");
  fprintf(file, "\n");
  for (const std::pair<std::string, RPC> &pair : rpcs) {
    const RPC &rpc = pair.second;
    std::stringstream stream;
    std::string return_type;
    if (rpc.schedule == "true") {
      if (rpc.return_value.c_type == "bool") {
        return_type = "void";
      } else {
        std::map<std::string, Struct>::iterator sit =
            structs.find(rpc.return_value.thrift_type);
        if (sit != structs.end()) {
          if (sit->second.fields[1].name == "return_value") {
            return_type = sit->second.fields[1].c_type;
          } else {
            Die("Could not find field 'return_value'.\n");
          }
        } else {
          Die("Could not find struct %s.\n",
              rpc.return_value.thrift_type.c_str());
        }
      }
    } else {
      return_type = rpc.return_value.c_type;
    }
    stream << return_type << " DefaultLocalArbiter::" << rpc.name << "(";
    for (unsigned i = 0; i < rpc.arguments.size(); i++) {
      Argument argument = rpc.arguments[i];
      if (i > 0) { stream << ", "; }
      stream << "const " << argument.c_type << " ";
      if (!IsPrimitiveType(argument.c_type)) { stream << "&"; }
      stream << argument.name;
    }
    stream << ") {\n";
    if (rpc.schedule == "true") {
      if (return_type != "void") {
        stream << "  " << rpc.return_value.c_type << " option;\n";
        stream << "  handler_->" << rpc.name << "(option";
        for (unsigned i = 0; i < rpc.arguments.size(); i++) {
          Argument argument = rpc.arguments[i];
          stream << ", " << argument.name;
        }
        stream << ");\n"
               << "  MESSAGE(\"Thread %d called "
               << rpc.name << "().\", ThreadID());\n";
        stream << "  if (!option.success) {\n"
               << "    //TODO(jsimsa): Die quietly.\n"
               << "  }\n"
               << "  return option.return_value;\n";
      } else {
        stream << "  bool success = handler_->" << rpc.name << "(";
        for (unsigned i = 0; i < rpc.arguments.size(); i++) {
          Argument argument = rpc.arguments[i];
          if (i > 0) { stream << ", "; }
          stream << argument.name;
        }
        stream << ");\n"
               << "  MESSAGE(\"Thread %d called "
               << rpc.name << "().\", ThreadID());\n";
        stream << "  if (!success) {\n"
               << "    //TODO(jsimsa): Die quietly.\n"
               << "  }\n";
      }
    } else {
      stream << "  ";
      std::map<std::string, Struct>::iterator sit =
          structs.find(rpc.return_value.thrift_type);
      if (sit != structs.end()) {
        stream << return_type << " return_value;\n"
               << "handler_->" << rpc.name << "(return_value";
        for (unsigned i = 0; i < rpc.arguments.size(); i++) {
          Argument argument = rpc.arguments[i];
          stream << ", " << argument.name;
        }
        stream << ");\n"
               << "  MESSAGE(\"Thread %d called "
               << rpc.name << "().\", ThreadID());\n";
        if (return_type != "void") {
          stream << "  return return_value;\n";
        }
      } else {
        if (return_type != "void") {
          stream << return_type << " return_value = ";
        }
        stream << "handler_->" << rpc.name << "(";
        for (unsigned i = 0; i < rpc.arguments.size(); i++) {
          Argument argument = rpc.arguments[i];
          if (i > 0) { stream << ", "; }
          stream << argument.name;
        }
        stream << ");\n"
               << "  MESSAGE(\"Thread %d called "
               << rpc.name << "().\", ThreadID());\n";
        if (return_type != "void") {
          stream << "  return return_value;\n";
        }
      }
    }
    stream << "}\n\n";
    fprintf(file, "%s", stream.str().c_str());
  }
  fprintf(file, "}  // namespace interposition\n");
  fprintf(file, "}  // namespace dbug\n");

  fclose(file);
  UpdateFile(name, tmp_name);
}

void GenerateDefaultRemoteArbiterHeader(std::map<std::string, Struct> structs,
                                        std::map<std::string, RPC> rpcs) {
  if (!Exists(SRCDIR "/include/dbug/interposition/gen")) {
    CreateDirectory(SRCDIR "/include/dbug/interposition/gen");
  }
  std::string name =
      SRCDIR "/include/dbug/interposition/gen/default-remote-arbiter.h";
  std::string tmp_name = name + ".tmp";
  FILE *file = fopen(tmp_name.c_str(), "w");
  if (file == NULL) { Die("Could not create file %s.\n", tmp_name.c_str()); }

  AddCopyright(file);
  fprintf(file, "// This is an automatically generated file. ");
  fprintf(file, "Any modifications will be lost.\n");
  fprintf(file, "\n");
  fprintf(file, "#include <set>\n");
  fprintf(file, "#include <string>\n");
  fprintf(file, "#include <vector>\n");
  fprintf(file, "\n");
  fprintf(file, "#include \"dbug/common/macros.h\"\n");
  fprintf(file, "#include \"dbug/interposition/arbiter.h\"\n");
  fprintf(file, "#include \"src/thrift/arbiter/arbiter_types.h\"\n");
  fprintf(file, "\n");
  fprintf(file, "namespace dbug {\n");
  fprintf(file, "namespace interposition {\n");
  fprintf(file, "\n");
  fprintf(file, "class DefaultRemoteArbiter : public Arbiter {\n");
  fprintf(file, " public:\n");
  fprintf(file, "  DefaultRemoteArbiter(std::string hostname, int port);\n");
  fprintf(file, "\n");
  for (const std::pair<std::string, RPC> &pair : rpcs) {
    const RPC &rpc = pair.second;
    std::stringstream stream;
    stream << "  virtual ";
    std::string return_type;
    if (rpc.schedule == "true") {
      if (rpc.return_value.c_type == "bool") {
        return_type = "void";
      } else {
        std::map<std::string, Struct>::iterator sit =
            structs.find(rpc.return_value.thrift_type);
        if (sit != structs.end()) {
          if (sit->second.fields[1].name == "return_value") {
            return_type = sit->second.fields[1].c_type;
          } else {
            Die("Could not find field 'return_value'.\n");
          }
        } else {
          Die("Could not find struct %s.\n",
              rpc.return_value.thrift_type.c_str());
        }
      }
    } else {
      return_type = rpc.return_value.c_type;
    }
    stream << return_type << " " << rpc.name << "(";
    for (unsigned i = 0; i < rpc.arguments.size(); i++) {
      Argument argument = rpc.arguments[i];
      if (i > 0) { stream << ", "; }
      stream << "const " << argument.c_type << " ";
      if (!IsPrimitiveType(argument.c_type)) { stream << "&"; }
      stream << argument.name;
    }
    stream << ");\n";
    fprintf(file, "%s", stream.str().c_str());
  }
  fprintf(file, "\n");
  fprintf(file, " protected:\n");
  fprintf(file, "  void DieQuietly();\n");
  fprintf(file, "\n");
  fprintf(file, "  std::string hostname_;\n");
  fprintf(file, "  int port_;\n");
  fprintf(file, "\n");
  fprintf(file, " private:\n");
  fprintf(file, "  DISALLOW_COPY_AND_ASSIGN(DefaultRemoteArbiter);\n");
  fprintf(file, "};\n");
  fprintf(file, "\n");
  fprintf(file, "}  // namespace interposition\n");
  fprintf(file, "}  // namespace dbug\n");

  fclose(file);
  UpdateFile(name, tmp_name);
}

void GenerateDefaultRemoteArbiterSource(std::map<std::string, Struct> structs,
                                        std::map<std::string, RPC> rpcs) {
  if (!Exists(SRCDIR "/src/interposition/gen")) {
    CreateDirectory(SRCDIR "/src/interposition/gen");
  }
  std::string name =
      SRCDIR "/src/interposition/gen/default-remote-arbiter.cc";
  std::string tmp_name = name + ".tmp";
  FILE *file = fopen(tmp_name.c_str(), "w");
  if (file == NULL) { Die("Could not create file %s.\n", tmp_name.c_str()); }

  AddCopyright(file);
  fprintf(file, "// This is an automatically generated file. ");
  fprintf(file, "Any modifications will be lost.\n");
  fprintf(file, "\n");
  fprintf(file, "#include \"dbug/interposition/default-remote-arbiter.h\"\n");
  fprintf(file, "\n");
  fprintf(file, "#include <boost/scoped_ptr.hpp>\n");
  fprintf(file, "#include <set>\n");
  fprintf(file, "#include <string>\n");
  fprintf(file, "#include <vector>\n");
  fprintf(file, "\n");
  fprintf(file, "#include \"dbug/interposition/arbiter-connection.h\"\n");
  fprintf(file, "#include \"dbug/interposition/logging.h\"\n");
  fprintf(file, "#include \"src/thrift/arbiter/arbiter_types.h\"\n");
  fprintf(file, "\n");
  fprintf(file, "namespace dbug {\n");
  fprintf(file, "namespace interposition {\n");
  fprintf(file, "\n");

  for (const std::pair<std::string, RPC> &pair : rpcs) {
    const RPC &rpc = pair.second;
    std::stringstream stream;
    std::string return_type;
    if (rpc.schedule == "true") {
      if (rpc.return_value.c_type == "bool") {
        return_type = "void";
      } else {
        std::map<std::string, Struct>::iterator sit =
            structs.find(rpc.return_value.thrift_type);
        if (sit != structs.end()) {
          if (sit->second.fields[1].name == "return_value") {
            return_type = sit->second.fields[1].c_type;
          } else {
            Die("Could not find field 'return_value'.\n");
          }
        } else {
          Die("Could not find struct %s.\n",
              rpc.return_value.thrift_type.c_str());
        }
      }
    } else {
      return_type = rpc.return_value.c_type;
    }
    stream << return_type << " DefaultRemoteArbiter::" << rpc.name << "(";
    for (unsigned i = 0; i < rpc.arguments.size(); i++) {
      Argument argument = rpc.arguments[i];
      if (i > 0) { stream << ", "; }
      stream << "const " << argument.c_type << " ";
      if (!IsPrimitiveType(argument.c_type)) { stream << "&"; }
      stream << argument.name;
    }
    stream << ") {\n";
    if (rpc.schedule == "true") {
      if (return_type != "void") {
        stream << "  " << rpc.return_value.c_type << " option;\n"
               << "  try {\n"
               << "    boost::scoped_ptr<ArbiterConnection>"
               << " connection(new ArbiterConnection(hostname_, port_));\n"
               << "    connection->client()->" << rpc.name << "(option";
        for (unsigned i = 0; i < rpc.arguments.size(); i++) {
          Argument argument = rpc.arguments[i];
          stream << ", " << argument.name;
        }
        stream << ");\n"
               << "    MESSAGE(\"Thread %d called " << rpc.name
               << "().\", ThreadID());\n"
               << "  } catch (const apache::thrift::TException &tx) {\n"
               << "    ERROR(\"Connection with the arbiter failed: %s\", "
               << "tx.what());\n"
               << "  }\n"
               << "  if (!option.success) { DieQuietly(); }\n"
               << "  return option.return_value;\n";
      } else {
        stream << "  bool success;\n"
               << "  try {\n"
               << "    boost::scoped_ptr<ArbiterConnection>"
               << " connection(new ArbiterConnection(hostname_, port_));\n"
               << "    success = connection->client()->" << rpc.name << "(";
        for (unsigned i = 0; i < rpc.arguments.size(); i++) {
          Argument argument = rpc.arguments[i];
          if (i > 0) { stream << ", "; }
          stream << argument.name;
        }
        stream << ");\n"
               << "    MESSAGE(\"Thread %d called " << rpc.name
               << "().\", ThreadID());\n"
               << "  } catch (const apache::thrift::TException &tx) {\n"
               << "    ERROR(\"Connection with the arbiter failed: %s\", "
               << "tx.what());\n"
               << "  }\n"
               << "  if (!success) { DieQuietly(); }\n";
      }
    } else {
      if (return_type != "void") {
        std::map<std::string, Struct>::iterator sit =
            structs.find(rpc.return_value.thrift_type);
        stream << "  ";
        if (sit != structs.end()) {
          stream << return_type << " return_value;\n"
                 << "  try {\n"
                 << "    boost::scoped_ptr<ArbiterConnection>"
                 << " connection(new ArbiterConnection(hostname_, port_));\n"
                 << "    connection->client()->"
                 << rpc.name << "(return_value";
          for (unsigned i = 0; i < rpc.arguments.size(); i++) {
            Argument argument = rpc.arguments[i];
            stream << ", " << argument.name;
          }
          stream << ");\n"
                 << "    MESSAGE(\"Thread %d called " << rpc.name
                 << "().\", ThreadID());\n"
                 << "  } catch (const apache::thrift::TException &tx) {\n"
                 << "    ERROR(\"Connection with the arbiter failed: %s\", "
                 << "tx.what());\n"
                 << "  }\n"
                 << "  return return_value;\n";
        } else {
          stream << return_type << " return_value;\n"
                 << "  try {\n"
                 << "    boost::scoped_ptr<ArbiterConnection>"
                 << " connection(new ArbiterConnection(hostname_, port_));\n"
                 << "    return_value = connection->client()->"
                 << rpc.name << "(";
          for (unsigned i = 0; i < rpc.arguments.size(); i++) {
            Argument argument = rpc.arguments[i];
            if (i > 0) { stream << ", "; }
            stream << argument.name;
          }
          stream << ");\n"
                 << "    MESSAGE(\"Thread %d called " << rpc.name
                 << "().\", ThreadID());\n"
                 << "  } catch (const apache::thrift::TException &tx) {\n"
                 << "    ERROR(\"Connection with the arbiter failed: %s\", "
                 << "tx.what());\n"
                 << "  }\n"
                 << "  return return_value;\n";
        }
      } else {
        stream << "  try {\n"
               << "    boost::scoped_ptr<ArbiterConnection>"
               << " connection(new ArbiterConnection(hostname_, port_));\n"
               << "    connection->client()->" << rpc.name << "(";
        for (unsigned i = 0; i < rpc.arguments.size(); i++) {
          Argument argument = rpc.arguments[i];
          if (i > 0) { stream << ", "; }
          stream << argument.name;
        }
        stream << ");\n"
               << "    MESSAGE(\"Thread %d called " << rpc.name
               << "().\", ThreadID());\n"
               << "  } catch (const apache::thrift::TException &tx) {\n"
               << "    ERROR(\"Connection with the arbiter failed: %s\", "
               << "tx.what());\n"
               << "  }\n";
      }
    }
    stream << "}\n\n";
    fprintf(file, "%s", stream.str().c_str());
  }
  fprintf(file, "}  // namespace interposition\n");
  fprintf(file, "}  // namespace dbug\n");

  fclose(file);
  UpdateFile(name, tmp_name);
}

void GenerateDefaultHandlerHeader(std::map<std::string, Struct> structs,
                                  std::map<std::string, RPC> rpcs) {
  if (!Exists(SRCDIR "/include/dbug/arbiter/gen")) {
    CreateDirectory(SRCDIR "/include/dbug/arbiter/gen");
  }
  std::string name =
      SRCDIR "/include/dbug/arbiter/gen/default-handler.h";
  std::string tmp_name = name + ".tmp";
  FILE *file = fopen(tmp_name.c_str(), "w");
  if (file == NULL) { Die("Could not create file %s.\n", tmp_name.c_str()); }

  AddCopyright(file);
  fprintf(file, "// This is an automatically generated file. ");
  fprintf(file, "Any modifications will be lost.\n");
  fprintf(file, "\n");
  fprintf(file, "#include <set>\n");
  fprintf(file, "#include <string>\n");
  fprintf(file, "#include <vector>\n");
  fprintf(file, "\n");
  fprintf(file, "#include \"dbug/common/macros.h\"\n");
  fprintf(file, "#include \"src/thrift/arbiter/arbiter.h\"\n");
  fprintf(file, "\n");
  fprintf(file, "namespace dbug {\n");
  fprintf(file, "namespace arbiter {\n");
  fprintf(file, "\n");
  fprintf(file, "class Server;\n");
  fprintf(file, "class Thread;\n");
  fprintf(file, "\n");
  fprintf(file, "class DefaultHandler : public dbug::thrift::arbiterIf {\n");
  fprintf(file, " public:\n");
  fprintf(file,
          "  explicit DefaultHandler(Server *server) : server_(server) {}\n");
  fprintf(file, "\n");
  for (const std::pair<std::string, RPC> &pair : rpcs) {
    const RPC &rpc = pair.second;
    std::stringstream stream;
    stream << "  virtual ";
    std::string return_type;
    std::map<std::string, Struct>::iterator sit =
        structs.find(rpc.return_value.thrift_type);
    if (sit != structs.end()) {
      stream << "void " << rpc.name << "("
             << rpc.return_value.c_type << " &return_value, ";
    } else {
      stream << rpc.return_value.c_type << " " << rpc.name << "(";
    }
    for (unsigned i = 0; i < rpc.arguments.size(); i++) {
      Argument argument = rpc.arguments[i];
      if (i > 0) { stream << ", "; }
      stream << "const " << argument.c_type << " ";
      if (!IsPrimitiveType(argument.c_type)) { stream << "&"; }
      stream << argument.name;
    }
    stream << ")";
    if (rpc.schedule != "true") { stream << " = 0"; }
    stream << ";\n";
    fprintf(file, "%s", stream.str().c_str());
  }
  fprintf(file, "\n");
  fprintf(file, " protected:\n");
  fprintf(file, "  Thread *GetRequestThread(const int tid);\n");
  fprintf(file, "\n");
  fprintf(file, "  Server *server_;\n");
  fprintf(file, "\n");
  fprintf(file, " private:\n");
  fprintf(file, "  DISALLOW_COPY_AND_ASSIGN(DefaultHandler);\n");
  fprintf(file, "};\n");
  fprintf(file, "\n");
  fprintf(file, "}  // namespace arbiter\n");
  fprintf(file, "}  // namespace dbug\n");

  fclose(file);
  UpdateFile(name, tmp_name);
}

void GenerateDefaultHandlerSource(std::map<std::string, Struct> structs,
                                  std::map<std::string, RPC> rpcs) {
  if (!Exists(SRCDIR "/src/arbiter/gen")) {
    CreateDirectory(SRCDIR "/src/arbiter/gen");
  }
  std::string name =
      SRCDIR "/src/arbiter/gen/default-handler.cc";
  std::string tmp_name = name + ".tmp";
  FILE *file = fopen(tmp_name.c_str(), "w");
  if (file == NULL) { Die("Could not create file %s.\n", tmp_name.c_str()); }

  AddCopyright(file);
  fprintf(file, "// This is an automatically generated file. ");
  fprintf(file, "Any modifications will be lost.\n");
  fprintf(file, "\n");
  fprintf(file, "#include \"dbug/arbiter/default-handler.h\"\n");
  fprintf(file, "\n");
  fprintf(file, "#include <boost/scoped_ptr.hpp>\n");
  fprintf(file, "#include <set>\n");
  fprintf(file, "#include <string>\n");
  fprintf(file, "#include <vector>\n");
  fprintf(file, "\n");
  fprintf(file, "#include \"dbug/arbiter/request-prototypes.h\"\n");
  fprintf(file, "#include \"dbug/arbiter/server.h\"\n");
  fprintf(file, "#include \"dbug/arbiter/thread.h\"\n");
  fprintf(file, "#include \"src/thrift/arbiter/arbiter_types.h\"\n");
  fprintf(file, "\n");
  fprintf(file, "namespace dbug {\n");
  fprintf(file, "namespace arbiter {\n");
  fprintf(file, "\n");

  for (const std::pair<std::string, RPC> &pair : rpcs) {
    const RPC &rpc = pair.second;
    if (rpc.schedule != "true") { continue; }
    std::stringstream stream;
    std::map<std::string, Struct>::iterator sit =
        structs.find(rpc.return_value.thrift_type);
    if (sit != structs.end()) {
      stream << "void DefaultHandler::" << rpc.name << "("
             << rpc.return_value.c_type << " &return_value, ";
    } else {
      stream << rpc.return_value.c_type
             << " DefaultHandler::" << rpc.name << "(";
    }
    for (unsigned i = 0; i < rpc.arguments.size(); i++) {
      Argument argument = rpc.arguments[i];
      if (i > 0) { stream << ", "; }
      stream << "const " << argument.c_type << " ";
      if (!IsPrimitiveType(argument.c_type)) { stream << "&"; }
      stream << argument.name;
    }
    stream << ") {\n"
           << "  Server::Lock lock(server_);\n"
           << "  Thread *thread = GetRequestThread(tid);\n"
           << "  if (thread) {\n"
           << "    using dbug::arbiter::" << rpc.name << ";\n"
           << "    boost::scoped_ptr<" << rpc.name << "> request(new "
           << rpc.name << "(server_, thread";
    if (sit != structs.end()) { stream << ", &return_value"; }
    for (unsigned i = 1; i < rpc.arguments.size(); i++) {
      Argument argument = rpc.arguments[i];
      stream << ", " << argument.name;
    }
    stream << "));\n"
           << "    request->PreSchedule();\n"
           << "    request->Schedule();\n";
    if (sit != structs.end()) {
      stream << "    return_value.success = !request->canceled();\n"
             << "  } else {\n"
             << "    return_value.success = false;\n"
             << "  }\n";
    } else {
      stream << "    return !request->canceled();\n"
             << "  } else {\n"
             << "    return false;\n"
             << "  }\n";
    }
    stream << "}\n\n";
    fprintf(file, "%s", stream.str().c_str());
  }

  fprintf(file, "}  // namespace arbiter\n");
  fprintf(file, "}  // namespace dbug\n");

  fclose(file);
  UpdateFile(name, tmp_name);
}

void GenerateRequestPrototypesHeader(std::map<std::string, Struct> structs,
                                     std::map<std::string, RPC> rpcs) {
  if (!Exists(SRCDIR "/include/dbug/arbiter/gen")) {
    CreateDirectory(SRCDIR "/include/dbug/arbiter/gen");
  }
  std::string name =
      SRCDIR "/include/dbug/arbiter/gen/request-prototypes.h";
  std::string tmp_name = name + ".tmp";
  FILE *file = fopen(tmp_name.c_str(), "w");
  if (file == NULL) { Die("Could not create file %s.\n", tmp_name.c_str()); }

  AddCopyright(file);
  fprintf(file, "// This is an automatically generated file. ");
  fprintf(file, "Any modifications will be lost.\n");
  fprintf(file, "\n");
  fprintf(file, "#include \"dbug/arbiter/request.h\"\n");
  fprintf(file, "\n");
  fprintf(file, "#include <set>\n");
  fprintf(file, "#include <string>\n");
  fprintf(file, "#include <vector>\n");
  fprintf(file, "\n");
  fprintf(file, "#include \"dbug/arbiter/server.h\"\n");
  fprintf(file, "#include \"dbug/arbiter/thread.h\"\n");
  fprintf(file, "#include \"src/thrift/arbiter/arbiter_types.h\"\n");
  fprintf(file, "\n");
  fprintf(file, "namespace dbug {\n");
  fprintf(file, "namespace arbiter {\n");
  fprintf(file, "\n");

  for (const std::pair<std::string, RPC> &pair : rpcs) {
    const RPC &rpc = pair.second;
    if (rpc.schedule != "true") { continue; }
    std::stringstream stream;
    stream << "class " << rpc.name << " : public Request {\n"
           << " public:\n"
           << "  " << rpc.name << "(Server *server, Thread *thread";
    std::map<std::string, Struct>::iterator sit =
        structs.find(rpc.return_value.thrift_type);
    if (sit != structs.end()) {
      stream << ", " << rpc.return_value.c_type << " *return_value";
    }

    // The for loop below skips the <tid> argument.
    for (unsigned i = 1; i < rpc.arguments.size(); i++) {
      Argument argument = rpc.arguments[i];
      stream << ", const " << argument.c_type << " ";
      if (!IsPrimitiveType(argument.c_type)) { stream << "&"; }
      stream << argument.name;
    }
    stream << ") : Request(server, thread, func, backtrace)";
    if (sit != structs.end()) { stream << ", return_value_(return_value)"; }

    // The for loop below skips the <tid>, <func>, and <backtrace> arguments.
    for (unsigned i = 3; i < rpc.arguments.size(); i++) {
      Argument argument = rpc.arguments[i];
      stream << ", " << argument.name << "_(" << argument.name << ")";
    }
    stream << " {}\n\n" << "  // accessors\n";
    if (sit != structs.end()) {
      Struct s = sit->second;
      stream << "  " << s.fields[1].c_type << " return_value() const { "
             << "return return_value_->return_value; }\n";
    }

    // The for loop below skips the <tid>, <func>, and <backtrace> arguments.
    for (unsigned i = 3; i < rpc.arguments.size(); i++) {
      Argument argument = rpc.arguments[i];
      stream << "  " << argument.c_type << " " << argument.name << "() const { "
             << " return " << argument.name << "_; }\n";
    }
    stream << "\n" << "  // mutators\n";

    if (sit != structs.end()) {
      Struct s = sit->second;
      stream << "  void set_return_value(const " << s.fields[1].c_type << " ";
      if (!IsPrimitiveType(s.fields[1].c_type)) { stream << "&"; }
      stream << "return_value) { return_value_->return_value = return_value; "
             << "}\n";
    }
    stream << "\n";

    if (rpc.preschedule == "true") {
      stream << "  virtual void PreSchedule();\n";
    }
    stream << "  virtual void Test(Set requests, GroupSet *groups);\n"
           << "  virtual std::list<std::string> ToString() const;\n\n"
           << " private:\n"
           << "  virtual void Execute();\n\n";

    if (sit != structs.end()) {
      stream << "  " << rpc.return_value.c_type << " *return_value_;\n";
    }

    // The for loop below skips the <tid>, <func>, and <backtrace> arguments.
    for (unsigned i = 3; i < rpc.arguments.size(); i++) {
      Argument argument = rpc.arguments[i];
      stream << "  " << argument.c_type << " " << argument.name << "_;\n";
    }
    stream << "\n"
           << "  DISALLOW_COPY_AND_ASSIGN(" << rpc.name << ");\n"
           << "};\n\n";
    fprintf(file, "%s", stream.str().c_str());
  }

  fprintf(file, "}  // namespace arbiter\n");
  fprintf(file, "}  // namespace dbug\n");

  fclose(file);
  UpdateFile(name, tmp_name);
}

int main(int argc, char *argv[]) {
  if (argc < 2) { Die("Usage: arbiter <api-description> ...\n"); }

  std::map<std::string, Enum> enums;
  std::map<std::string, Struct> structs;
  std::map<std::string, RPC> rpcs;

  for (int i = 1; i < argc; i++) {
    xmlDoc *doc = xmlReadFile(argv[i], NULL, 0);
    if (doc == NULL) { Die("Could not parse the XML file %s.\n", argv[i]); }

    xmlNode *root_node = xmlDocGetRootElement(doc);
    if (root_node->type != XML_ELEMENT_NODE ||
        xmlStrcmp(root_node->name,
                  reinterpret_cast<const xmlChar *>("arbiter")) != 0) {
      Die("Error parsing the <arbiter> element.\n");
    }
    for (xmlNode *child = root_node->children; child; child = child->next) {
      if (child->type == XML_ELEMENT_NODE) {
        if (xmlStrcmp(child->name,
                      reinterpret_cast<const xmlChar *>("enum")) == 0) {
          Enum e = ParseEnumNode(child);
          enums[e.name] = e;
        }
        if (xmlStrcmp(child->name,
                      reinterpret_cast<const xmlChar *>("struct")) == 0) {
          Struct s = ParseStructNode(child);
          structs[s.name] = s;
        }
        if (xmlStrcmp(child->name,
                      reinterpret_cast<const xmlChar *>("rpc")) == 0) {
          RPC r = ParseRPCNode(child);
          rpcs[r.name] = r;
        }
      }
    }
    xmlFreeDoc(doc);
    xmlCleanupParser();
  }

  // Thrift
  GenerateArbiterThrift(enums, structs, rpcs);

  // Interposition
  GenerateArbiterInterfaceHeader(structs, rpcs);
  GenerateDefaultLocalArbiterHeader(structs, rpcs);
  GenerateDefaultLocalArbiterSource(structs, rpcs);
  GenerateDefaultRemoteArbiterHeader(structs, rpcs);
  GenerateDefaultRemoteArbiterSource(structs, rpcs);

  // Arbiter
  GenerateDefaultHandlerHeader(structs, rpcs);
  GenerateDefaultHandlerSource(structs, rpcs);
  GenerateRequestPrototypesHeader(structs, rpcs);

  return 0;
}
