/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

// TODO(jsimsa): Figure out if it is possible to annotate function
// pointer with the "no return" attribute to avoid the need for
// assert(0) in wrappers for functions that do not return.

#include <libxml/parser.h>
#include <libxml/tree.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>

#include <boost/algorithm/string.hpp>
#include <list>
#include <set>
#include <sstream>
#include <string>
#include <vector>

struct Argument {
  std::string type;
  std::string name;
  std::string format;
};

struct ReturnValue {
  std::string type;
  std::string format;
};

struct Function {
  std::string name;
  std::string generate;
  std::string noreturn;
  std::vector<std::string> headers;
  std::vector<Argument> arguments;
  ReturnValue return_value;
};

struct Group {
  std::string name;
  std::string interpose;
  std::vector<Function> functions;
};

const char *copyright =
    "// Authors: Jiri Simsa\n"
    "// Copyright (c) 2013, Carnegie Mellon University.\n"
    "// All rights reserved.\n"
    "//\n"
    "// Redistribution and use in source and binary forms, with or without\n"
    "// modification, are permitted provided that the following conditions\n"
    "// are met:\n"
    "// 1. Redistributions of source code must retain the above copyright\n"
    "//    notice, this list of conditions and the following disclaimer.\n"
    "// 2. Redistributions in binary form must reproduce the above copyright\n"
    "//    notice, this list of conditions and "
    "the following disclaimer in the\n"
    "//    documentation and/or other materials "
    "provided with the distribution.\n"
    "// 3. Neither the name of the University "
    "nor the names of its contributors\n"
    "//    may be used to endorse or promote "
    "products derived from this software\n"
    "//    without specific prior written permission.\n"
    "//\n"
    "// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS\n"
    "// ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT\n"
    "// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR\n"
    "// A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT\n"
    "// HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,\n"
    "// INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,\n"
    "// BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS\n"
    "// OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED\n"
    "// AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT\n"
    "// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY\n"
    "// WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE\n"
    "// POSSIBILITY OF SUCH DAMAGE.\n";

void Die(const char *format, ...) {
  va_list ap;
  va_start(ap, format);
  vfprintf(stderr, format, ap);
  va_end(ap);
  exit(-1);
}

bool Exists(const char *path) { return (access(path, F_OK) != -1); }

void CreateDirectory(const char *path) {
  int ret = mkdir(path, S_IRWXU | S_IRWXG | S_IROTH | S_IXOTH);
  if (ret == -1) { Die("Could not create directory %s.\n", path); }
}

std::string ConvertToHeaderGuard(std::string name) {
  for (unsigned i = 0; i < name.length(); i++) {
    if (name[i] >= 'a' && name[i] <= 'z') { name[i] += 'A' - 'a'; }
    if (name[i] < '0' || (name[i] > '9' && name[i] < 'A') || name[i] > 'Z') {
      name[i] = '_';
    }
  }
  return name;
}

void UpdateFile(std::string previous, std::string current) {
  // Overwrite the previous version only if it is different from the
  // current version to avoid recompilation of unchanged files.
  std::stringstream stream;
  stream << "diff -q " << previous.c_str() << " " << current.c_str();
  if (!Exists(previous.c_str()) || system(stream.str().c_str())) {
    int ret = rename(current.c_str(), previous.c_str());
    if (ret == -1) {
      Die("Could not move %s to %s.\n", current.c_str(), previous.c_str());
    } else {
      printf("Updated file %s\n", previous.c_str());
    }
  } else {
    printf("File %s is up-to-date.\n", previous.c_str());
    int ret = unlink(current.c_str());
    if (ret == -1) { Die("Could not remove %s.\n", current.c_str()); }
  }
}

Function ParseFunctionNode(xmlNode *node) {
  Function function;
  const xmlChar *name_xml = reinterpret_cast<const xmlChar *>("name");
  if (xmlHasProp(node, name_xml)) {
    xmlChar *prop = xmlGetProp(node, name_xml);
    function.name = reinterpret_cast<char *>(prop);
    xmlFree(prop);
  } else {
    Die("The name attribute of the <function> element is missing.\n");
  }
  const xmlChar *generate_xml = reinterpret_cast<const xmlChar *>("generate");
  if (xmlHasProp(node, generate_xml)) {
    xmlChar *prop = xmlGetProp(node, generate_xml);
    function.generate = reinterpret_cast<char *>(prop);
    xmlFree(prop);
    if (function.generate != "none" &&
        function.generate != "internal" &&
        function.generate != "all") {
      Die("Unsupported value %s for the generate attribute "
          "of the <function> element.\n", function.generate.c_str());
    }
  } else {
    function.generate = "internal";
  }
  const xmlChar *noreturn_xml = reinterpret_cast<const xmlChar *>("noreturn");
  if (xmlHasProp(node, noreturn_xml)) {
    xmlChar *prop = xmlGetProp(node, noreturn_xml);
    function.noreturn = reinterpret_cast<char *>(prop);
    xmlFree(prop);
    if (function.noreturn != "true" &&
        function.noreturn != "false") {
      Die("Unsupported value %s for the noreturn attribute "
          "of the <function> element.\n", function.noreturn.c_str());
    }
  } else {
    function.noreturn = "false";
  }
  for (xmlNode *child = node->children; child; child = child->next) {
    if (child->type == XML_ELEMENT_NODE) {
      if (xmlStrcmp(child->name,
                    reinterpret_cast<const xmlChar *>("header")) == 0) {
        const xmlChar *name_xml = reinterpret_cast<const xmlChar *>("name");
        if (xmlHasProp(child, name_xml)) {
          xmlChar *prop = xmlGetProp(child, name_xml);
          std::string name(reinterpret_cast<char *>(prop));
          xmlFree(prop);
          function.headers.push_back(name);
        } else {
          Die("The name attribute of the <header> element is missing.\n");
        }
      }
      if (xmlStrcmp(child->name,
                    reinterpret_cast<const xmlChar *>("argument")) == 0) {
        Argument argument;
        const xmlChar *type_xml = reinterpret_cast<const xmlChar *>("type");
        if (xmlHasProp(child, type_xml)) {
          xmlChar *prop = xmlGetProp(child, type_xml);
          argument.type = reinterpret_cast<char *>(prop);
          xmlFree(prop);
        } else {
          Die("The type attribute of the <argument> element is missing.\n");
        }
        const xmlChar *name_xml = reinterpret_cast<const xmlChar *>("name");
        if (xmlHasProp(child, name_xml)) {
          xmlChar *prop = xmlGetProp(child, name_xml);
          argument.name = reinterpret_cast<char *>(prop);
          xmlFree(prop);
        } else {
          Die("The name attribute of the <argument> element is missing.\n");
        }
        const xmlChar *format_xml = reinterpret_cast<const xmlChar *>("format");
        if (xmlHasProp(child, format_xml)) {
          xmlChar *prop = xmlGetProp(child, format_xml);
          argument.format = reinterpret_cast<char *>(prop);
          xmlFree(prop);
        } else {
          Die("The format attribute of the <argument> element is missing.\n");
        }
        function.arguments.push_back(argument);
      }
      if (xmlStrcmp(child->name,
                    reinterpret_cast<const xmlChar *>("return_value")) == 0) {
        ReturnValue return_value;
        const xmlChar *type_xml = reinterpret_cast<const xmlChar *>("type");
        if (xmlHasProp(child, type_xml)) {
          xmlChar *prop = xmlGetProp(child, type_xml);
          return_value.type = reinterpret_cast<char *>(prop);
          xmlFree(prop);
        } else {
          Die("The type attribute of the <return_value> element is missing.\n");
        }
        const xmlChar *format_xml = reinterpret_cast<const xmlChar *>("format");
        if (xmlHasProp(child, format_xml)) {
          xmlChar *prop = xmlGetProp(child, format_xml);
          return_value.format = reinterpret_cast<char *>(prop);
          xmlFree(prop);
        } else {
          Die("The format attribute of the "
              "<return_value> element is missing.\n");
        }
        function.return_value = return_value;
      }
    }
  }
  return function;
}

Group ParseGroupNode(xmlNode *node) {
  Group group;
  const xmlChar *name_xml = reinterpret_cast<const xmlChar *>("name");
  if (xmlHasProp(node, name_xml)) {
    xmlChar *prop = xmlGetProp(node, name_xml);
    group.name = reinterpret_cast<char *>(prop);
    xmlFree(prop);
  } else {
    Die("The name attribute of the <group> element is missing.\n");
  }
  const xmlChar *interpose_xml =
      reinterpret_cast<const xmlChar *>("interpose");
  if (xmlHasProp(node, interpose_xml)) {
    xmlChar *prop = xmlGetProp(node, interpose_xml);
    group.interpose = reinterpret_cast<char *>(prop);
    xmlFree(prop);
    if (group.interpose != "false" &&
        group.interpose != "true") {
      Die("Unsupported value %s for the interpose attribute "
          "of the <group> element.\n", group.interpose.c_str());
    }
  } else {
    group.interpose = "true";
  }
  for (xmlNode *child = node->children; child; child = child->next) {
    if (child->type == XML_ELEMENT_NODE) {
      if (xmlStrcmp(child->name,
                    reinterpret_cast<const xmlChar *>("function")) == 0) {
        group.functions.push_back(ParseFunctionNode(child));
      }
    }
  }
  return group;
}

void AddCopyright(FILE *file) { fprintf(file, "%s\n", copyright); }

void GenerateTypesHeader(std::list<Group> groups) {
  if (!Exists(SRCDIR "/include/dbug/interposition/gen")) {
    CreateDirectory(SRCDIR "/include/dbug/interposition/gen");
  }
  std::string name = SRCDIR "/include/dbug/interposition/gen/types.h";
  std::string tmp_name = name + ".tmp";
  FILE *file = fopen(tmp_name.c_str(), "w");
  if (file == NULL) { Die("Could not create file %s.\n", tmp_name.c_str()); }

  AddCopyright(file);
  fprintf(file, "// This is an automatically generated file. ");
  fprintf(file, "Any modifications will be lost.\n");
  fprintf(file, "\n");
  fprintf(file, "#if HAVE_CONFIG_H\n");
  fprintf(file, "#  include \"dbug-config.h\"\n");
  fprintf(file, "#endif\n");
  fprintf(file, "\n");
  fprintf(file, "#include \"dbug/common/macros.h\"\n");
  fprintf(file, "#include \"dbug/common/system.h\"\n");
  fprintf(file, "\n");

  // Include appropriate header files.
  std::set<std::string> headers;
  for (const Group &group : groups) {
    for (const Function &function : group.functions) {
      for (const std::string &header : function.headers) {
        if (headers.find(header) == headers.end()) {
          headers.insert(header);
        }
      }
    }
  }
  for (const std::string &header : headers) {
    std::stringstream stream;
    stream << "#if DBUG_HAVE_" << ConvertToHeaderGuard(header) << "\n"
           << "#  include <" << header << ">\n"
           << "#endif\n\n";
    fprintf(file, "%s", stream.str().c_str());
  }

  // Generate function and argument type and macros.
  for (const Group &group : groups) {
    for (const Function &function : group.functions) {
      std::stringstream stream;
      std::string upper_case_name(function.name);
      boost::to_upper(upper_case_name);
      stream << "#if DBUG_HAVE_" << upper_case_name << "\n"
             << "#define " << upper_case_name << "_ARGS_TYPE ";
      if (!function.arguments.empty()) {
        for (unsigned i = 0; i < function.arguments.size(); i++) {
          Argument argument = function.arguments[i];
          if (i > 0) { stream << ", "; }
          stream << argument.type;
        }
      } else {
        stream << "void";
      }
      stream << "\n"
             << "#define " << upper_case_name << "_TYPE "
             << function.return_value.type << " (*)("
             << upper_case_name << "_ARGS_TYPE)\n"
             << "#define INTERNAL_" << upper_case_name << "_ARGS_TYPE ";
      if (!function.arguments.empty()) {
        for (unsigned i = 0; i < function.arguments.size(); i++) {
          Argument argument = function.arguments[i];
          if (i > 0) { stream << ", "; }
          if (argument.type != "...") {
            stream << argument.type;
          } else {
            stream << "va_list";
          }
        }
      } else {
        stream << "void";
      }
      stream << "\n"
             << "#define INTERNAL_" << upper_case_name << "_TYPE "
             << function.return_value.type << " (*)(INTERNAL_"
             << upper_case_name << "_ARGS_TYPE)\n"
             << "#endif\n\n";
      fprintf(file, "%s", stream.str().c_str());
    }
  }

  fclose(file);
  UpdateFile(name, tmp_name);
}

void GenerateHandlerHeader(std::list<Group> groups) {
  if (!Exists(SRCDIR "/include/dbug/interposition/gen")) {
    CreateDirectory(SRCDIR "/include/dbug/interposition/gen");
  }
  std::string name = SRCDIR "/include/dbug/interposition/gen/handler.h";
  std::string tmp_name = name + ".tmp";
  FILE *file = fopen(tmp_name.c_str(), "w");
  if (file == NULL) { Die("Could not create file %s.\n", tmp_name.c_str()); }

  AddCopyright(file);
  fprintf(file, "// This is an automatically generated file. ");
  fprintf(file, "Any modifications will be lost.\n");
  fprintf(file, "\n");
  fprintf(file, "#if HAVE_CONFIG_H\n");
  fprintf(file, "#  include \"dbug-config.h\"\n");
  fprintf(file, "#endif\n");
  fprintf(file, "\n");
  fprintf(file, "#include \"dbug/common/macros.h\"\n");
  fprintf(file, "#include \"dbug/common/system.h\"\n");
  fprintf(file, "#include \"dbug/interposition/types.h\"\n");
  fprintf(file, "\n");

  // Include appropriate header files.
  std::set<std::string> headers;
  for (const Group &group : groups) {
    for (const Function &function : group.functions) {
      for (const std::string &header : function.headers) {
        if (headers.find(header) == headers.end()) {
          headers.insert(header);
        }
      }
    }
  }
  for (const std::string &header : headers) {
    std::stringstream stream;
    stream << "#if DBUG_HAVE_" << ConvertToHeaderGuard(header) << "\n"
           << "#  include <" << header << ">\n"
           << "#endif\n\n";
    fprintf(file, "%s", stream.str().c_str());
  }

  // Generate appropriate prototypes.
  for (const Group &group : groups) {
    for (const Function &function : group.functions) {
      std::stringstream stream;
      std::string upper_case_name(function.name);
      boost::to_upper(upper_case_name);
      stream << "#if DBUG_HAVE_" << upper_case_name << "\n";
      if (function.generate != "all") {
        stream << "  static " << function.return_value.type
               << " " << function.name << "(INTERNAL_"
               << upper_case_name << "_ARGS_TYPE)";
        if (function.noreturn == "true") { stream << " DBUG_NORETURN"; }
        stream << ";\n";
      }
      stream << "  static " << function.return_value.type
             << " " << function.name << "_wrapper(INTERNAL_"
             << upper_case_name << "_ARGS_TYPE)";
      if (function.noreturn == "true") { stream << " DBUG_NORETURN"; }
      stream << ";\n"
             << "#endif\n\n";
      fprintf(file, "%s", stream.str().c_str());
    }
  }

  fclose(file);
  UpdateFile(name, tmp_name);
}

void GenerateOriginalHeader(std::list<Group> groups) {
  if (!Exists(SRCDIR "/include/dbug/interposition/gen")) {
    CreateDirectory(SRCDIR "/include/dbug/interposition/gen");
  }
  std::string name = SRCDIR "/include/dbug/interposition/gen/original.h";
  std::string tmp_name = name + ".tmp";
  FILE *file = fopen(tmp_name.c_str(), "w");
  if (file == NULL) { Die("Could not create file %s.\n", tmp_name.c_str()); }

  AddCopyright(file);
  fprintf(file, "// This is an automatically generated file. ");
  fprintf(file, "Any modifications will be lost.\n");
  fprintf(file, "\n");
  fprintf(file, "#if HAVE_CONFIG_H\n");
  fprintf(file, "#  include \"dbug-config.h\"\n");
  fprintf(file, "#endif\n");
  fprintf(file, "\n");
  fprintf(file, "#include \"dbug/common/macros.h\"\n");
  fprintf(file, "#include \"dbug/common/system.h\"\n");
  fprintf(file, "#include \"dbug/interposition/types.h\"\n");
  fprintf(file, "\n");

  // Include appropriate header files.
  std::set<std::string> headers;
  for (const Group &group : groups) {
    for (const Function &function : group.functions) {
      for (const std::string &header : function.headers) {
        if (headers.find(header) == headers.end()) {
          headers.insert(header);
        }
      }
    }
  }
  for (const std::string &header : headers) {
    std::stringstream stream;
    stream << "#if DBUG_HAVE_" << ConvertToHeaderGuard(header) << "\n"
           << "#  include <" << header << ">\n"
           << "#endif\n\n";
    fprintf(file, "%s", stream.str().c_str());
  }

  // Declare pointers to system and library calls.
  fprintf(file, "namespace dbug {\n");
  fprintf(file, "namespace interposition {\n");
  fprintf(file, "\n");
  fprintf(file, "class Original {\n");
  fprintf(file, " public:\n");
  for (const Group &group : groups) {
    for (const Function &function : group.functions) {
      std::stringstream stream;
      std::string upper_case_name(function.name);
      boost::to_upper(upper_case_name);
      stream << "#if DBUG_HAVE_" << upper_case_name << "\n"
             << "  static " << function.return_value.type
             << " (*" << function.name << ")("
             << upper_case_name << "_ARGS_TYPE);\n"
             << "#endif\n\n";
      fprintf(file, "%s", stream.str().c_str());
    }
  }
  fprintf(file, "  static void Initialize();\n");
  fprintf(file, "\n");
  fprintf(file, " private:\n");
  fprintf(file, "  DISALLOW_ALL_DEFAULT(Original);\n");
  fprintf(file, "};\n");
  fprintf(file, "\n");
  fprintf(file, "}  // namespace interposition\n");
  fprintf(file, "}  // namespace dbug\n");

  fclose(file);
  UpdateFile(name, tmp_name);
}

void GenerateGroupSource(Group group) {
  if (!Exists(SRCDIR "/src/interposition/gen")) {
    CreateDirectory(SRCDIR "/src/interposition/gen");
  }
  std::stringstream stream;
  stream << SRCDIR << "/src/interposition/gen/" << group.name << ".cc";
  std::string name = stream.str();
  std::string tmp_name = name + ".tmp";
  FILE *file = fopen(tmp_name.c_str(), "w");
  if (file == NULL) { Die("Could not create file %s.\n", tmp_name.c_str()); }

  AddCopyright(file);
  fprintf(file, "// This is an automatically generated file. ");
  fprintf(file, "Any modifications will be lost.\n");
  fprintf(file, "\n");
  fprintf(file, "#include <dlfcn.h>\n");
  fprintf(file, "\n");
  fprintf(file, "#if HAVE_CONFIG_H\n");
  fprintf(file, "#  include \"dbug-config.h\"\n");
  fprintf(file, "#endif\n");
  fprintf(file, "\n");
  fprintf(file, "#include \"dbug/common/macros.h\"\n");
  fprintf(file, "#include \"dbug/common/system.h\"\n");
  fprintf(file, "#include \"dbug/interposition/handler.h\"\n");
  fprintf(file, "#include \"dbug/interposition/logging.h\"\n");
  fprintf(file, "#include \"dbug/interposition/original.h\"\n");
  fprintf(file, "#include \"dbug/interposition/types.h\"\n");
  fprintf(file, "\n");

  // Include appropriate header files.
  std::set<std::string> headers;
  for (const Function &function : group.functions) {
    for (const std::string &header : function.headers) {
      if (headers.find(header) == headers.end()) {
        headers.insert(header);
      }
    }
  }
  for (const std::string &header : headers) {
    std::stringstream stream;
    stream << "#if DBUG_HAVE_" << ConvertToHeaderGuard(header) << "\n"
           << "#  include <" << header << ">\n"
           << "#endif\n\n";
    fprintf(file, "%s", stream.str().c_str());
  }

  // Generate function wrappers.
  fprintf(file, "namespace dbug {\n");
  fprintf(file, "namespace interposition {\n");
  for (const Function &function : group.functions) {
    std::stringstream stream;
    std::string upper_case_name(function.name);
    boost::to_upper(upper_case_name);
    stream << "#if DBUG_HAVE_" << upper_case_name + "\n"
           << function.return_value.type << " "
           << "Handler::" << function.name << "_wrapper(";
    std::stringstream format_stream, name_stream;
    if (!function.arguments.empty()) {
      for (unsigned i = 0; i < function.arguments.size(); i++) {
        Argument argument = function.arguments[i];
        if (argument.type == "void") { continue; }
        if (i > 0) {
          format_stream << ", ";
          name_stream << ", ";
        }
        if (argument.type != "...") {
          format_stream << argument.format;
          name_stream << argument.name;
        } else {
          format_stream << "%p";
          name_stream << "ap";
        }
      }
      for (unsigned i = 0; i < function.arguments.size(); i++) {
        Argument argument = function.arguments[i];
        if (i > 0) { stream << ", "; }
        if (argument.type == "...") {
          stream << "va_list ap";
        } else {
          size_t position = argument.type.find("(*)");
          if (position != std::string::npos) {
            stream << argument.type.substr(0, position) << "(* "
                   << argument.name
                   << argument.type.substr(position + 2);
          } else {
            stream << argument.type << " " << argument.name;
          }
        }
      }
    }
    stream << ") {\n";
    if (function.generate == "none") {
      stream << "  ";
      if (function.return_value.type != "void") { stream << "return "; }
      stream << "Handler::" << function.name
             << "(" << name_stream.str() << ");\n";
    } else {
      stream << "  DBUG_REQUIRE_SYMBOL(Original::"
             << function.name << ", " << upper_case_name << "_TYPE, \""
             << function.name << "\");\n";
      if (function.return_value.type != "void") {
        stream << "  " << function.return_value.type << " ret;\n";
      }
      stream << "  if (Interpose()) {\n"
             << "    Interposition off(__FUNCTION__);\n"
             << "    ";
      if (function.return_value.type != "void") { stream << "ret = "; }
      if (function.generate == "all") { stream << "Original::"; }
      if (function.generate == "internal") { stream << "Handler::"; }
      stream << function.name << "(" << name_stream.str() << ");\n";
      if (function.return_value.type != "void") {
        stream << "    DEBUG_EXTERNAL(\"" << format_stream.str() << "\", "
               << "\"" << function.return_value.format << "\"";
        if (!function.arguments.empty()) {
          stream << ", " << name_stream.str();
        }
        stream << ", ret);\n"
               << "    return ret;\n";
      } else {
        stream << "    DEBUG_EXTERNAL(\"" << format_stream.str()
               << "\", \"void\"";
        if (!function.arguments.empty()) {
          stream << ", " << name_stream.str();
        }
        stream << ");\n"
               << "    return;\n";
      }
      stream << "  } else {\n"
             << "    ";
      if (function.return_value.type != "void") { stream << "ret = "; }
      stream << "Original::" << function.name << "("
             << name_stream.str() << ");\n";
      if (function.return_value.type != "void") {
        stream << "    DEBUG_INTERNAL(\"" << format_stream.str()
               << "\", \"" + function.return_value.format + "\"";
        if (!function.arguments.empty()) {
          stream << ", " << name_stream.str();
        }
        stream << ", ret);\n"
               << "    return ret;\n";
      } else {
        stream << "    DEBUG_INTERNAL(\"" << format_stream.str()
               << "\", \"void\"";
        if (!function.arguments.empty()) {
          stream << ", " << name_stream.str();
        }
        stream << ");\n"
               << "    return;\n";
      }
      stream << "  }\n";
    }
    stream << "}\n" << "#endif\n\n";
    fprintf(file, "%s", stream.str().c_str());
  }
  fprintf(file, "}  // namespace interposition\n");
  fprintf(file, "}  // namespace dbug\n");

  fclose(file);
  UpdateFile(name, tmp_name);
}

void GenerateInterposeSource(std::list<Group> groups) {
  if (!Exists(SRCDIR "/src/interposition/gen")) {
    CreateDirectory(SRCDIR "/src/interposition/gen");
  }
  std::string name = SRCDIR "/src/interposition/gen/interpose.cc";
  std::string tmp_name = name + ".tmp";
  FILE *file = fopen(tmp_name.c_str(), "w");
  if (file == NULL) { Die("Could not create file %s.\n", tmp_name.c_str()); }

  AddCopyright(file);
  fprintf(file, "// This is an automatically generated file. ");
  fprintf(file, "Any modifications will be lost.\n");
  fprintf(file, "\n");
  fprintf(file, "#if HAVE_CONFIG_H\n");
  fprintf(file, "#  include \"dbug-config.h\"\n");
  fprintf(file, "#endif\n");
  fprintf(file, "\n");
  fprintf(file, "#include \"dbug/common/macros.h\"\n");
  fprintf(file, "#include \"dbug/common/system.h\"\n");
  fprintf(file, "#include \"dbug/interposition/handler.h\"\n");
  fprintf(file, "\n");

  // Include appropriate header files.
  std::set<std::string> headers;
  for (const Group &group : groups) {
    for (const Function &function : group.functions) {
      for (const std::string &header : function.headers) {
        if (headers.find(header) == headers.end()) {
          headers.insert(header);
        }
      }
    }
  }
  for (const std::string &header : headers) {
    std::stringstream stream;
    stream << "#if DBUG_HAVE_" << ConvertToHeaderGuard(header) << "\n"
           << "#  include <" << header << ">\n"
           << "#endif\n\n";
    fprintf(file, "%s", stream.str().c_str());
  }

  // Define wrappers for interposed functions.
  for (const Group &group : groups) {
    if (group.interpose == "false") { continue; }
    for (const Function &function : group.functions) {
      std::stringstream stream;
      std::string upper_case_name(function.name);
      boost::to_upper(upper_case_name);
      stream << "#if DBUG_HAVE_" << upper_case_name + "\n"
             << "DBUG_PUBLIC " << function.return_value.type
             << " " << function.name << "(";
      std::vector<Argument> args = function.arguments;
      for (unsigned i = 0; i < args.size(); i++) {
        Argument argument = args[i];
        if (i > 0) { stream << ", "; }
        size_t position = argument.type.find("(*)");
        if (position != std::string::npos) {
          stream << argument.type.substr(0, position) << "(* "
                 << argument.name
                 << argument.type.substr(position + 2);
        } else {
          stream << argument.type << " " << argument.name;
        }
      }
      stream << ") {\n";
      if (!args.empty() && args.back().type == "...") {
        stream << "  va_list ap;\n"
               << "  if (dbug::interposition::Handler::Interpose()) {\n"
               << "    dbug::interposition::Handler::Disable(NULL);\n"
               << "    va_start(ap, " << args[args.size() - 2].name << ");\n"
               << "    dbug::interposition::Handler::Enable();\n"
               << "  } else {\n"
               << "    va_start(ap, " << args[args.size() - 2].name << ");\n"
               << "  }\n";
      }
      if (function.return_value.type != "void") {
        stream << "  " << function.return_value.type + " __ret;\n"
               << "  __ret = dbug::interposition::Handler::"
               << function.name + "_wrapper(";
      } else {
        stream << "  dbug::interposition::Handler::"
               << function.name + "_wrapper(";
      }
      for (unsigned i = 0; i < args.size(); i++) {
        Argument argument = args[i];
        if (i > 0) { stream << ", "; }
        if (argument.type != "...") {
          stream << argument.name;
        } else {
          stream << "ap";
        }
      }
      stream << ");\n";
      if (!args.empty() && args.back().type == "...") {
        stream << "  if (dbug::interposition::Handler::Interpose()) {\n"
               << "    dbug::interposition::Handler::Disable(NULL);\n"
               << "    va_end(ap);\n"
               << "    dbug::interposition::Handler::Enable();\n"
               << "  } else {\n"
               << "    va_end(ap);\n"
               << "  }\n";
      }
      if (function.return_value.type != "void") {
        stream << "  return __ret;\n";
      }
      stream << "}\n" << "#endif\n\n";
      fprintf(file, "%s", stream.str().c_str());
    }
  }

  fclose(file);
  UpdateFile(name, tmp_name);
}

void GenerateOriginalSource(std::list<Group> groups) {
  if (!Exists(SRCDIR "/src/interposition/gen")) {
    CreateDirectory(SRCDIR "/src/interposition/gen");
  }
  std::string name = SRCDIR "/src/interposition/gen/original.cc";
  std::string tmp_name = name + ".tmp";
  FILE *file = fopen(tmp_name.c_str(), "w");
  if (file == NULL) { Die("Could not create file %s.\n", tmp_name.c_str()); }

  AddCopyright(file);
  fprintf(file, "// This is an automatically generated file. ");
  fprintf(file, "Any modifications will be lost.\n");
  fprintf(file, "\n");
  fprintf(file, "#include <dlfcn.h>\n");
  fprintf(file, "\n");
  fprintf(file, "#if HAVE_CONFIG_H\n");
  fprintf(file, "#  include \"dbug-config.h\"\n");
  fprintf(file, "#endif\n");
  fprintf(file, "\n");
  fprintf(file, "#include \"dbug/common/macros.h\"\n");
  fprintf(file, "#include \"dbug/common/system.h\"\n");
  fprintf(file, "#include \"dbug/interposition/logging.h\"\n");
  fprintf(file, "#include \"dbug/interposition/original.h\"\n");
  fprintf(file, "\n");

  // Include appropriate header files.
  std::set<std::string> headers;
  for (const Group &group : groups) {
    for (const Function &function : group.functions) {
      for (const std::string &header : function.headers) {
        if (headers.find(header) == headers.end()) {
          headers.insert(header);
        }
      }
    }
  }
  for (const std::string &header : headers) {
    std::stringstream stream;
    stream << "#if DBUG_HAVE_" << ConvertToHeaderGuard(header) << "\n"
           << "#  include <" << header << ">\n"
           << "#endif\n\n";
    fprintf(file, "%s", stream.str().c_str());
  }

  // Define pointers to system and library calls.
  fprintf(file, "namespace dbug {\n");
  fprintf(file, "namespace interposition {\n");
  fprintf(file, "\n");
  for (const Group &group : groups) {
    for (const Function &function : group.functions) {
      std::stringstream stream;
      std::string upper_case_name(function.name);
      boost::to_upper(upper_case_name);
      stream << "#if DBUG_HAVE_" << upper_case_name + "\n"
             << function.return_value.type << " "
             << "(* Original::" << function.name << ")("
             << upper_case_name << "_ARGS_TYPE) = NULL;\n"
             << "#endif\n\n";
      fprintf(file, "%s", stream.str().c_str());
    }
  }
  fprintf(file, "void Original::Initialize() {\n");
  for (const Group &group : groups) {
    for (const Function &function : group.functions) {
      std::stringstream stream;
      std::string upper_case_name(function.name);
      boost::to_upper(upper_case_name);
      stream << "#if DBUG_HAVE_" << upper_case_name + "\n"
             << "  DBUG_REQUIRE_SYMBOL(" << function.name
             << ", " << upper_case_name << "_TYPE, "
             << "\"" << function.name << "\");\n" << "#endif\n\n";
      fprintf(file, "%s", stream.str().c_str());
    }
  }
  fprintf(file, "}\n\n");
  fprintf(file, "}  // namespace interposition\n");
  fprintf(file, "}  // namespace dbug\n");

  fclose(file);
  UpdateFile(name, tmp_name);
}

int main(int argc, char *argv[]) {
  if (argc < 2) { Die("Usage: interposition <api-description>\n"); }

  std::list<Group> groups;

  for (int i = 1; i < argc; i++) {
    xmlDoc *doc = xmlReadFile(argv[i], NULL, 0);
    if (doc == NULL) { Die("Could not parse the XML file %s.\n", argv[i]); }

    xmlNode *root_node = xmlDocGetRootElement(doc);
    if (root_node->type != XML_ELEMENT_NODE ||
        xmlStrcmp(root_node->name,
                  reinterpret_cast<const xmlChar *>("interposition")) != 0) {
      Die("Error parsing the <interposition> element.\n");
    }
    for (xmlNode *child = root_node->children; child; child = child->next) {
      if (child->type == XML_ELEMENT_NODE) {
        if (xmlStrcmp(child->name,
                      reinterpret_cast<const xmlChar *>("group")) == 0) {
          groups.push_back(ParseGroupNode(child));
        }
      }
    }
    xmlFreeDoc(doc);
    xmlCleanupParser();
  }

  // Generate headers files.
  GenerateTypesHeader(groups);
  GenerateHandlerHeader(groups);
  GenerateOriginalHeader(groups);

  // Generate source code files.
  for (const Group &group : groups) { GenerateGroupSource(group); }
  GenerateInterposeSource(groups);
  GenerateOriginalSource(groups);
  return 0;
}
