/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "dbug/common/big-float.h"

#include <gmp.h>
#include <mpfr.h>

namespace dbug {

void BigFloat::Initialize() {}

void BigFloat::Finalize() { mpfr_free_cache(); }

BigFloat::BigFloat() {
  mpfr_init(value_);
}

BigFloat::BigFloat(const BigFloat &bf) {
  mpfr_init(value_);
  mpfr_set(value_, bf.value_, MPFR_RNDN);
}

BigFloat::BigFloat(double d) {
  mpfr_init(value_);
  mpfr_set_d(value_, d, MPFR_RNDN);
}

BigFloat::BigFloat(int i) {
  mpfr_init(value_);
  mpfr_set_si(value_, i, MPFR_RNDN);
}

BigFloat::~BigFloat() {
  mpfr_clear(value_);
}

BigFloat &BigFloat::operator=(const BigFloat &rhs) {
  mpfr_set(value_, rhs.value_, MPFR_RNDN);
  return *this;
}

BigFloat BigFloat::operator+(const BigFloat &rhs) {
  BigFloat result;
  mpfr_add(result.value_, value_, rhs.value_, MPFR_RNDN);
  return result;
}

BigFloat BigFloat::operator-(const BigFloat &rhs) {
  BigFloat result;
  mpfr_sub(result.value_, value_, rhs.value_, MPFR_RNDN);
  return result;
}

BigFloat BigFloat::operator*(const BigFloat &rhs) {
  BigFloat result;
  mpfr_mul(result.value_, value_, rhs.value_, MPFR_RNDN);
  return result;
}

BigFloat BigFloat::operator/(const BigFloat &rhs) {
  BigFloat result;
  mpfr_div(result.value_, value_, rhs.value_, MPFR_RNDN);
  return result;
}

bool BigFloat::operator<(const BigFloat &rhs) {
  return (mpfr_cmp(value_, rhs.value_) < 0);
}

bool BigFloat::operator<=(const BigFloat &rhs) {
  return (mpfr_cmp(value_, rhs.value_) <= 0);
}

bool BigFloat::operator==(const BigFloat &rhs) {
  return (mpfr_cmp(value_, rhs.value_) == 0);
}

bool BigFloat::operator!=(const BigFloat &rhs) {
  return (mpfr_cmp(value_, rhs.value_) != 0);
}

bool BigFloat::operator>(const BigFloat &rhs) {
  return (mpfr_cmp(value_, rhs.value_) > 0);
}

bool BigFloat::operator>=(const BigFloat &rhs) {
  return (mpfr_cmp(value_, rhs.value_) >= 0);
}

BigFloat BigFloat::sqrt(const BigFloat &bf) {
  BigFloat result;
  mpfr_sqrt(result.value_, bf.value_, MPFR_RNDN);
  return result;
}

std::string BigFloat::ToString() const {
  char buffer[256];
  mpfr_sprintf(buffer, "%.6RG", value_);
  std::string result(buffer);
  return result;
}

}  // namespace dbug
