/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "dbug/explorer/edge.h"

#include <list>
#include <sstream>
#include <string>

#include "dbug/common/system.h"
#include "dbug/common/util.h"
#include "dbug/explorer/logging.h"
#include "dbug/explorer/node.h"
#include "src/thrift/arbiter/arbiter_types.h"

using dbug::thrift::Location;
using dbug::thrift::Thread;

namespace dbug {
namespace explorer {

Edge::Edge(Node *from,
           Node *to,
           std::vector<Thread> threads,
           bool enabled,
           std::set<std::string> read_set,
           std::set<std::string> write_set)
    : enabled_(enabled),
      from_(from),
      read_set_(read_set),
      to_(to),
      threads_(threads),
      write_set_(write_set) {
  for (unsigned i = 0; i < threads_.size(); i++) {
    Thread thread = threads_[i];
    clock_.JoinWith(thread.clock);
  }
}

// accessors

Clock Edge::clock() const { return clock_; }

bool Edge::enabled() const { return enabled_; }

Node *Edge::from() const { return from_; }

std::set<std::string> Edge::read_set() const { return read_set_; }

std::vector<Thread> Edge::threads() const { return threads_; }

Node *Edge::to() const { return to_; }

std::set<std::string> Edge::write_set() const { return write_set_; }

// mutators

void Edge::set_read_set(const std::set<std::string> &set) { read_set_ = set; }

void Edge::set_write_set(const std::set<std::string> &set) { write_set_ = set; }

// public methods

void Edge::Print(Logging::Verbosity verbosity) const {
  const Logging::Config *config = Logging::get_channel(Logging::EXPLORER);
  if (config->verbosity() < verbosity) { return; }
  for (const std::string &line : ToString()) {
    LOG(verbosity, "%s", line.c_str());
  }
}

std::list<std::string> Edge::ToString() const {
  std::list<std::string> result;
  std::stringstream stream;
  stream << "  Edge:";
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "   from: " << from_->id();
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "   to: " << to_->id();
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "   enabled: " << (enabled_ ? "yes" : "no");
  result.push_back(stream.str());
  stream.str(std::string());
  for (unsigned i = 0; i < threads_.size(); i++) {
    stream << "   thread: ";
    result.push_back(stream.str());
    stream.str(std::string());
    stream << "    id: " << threads_[i].id;
    result.push_back(stream.str());
    stream.str(std::string());
    stream << "    function: " << threads_[i].fname;
    result.push_back(stream.str());
    stream.str(std::string());
    std::vector<Location> backtrace = threads_[i].backtrace;
    for (unsigned j = 0; j < backtrace.size(); j++) {
      stream << "    backtrace: "
             << backtrace[j].func_name
             << "() at " << System::Basename(backtrace[j].file_name.c_str())
             << ":" << backtrace[j].line;
      result.push_back(stream.str());
      stream.str(std::string());
    }
    stream << "    clock: " << Util::VectorToString<int>(threads_[i].clock);
    result.push_back(stream.str());
    stream.str(std::string());
  }
  stream << "   read set: " << Util::SetToString<std::string>(read_set_);
  result.push_back(stream.str());
  stream.str(std::string());
  stream << "   write set: " << Util::SetToString<std::string>(write_set_);
  result.push_back(stream.str());
  stream.str(std::string());
  return result;
}

}  // namespace explorer
}  // namespace dbug
