/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "dbug/explorer/handler.h"

#include <thrift/protocol/TBinaryProtocol.h>
#include <thrift/server/TThreadedServer.h>
#include <thrift/transport/TServerSocket.h>
#include <thrift/transport/TTransportUtils.h>

#include <boost/scoped_ptr.hpp>
#include <boost/shared_ptr.hpp>
#include <stack>
#include <vector>

#include "dbug/common/util.h"
#include "dbug/explorer/edge.h"
#include "dbug/explorer/logging.h"
#include "dbug/explorer/node.h"
#include "dbug/explorer/server.h"
#include "src/thrift/explorer/explorer.h"

namespace dbug {
namespace explorer {

void *handler_thread(void *args) {
  Server *explorer = reinterpret_cast<Server *>(args);
  boost::shared_ptr<Handler> handler(new Handler(explorer));
  boost::shared_ptr<apache::thrift::TProcessor>
      processor(new dbug::thrift::explorerProcessor(handler));
  boost::shared_ptr<apache::thrift::transport::TServerTransport>
      transport(new apache::thrift::transport::TServerSocket(explorer->port()));
  boost::shared_ptr<apache::thrift::transport::TTransportFactory>
      t_factory(new apache::thrift::transport::TBufferedTransportFactory());
  boost::shared_ptr<apache::thrift::protocol::TProtocolFactory>
      p_factory(new apache::thrift::protocol::TBinaryProtocolFactory());
  boost::scoped_ptr<apache::thrift::server::TThreadedServer>
      server(new apache::thrift::server::TThreadedServer(
          processor, transport, t_factory, p_factory));
  boost::shared_ptr<apache::thrift::server::TServerEventHandler>
      eventHandler(new Server::EventHandler(explorer));
  explorer->set_thrift_server(server.get());
  server->setServerEventHandler(eventHandler);
  server->serve();
  return NULL;
}

Handler::Handler(Server *server) : server_(server) {}

Handler::~Handler() {}

// public methods

void Handler::NewProcessNotification() {
  Server::Lock lock(server_);
  server_->increment_process_started();
}

void Handler::RegisterArbiter(std::vector<dbug::thrift::Move> &schedule) { // NOLINT
  Server::Lock lock(server_);
  server_->set_arbiter_started(true);
  // Generate an exploration strategy.
  schedule = server_->frontier()->GenerateSchedule();
  // Add the prefix to it.
  std::vector<dbug::thrift::Move> prefix = server_->prefix();
  schedule.insert(schedule.begin(), prefix.begin(), prefix.end());
}

void Handler::UnregisterArbiter(const std::vector<dbug::thrift::Node> &branch) {
  Server::Lock lock(server_);
  server_->set_arbiter_finished(true);
  std::vector<dbug::thrift::Move> prefix = server_->prefix();
  unsigned prefix_size = prefix.size();
  Node *current_node = server_->execution_tree();
  std::map<int, Node *> back_pointers;
  // Map the explored branch into the execution tree.
  for (unsigned i = 0; i < branch.size(); i++) {
    if (i < prefix_size) {
      ASSERT(prefix[i].choice_id == branch[i].choice_id,
             "Arbiter did not follow the strategy.");
    } else {
      dbug::thrift::Node node = branch[i];
      // Update <back_pointers>.
      for (unsigned j = 0; j < node.choices.size(); j++) {
        dbug::thrift::Choice choice = node.choices[j];
        if (choice.id == node.choice_id) {
          for (unsigned k = 0; k < choice.threads.size(); k++) {
            dbug::thrift::Thread thread = choice.threads[k];
            back_pointers[thread.id] = current_node;
          }
          break;
        }
      }
      PrintThriftNode(Logging::DEBUG, node);
      if (current_node->visited()) {
        if (current_node->nondet_events() != node.nondet_events) {
          PrintComparison(Logging::WARNING, current_node, node);
          ASSERT(current_node->nondet_events() == node.nondet_events,
                 "Divergence encountered: different non-deterministic events.");
        }
        for (unsigned j = 0; j < node.choices.size(); j++) {
          dbug::thrift::Choice choice = node.choices[j];
          Edge *edge = current_node->GetEdge(choice.id);
          if (!edge) {
            PrintComparison(Logging::WARNING, current_node, node);
            ASSERT(edge, "Divergence encountered: a missing edge.");
          }
          if (edge->threads() != choice.threads) {
            PrintComparison(Logging::WARNING, current_node, node);
            ASSERT(edge->threads() == choice.threads,
                   "Divergence encountered: different edge threads.");
          }
          if (edge->enabled() != choice.enabled) {
            PrintComparison(Logging::WARNING, current_node, node);
            ASSERT(edge->enabled() == choice.enabled,
                   "Divergence encountered: different edge status.");
          }
          // TODO(jsimsa): The line below is a temporary hack for the
          // SOSP experiments.
          edge->set_read_set(choice.read_set);
          if (edge->read_set() != choice.read_set) {
            PrintComparison(Logging::WARNING, current_node, node);
            ASSERT(edge->read_set() == choice.read_set,
                   "Divergence encountered: different read set.");
          }
          // TODO(jsimsa): The line below is a temporary hack for the
          // SOSP experiments.
          edge->set_write_set(choice.write_set);
          if (edge->write_set() != choice.write_set) {
            PrintComparison(Logging::WARNING, current_node, node);
            ASSERT(edge->write_set() == choice.write_set,
                   "Divergence encountered: different write set.");
          }
        }
      } else {
        current_node->set_visited(true);
        current_node->set_nondet_events(node.nondet_events);
        for (unsigned j = 0; j < node.choices.size(); j++) {
          dbug::thrift::Choice choice = node.choices[j];
          Node *new_node = new Node(choice.id, current_node);
          Edge *new_edge =
              new Edge(current_node, new_node, choice.threads,
                       choice.enabled, choice.read_set, choice.write_set);
          current_node->add_edge(choice.id, new_edge);
        }
        // Account for non-deterministic choices.
        Edge *edge1 = current_node->GetEdge(node.choice_id);
        if (edge1) {
          ASSERT(edge1->enabled(), "Edge is not enabled.");
          std::set<int> ids1;
          for (const Thread &thread : edge1->threads()) {
            ids1.insert(thread.id);
          }
          for (std::pair<int, const Edge *> pair : current_node->edges()) {
            int choice_id = pair.first;
            const Edge *edge2 = pair.second;
            if (choice_id != node.choice_id) {
              std::set<int> ids2;
              for (const Thread &thread : edge2->threads()) {
                ids2.insert(thread.id);
              }
              if (ids1 == ids2) {
                ASSERT(edge2->enabled(), "Edge is not enabled.");
                Node *node2 = edge2->to();
                server_->frontier()->add_node(node2);
                node2->set_seeded(true);
              }
            }
          }
        }
      }
      if (node.choice_id == -1) {
        if (current_node->edges().size() > 0) {
          server_->increment_num_errors();
        }
        break;
      }
      current_node = current_node->FollowEdge(node.choice_id);
      if (!current_node->visited()) {
        current_node->set_back_pointers(back_pointers);
      }
      ASSERT(current_node, "Edge of thread %d is missing.", node.choice_id);
    }
  }
  current_node->set_visited(true);

  UpdateExplorationSeeds(current_node);
  UpdateSizeEstimate(current_node);
  UpdateExplorationStatus(current_node);
}

// private methods

bool Handler::AreConcurrent(const Edge *edge1, const Edge *edge2) const {
  return (edge1->clock().CompareWith(edge2->clock()) == Clock::CONCURRENT);
}

bool Handler::AreDependent(const Edge *edge1, const Edge *edge2) const {
  std::set<std::string> rw_set, wr_set, ww_set;
  std::set<std::string> read_set1 = edge1->read_set();
  std::set<std::string> read_set2 = edge2->read_set();
  std::set<std::string> write_set1 = edge1->write_set();
  std::set<std::string> write_set2 = edge2->write_set();
  std::set_intersection(read_set1.begin(), read_set1.end(),
                        write_set2.begin(), write_set2.end(),
                        std::inserter(rw_set, rw_set.end()));
  std::set_intersection(read_set2.begin(), read_set2.end(),
                        write_set1.begin(), write_set1.end(),
                        std::inserter(wr_set, wr_set.end()));
  std::set_intersection(write_set1.begin(), write_set1.end(),
                        write_set2.begin(), write_set2.end(),
                        std::inserter(ww_set, ww_set.end()));
  return (!rw_set.empty() || !wr_set.empty() || !ww_set.empty());
}

void Handler::ExploreEdgeFromNode(const Edge *edge, Node *node) {
  // Identify all edges originating from <node> that happened before
  // or are equal to <edge>.
  std::list<const Edge *> edges;
  for (std::pair<int, const Edge *> pair : node->edges()) {
    const Edge *edge2 = pair.second;
    if (edge2->enabled()) {
      Clock::Order order = edge2->clock().CompareWith(edge->clock());
      if (order == Clock::BEFORE || order == Clock::EQUAL) {
        edges.push_back(edge2);
      }
    }
  }
  DEBUG("ADDING");
  if (edges.empty()) {
    // TODO(jsimsa): When DPOR fails to identify an edge originating
    // from <node> that happened before or is equal to <edge>, all
    // edges originating from <node> are enabled. This decreases the
    // efficiency of DPOR and should be avoided whenever possible.
    DEBUG("No enabled edges leading to the target edge found.");
    DEBUG("Target edge:");
    edge->Print(Logging::DEBUG);
    DEBUG("Edges:");
    int nseeded = 0;
    int nenabled = 0;
    for (std::pair<int, const Edge *> pair : node->edges()) {
      const Edge *edge2 = pair.second;
      Node *node2 = edge2->to();
      edge2->Print(Logging::DEBUG);
      if (edge2->enabled()) {
        nenabled++;
        if (!node2->seeded() && !node2->visited()) {
          server_->frontier()->add_node(node2);
          node2->set_seeded(true);
          nseeded++;
        }
      }
    }
    ASSERT(nenabled > nseeded, "Too many enabled edges.");
    if (nseeded > 0) {
      BigFloat delta(static_cast<double>(nenabled - nseeded) /
                     static_cast<double>(nenabled));
      UpdateBranchProbability(node, delta);
    }
  } else {
    // Make sure that at least one of the identified edges has been or
    // will be explored.
    for (const Edge *edge2 : edges) {
      Node *node2 = edge2->to();
      if (node2->seeded() || node2->visited()) { return; }
    }
    int nenabled = 0;
    for (std::pair<int, const Edge *> pair : node->edges()) {
      const Edge *edge2 = pair.second;
      const Node *child = edge2->to();
      if (edge2->enabled() && (child->seeded() || child->visited())) {
        nenabled++;
      }
    }
    const Edge *edge2 = edges.front();
    Node *node2 = edge2->to();
    edge2->Print(Logging::DEBUG);
    ASSERT(edge2->enabled(), "Edge is not enabled.");
    ASSERT(!node2->seeded(), "Node is already seeded.");
    ASSERT(!node2->visited(), "Node has already been visited.");
    server_->frontier()->add_node(node2);
    node2->set_seeded(true);
    ASSERT(nenabled > 0, "No enabled edges.");
    BigFloat delta(static_cast<double>(nenabled) /
                   static_cast<double>(nenabled + 1));
    UpdateBranchProbability(node, delta);
  }
}

bool Handler::PossiblyCoenabled(const Edge *edge1, const Edge *edge2) const {
  std::set<std::string> funcs1, funcs2;
  for (Thread thread : edge1->threads()) { funcs1.insert(thread.fname); }
  for (Thread thread : edge2->threads()) { funcs2.insert(thread.fname); }
  std::set<std::string> unlock_set;
  unlock_set.insert("pthread_mutex_unlock");
  std::set<std::string> lock_set;
  lock_set.insert("pthread_mutex_lock");
  if ((funcs1 == unlock_set && funcs2 == lock_set) ||
      (funcs2 == unlock_set && funcs1 == lock_set)) {
    return edge1->write_set() != edge2->write_set();
  }
  std::set<std::string> exit_set;
  exit_set.insert("exit");
  std::set<std::string> signal_set;
  signal_set.insert("signal");
  std::set<std::string> waitpid_set;
  waitpid_set.insert("waitpid");
  if ((funcs1 == waitpid_set && (funcs2 == exit_set || funcs2 == signal_set)) ||
      (funcs2 == waitpid_set && (funcs1 == exit_set || funcs1 == signal_set))) {
    std::set<std::string> write_set1 = edge1->write_set();
    std::set<std::string> write_set2 = edge2->write_set();
    std::set<std::string> intersection;
    std::set_intersection(write_set1.begin(), write_set1.end(),
                          write_set2.begin(), write_set2.end(),
                          std::inserter(intersection, intersection.end()));
    return intersection.empty();
  }
  return true;
}

void Handler::PrintComparison(Logging::Verbosity verbosity,
                              const Node *node,
                              const dbug::thrift::Node &thrift_node) const {
  const Logging::Config *config = Logging::get_channel(Logging::EXPLORER);
  if (config->verbosity() < verbosity) { return; }
  if (node->parent()) {
    LOG(verbosity, "--PARENT--");
    node->parent()->Print(verbosity);
  }
  LOG(verbosity, "--PREVIOUS--");
  node->Print(verbosity);
  LOG(verbosity, "--CURRENT--");
  PrintThriftNode(verbosity, thrift_node);
}

void Handler::PrintThriftNode(Logging::Verbosity verbosity,
                              const dbug::thrift::Node &node) const {
  const Logging::Config *config = Logging::get_channel(Logging::EXPLORER);
  if (config->verbosity() < verbosity) { return; }
  for (unsigned j = 0; j < node.choices.size(); j++) {
    LOG(verbosity, " Choice:");
    dbug::thrift::Choice choice = node.choices[j];
    LOG(verbosity, "  ID: %d", choice.id);
    LOG(verbosity, "  Enabled: %s", choice.enabled ? "yes" : "no");
    for (unsigned k = 0; k < choice.threads.size(); k++) {
      dbug::thrift::Thread thread = choice.threads[k];
      LOG(verbosity, "    id: %d", thread.id);
      LOG(verbosity, "    function: %s", thread.fname.c_str());
      for (unsigned l = 0; l < thread.backtrace.size(); l++) {
        dbug::thrift::Location location = thread.backtrace[l];
        LOG(verbosity,
            "    backtrace: %s() at %s:%d",
            location.func_name.c_str(),
            System::Basename(location.file_name.c_str()),
            location.line);
      }
      LOG(verbosity,
          "    clock: %s",
          Util::VectorToString<int>(thread.clock).c_str());
    }
    LOG(verbosity,
        "  Read set: %s",
        Util::SetToString<std::string>(choice.read_set).c_str());
    LOG(verbosity,
        "  Write set: %s",
        Util::SetToString<std::string>(choice.write_set).c_str());
  }
  LOG(verbosity, " Next: %d", node.choice_id);
  LOG(verbosity,
      " Events: %s",
      Util::VectorToString<int>(node.nondet_events).c_str());
}

void Handler::UpdateBranchProbability(Node *node, const BigFloat &delta) {
  std::stack<Node *> stack;
  stack.push(node);
  while (!stack.empty()) {
    Node *node = stack.top();
    stack.pop();
    for (std::pair<int, const Edge *> pair : node->edges()) {
      const Edge *edge = pair.second;
      Node *child = edge->to();
      if (child->probability() != BigFloat(0)) {
        child->set_probability(child->probability() * delta);
        child->set_estimate(child->estimate() / delta);
        stack.push(child);
      }
    }
  }
}

void Handler::UpdateExplorationSeeds(Node *leaf_node) {
  Node *current_node = leaf_node;
  for (; current_node; current_node = current_node->parent()) {
    if (current_node->seeds_computed()) { break; }
    if (server_->use_dpor()) {
      Node *parent = current_node->parent();
      if (!parent) { break; }
      Edge *parent_edge = parent->GetEdge(current_node->id());
      for (std::pair<int, Edge *> pair : current_node->edges()) {
        Edge *edge1 = pair.second;
        // If <edge1> and <parent_edge> are independent and
        // concurrent, then <edge1> is enabled at <parent> and will be
        // checked against predecessors of <edge1> in subsequent
        // iterations of the main for loop.
        if (!AreDependent(edge1, parent_edge) &&
            AreConcurrent(edge1, parent_edge)) { continue; }
        // Otherwise, check if there exists an edge <edge2> on the
        // branch leading to <current_node> such that <edge1> and
        // <edge2> are dependent, concurrent, and possibly
        // co-enabled. As an optimization, we create projections of
        // the branch for each thread and scan each projection
        // individually.
        DEBUG("CHECKING:");
        edge1->Print(Logging::DEBUG);
        DEBUG("AGAINST:");
        for (std::pair<int, Node *> pair : current_node->back_pointers()) {
          int tid = pair.first;
          Node *node = pair.second;
          while (1) {
            Edge *edge2 = node->GetThreadEdge(tid);
            ASSERT(edge2, "Edge of thread %d does not exist.", tid);
            edge2->Print(Logging::DEBUG);
            // A check of a projection can terminate early as soon as
            // an edge that happened before <edge1> is encountered as
            // the edges of a projection are in a happens-before
            // relationship.
            if (edge1->clock().CompareWith(edge2->clock()) == Clock::AFTER) {
              break;
            }
            if (AreDependent(edge1, edge2) &&
                AreConcurrent(edge1, edge2) &&
                PossiblyCoenabled(edge1, edge2)) {
              // If <edge1> and <edge2> are dependent, concurrent, and
              // possibly coenabled, make sure to seed a branch
              // along which <edge1> could occur before <edge2>.
              ExploreEdgeFromNode(edge1, node);
              break;
            }
            std::map<int, Node *> back_pointers = node->back_pointers();
            if (back_pointers.find(tid) == back_pointers.end()) { break; }
            node = back_pointers[tid];
            ASSERT(node, "Node does not exist.");
          }
        }
      }
    } else {
      for (std::pair<int, const Edge *> pair : current_node->edges()) {
        const Edge *edge = pair.second;
        Node *node = edge->to();
        if (edge->enabled() && (!node->visited() && !node->seeded())) {
          server_->frontier()->add_node(node);
          node->set_seeded(true);
        }
      }
    }
    current_node->set_seeds_computed(true);
  }
}

void Handler::UpdateExplorationStatus(Node *leaf_node) {
  Node *current_node = leaf_node;
  for (; current_node; current_node = current_node->parent()) {
    int n = 0;  // The number of children that will be (or has been) explored.
    int nexplored = 0;  // The number of children that has been explored.
    for (std::pair<int, const Edge *> pair : current_node->edges()) {
      const Edge *edge = pair.second;
      const Node *node = edge->to();
      if (edge->enabled() && (node->seeded() || node->visited())) { n++; }
      if (edge->enabled() && (node->explored())) { nexplored++; }
    }
    if (n == nexplored) {
      current_node->set_explored(true);
      current_node->DeallocateSubtree();
    } else {
      break;
    }
  }
}

void Handler::UpdateSizeEstimate(Node *leaf_node) {
  // 1) Compute the probability and estimate of the explored branch
  // assuming uniform probability over edges.
  Node *current_node = leaf_node->parent();
  BigFloat estimate(1);
  BigFloat probability(1);
  for (; current_node; current_node = current_node->parent()) {
    // Compute the number of children that has been (or will be) explored.
    int n = 0;
    for (std::pair<int, const Edge *> pair : current_node->edges()) {
      const Edge *edge = pair.second;
      Node *child = edge->to();
      if (edge->enabled() && (child->visited() || child->seeded())) { n++; }
    }
    estimate = estimate * BigFloat(n);
    probability = probability / BigFloat(n);
  }
  // 2) Update the estimate and probability for nodes of the explored branch.
  leaf_node->set_num_explored_branches(1);
  leaf_node->set_estimate(estimate);
  leaf_node->set_probability(probability);
  current_node = leaf_node->parent();
  for (; current_node; current_node = current_node->parent()) {
    int explored_sum = 0;
    BigFloat probability_sum(0);
    BigFloat estimate_sum(0);
    for (std::pair<int, const Edge *> pair : current_node->edges()) {
      const Edge *edge = pair.second;
      Node *child_node = edge->to();
      explored_sum += child_node->num_explored_branches();
      probability_sum = probability_sum + child_node->probability();
      estimate_sum = estimate_sum + child_node->estimate();
    }
    current_node->set_num_explored_branches(explored_sum);
    current_node->set_estimate(estimate_sum);
    current_node->set_probability(probability_sum);
  }
}

}  // namespace explorer
}  // namespace dbug
