/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "dbug/explorer/server.h"

#include <assert.h>
#include <libxml/parser.h>
#include <libxml/tree.h>

#include <boost/scoped_ptr.hpp>
#include <libconfig.h++>
#include <limits>
#include <list>
#include <map>
#include <sstream>
#include <string>
#include <vector>

#include "dbug/common/command.h"
#include "dbug/common/configuration.h"
#include "dbug/common/system.h"
#include "dbug/common/time.h"
#include "dbug/common/timer.h"
#include "dbug/common/util.h"
#include "dbug/arbiter/server.h"
#include "dbug/explorer/edge.h"
#include "dbug/explorer/handler.h"
#include "dbug/explorer/logging.h"
#include "dbug/explorer/node.h"

using libconfig::Config;
using libconfig::Setting;

namespace dbug {
namespace explorer {

// Server::EventHandler class

Server::EventHandler::EventHandler(Server *server) : server_(server) {}

void Server::EventHandler::preServe() {
  Lock lock(server_);
  server_->set_handler_started(true);
}

// Server::Lock class

Server::Lock::Lock(Server *server) : server_(server) {
  System::MutexLock(&server_->mutex_);
}

Server::Lock::~Lock() {
  System::CVarSignal(&server_->cvar_);
  System::MutexUnlock(&server_->mutex_);
}

// Server class

Server::Server(const char *config_path)
    : arbiter_finished_(false),
      arbiter_port_(Configuration::arbiter_port()),
      arbiter_started_(false),
      config_path_(config_path),
      execution_tree_(new Node(0, NULL)),
      frontier_(new Frontier(execution_tree_)),
      handler_started_(false),
      init_(""),
      interposition_(BINDIR "/dbug-interposition"),
      log_dir_(System::GetEnvironmentVariable("PWD")),
      max_iterations_(0),
      num_errors_(0),
      num_iterations_(0),
      num_processes_started_(0),
      port_(Configuration::explorer_port()),
      timeout_(0),
      use_dpor_(false) {
  arbiter_ = new Command();
  arbiter_->set_program(BINDIR "/dbug-arbiter");
  execution_tree_->set_seeded(true);
  System::CVarInit(&cvar_, NULL);
  System::MutexInit(&mutex_, NULL);
  System::MutexLock(&mutex_);
}

Server::~Server() {
  delete arbiter_;
  std::list<Command *>::iterator it;
  for (it = commands_.begin(); it != commands_.end(); ++it) {
    Command *command = *it;
    delete command;
  }
  commands_.clear();
  System::CVarDestroy(&cvar_);
  delete execution_tree_;
  System::MutexUnlock(&mutex_);
  System::MutexDestroy(&mutex_);
}

// accessors

Node *Server::execution_tree() const { return execution_tree_; }

Frontier *Server::frontier() const { return frontier_.get(); }

int Server::port() const { return port_; }

std::vector<Move> Server::prefix() const { return prefix_; }

bool Server::use_dpor() const { return use_dpor_; }

// mutators

void Server::set_arbiter_finished(bool flag) { arbiter_finished_ = flag; }

void Server::set_handler_started(bool flag) { handler_started_ = flag; }

void Server::increment_num_errors() { num_errors_++; }

void Server::increment_process_started() { num_processes_started_++; }

void Server::set_arbiter_started(bool flag) { arbiter_started_ = flag; }

void Server::set_thrift_server(TThreadedServer *server) {
  thrift_server_ = server;
}

// public methods

void Server::Configure() {
  xmlDoc *doc = xmlReadFile(config_path_.c_str(), NULL, 0);
  if (doc == NULL) {
    ERROR("Could not parse the XML file %s.", config_path_.c_str());
  }
  xmlNode *root_node = xmlDocGetRootElement(doc);
  if (root_node->type != XML_ELEMENT_NODE ||
      xmlStrcmp(root_node->name,
                reinterpret_cast<const xmlChar *>("config")) != 0) {
    ERROR("Error parsing the root element.");
  }
  for (xmlNode *node = root_node->children; node; node = node->next) {
    if (node->type == XML_ELEMENT_NODE) {
      if (xmlStrcmp(node->name,
                    reinterpret_cast<const xmlChar *>("arbiter")) == 0) {
        const xmlChar *arguments =
            reinterpret_cast<const xmlChar *>("arguments");
        if (xmlHasProp(node, arguments)) {
          xmlChar *prop = xmlGetProp(node, arguments);
          std::stringstream stream(reinterpret_cast<char *>(prop));
          xmlFree(prop);
          std::string str;
          if (!(stream >> str)) { ERROR("Error parsing the arbiter element."); }
          arbiter_->add_arg(str.c_str());
          while (stream >> str) { arbiter_->add_arg(str.c_str()); }
          DEBUG("Setting arbiter arguments to: %s",
                arbiter_->ToString().c_str());
        }
        const xmlChar *port = reinterpret_cast<const xmlChar *>("port");
        if (xmlHasProp(node, port)) {
          xmlChar *prop = xmlGetProp(node, port);
          std::stringstream stream(reinterpret_cast<char *>(prop));
          xmlFree(prop);
          if (!(stream >> arbiter_port_)) {
            ERROR("Error parsing the arbiter element.");
          }
        }
      }
      if (xmlStrcmp(node->name,
                    reinterpret_cast<const xmlChar *>("explorer")) == 0) {
        const xmlChar *dpor = reinterpret_cast<const xmlChar *>("dpor");
        if (xmlHasProp(node, dpor)) {
          xmlChar *prop = xmlGetProp(node, dpor);
          std::string string = reinterpret_cast<char *>(prop);
          xmlFree(prop);
          if (string == "true") { use_dpor_ = true; }
        }
        const xmlChar *log_dir = reinterpret_cast<const xmlChar *>("log_dir");
        if (xmlHasProp(node, log_dir)) {
          xmlChar *prop = xmlGetProp(node, log_dir);
          std::stringstream stream(reinterpret_cast<char *>(prop));
          xmlFree(prop);
          if (!(stream >> log_dir_)) {
            ERROR("Error parsing the explorer element.");
          }
        }
        const xmlChar *max_fragments =
            reinterpret_cast<const xmlChar *>("max_fragments");
        if (xmlHasProp(node, max_fragments)) {
          xmlChar *prop = xmlGetProp(node, max_fragments);
          std::stringstream stream(reinterpret_cast<char *>(prop));
          xmlFree(prop);
          int max_fragments;
          if (!(stream >> max_fragments)) {
            ERROR("Error parsing the explorer element.");
          }
          frontier_->set_max_fragments(max_fragments);
        }
        const xmlChar *max_iterations =
            reinterpret_cast<const xmlChar *>("max_iterations");
        if (xmlHasProp(node, max_iterations)) {
          xmlChar *prop = xmlGetProp(node, max_iterations);
          std::stringstream stream(reinterpret_cast<char *>(prop));
          xmlFree(prop);
          if (!(stream >> max_iterations_)) {
            ERROR("Error parsing the explorer element.");
          }
        }
        const xmlChar *port = reinterpret_cast<const xmlChar *>("port");
        if (xmlHasProp(node, port)) {
          xmlChar *prop = xmlGetProp(node, port);
          std::stringstream stream(reinterpret_cast<char *>(prop));
          xmlFree(prop);
          if (!(stream >> port_)) {
            ERROR("Error parsing the explorer element.");
          }
        }
        const xmlChar *strategy = reinterpret_cast<const xmlChar *>("strategy");
        if (xmlHasProp(node, strategy)) {
          xmlChar *prop = xmlGetProp(node, strategy);
          std::stringstream stream(reinterpret_cast<char *>(prop));
          xmlFree(prop);
          std::string strategy;
          if (!(stream >> strategy)) {
            ERROR("Error parsing the explorer element.");
          }
          if (strategy == "random") {
            frontier_->set_strategy(Frontier::Strategy::RANDOM);
          }
        }
        const xmlChar *timeout = reinterpret_cast<const xmlChar *>("timeout");
        if (xmlHasProp(node, timeout)) {
          xmlChar *prop = xmlGetProp(node, timeout);
          std::stringstream stream(reinterpret_cast<char *>(prop));
          xmlFree(prop);
          if (!(stream >> timeout_)) {
            ERROR("Error parsing the explorer element.");
          }
        }
      }
      if (xmlStrcmp(node->name,
                    reinterpret_cast<const xmlChar *>("init")) == 0) {
        const xmlChar *command = reinterpret_cast<const xmlChar *>("command");
        if (xmlHasProp(node, command)) {
          xmlChar *prop = xmlGetProp(node, command);
          init_ = reinterpret_cast<char *>(prop);
          xmlFree(prop);
          DEBUG("Setting init command to: %s", init_.c_str());
        } else {
          ERROR("The command attribute of the init element is missing.");
        }
      }
      if (xmlStrcmp(node->name,
                    reinterpret_cast<const xmlChar *>("interposition")) == 0) {
        const xmlChar *command = reinterpret_cast<const xmlChar *>("command");
        if (xmlHasProp(node, command)) {
          xmlChar *prop = xmlGetProp(node, command);
          interposition_ = reinterpret_cast<char *>(prop);
          xmlFree(prop);
          DEBUG("Setting interposition command to: %s", interposition_.c_str());
        } else {
          ERROR("The command attribute of the "
                "interposition element is missing.");
        }
      }
      if (xmlStrcmp(node->name,
                    reinterpret_cast<const xmlChar *>("program")) == 0) {
        const xmlChar *command = reinterpret_cast<const xmlChar *>("command");
        if (xmlHasProp(node, command)) {
          xmlChar *prop = xmlGetProp(node, command);
          std::stringstream stream(reinterpret_cast<char *>(prop));
          xmlFree(prop);
          std::string str;
          if (!(stream >> str)) { ERROR("Error parsing the program element."); }
          Command *command = new Command();
          command->set_program(interposition_);
          command->add_arg(str.c_str());
          command->add_binary(str);
          while (stream >> str) { command->add_arg(str.c_str()); }
          const xmlChar *input = reinterpret_cast<const xmlChar *>("input");
          if (xmlHasProp(node, input)) {
            xmlChar *prop = xmlGetProp(node, input);
            command->set_input(reinterpret_cast<char *>(prop));
            xmlFree(prop);
          }
          const xmlChar *output = reinterpret_cast<const xmlChar *>("output");
          if (xmlHasProp(node, output)) {
            xmlChar *prop = xmlGetProp(node, output);
            command->set_output(reinterpret_cast<char *>(prop));
            xmlFree(prop);
          }
          commands_.push_back(command);
          DEBUG("Adding new program command: %s", command->ToString().c_str());
        } else {
          ERROR("The command attribute of the program element is missing.");
        }
      }
      if (xmlStrcmp(node->name,
                    reinterpret_cast<const xmlChar *>("mpi_program")) == 0) {
        Command *mpi_command = new Command();
        const xmlChar *mpirun = reinterpret_cast<const xmlChar *>("mpirun");
        if (xmlHasProp(node, mpirun)) {
          xmlChar *prop = xmlGetProp(node, mpirun);
          std::stringstream stream(reinterpret_cast<char *>(prop));
          xmlFree(prop);
          std::string str;
          if (!(stream >> str)) {
            ERROR("Error parsing the mpi_program element.");
          }
          mpi_command->set_program(str.c_str());
          mpi_command->add_binary(str);
          while (stream >> str) { mpi_command->add_arg(str.c_str()); }
        } else {
          ERROR("The mpirun attribute of the mpi_program element is missing.");
        }
        const xmlChar *np = reinterpret_cast<const xmlChar *>("np");
        if (xmlHasProp(node, np)) {
          xmlChar *prop = xmlGetProp(node, np);
          std::stringstream stream(reinterpret_cast<char *>(prop));
          xmlFree(prop);
          std::string str;
          if (!(stream >> str)) {
            ERROR("Error parsing the mpi_program element.");
          }
          mpi_command->add_arg("-n");
          mpi_command->add_arg(str.c_str());
          mpi_command->set_num_processes(System::StringToInt(str.c_str()));
        } else {
          ERROR("The np attribute of the mpi_program element is missing.");
        }
        const xmlChar *command = reinterpret_cast<const xmlChar *>("command");
        if (xmlHasProp(node, command)) {
          xmlChar *prop = xmlGetProp(node, command);
          std::stringstream stream(reinterpret_cast<char *>(prop));
          xmlFree(prop);
          std::string str;
          if (!(stream >> str)) {
            ERROR("Error parsing the mpi_program element.");
          }
          mpi_command->add_arg(BINDIR "/dbug-interposition");
          mpi_command->add_arg(str.c_str());
          mpi_command->add_binary(str);
          while (stream >> str) { mpi_command->add_arg(str.c_str()); }
        } else {
          ERROR("The command attribute of the mpi_program element is missing.");
        }
        commands_.push_back(mpi_command);
        DEBUG("Adding new program command: %s",
              mpi_command->ToString().c_str());
      }
      if (xmlStrcmp(node->name,
                    reinterpret_cast<const xmlChar *>("prefix")) == 0) {
        const xmlChar *path = reinterpret_cast<const xmlChar *>("path");
        if (xmlHasProp(node, path)) {
          xmlChar *prop = xmlGetProp(node, path);
          std::string filename = reinterpret_cast<char *>(prop);
          xmlFree(prop);
          System::File *prefix = System::FileOpen(filename.c_str(), "r");
          int nmoves;
          System::FileScan(prefix, "%d\n", &nmoves);
          prefix_.clear();
          for (int i = 0; i < nmoves; i++) {
            dbug::thrift::Move move;
            System::FileScan(prefix, "%d\n", &move.choice_id);
            prefix_.push_back(move);
          }
          System::FileClose(prefix);
        } else {
          ERROR("The path attribute of the prefix element is missing.");
        }
      }
    }
  }
  xmlFreeDoc(doc);
  xmlCleanupParser();
}

void Server::Explore() {
  // Make sure that the files pointed to by the config file exist.
  CheckForFiles();

  // Clean-up after previous exploration.
  CleanUpLogs();

  while (!handler_started_) {
    DEBUG("Waiting for the handler to start.");
    System::CVarWait(&cvar_, &mutex_);
  }

  // Enumerate different schedules.
  Timer exploration_timer;
  while (!execution_tree_->explored() &&
         (timeout_ == 0 || exploration_timer.value().tv_sec < timeout_) &&
         (max_iterations_ == 0 || num_iterations_ < max_iterations_)) {
    std::vector<System::Process> children;
    Timer iteration_timer;
    frontier_->Partition();
    arbiter_started_ = false;
    num_processes_started_ = 0;
    arbiter_finished_ = false;
    MESSAGE("Iteration #%d", ++num_iterations_);
    // Run the initialization command.
    if (init_ != "") { System::ExecuteShellCommand(init_.c_str()); }
    // Start the arbiter on the background.
    DEBUG("Starting arbiter: %s", arbiter_->ToString().c_str());
    System::Process child = System::Fork();
    if (child == 0) {
      std::map<std::string, std::string> environment;
      // Make sure arbiter uses a log file.
      arbiter_->add_arg("--use_log_file");
      // Set explorer port.
      { std::stringstream stream;
        stream << port_;
        arbiter_->add_arg("--explorer_port");
        arbiter_->add_arg(stream.str().c_str()); }
      // Set arbiter port.
      { std::stringstream stream;
        stream << arbiter_port_;
        arbiter_->add_arg("--arbiter_port");
        arbiter_->add_arg(stream.str().c_str()); }
      if (frontier_->strategy() == Frontier::Strategy::RANDOM) {
        // Set arbiter scheduler mode.
        std::stringstream stream;
        stream << dbug::arbiter::Server::Mode::RANDOM;
        arbiter_->add_arg("--scheduler_mode");
        arbiter_->add_arg(stream.str().c_str());
      }
      // Set logging directory.
      environment["DBUG_LOG_DIR"] = log_dir_;
      Util::StartCommand(arbiter_, environment);
    }
    children.push_back(child);
    while (!arbiter_started_) {
      DEBUG("Waiting for the arbiter to start.");
      System::CVarWait(&cvar_, &mutex_);
    }
    // Start the commands.
    int i = 1;
    for (const Command *command : commands_) {
      DEBUG("Starting command #%d: %s", i++, command->ToString().c_str());
      child = System::Fork();
      if (child == 0) {
        std::map<std::string, std::string> environment;
        // Set log directory.
        environment["DBUG_LOG_DIR"] = log_dir_;
        // Set arbiter port.
        { std::stringstream stream;
          stream << arbiter_port_;
          environment["DBUG_PORT"] = stream.str(); }
        Util::StartCommand(command, environment);
      }
      children.push_back(child);
      while (num_processes_started_ < command->num_processes()) {
        DEBUG("Waiting for a process to start.");
        System::CVarWait(&cvar_, &mutex_);
      }
    }
    while (!arbiter_finished_) {
      DEBUG("Waiting for the arbiter to finish.");
      System::CVarWait(&cvar_, &mutex_);
    }
    ComputeEstimate();
    for (unsigned i = 0; i < children.size(); i++) {
      DEBUG("Collecting child %d.", children[i]);
      System::Wait(children[i], NULL, 0);
    }
    CollectLogs(num_iterations_);
    System::TimeValue time = iteration_timer.value();
    MESSAGE("Iteration #%d lasted %d.%03d seconds",
            num_iterations_, time.tv_sec, time.tv_usec / Time::kM2U);
  }
  if (execution_tree_->explored()) {
    MESSAGE("Finished exploration after %d iterations.", num_iterations_);
  } else {
    MESSAGE("Terminated early after %d iterations.", num_iterations_);
  }
  System::TimeValue time = exploration_timer.value();
  MESSAGE("Exploration lasted %d.%03d seconds",
          time.tv_sec, time.tv_usec / Time::kM2U);
  WriteResults();
}

void Server::Stop() { thrift_server_->stop(); }

// private methods

void Server::CheckForFiles() {
  if (!System::FileExists(BINDIR "/dbug-arbiter")) {
    ERROR("Arbiter binary %s does not exist.", BINDIR "/dbug-arbiter");
  }
  if (!System::FileExists(BINDIR "/dbug-interposition")) {
    ERROR("Interposition binary %s does not exist.",
          BINDIR "/dbug-interposition");
  }
  for (const Command *command : commands_) {
    for (const std::string binary : command->binaries()) {
      if (!System::FileExists(binary.c_str())) {
        ERROR("Binary %s does not exist.", binary.c_str());
      }
    }
  }
}

void Server::CleanUpLogs() {
  if (System::FileExists((log_dir_ + "/dbug-arbiter.log").c_str())) {
    System::Remove((log_dir_ + "/dbug-arbiter.log").c_str());
  }
  if (System::FileExists((log_dir_ + "/dbug-arbiter.strategy").c_str())) {
    System::Remove((log_dir_ + "/dbug-arbiter.strategy").c_str());
  }
  if (System::FileExists((log_dir_ + "/dbug-interposition.log").c_str())) {
    System::Remove((log_dir_ + "/dbug-interposition.log").c_str());
  }
  if (System::FileExists((log_dir_ + "/dbug-logs").c_str())) {
    System::Remove((log_dir_ + "/dbug-logs").c_str());
  }
  System::CreateDirectory((log_dir_ + "/dbug-logs").c_str());
}

void Server::CollectLogs(int n) {
  std::stringstream arbiter;
  arbiter << log_dir_ << "/dbug-logs/dbug-arbiter-" << n << ".log";
  System::Rename((log_dir_ + "/dbug-arbiter.log").c_str(),
                 arbiter.str().c_str());
  std::stringstream strategy;
  strategy << log_dir_ << "/dbug-logs/dbug-arbiter-" << n << ".strategy";
  System::Rename((log_dir_ + "/dbug-arbiter.strategy").c_str(),
                 strategy.str().c_str());
  std::stringstream interposition;
  interposition << log_dir_ << "/dbug-logs/dbug-interposition-" << n << ".log";
  System::Rename((log_dir_ + "/dbug-interposition.log").c_str(),
                 interposition.str().c_str());
}

void Server::ComputeEstimate() {
  BigFloat zero(0);
  BigFloat estimate(execution_tree_->estimate());
  BigFloat n(execution_tree_->num_explored_branches());
  BigFloat probability(execution_tree_->probability());
  BigFloat mean(n / probability);
  BigFloat variance(estimate / probability - mean * mean);
  BigFloat sigma(BigFloat::sqrt(variance));
  BigFloat weight(BigFloat(1) - n / mean);
  BigFloat lower_bound(n + weight * (mean - BigFloat(2) * sigma));
  if (lower_bound < zero) { lower_bound = zero; }
  BigFloat upper_bound(n + weight * (mean + BigFloat(2) * sigma));
  if (upper_bound < zero) { upper_bound = zero; }
  MESSAGE("Estimated number of iterations %s [%s; %s]",
          mean.ToString().c_str(),
          lower_bound.ToString().c_str(),
          upper_bound.ToString().c_str());
}

void Server::WriteResults() {
  Config config;
  Setting &root = config.getRoot();
  root.add("explored", Setting::TypeInt) = num_iterations_;
  root.add("errors", Setting::TypeInt) = num_errors_;
  config.writeFile("dbug.results");
}

}  // namespace explorer
}  // namespace dbug
