/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "dbug/interposition/handler.h"

#include <pthread.h>
#include <semaphore.h>

#include <map>
#include <string>

#include "dbug/arbiter/process.h"
#include "dbug/arbiter/thread.h"
#include "dbug/common/configuration.h"
#include "dbug/common/init.h"
#include "dbug/common/system.h"
#include "dbug/common/util.h"
#include "dbug/interposition/arbiter.h"
#include "dbug/interposition/local-arbiter.h"
#include "dbug/interposition/logging.h"
#include "dbug/interposition/original.h"
#include "dbug/interposition/remote-arbiter.h"
#include "src/thrift/arbiter/arbiter_types.h"

using dbug::System;
using dbug::Util;
using dbug::arbiter::Thread;

namespace dbug {
namespace interposition {

const void *Handler::kEnabled = reinterpret_cast<void *>(0x1);

int Handler::Choose(int n) {
  Interposition off(NULL);
  return arbiter_->Choose(arbiter_->ThreadID(),
                          "dbug_choose",
                          Util::GetBacktrace(),
                          n);
}

void Handler::Detach() {
  Disable(NULL);
  arbiter_->UnregisterThread();
  EnableFork();
  EnableCancellation();
}

void Handler::Disable(const void *value) {
  DisableInterposition(value);
  DisableCancellation();
  DisableFork();
}

void Handler::DisableCancellation() {
  System::ThreadSetCancelState(
      PTHREAD_CANCEL_DISABLE,
      reinterpret_cast<int *>(System::KeyGetSpecific(cancel_state_key_)));
}

void Handler::DisableFork() { System::RWLockReadLock(&fork_rwlock_); }

void Handler::DisableInterposition(const void *value) {
  // We need to call the original implementation of
  // pthread_setspecific directly. Using the System::KeySetSpecific()
  // wrapper leads to infinite recursion.
  Original::pthread_setspecific(interpose_key_, value);
}

void Handler::Enable() {
  EnableFork();
  EnableCancellation();
  EnableInterposition();
}

void Handler::EnableCancellation() {
  System::ThreadSetCancelState(*(reinterpret_cast<int *>(
      System::KeyGetSpecific(cancel_state_key_))), NULL);
}

void Handler::EnableFork() { System::RWLockUnlock(&fork_rwlock_); }

void Handler::EnableInterposition() {
  System::KeySetSpecific(interpose_key_, kEnabled);
}

void Handler::ExportResourceIdentifiers() {
  System::MutexLock(&epoll_map_mutex_);
  std::stringstream epoll_str;
  for (std::pair<int, std::string> pair : epoll_map_) {
    epoll_str << pair.first << " " << pair.second << " ";
  }
  System::SetEnvironmentVariable("DBUG_EPOLL", epoll_str.str().c_str(), 1);
  System::MutexUnlock(&epoll_map_mutex_);
  System::MutexLock(&file_map_mutex_);
  std::stringstream file_str;
  for (std::pair<int, std::string> pair : file_map_) {
    file_str << pair.first << " " << pair.second << " ";
  }
  System::SetEnvironmentVariable("DBUG_FILE", file_str.str().c_str(), 1);
  System::MutexUnlock(&file_map_mutex_);
  System::MutexLock(&pipe_map_mutex_);
  std::stringstream pipe_str;
  for (std::pair<int, std::string> pair : pipe_map_) {
    pipe_str << pair.first << " " << pair.second << " ";
  }
  System::SetEnvironmentVariable("DBUG_PIPE", pipe_str.str().c_str(), 1);
  System::MutexUnlock(&pipe_map_mutex_);
  System::MutexLock(&socket_map_mutex_);
  std::stringstream socket_str;
  for (std::pair<int, std::string> pair : socket_map_) {
    socket_str << pair.first << " " << pair.second << " ";
  }
  System::SetEnvironmentVariable("DBUG_SOCKET", socket_str.str().c_str(), 1);
  System::MutexUnlock(&socket_map_mutex_);
}

void Handler::Finalize() {
  DEBUG("Finalizing dbug interposition handler.");
  bool expected = false;
  bool desired = true;
  if (finalized_.compare_exchange_strong(expected, desired)) {
    arbiter_->ProcessTerminate(arbiter_->ThreadID(),
                               "exit",
                               Util::GetBacktrace(),
                               System::GetProcessID());
    while (arbiter_->nrunning_threads() > 1) { System::Yield(); }
    delete arbiter_;
    System::KeyDelete(cancel_state_key_);
    System::KeyDelete(interpose_key_);
  }
}

void Handler::IgnoreMutex(pthread_mutex_t *mutex) {
  Interposition off(NULL);
  System::MutexLock(&ignore_mutex_set_mutex_);
  ignore_mutex_set_.insert(mutex);
  System::MutexUnlock(&ignore_mutex_set_mutex_);
}

void Handler::ImportResourceIdentifiers() {
  System::MutexLock(&epoll_map_mutex_);
  if (System::GetEnvironmentVariable("DBUG_EPOLL")) {
    std::stringstream epoll_str;
    epoll_str << System::GetEnvironmentVariable("DBUG_EPOLL");
    int fd;
    std::string rid;
    while (epoll_str >> fd && epoll_str >> rid) { epoll_map_[fd] = rid; }
  }
  System::MutexUnlock(&epoll_map_mutex_);
  System::MutexLock(&file_map_mutex_);
  if (System::GetEnvironmentVariable("DBUG_FILE")) {
    std::stringstream file_str;
    file_str << System::GetEnvironmentVariable("DBUG_FILE");
    int fd;
    std::string rid;
    while (file_str >> fd && file_str >> rid) { file_map_[fd] = rid; }
  }
  System::MutexUnlock(&file_map_mutex_);
  System::MutexLock(&pipe_map_mutex_);
  if (System::GetEnvironmentVariable("DBUG_PIPE")) {
    std::stringstream pipe_str;
    pipe_str << System::GetEnvironmentVariable("DBUG_PIPE");
    int fd;
    std::string rid;
    while (pipe_str >> fd && pipe_str >> rid) { pipe_map_[fd] = rid; }
  }
  System::MutexUnlock(&pipe_map_mutex_);
  System::MutexLock(&socket_map_mutex_);
  if (System::GetEnvironmentVariable("DBUG_SOCKET")) {
    std::stringstream socket_str;
    socket_str << System::GetEnvironmentVariable("DBUG_SOCKET");
    int fd;
    std::string rid;
    while (socket_str >> fd && socket_str >> rid) { socket_map_[fd] = rid; }
  }
  System::MutexUnlock(&socket_map_mutex_);
}

void Handler::Initialize() {
  DEBUG("Initializing dbug interposition handler.");
  initialized_ = false;
  finalized_ = false;

  // Install custom dBug handlers for all signals except for:
  // 1) SIGKILL - cannot be intercepted
  // 2) SIGSTOP - cannot be intercepted
  // 3) SIGCONT - handler installation is postponed until SIGTSTP, SIGTTIN,
  //              or SIGTTOU is received.

  InstallHandler(SIGHUP, TerminationHandler);
  InstallHandler(SIGINT, TerminationHandler);
  InstallHandler(SIGQUIT, TerminationHandler);
  InstallHandler(SIGILL, TerminationHandler);
  InstallHandler(SIGABRT, TerminationHandler);
  InstallHandler(SIGBUS, TerminationHandler);
  InstallHandler(SIGFPE, TerminationHandler);
  InstallHandler(SIGUSR1, TerminationHandler);
  InstallHandler(SIGSEGV, TerminationHandler);
  InstallHandler(SIGUSR2, TerminationHandler);
  InstallHandler(SIGPIPE, TerminationHandler);
  InstallHandler(SIGALRM, TerminationHandler);
  InstallHandler(SIGTERM, TerminationHandler);
  InstallHandler(SIGTSTP, StopHandler);
  InstallHandler(SIGTTIN, StopHandler);
  InstallHandler(SIGTTOU, StopHandler);
  InstallHandler(SIGPOLL, TerminationHandler);
  InstallHandler(SIGPROF, TerminationHandler);
  InstallHandler(SIGSYS, TerminationHandler);
  InstallHandler(SIGTRAP, TerminationHandler);
  InstallHandler(SIGVTALRM, TerminationHandler);
  InstallHandler(SIGXCPU, TerminationHandler);
  InstallHandler(SIGXFSZ, TerminationHandler);

  cancel_state_ = PTHREAD_CANCEL_ENABLE;
  System::KeyCreate(&cancel_state_key_, NULL);
  System::KeyCreate(&interpose_key_, NULL);
  System::KeySetSpecific(cancel_state_key_, &cancel_state_);

  System::MutexInit(&barrier_map_mutex_, NULL);
  System::MutexInit(&cvar_map_mutex_, NULL);
  System::MutexInit(&epoll_map_mutex_, NULL);
  System::MutexInit(&file_map_mutex_, NULL);
  System::MutexInit(&mpi_comms_map_mutex_, NULL);
  System::MutexInit(&mpi_message_map_mutex_, NULL);
  System::MutexInit(&mpi_threads_mutex_, NULL);
  System::MutexInit(&mutex_map_mutex_, NULL);
  System::MutexInit(&nsem_map_mutex_, NULL);
  System::MutexInit(&pipe_map_mutex_, NULL);
  System::MutexInit(&rwlock_map_mutex_, NULL);
  System::MutexInit(&socket_map_mutex_, NULL);
  System::MutexInit(&spin_map_mutex_, NULL);
  System::MutexInit(&usem_map_mutex_, NULL);
  System::MutexInit(&ignore_mutex_set_mutex_, NULL);
  System::RWLockInit(&fork_rwlock_, NULL);

  // Disable fork because the pre-condition to open connections to
  // arbiter is that fork is disabled.
  DisableFork();

  std::string hostname = dbug::Configuration::arbiter_hostname();
  if (dbug::System::GetEnvironmentVariable("DBUG_HOSTNAME") != NULL) {
    hostname = dbug::System::GetEnvironmentVariable("DBUG_HOSTNAME");
  }
  int port = dbug::Configuration::arbiter_port();
  if (dbug::System::GetEnvironmentVariable("DBUG_PORT") != NULL) {
    port = System::StringToInt(
        dbug::System::GetEnvironmentVariable("DBUG_PORT"));
  }

#ifdef DBUG_MT
  arbiter_ = new dbug::interposition::LocalArbiter();
#else
  arbiter_ = new dbug::interposition::RemoteArbiter(hostname, port);
#endif

  // Check if the initial thread has been already registered
  // with the arbiter.
  if (System::GetEnvironmentVariable("DBUG_PID") != NULL &&
      System::GetEnvironmentVariable("DBUG_TID") != NULL) {
    arbiter_->CompleteThreadRegistration(
        System::StringToInt(System::GetEnvironmentVariable("DBUG_PID")),
        System::StringToInt(System::GetEnvironmentVariable("DBUG_TID")));
    System::UnsetEnvironmentVariable("DBUG_PID");
    System::UnsetEnvironmentVariable("DBUG_TID");
  } else {
    arbiter_->RegisterThread();
  }

  ImportResourceIdentifiers();

  clock_ = System::Clock();
  System::GetTimeOfDay(&gettimeofday_, NULL);
  time_ = System::Time(NULL);

  initialized_ = true;
  Enable();
}

bool Handler::Interpose() {
  // We need to call the original implementation of
  // pthread_getspecific directly. Using the System::KeyGetSpecific()
  // wrapper leads to infinite recursion.
  return (initialized_ &&
          Original::pthread_getspecific(interpose_key_) == kEnabled);
}

void Handler::Off() { DisableInterposition(NULL); }

void Handler::On() { EnableInterposition(); }

int Handler::RecordNonDeterminism(int event) {
  Interposition off(NULL);
  return arbiter_->RecordNonDeterminism(event);
}

// private static methods

void Handler::ContinuationHandler(int signal) {
  Interposition off(NULL);
  WARNING("Signal %s received", Util::SignalToString(signal).c_str());
  arbiter_->ProcessSignal(arbiter_->ThreadID(),
                          Util::SignalToString(signal),
                          Util::GetBacktrace(),
                          System::GetProcessID(),
                          signal);
  // Reinstall stop handlers.
  System::ControlSignalAction(SIGTSTP, &new_actions_[SIGTSTP], NULL);
  System::ControlSignalAction(SIGTTIN, &new_actions_[SIGTTIN], NULL);
  System::ControlSignalAction(SIGTTOU, &new_actions_[SIGTTOU], NULL);
  System::ControlSignalAction(signal, &old_actions_[signal], NULL);
  System::Kill(System::GetProcessID(), signal);
}

void Handler::InstallHandler(int sig, System::SignalHandler handler) {
  new_actions_[sig].sa_handler = handler;
  System::SignalSetEmpty(&new_actions_[sig].sa_mask);
  new_actions_[sig].sa_flags = 0;
  if (System::ControlSignalAction(sig, NULL, &old_actions_[sig])) {
    ERROR("Retrieving default %s handler.", Util::SignalToString(sig).c_str());
  }
  if (old_actions_[sig].sa_handler != SIG_IGN) {
    if (System::ControlSignalAction(sig, &new_actions_[sig], NULL)) {
      ERROR("Installing dBug %s handler.", Util::SignalToString(sig).c_str());
    } else {
      DEBUG("dBug %s handler installed.", Util::SignalToString(sig).c_str());
    }
  } else {
    DEBUG("Signal %s is ignored.", Util::SignalToString(sig).c_str());
  }
  return;
}

void Handler::ThreadActive(pthread_t tid) {
  Interposition off(NULL);
  arbiter_->SetThreadState(arbiter_->ThreadID(),
                           arbiter_->ThreadID(tid),
                           Thread::PREVIOUS);
}

void Handler::ThreadWaiting() {
  Interposition off(NULL);
  arbiter_->SetThreadState(arbiter_->ThreadID(),
                           arbiter_->ThreadID(),
                           Thread::WAITING);
}

void Handler::Reset() {
  barrier_map_.clear();
  System::MutexInit(&barrier_map_mutex_, NULL);
  cvar_map_.clear();
  System::MutexInit(&cvar_map_mutex_, NULL);
  epoll_map_.clear();
  System::MutexInit(&epoll_map_mutex_, NULL);
  // Do not clear the file descriptor map.
  // file_map_.clear();
  System::MutexInit(&file_map_mutex_, NULL);
  mpi_comms_map_.clear();
  System::MutexInit(&mpi_comms_map_mutex_, NULL);
  mpi_message_map_.clear();
  System::MutexInit(&mpi_message_map_mutex_, NULL);
  mpi_threads_.clear();
  System::MutexInit(&mpi_threads_mutex_, NULL);
  mutex_map_.clear();
  System::MutexInit(&mutex_map_mutex_, NULL);
  nsem_map_.clear();
  System::MutexInit(&nsem_map_mutex_, NULL);
  // Do not clear the pipe descriptor map.
  // pipe_map_.clear();
  System::MutexInit(&pipe_map_mutex_, NULL);
  rwlock_map_.clear();
  System::MutexInit(&rwlock_map_mutex_, NULL);
  // Do not clear the socket descriptor map.
  // socket_map_.clear();
  System::MutexInit(&socket_map_mutex_, NULL);
  spin_map_.clear();
  System::MutexInit(&spin_map_mutex_, NULL);
  usem_map_.clear();
  System::MutexInit(&usem_map_mutex_, NULL);
  ignore_mutex_set_.clear();
  System::MutexInit(&ignore_mutex_set_mutex_, NULL);
  arbiter_->Reset();
}

void Handler::StopHandler(int signal) {
  Interposition off(NULL);
  WARNING("Signal %s received", Util::SignalToString(signal).c_str());
  arbiter_->ProcessSignal(arbiter_->ThreadID(),
                          Util::SignalToString(signal),
                          Util::GetBacktrace(),
                          System::GetProcessID(),
                          signal);
  // Install continuation handler.
  InstallHandler(SIGCONT, ContinuationHandler);
  System::ControlSignalAction(signal, &old_actions_[signal], NULL);
  System::Kill(System::GetProcessID(), signal);
}

void Handler::TerminationHandler(int signal) {
  Disable(NULL);
  WARNING("Signal %s received", Util::SignalToString(signal).c_str());
  bool expected = false;
  bool desired = true;
  if (finalized_.compare_exchange_strong(expected, desired)) {
    arbiter_->ProcessSignal(arbiter_->ThreadID(),
                            Util::SignalToString(signal),
                            Util::GetBacktrace(),
                            System::GetProcessID(),
                            signal);
    while (arbiter_->nrunning_threads() > 1) { System::Yield(); }
    delete arbiter_;
    System::KeyDelete(cancel_state_key_);
    System::KeyDelete(interpose_key_);
  }
  System::ControlSignalAction(signal, &old_actions_[signal], NULL);
  System::Kill(System::GetProcessID(), signal);
}

// private static fields

Arbiter *Handler::arbiter_ = NULL;
std::map<pthread_barrier_t *, std::string> Handler::barrier_map_;
System::Mutex Handler::barrier_map_mutex_;
int Handler::cancel_state_;
System::ThreadKey Handler::cancel_state_key_;
System::ClockType Handler::clock_;
std::map<pthread_cond_t *, std::string> Handler::cvar_map_;
System::Mutex Handler::cvar_map_mutex_;
std::map<int, std::string> Handler::epoll_map_;
int Handler::dbug_comm_internal_;
int Handler::dbug_comm_world_;
int Handler::dbug_group_world_;
System::Mutex Handler::epoll_map_mutex_;
std::map<int, std::string> Handler::file_map_;
System::Mutex Handler::file_map_mutex_;
std::atomic_bool Handler::finalized_;
System::RWLock Handler::fork_rwlock_;
System::TimeValue Handler::gettimeofday_;
std::set<pthread_mutex_t *> Handler::ignore_mutex_set_;
System::Mutex Handler::ignore_mutex_set_mutex_;
std::atomic_bool Handler::initialized_;
System::ThreadKey Handler::interpose_key_;
std::map<int, int> Handler::mpi_comms_map_;
System::Mutex Handler::mpi_comms_map_mutex_;
int Handler::mpi_group_world_;
std::map<void *, Handler::MPIMessage> Handler::mpi_message_map_;
System::Mutex Handler::mpi_message_map_mutex_;
int Handler::mpi_rank_;
std::set<System::Thread> Handler::mpi_threads_;
System::Mutex Handler::mpi_threads_mutex_;
std::map<pthread_mutex_t *, std::string> Handler::mutex_map_;
System::Mutex Handler::mutex_map_mutex_;
System::SignalAction Handler::new_actions_[NSIG];
std::map<sem_t *, std::string> Handler::nsem_map_;
System::Mutex Handler::nsem_map_mutex_;
System::SignalAction Handler::old_actions_[NSIG];
std::map<int, std::string> Handler::pipe_map_;
System::Mutex Handler::pipe_map_mutex_;
std::map<pthread_rwlock_t *, std::string> Handler::rwlock_map_;
System::Mutex Handler::rwlock_map_mutex_;
std::map<int, std::string> Handler::socket_map_;
System::Mutex Handler::socket_map_mutex_;
std::map<pthread_spinlock_t *, std::string> Handler::spin_map_;
System::Mutex Handler::spin_map_mutex_;
std::map<sem_t *, int> Handler::usem_map_;
System::Mutex Handler::usem_map_mutex_;
System::TimeType Handler::time_;

}  // namespace interposition
}  // namespace dbug
