/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "dbug/common/util.h"
#include "dbug/interposition/handler.h"
#include "dbug/interposition/logging.h"
#include "dbug/interposition/original.h"
#include "src/interposition/gen/mem-mgmt.cc"

namespace dbug {
namespace interposition {

#if DBUG_HAVE_CALLOC
void *Handler::calloc(size_t count, size_t size) {
  // Note: dlsym() calls calloc to store an error message string, if
  // the pointer to calloc() has not been retrieved yet, return NULL.
  if (Original::calloc == NULL) { return NULL; }

  if (Interpose()) {
    Interposition off(__FUNCTION__);
    switch (arbiter_->Choose(arbiter_->ThreadID(),
                             "calloc",
                             Util::GetBacktrace(),
                             2)) {
      case 1:
        // Simulate the lack of memory.
        errno = ENOMEM;
        DEBUG_EXTERNAL("%d, %d", "%p", count, size, NULL);
        return NULL;
      default:
        void *ret = Original::calloc(count, size);
        DEBUG_EXTERNAL("%d, %d", "%p", count, size, ret);
        return ret;
    }
  } else {
    DBUG_REQUIRE_SYMBOL(Original::calloc, CALLOC_TYPE, "calloc");
    void *ret = Original::calloc(count, size);
    DEBUG_INTERNAL("%d, %d", "%p", count, size, ret);
    return ret;
  }
}
#endif

#if DBUG_HAVE_FREE
void Handler::free(void *ptr) {
  if (Interpose()) {
    Interposition off(__FUNCTION__);
    Original::free(ptr);
    DEBUG_EXTERNAL("%p", "void", ptr);
  } else {
    DBUG_REQUIRE_SYMBOL(Original::free, FREE_TYPE, "free");
    Original::free(ptr);
    DEBUG_INTERNAL("%p", "void", ptr);
  }
}
#endif

#if DBUG_HAVE_MALLOC
void *Handler::malloc(size_t size) {
  switch (arbiter_->Choose(arbiter_->ThreadID(),
                           "malloc",
                           Util::GetBacktrace(),
                           2)) {
    case 1:
      // Simulate the lack of memory.
      errno = ENOMEM;
      return NULL;
    default:
      return Original::malloc(size);
  }
}
#endif

#if DBUG_HAVE_REALLOC
void *Handler::realloc(void *ptr, size_t size) {
  switch (arbiter_->Choose(arbiter_->ThreadID(),
                           "realloc",
                           Util::GetBacktrace(),
                           2)) {
    case 1:
      // Simulate the lack of memory.
      errno = ENOMEM;
      return NULL;
    default:
      return Original::realloc(ptr, size);
  }
}
#endif

}  // namespace interposition
}  // namespace dbug
