/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

// TODO(jsimsa): Inter-communicators.

// TODO(jsimsa): Persistent P2P communication.

// TODO(jsimsa): One-sided communication.

// TODO(jsimsa): Non-blocking collectives.

// TODO(jsimsa): Record status upon return from PMPI_Test* and PMPI_Wait*.

#include <mpi.h>

#include <boost/scoped_array.hpp>

#include "dbug/arbiter/mpi-collective.h"
#include "dbug/common/util.h"
#include "dbug/common/system.h"
#include "dbug/interposition/handler.h"
#include "dbug/interposition/logging.h"
#include "dbug/interposition/original.h"
#include "src/interposition/gen/pmpi.cc"

namespace dbug {
namespace interposition {

struct SendArguments {
  char *buffer;
  int count;
  MPI_Datatype datatype;
  int destination;
  int tag;
  MPI_Comm communicator;
};

struct IMRecvArguments {
  void *buffer;
  int count;
  MPI_Datatype datatype;
  int source;
  int tag;
  MPI_Comm communicator;
  MPI_Comm dbug_communicator;
  MPI_Message *message;
  MPI_Request *request;
  int pid;
  int tid;
  std::vector<dbug::thrift::Location> backtrace;
};

struct IRecvArguments {
  void *buffer;
  int count;
  MPI_Datatype datatype;
  int source;
  int tag;
  MPI_Comm communicator;
  MPI_Comm dbug_communicator;
  MPI_Request *request;
  int pid;
  int tid;
  std::vector<dbug::thrift::Location> backtrace;
};

struct ISendArguments {
  const void *buffer;
  int count;
  MPI_Datatype datatype;
  int destination;
  int tag;
  MPI_Comm communicator;
  MPI_Comm dbug_communicator;
  MPI_Request *request;
  int pid;
  int tid;
  std::vector<dbug::thrift::Location> backtrace;
};

void *Handler::SendHelper(void *args) {
  SendArguments *send_args = reinterpret_cast<SendArguments *>(args);
  Original::PMPI_Ssend(send_args->buffer,
                       send_args->count,
                       send_args->datatype,
                       send_args->destination,
                       send_args->tag,
                       send_args->communicator);
  if (System::MutexTryLock(&mpi_threads_mutex_) == 0) {
    mpi_threads_.erase(System::ThreadSelf());
    System::ThreadDetach(System::ThreadSelf());
    System::MutexUnlock(&mpi_threads_mutex_);
  }
  delete send_args->buffer;
  delete send_args;
  return NULL;
}

void *SendHelperWrapper(void *args) {
  return Handler::SendHelper(args);
}

void GenerateInternalRanks(MPI_Comm comm,
                           MPI_Comm comm_internal,
                           int *internal_ranks) {
  int comm_size;
  MPI_Comm_size(comm, &comm_size);
  MPI_Group comm_group, internal_comm_group;
  Original::PMPI_Comm_group(comm, &comm_group);
  Original::PMPI_Comm_group(comm_internal, &internal_comm_group);
  boost::scoped_array<int> ranks(new int[comm_size]);
  for (int i = 0; i < comm_size; i++) { ranks[i] = i; }
  MPI_Group_translate_ranks(comm_group,
                            comm_size,
                            ranks.get(),
                            internal_comm_group,
                            internal_ranks);
  MPI_Group_free(&comm_group);
  MPI_Group_free(&internal_comm_group);
}

#ifdef DBUG_HAVE_PMPI_ABORT
int Handler::PMPI_Abort(MPI_Comm comm,
                        int errorcode) {
  System::MutexLock(&mpi_comms_map_mutex_);
  int dbug_comm = mpi_comms_map_[comm];
  System::MutexUnlock(&mpi_comms_map_mutex_);
  arbiter_->MPIAbort(arbiter_->ThreadID(),
                     "MPI_Abort",
                     Util::GetBacktrace(),
                     dbug_comm);
  if (comm == MPI_COMM_WORLD) { comm = dbug_comm_world_; }
  return Original::PMPI_Abort(comm, errorcode);
}
#endif

#ifdef DBUG_HAVE_PMPI_ALLGATHER
int Handler::PMPI_Allgather(const void *sendbuf,
                            int sendcount,
                            MPI_Datatype sendtype,
                            void *recvbuf,
                            int recvcount,
                            MPI_Datatype recvtype,
                            MPI_Comm comm) {
  int origin;
  MPI_Comm_rank(comm, &origin);
  int comm_size;
  MPI_Comm_size(comm, &comm_size);
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  System::MutexLock(&mpi_comms_map_mutex_);
  int dbug_comm = mpi_comms_map_[comm];
  System::MutexUnlock(&mpi_comms_map_mutex_);
  arbiter_->MPIAllToAll(arbiter_->ThreadID(),
                        "MPI_Allgather",
                        Util::GetBacktrace(),
                        origin,
                        comm_size,
                        dbug::arbiter::mpi::AllToAll::ALLGATHER,
                        dbug_comm,
                        world_rank);
  if (comm == MPI_COMM_WORLD) { comm = dbug_comm_world_; }
  return Original::PMPI_Allgather(sendbuf,
                                  sendcount,
                                  sendtype,
                                  recvbuf,
                                  recvcount,
                                  recvtype,
                                  comm);
}
#endif

#ifdef DBUG_HAVE_PMPI_IALLGATHER
int Handler::PMPI_Iallgather(const void *sendbuf,
                             int sendcount,
                             MPI_Datatype sendtype,
                             void *recvbuf,
                             int recvcount,
                             MPI_Datatype recvtype,
                             MPI_Comm comm,
                             MPI_Request *request) {
  WARNING("dbug currently does not support non-blocking collectives.");
  return Original::PMPI_Iallgather(sendbuf,
                                   sendcount,
                                   sendtype,
                                   recvbuf,
                                   recvcount,
                                   recvtype,
                                   comm,
                                   request);
}
#endif

#ifdef DBUG_HAVE_PMPI_ALLGATHERV
int Handler::PMPI_Allgatherv(const void *sendbuf,
                             int sendcount,
                             MPI_Datatype sendtype,
                             void *recvbuf,
                             const int *recvcounts,
                             const int *displs,
                             MPI_Datatype recvtype,
                             MPI_Comm comm) {
  int origin;
  MPI_Comm_rank(comm, &origin);
  int comm_size;
  MPI_Comm_size(comm, &comm_size);
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  System::MutexLock(&mpi_comms_map_mutex_);
  int dbug_comm = mpi_comms_map_[comm];
  System::MutexUnlock(&mpi_comms_map_mutex_);
  arbiter_->MPIAllToAll(arbiter_->ThreadID(),
                        "MPI_Allgatherv",
                        Util::GetBacktrace(),
                        origin,
                        comm_size,
                        dbug::arbiter::mpi::AllToAll::ALLGATHERV,
                        dbug_comm,
                        world_rank);
  if (comm == MPI_COMM_WORLD) { comm = dbug_comm_world_; }
  return Original::PMPI_Allgatherv(sendbuf,
                                   sendcount,
                                   sendtype,
                                   recvbuf,
                                   recvcounts,
                                   displs,
                                   recvtype,
                                   comm);
}
#endif

#ifdef DBUG_HAVE_PMPI_IALLGATHERV
int Handler::PMPI_Iallgatherv(const void *sendbuf,
                              int sendcount,
                              MPI_Datatype sendtype,
                              void *recvbuf,
                              const int *recvcounts,
                              const int *displs,
                              MPI_Datatype recvtype,
                              MPI_Comm comm,
                              MPI_Request *request) {
  WARNING("dbug currently does not support non-blocking collectives.");
  return Original::PMPI_Iallgatherv(sendbuf,
                                    sendcount,
                                    sendtype,
                                    recvbuf,
                                    recvcounts,
                                    displs,
                                    recvtype,
                                    comm,
                                    request);
}
#endif

#ifdef DBUG_HAVE_PMPI_ALLREDUCE
int Handler::PMPI_Allreduce(const void *sendbuf,
                            void *recvbuf,
                            int count,
                            MPI_Datatype datatype,
                            MPI_Op op,
                            MPI_Comm comm) {
  int origin;
  MPI_Comm_rank(comm, &origin);
  int comm_size;
  MPI_Comm_size(comm, &comm_size);
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  System::MutexLock(&mpi_comms_map_mutex_);
  int dbug_comm = mpi_comms_map_[comm];
  System::MutexUnlock(&mpi_comms_map_mutex_);
  arbiter_->MPIAllToAll(arbiter_->ThreadID(),
                        "MPI_Allreduce",
                        Util::GetBacktrace(),
                        origin,
                        comm_size,
                        dbug::arbiter::mpi::AllToAll::ALLREDUCE,
                        dbug_comm,
                        world_rank);
  if (comm == MPI_COMM_WORLD) { comm = dbug_comm_world_; }
  return Original::PMPI_Allreduce(sendbuf,
                                  recvbuf,
                                  count,
                                  datatype,
                                  op,
                                  comm);
}
#endif

#ifdef DBUG_HAVE_PMPI_IALLREDUCE
int Handler::PMPI_Iallreduce(const void *sendbuf,
                             void *recvbuf,
                             int count,
                             MPI_Datatype datatype,
                             MPI_Op op,
                             MPI_Comm comm,
                             MPI_Request *request) {
  WARNING("dbug currently does not support non-blocking collectives.");
  return Original::PMPI_Iallreduce(sendbuf,
                                   recvbuf,
                                   count,
                                   datatype,
                                   op,
                                   comm,
                                   request);
}
#endif

#ifdef DBUG_HAVE_PMPI_ALLTOALL
int Handler::PMPI_Alltoall(const void *sendbuf,
                           int sendcount,
                           MPI_Datatype sendtype,
                           void *recvbuf,
                           int recvcount,
                           MPI_Datatype recvtype,
                           MPI_Comm comm) {
  int origin;
  MPI_Comm_rank(comm, &origin);
  int comm_size;
  MPI_Comm_size(comm, &comm_size);
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  System::MutexLock(&mpi_comms_map_mutex_);
  int dbug_comm = mpi_comms_map_[comm];
  System::MutexUnlock(&mpi_comms_map_mutex_);
  arbiter_->MPIAllToAll(arbiter_->ThreadID(),
                        "MPI_Alltoall",
                        Util::GetBacktrace(),
                        origin,
                        comm_size,
                        dbug::arbiter::mpi::AllToAll::ALLTOALL,
                        dbug_comm,
                        world_rank);
  if (comm == MPI_COMM_WORLD) { comm = dbug_comm_world_; }
  return Original::PMPI_Alltoall(sendbuf,
                                 sendcount,
                                 sendtype,
                                 recvbuf,
                                 recvcount,
                                 recvtype,
                                 comm);
}
#endif

#ifdef DBUG_HAVE_PMPI_IALLTOALL
int Handler::PMPI_Ialltoall(const void *sendbuf,
                            int sendcount,
                            MPI_Datatype sendtype,
                            void *recvbuf,
                            int recvcount,
                            MPI_Datatype recvtype,
                            MPI_Comm comm,
                            MPI_Request *request) {
  WARNING("dbug currently does not support non-blocking collectives.");
  return Original::PMPI_Ialltoall(sendbuf,
                                  sendcount,
                                  sendtype,
                                  recvbuf,
                                  recvcount,
                                  recvtype,
                                  comm,
                                  request);
}
#endif

#ifdef DBUG_HAVE_PMPI_ALLTOALLV
int Handler::PMPI_Alltoallv(const void *sendbuf,
                            const int *sendcounts,
                            const int *sdispls,
                            MPI_Datatype sendtype,
                            void *recvbuf,
                            const int *recvcounts,
                            const int *rdispls,
                            MPI_Datatype recvtype,
                            MPI_Comm comm) {
  int origin;
  MPI_Comm_rank(comm, &origin);
  int comm_size;
  MPI_Comm_size(comm, &comm_size);
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  System::MutexLock(&mpi_comms_map_mutex_);
  int dbug_comm = mpi_comms_map_[comm];
  System::MutexUnlock(&mpi_comms_map_mutex_);
  arbiter_->MPIAllToAll(arbiter_->ThreadID(),
                        "MPI_Alltoallv",
                        Util::GetBacktrace(),
                        origin,
                        comm_size,
                        dbug::arbiter::mpi::AllToAll::ALLTOALLV,
                        dbug_comm,
                        world_rank);
  if (comm == MPI_COMM_WORLD) { comm = dbug_comm_world_; }
  return Original::PMPI_Alltoallv(sendbuf,
                                  sendcounts,
                                  sdispls,
                                  sendtype,
                                  recvbuf,
                                  recvcounts,
                                  rdispls,
                                  recvtype,
                                  comm);
}
#endif

#ifdef DBUG_HAVE_PMPI_IALLTOALLV
int Handler::PMPI_Ialltoallv(const void *sendbuf,
                             const int *sendcounts,
                             const int *sdispls,
                             MPI_Datatype sendtype,
                             void *recvbuf,
                             const int *recvcounts,
                             const int *rdispls,
                             MPI_Datatype recvtype,
                             MPI_Comm comm,
                             MPI_Request *request) {
  WARNING("dbug currently does not support non-blocking collectives.");
  return Original::PMPI_Ialltoallv(sendbuf,
                                   sendcounts,
                                   sdispls,
                                   sendtype,
                                   recvbuf,
                                   recvcounts,
                                   rdispls,
                                   recvtype,
                                   comm,
                                   request);
}
#endif

#ifdef DBUG_HAVE_PMPI_ALLTOALLW
int Handler::PMPI_Alltoallw(const void *sendbuf,
                            const int *sendcounts,
                            const int *sdispls,
                            const MPI_Datatype *sendtypes,
                            void *recvbuf,
                            const int *recvcounts,
                            const int *rdispls,
                            const MPI_Datatype *recvtypes,
                            MPI_Comm comm) {
  int origin;
  MPI_Comm_rank(comm, &origin);
  int comm_size;
  MPI_Comm_size(comm, &comm_size);
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  System::MutexLock(&mpi_comms_map_mutex_);
  int dbug_comm = mpi_comms_map_[comm];
  System::MutexUnlock(&mpi_comms_map_mutex_);
  arbiter_->MPIAllToAll(arbiter_->ThreadID(),
                        "MPI_Alltoallw",
                        Util::GetBacktrace(),
                        origin,
                        comm_size,
                        dbug::arbiter::mpi::AllToAll::ALLTOALLW,
                        dbug_comm,
                        world_rank);
  if (comm == MPI_COMM_WORLD) { comm = dbug_comm_world_; }
  return Original::PMPI_Alltoallw(sendbuf,
                                  sendcounts,
                                  sdispls,
                                  sendtypes,
                                  recvbuf,
                                  recvcounts,
                                  rdispls,
                                  recvtypes,
                                  comm);
}
#endif

#ifdef DBUG_HAVE_PMPI_IALLTOALLW
int Handler::PMPI_Ialltoallw(const void *sendbuf,
                             const int *sendcounts,
                             const int *sdispls,
                             const MPI_Datatype *sendtypes,
                             void *recvbuf,
                             const int *recvcounts,
                             const int *rdispls,
                             const MPI_Datatype *recvtypes,
                             MPI_Comm comm,
                             MPI_Request *request) {
  WARNING("dbug currently does not support non-blocking collectives.");
  return Original::PMPI_Ialltoallw(sendbuf,
                                   sendcounts,
                                   sdispls,
                                   sendtypes,
                                   recvbuf,
                                   recvcounts,
                                   rdispls,
                                   recvtypes,
                                   comm,
                                   request);
}
#endif

#ifdef DBUG_HAVE_PMPI_BARRIER
int Handler::PMPI_Barrier(MPI_Comm comm) {
  int origin;
  MPI_Comm_rank(comm, &origin);
  int comm_size;
  MPI_Comm_size(comm, &comm_size);
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  System::MutexLock(&mpi_comms_map_mutex_);
  int dbug_comm = mpi_comms_map_[comm];
  System::MutexUnlock(&mpi_comms_map_mutex_);
  arbiter_->MPIAllToAll(arbiter_->ThreadID(),
                        "MPI_Barrier",
                        Util::GetBacktrace(),
                        origin,
                        comm_size,
                        dbug::arbiter::mpi::AllToAll::BARRIER,
                        dbug_comm,
                        world_rank);
  if (comm == MPI_COMM_WORLD) { comm = dbug_comm_world_; }
  return Original::PMPI_Barrier(comm);
}
#endif

#ifdef DBUG_HAVE_PMPI_IBARRIER
int Handler::PMPI_Ibarrier(MPI_Comm comm, MPI_Request *request) {
  WARNING("dbug currently does not support non-blocking collectives.");
  return Original::PMPI_Ibarrier(comm, request);
}
#endif

#ifdef DBUG_HAVE_PMPI_BCAST
int Handler::PMPI_Bcast(void *buf,
                        int count,
                        MPI_Datatype datatype,
                        int root,
                        MPI_Comm comm) {
  int origin;
  MPI_Comm_rank(comm, &origin);
  int comm_size;
  MPI_Comm_size(comm, &comm_size);
  int type_size;
  MPI_Type_size(datatype, &type_size);
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  System::MutexLock(&mpi_comms_map_mutex_);
  int dbug_comm = mpi_comms_map_[comm];
  System::MutexUnlock(&mpi_comms_map_mutex_);
  arbiter_->MPIOneToAll(arbiter_->ThreadID(),
                        "MPI_Bcast",
                        Util::GetBacktrace(),
                        origin,
                        comm_size,
                        root,
                        dbug::arbiter::mpi::OneToAll::BCAST,
                        dbug_comm,
                        world_rank);
  if (comm == MPI_COMM_WORLD) { comm = dbug_comm_world_; }
  boost::scoped_array<int> internal_ranks(new int[comm_size]);
  GenerateInternalRanks(comm, dbug_comm_internal_, internal_ranks.get());
  if (origin == root) {
    for (int i = 0; i < comm_size; i++) {
      if (i == root) { continue; }
      char *tmp_buffer = new char[count * type_size];
      System::CopyBytes(buf, tmp_buffer, count * type_size);
      SendArguments *args = new SendArguments();
      args->buffer = tmp_buffer;
      args->count = count;
      args->datatype = datatype;
      args->destination = internal_ranks[i];
      args->tag = 0;
      args->communicator = dbug_comm_internal_;
      System::Thread tid;
      System::ThreadCreate(&tid, NULL, SendHelperWrapper, args);
      System::MutexLock(&mpi_threads_mutex_);
      mpi_threads_.insert(tid);
      System::MutexUnlock(&mpi_threads_mutex_);
    }
    return MPI_SUCCESS;
  } else {
    MPI_Status status;
    return Original::PMPI_Recv(buf,
                               count,
                               datatype,
                               internal_ranks[root],
                               0,
                               dbug_comm_internal_,
                               &status);
  }
}
#endif

#ifdef DBUG_HAVE_PMPI_IBCAST
int Handler::PMPI_Ibcast(void *buf,
                         int count,
                         MPI_Datatype datatype,
                         int root,
                         MPI_Comm comm,
                         MPI_Request *request) {
  WARNING("dbug currently does not support non-blocking collectives.");
  return Original::PMPI_Ibcast(buf,
                               count,
                               datatype,
                               root,
                               comm,
                               request);
}
#endif

#ifdef DBUG_HAVE_PMPI_BSEND
int Handler::PMPI_Bsend(const void *buf,
                        int count,
                        MPI_Datatype datatype,
                        int dest,
                        int tag,
                        MPI_Comm comm) {
  int origin;
  MPI_Comm_rank(comm, &origin);
  int type_size;
  MPI_Type_size(datatype, &type_size);
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  System::MutexLock(&mpi_comms_map_mutex_);
  int dbug_comm = mpi_comms_map_[comm];
  System::MutexUnlock(&mpi_comms_map_mutex_);
  arbiter_->MPIBufferedSend(arbiter_->ThreadID(),
                            "MPI_Bsend",
                            Util::GetBacktrace(),
                            origin,
                            type_size * count,
                            dest,
                            tag,
                            dbug_comm,
                            world_rank);
  if (comm == MPI_COMM_WORLD) { comm = dbug_comm_world_; }
  return Original::PMPI_Bsend(buf,
                              count,
                              datatype,
                              dest,
                              tag,
                              comm);
}
#endif

#ifdef DBUG_HAVE_PMPI_IBSEND
void *Handler::IBSendHelper(void *args) {
  Handler::DisableFork();
  ISendArguments *send_args = reinterpret_cast<ISendArguments *>(args);
  arbiter_->CompleteThreadRegistration(send_args->pid,
                                       send_args->tid);
  int origin;
  MPI_Comm_rank(send_args->communicator, &origin);
  int type_size;
  MPI_Type_size(send_args->datatype, &type_size);
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  arbiter_->MPIBufferedSend(send_args->tid,
                            "MPI_Ibsend",
                            send_args->backtrace,
                            origin,
                            type_size * send_args->count,
                            send_args->destination,
                            send_args->tag,
                            send_args->dbug_communicator,
                            world_rank);
  Original::PMPI_Bsend(send_args->buffer,
                       send_args->count,
                       send_args->datatype,
                       send_args->destination,
                       send_args->tag,
                       send_args->communicator);
  std::string request_id =
      Util::CreateResourceID("mpi_request",
                             arbiter_->ProcessID(),
                             reinterpret_cast<System::UInt64>
                             (send_args->request));
  arbiter_->MPIRequestComplete(arbiter_->ThreadID(),
                               request_id);
  arbiter_->UnregisterThread();
  if (System::MutexTryLock(&mpi_threads_mutex_) == 0) {
    mpi_threads_.erase(System::ThreadSelf());
    System::ThreadDetach(System::ThreadSelf());
    System::MutexUnlock(&mpi_threads_mutex_);
  }
  delete send_args;
  Handler::EnableFork();
  return NULL;
}

void *IBSendHelperWrapper(void *args) { return Handler::IBSendHelper(args); }

int Handler::PMPI_Ibsend(const void *buf,
                         int count,
                         MPI_Datatype datatype,
                         int dest,
                         int tag,
                         MPI_Comm comm,
                         MPI_Request *request) {
  int type_size;
  MPI_Type_size(datatype, &type_size);
  System::MutexLock(&mpi_comms_map_mutex_);
  int dbug_comm = mpi_comms_map_[comm];
  System::MutexUnlock(&mpi_comms_map_mutex_);
  if (comm == MPI_COMM_WORLD) { comm = dbug_comm_world_; }
  ISendArguments *args = new ISendArguments();
  args->buffer = buf;
  args->count = count;
  args->datatype = datatype;
  args->destination = dest;
  args->tag = tag;
  args->communicator = comm;
  args->dbug_communicator = dbug_comm;
  args->request = request;
  arbiter_->PostThreadRegistration(arbiter_->ThreadID(),
                                   "MPI_Ibsend",
                                   Util::GetBacktrace(),
                                   false,
                                   &args->pid,
                                   &args->tid);
  args->backtrace = Util::GetBacktrace();
  System::Thread tid;
  System::ThreadCreate(&tid, NULL, IBSendHelperWrapper, args);
  System::MutexLock(&mpi_threads_mutex_);
  mpi_threads_.insert(tid);
  System::MutexUnlock(&mpi_threads_mutex_);
  arbiter_->AdvanceClock(arbiter_->ThreadID(), args->tid);
  return MPI_SUCCESS;
}
#endif

#ifdef DBUG_HAVE_PMPI_BUFFER_DETACH
int Handler::PMPI_Buffer_detach(void *buffer,
                                int *size) {
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  arbiter_->MPIBufferDetach(arbiter_->ThreadID(),
                            "MPI_Buffer_detach",
                            Util::GetBacktrace(),
                            world_rank);
  return Original::PMPI_Buffer_detach(buffer, size);
}
#endif

#ifdef DBUG_HAVE_PMPI_CART_CREATE
int Handler::PMPI_Cart_create(MPI_Comm comm,
                              int ndims,
                              const int *dims,
                              const int *periods,
                              int reorder,
                              MPI_Comm *new_comm) {
  int origin;
  MPI_Comm_rank(comm, &origin);
  int comm_size;
  MPI_Comm_size(comm, &comm_size);
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  int grid_size = 1;
  for (int i = 0; i < ndims; i++) { grid_size *= dims[i]; }
  if (grid_size <= comm_size) {
    std::set<int> group_set;
    for (int i = 0; i < comm_size; i++) { group_set.insert(i); }
    System::MutexLock(&mpi_comms_map_mutex_);
    int dbug_comm = mpi_comms_map_[comm];
    System::MutexUnlock(&mpi_comms_map_mutex_);
    int new_dbug_comm =
        arbiter_->MPICreateCommunicator(arbiter_->ThreadID(),
                                        "MPI_Comm_create",
                                        Util::GetBacktrace(),
                                        origin,
                                        group_set,
                                        0,
                                        "everyone",
                                        dbug_comm,
                                        world_rank);
    if (comm == MPI_COMM_WORLD) { comm = dbug_comm_world_; }
    int ret = Original::PMPI_Cart_create(comm, ndims, dims,
                                         periods, reorder, new_comm);
    System::MutexLock(&mpi_comms_map_mutex_);
    mpi_comms_map_[*new_comm] = new_dbug_comm;
    System::MutexUnlock(&mpi_comms_map_mutex_);
    return ret;
  } else {
    return Original::PMPI_Cart_create(comm, ndims, dims,
                                      periods, reorder, new_comm);
  }
}
#endif

#ifdef DBUG_HAVE_PMPI_CART_SUB
int Handler::PMPI_Cart_sub(MPI_Comm comm,
                           const int *remain_dims,
                           MPI_Comm *new_comm) {
  std::stringstream stream;
  int ndims;
  MPI_Cartdim_get(comm, &ndims);
  int dims[ndims], periods[ndims], coords[ndims];
  MPI_Cart_get(comm, ndims, dims, periods, coords);
  for (int i = 0; i < ndims; i++) {
    if (i > 0) { stream << "-"; }
    if (remain_dims[i]) {
      stream << "*";
    } else {
      stream << coords[i];
    }
  }
  int origin;
  MPI_Comm_rank(comm, &origin);
  int comm_size;
  MPI_Comm_size(comm, &comm_size);
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  std::set<int> group;
  for (int i = 0; i < comm_size; i++) { group.insert(i); }
  System::MutexLock(&mpi_comms_map_mutex_);
  int dbug_comm = mpi_comms_map_[comm];
  System::MutexUnlock(&mpi_comms_map_mutex_);
  int new_dbug_comm =
      arbiter_->MPICreateCommunicator(arbiter_->ThreadID(),
                                      "MPI_Comm_sub",
                                      Util::GetBacktrace(),
                                      origin,
                                      group,
                                      0,
                                      stream.str(),
                                      dbug_comm,
                                      world_rank);
  if (comm == MPI_COMM_WORLD) { comm = dbug_comm_world_; }
  int ret = Original::PMPI_Cart_sub(comm, remain_dims, new_comm);
  System::MutexLock(&mpi_comms_map_mutex_);
  mpi_comms_map_[*new_comm] = new_dbug_comm;
  System::MutexUnlock(&mpi_comms_map_mutex_);
  return ret;
}
#endif


#ifdef DBUG_HAVE_PMPI_COMM_COMPARE
int Handler::PMPI_Comm_compare(MPI_Comm comm1,
                               MPI_Comm comm2,
                               int *result) {
  if (comm1 == MPI_COMM_WORLD) { comm1 = dbug_comm_world_; }
  return Original::PMPI_Comm_compare(comm1, comm2, result);
}
#endif

#ifdef DBUG_HAVE_PMPI_COMM_CREATE
int Handler::PMPI_Comm_create(MPI_Comm comm,
                              MPI_Group group,
                              MPI_Comm *new_comm) {
  int origin;
  MPI_Comm_rank(comm, &origin);
  int comm_size;
  MPI_Comm_size(comm, &comm_size);
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  if (origin != MPI_UNDEFINED) {
    System::MutexLock(&mpi_comms_map_mutex_);
    int dbug_comm = mpi_comms_map_[comm];
    System::MutexUnlock(&mpi_comms_map_mutex_);
    std::set<int> group_set;
    for (int i = 0; i < comm_size; i++) { group_set.insert(i); }
    int new_dbug_comm =
        arbiter_->MPICreateCommunicator(arbiter_->ThreadID(),
                                        "MPI_Comm_create",
                                        Util::GetBacktrace(),
                                        origin,
                                        group_set,
                                        0,
                                        "everyone",
                                        dbug_comm,
                                        world_rank);
    if (comm == MPI_COMM_WORLD) { comm = dbug_comm_world_; }
    int ret = Original::PMPI_Comm_create(comm, group, new_comm);
    System::MutexLock(&mpi_comms_map_mutex_);
    mpi_comms_map_[*new_comm] = new_dbug_comm;
    System::MutexUnlock(&mpi_comms_map_mutex_);
    return ret;
  } else {
    return Original::PMPI_Comm_create(comm, group, new_comm);
  }
}
#endif

#ifdef DBUG_HAVE_PMPI_COMM_CREATE_GROUP
int Handler::PMPI_Comm_create_group(MPI_Comm comm,
                                    MPI_Group group,
                                    int tag,
                                    MPI_Comm *new_comm) {
  int origin;
  MPI_Comm_rank(comm, &origin);
  int comm_size;
  MPI_Comm_size(comm, &comm_size);
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  if (origin != MPI_UNDEFINED) {
    System::MutexLock(&mpi_comms_map_mutex_);
    int dbug_comm = mpi_comms_map_[comm];
    System::MutexUnlock(&mpi_comms_map_mutex_);
    if (comm == MPI_COMM_WORLD) { comm = dbug_comm_world_; }
    std::set<int> group_set;
    int ranks[comm_size], group_ranks[comm_size];
    MPI_Group comm_group;
    Original::PMPI_Comm_group(comm, &comm_group);
    for (int i = 0; i < comm_size; i++) { ranks[i] = i; }
    MPI_Group_translate_ranks(comm_group, comm_size, ranks, group, group_ranks);
    for (int i = 0; i < comm_size; i++) {
      if (group_ranks[i] != MPI_UNDEFINED) { group_set.insert(i); }
    }
    MPI_Group_free(&comm_group);
    int new_dbug_comm =
        arbiter_->MPICreateCommunicator(arbiter_->ThreadID(),
                                        "MPI_Comm_create_group",
                                        Util::GetBacktrace(),
                                        origin,
                                        group_set,
                                        tag,
                                        "everyone",
                                        dbug_comm,
                                        world_rank);
    int ret = Original::PMPI_Comm_create_group(comm, group, tag, new_comm);
    System::MutexLock(&mpi_comms_map_mutex_);
    mpi_comms_map_[*new_comm] = new_dbug_comm;
    System::MutexUnlock(&mpi_comms_map_mutex_);
    return ret;
  } else {
    return Original::PMPI_Comm_create_group(comm, group, tag, new_comm);
  }
}
#endif

#ifdef DBUG_HAVE_PMPI_COMM_DUP
int Handler::PMPI_Comm_dup(MPI_Comm comm, MPI_Comm *new_comm) {
  int origin;
  MPI_Comm_rank(comm, &origin);
  int comm_size;
  MPI_Comm_size(comm, &comm_size);
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  System::MutexLock(&mpi_comms_map_mutex_);
  int dbug_comm = mpi_comms_map_[comm];
  System::MutexUnlock(&mpi_comms_map_mutex_);
  std::set<int> group;
  for (int i = 0; i < comm_size; i++) { group.insert(i); }
  int new_dbug_comm =
      arbiter_->MPICreateCommunicator(arbiter_->ThreadID(),
                                      "MPI_Comm_dup",
                                      Util::GetBacktrace(),
                                      origin,
                                      group,
                                      0,
                                      "everyone",
                                      dbug_comm,
                                      world_rank);
  if (comm == MPI_COMM_WORLD) { comm = dbug_comm_world_; }
  int ret = Original::PMPI_Comm_dup(comm, new_comm);
  System::MutexLock(&mpi_comms_map_mutex_);
  mpi_comms_map_[*new_comm] = new_dbug_comm;
  System::MutexUnlock(&mpi_comms_map_mutex_);
  return ret;
}
#endif

#ifdef DBUG_HAVE_PMPI_COMM_DUP_WITH_INFO
int Handler::PMPI_Comm_dup_with_info(MPI_Comm comm,
                                     MPI_Info info,
                                     MPI_Comm *new_comm) {
  int origin;
  MPI_Comm_rank(comm, &origin);
  int comm_size;
  MPI_Comm_size(comm, &comm_size);
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  System::MutexLock(&mpi_comms_map_mutex_);
  int dbug_comm = mpi_comms_map_[comm];
  System::MutexUnlock(&mpi_comms_map_mutex_);
  std::set<int> group;
  for (int i = 0; i < comm_size; i++) { group.insert(i); }
  int new_dbug_comm =
      arbiter_->MPICreateCommunicator(arbiter_->ThreadID(),
                                      "MPI_Comm_dup_with_info",
                                      Util::GetBacktrace(),
                                      origin,
                                      group,
                                      0,
                                      "everyone",
                                      dbug_comm,
                                      world_rank);
  if (comm == MPI_COMM_WORLD) { comm = dbug_comm_world_; }
  int ret = Original::PMPI_Comm_dup_with_info(comm, info, new_comm);
  System::MutexLock(&mpi_comms_map_mutex_);
  mpi_comms_map_[*new_comm] = new_dbug_comm;
  System::MutexUnlock(&mpi_comms_map_mutex_);
  return ret;
}
#endif

#ifdef DBUG_HAVE_PMPI_COMM_IDUP
int Handler::PMPI_Comm_idup(MPI_Comm comm,
                            MPI_Comm *new_comm,
                            MPI_Request *request) {
  WARNING("dbug currently does not support non-blocking collectives.");
  return Original::PMPI_Comm_idup(comm, new_comm, request);
}
#endif

#ifdef DBUG_HAVE_PMPI_COMM_FREE
int Handler::PMPI_Comm_free(MPI_Comm *comm) {
  if (*comm == MPI_COMM_WORLD) { *comm = dbug_comm_world_; }
  System::MutexLock(&mpi_comms_map_mutex_);
  mpi_comms_map_.erase(*comm);
  System::MutexUnlock(&mpi_comms_map_mutex_);
  return Original::PMPI_Comm_free(comm);
}
#endif

#ifdef DBUG_HAVE_PMPI_COMM_GROUP
int Handler::PMPI_Comm_group(MPI_Comm comm, MPI_Group *group) {
  if (comm == MPI_COMM_WORLD) { comm = dbug_comm_world_; }
  return Original::PMPI_Comm_group(comm, group);
}
#endif

#ifdef DBUG_HAVE_PMPI_COMM_RANK
int Handler::PMPI_Comm_rank(MPI_Comm comm, int *rank) {
  if (comm == MPI_COMM_WORLD) { comm = dbug_comm_world_; }
  return Original::PMPI_Comm_rank(comm, rank);
}
#endif

#ifdef DBUG_HAVE_PMPI_COMM_SPLIT
int Handler::PMPI_Comm_split(MPI_Comm comm,
                             int color,
                             int key,
                             MPI_Comm *new_comm) {
  int origin;
  MPI_Comm_rank(comm, &origin);
  int comm_size;
  MPI_Comm_size(comm, &comm_size);
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  System::MutexLock(&mpi_comms_map_mutex_);
  int dbug_comm = mpi_comms_map_[comm];
  System::MutexUnlock(&mpi_comms_map_mutex_);
  std::set<int> group;
  for (int i = 0; i < comm_size; i++) { group.insert(i); }
  std::stringstream color_stream;
  color_stream << "color-" << color;
  int new_dbug_comm =
      arbiter_->MPICreateCommunicator(arbiter_->ThreadID(),
                                      "MPI_Comm_split",
                                      Util::GetBacktrace(),
                                      origin,
                                      group,
                                      0,
                                      color_stream.str(),
                                      dbug_comm,
                                      world_rank);
  if (comm == MPI_COMM_WORLD) { comm = dbug_comm_world_; }
  int ret = Original::PMPI_Comm_split(comm, color, key, new_comm);
  System::MutexLock(&mpi_comms_map_mutex_);
  mpi_comms_map_[*new_comm] = new_dbug_comm;
  System::MutexUnlock(&mpi_comms_map_mutex_);
  return ret;
}
#endif

#ifdef DBUG_HAVE_PMPI_COMM_SPLIT_TYPE
int Handler::PMPI_Comm_split_type(MPI_Comm comm,
                                  int split_type,
                                  int key,
                                  MPI_Info info,
                                  MPI_Comm *new_comm) {
  WARNING("dbug currently does not support MPI_Comm_split_type.");
  return Original::PMPI_Comm_split_type(comm, split_type, key, info, new_comm);
}
#endif

#ifdef DBUG_HAVE_PMPI_EXSCAN
int Handler::PMPI_Exscan(const void *sendbuf,
                         void *recvbuf,
                         int count,
                         MPI_Datatype datatype,
                         MPI_Op op,
                         MPI_Comm comm) {
  int origin;
  MPI_Comm_rank(comm, &origin);
  int comm_size;
  MPI_Comm_size(comm, &comm_size);
  int type_size;
  MPI_Type_size(datatype, &type_size);
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  System::MutexLock(&mpi_comms_map_mutex_);
  int dbug_comm = mpi_comms_map_[comm];
  System::MutexUnlock(&mpi_comms_map_mutex_);
  arbiter_->MPIScan(arbiter_->ThreadID(),
                    "MPI_Exscan",
                    Util::GetBacktrace(),
                    origin,
                    comm_size,
                    dbug::arbiter::mpi::Scan::EXCLUSIVE,
                    dbug_comm,
                    world_rank);
  if (comm == MPI_COMM_WORLD) { comm = dbug_comm_world_; }
  boost::scoped_array<int> internal_ranks(new int[comm_size]);
  GenerateInternalRanks(comm, dbug_comm_internal_, internal_ranks.get());
  char *tmp_buffer = new char[count * type_size];
  if (origin == 0) {
    System::CopyBytes(sendbuf, tmp_buffer, count * type_size);
  } else {
    MPI_Status status;
    Original::PMPI_Recv(recvbuf,
                        count,
                        datatype,
                        internal_ranks[origin - 1],
                        0,
                        dbug_comm_internal_,
                        &status);
    System::CopyBytes(recvbuf, tmp_buffer, count * type_size);
    MPI_Reduce_local(sendbuf, tmp_buffer, count, datatype, op);
  }
  if (origin < comm_size - 1) {
    SendArguments *args = new SendArguments();
    args->buffer = tmp_buffer;
    args->count = count;
    args->datatype = datatype;
    args->destination = internal_ranks[origin + 1];
    args->tag = 0;
    args->communicator = dbug_comm_internal_;
    System::Thread tid;
    System::ThreadCreate(&tid, NULL, SendHelperWrapper, args);
    System::MutexLock(&mpi_threads_mutex_);
    mpi_threads_.insert(tid);
    System::MutexUnlock(&mpi_threads_mutex_);
  }
  return MPI_SUCCESS;
}
#endif

#ifdef DBUG_HAVE_PMPI_IEXSCAN
int Handler::PMPI_Iexscan(const void *sendbuf,
                          void *recvbuf,
                          int count,
                          MPI_Datatype datatype,
                          MPI_Op op,
                          MPI_Comm comm,
                          MPI_Request *request) {
  WARNING("dbug currently does not support non-blocking collectives.");
  return Original::PMPI_Iexscan(sendbuf,
                                recvbuf,
                                count,
                                datatype,
                                op,
                                comm,
                                request);
}
#endif

#ifdef DBUG_HAVE_PMPI_FINALIZE
int Handler::PMPI_Finalize() {
  DBUG_REQUIRE_SYMBOL(Original::PMPI_Finalize,
                      PMPI_FINALIZE_TYPE,
                      "PMPI_Finalize");
  int ret;
  if (Interpose()) {
    Disable(NULL);
    arbiter_->UnregisterThread();
    System::MutexLock(&mpi_threads_mutex_);
    for (System::Thread tid : mpi_threads_) { System::ThreadJoin(tid, NULL); }
    System::MutexUnlock(&mpi_threads_mutex_);
    Original::PMPI_Comm_free(&dbug_comm_internal_);
    Original::PMPI_Comm_free(&dbug_comm_world_);
    PMPI_Group_free(&dbug_group_world_);
    PMPI_Group_free(&mpi_group_world_);
    ret = Original::PMPI_Finalize();
    EnableFork();
    EnableCancellation();
    DEBUG_EXTERNAL("", "%d", ret);
    return ret;
  } else {
    assert(0);
  }
}
#endif

#ifdef DBUG_HAVE_PMPI_GATHER
int Handler::PMPI_Gather(const void *sendbuf,
                         int sendcount,
                         MPI_Datatype sendtype,
                         void *recvbuf,
                         int recvcount,
                         MPI_Datatype recvtype,
                         int root,
                         MPI_Comm comm) {
  int origin;
  MPI_Comm_rank(comm, &origin);
  int comm_size;
  MPI_Comm_size(comm, &comm_size);
  int recvtype_size;
  MPI_Type_size(recvtype, &recvtype_size);
  int sendtype_size;
  MPI_Type_size(recvtype, &sendtype_size);
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  System::MutexLock(&mpi_comms_map_mutex_);
  int dbug_comm = mpi_comms_map_[comm];
  System::MutexUnlock(&mpi_comms_map_mutex_);
  arbiter_->MPIAllToOne(arbiter_->ThreadID(),
                        "MPI_Gather",
                        Util::GetBacktrace(),
                        origin,
                        comm_size,
                        root,
                        dbug::arbiter::mpi::AllToOne::GATHER,
                        dbug_comm,
                        world_rank);
  if (comm == MPI_COMM_WORLD) { comm = dbug_comm_world_; }
  boost::scoped_array<int> internal_ranks(new int[comm_size]);
  GenerateInternalRanks(comm, dbug_comm_internal_, internal_ranks.get());
  char *tmp_buffer = new char[sendcount * sendtype_size];
  System::CopyBytes(sendbuf, tmp_buffer, sendcount * sendtype_size);
  SendArguments *args = new SendArguments();
  args->buffer = tmp_buffer;
  args->count = sendcount;
  args->datatype = sendtype;
  args->destination = internal_ranks[root];
  args->tag = 0;
  args->communicator = dbug_comm_internal_;
  System::Thread tid;
  System::ThreadCreate(&tid, NULL, SendHelperWrapper, args);
  System::MutexLock(&mpi_threads_mutex_);
  mpi_threads_.insert(tid);
  System::MutexUnlock(&mpi_threads_mutex_);
  if (origin == root) {
    for (int i = 0; i < comm_size; i++) {
      MPI_Status status;
      void *buffer = reinterpret_cast<char *>(recvbuf) +
          i * recvcount * recvtype_size;
      Original::PMPI_Recv(buffer,
                          recvcount,
                          recvtype,
                          internal_ranks[i],
                          0,
                          dbug_comm_internal_,
                          &status);
    }
  }
  return MPI_SUCCESS;
}
#endif

#ifdef DBUG_HAVE_PMPI_IGATHER
int Handler::PMPI_Igather(const void *sendbuf,
                          int sendcount,
                          MPI_Datatype sendtype,
                          void *recvbuf,
                          int recvcount,
                          MPI_Datatype recvtype,
                          int root,
                          MPI_Comm comm,
                          MPI_Request *request) {
  WARNING("dbug currently does not support non-blocking collectives.");
  return Original::PMPI_Igather(sendbuf,
                                sendcount,
                                sendtype,
                                recvbuf,
                                recvcount,
                                recvtype,
                                root,
                                comm,
                                request);
}
#endif

#ifdef DBUG_HAVE_PMPI_GATHERV
int Handler::PMPI_Gatherv(const void *sendbuf,
                          int sendcount,
                          MPI_Datatype sendtype,
                          void *recvbuf,
                          const int *recvcounts,
                          const int *displs,
                          MPI_Datatype recvtype,
                          int root,
                          MPI_Comm comm) {
  int origin;
  MPI_Comm_rank(comm, &origin);
  int comm_size;
  MPI_Comm_size(comm, &comm_size);
  int recvtype_size;
  MPI_Type_size(recvtype, &recvtype_size);
  int sendtype_size;
  MPI_Type_size(recvtype, &sendtype_size);
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  System::MutexLock(&mpi_comms_map_mutex_);
  int dbug_comm = mpi_comms_map_[comm];
  System::MutexUnlock(&mpi_comms_map_mutex_);
  arbiter_->MPIAllToOne(arbiter_->ThreadID(),
                        "MPI_Gatherv",
                        Util::GetBacktrace(),
                        origin,
                        comm_size,
                        root,
                        dbug::arbiter::mpi::AllToOne::GATHERV,
                        dbug_comm,
                        world_rank);
  if (comm == MPI_COMM_WORLD) { comm = dbug_comm_world_; }
  boost::scoped_array<int> internal_ranks(new int[comm_size]);
  GenerateInternalRanks(comm, dbug_comm_internal_, internal_ranks.get());
  char *tmp_buffer = new char[sendcount * sendtype_size];
  System::CopyBytes(sendbuf, tmp_buffer, sendcount * sendtype_size);
  SendArguments *args = new SendArguments();
  args->buffer = tmp_buffer;
  args->count = sendcount;
  args->datatype = sendtype;
  args->destination = internal_ranks[root];
  args->tag = 0;
  args->communicator = dbug_comm_internal_;
  System::Thread tid;
  System::ThreadCreate(&tid, NULL, SendHelperWrapper, args);
  System::MutexLock(&mpi_threads_mutex_);
  mpi_threads_.insert(tid);
  System::MutexUnlock(&mpi_threads_mutex_);
  if (origin == root) {
    for (int i = 0; i < comm_size; i++) {
      MPI_Status status;
      void *buffer = reinterpret_cast<char *>(recvbuf) +
          displs[i] * recvtype_size;
      Original::PMPI_Recv(buffer,
                          recvcounts[i],
                          recvtype,
                          internal_ranks[i],
                          0,
                          dbug_comm_internal_,
                          &status);
    }
  }
  return MPI_SUCCESS;
}
#endif

#ifdef DBUG_HAVE_PMPI_IGATHERV
int Handler::PMPI_Igatherv(const void *sendbuf,
                           int sendcount,
                           MPI_Datatype sendtype,
                           void *recvbuf,
                           const int *recvcounts,
                           const int *displs,
                           MPI_Datatype recvtype,
                           int root,
                           MPI_Comm comm,
                           MPI_Request *request) {
  WARNING("dbug currently does not support non-blocking collectives.");
  return Original::PMPI_Igatherv(sendbuf,
                                 sendcount,
                                 sendtype,
                                 recvbuf,
                                 recvcounts,
                                 displs,
                                 recvtype,
                                 root,
                                 comm,
                                 request);
}
#endif

#ifdef DBUG_HAVE_PMPI_INIT
int Handler::PMPI_Init(int *argc,
                       char ***argv) {
  int provided;
  return Handler::PMPI_Init_thread(argc, argv, MPI_THREAD_MULTIPLE, &provided);
}
#endif

#ifdef DBUG_HAVE_PMPI_INIT_THREAD
int Handler::PMPI_Init_thread(int *argc,
                              char ***argv,
                              int /* required */,
                              int *provided) {
  if (arbiter_->nrunning_threads() != 1) {
    WARNING("MPI should be initialized before creating threads.");
  }
  int ret = Original::PMPI_Init_thread(argc,
                                       argv,
                                       MPI_THREAD_MULTIPLE,
                                       provided);
  if (*provided != MPI_THREAD_MULTIPLE) {
    WARNING("MPI library does not support concurrent MPI calls.");
  }
  int mpi_rank;
  Original::PMPI_Comm_rank(MPI_COMM_WORLD, &mpi_rank);
  arbiter_->MPIInit(arbiter_->ThreadID(),
                    MPI_ANY_SOURCE,
                    MPI_ANY_TAG,
                    MPI_PROC_NULL,
                    mpi_rank);
  // Issue a barrier to make sure every process have registered their
  // rank with the arbiter.
  int world_size;
  MPI_Comm_size(MPI_COMM_WORLD, &world_size);
  arbiter_->InternalBarrier(arbiter_->ThreadID(),
                            "MPI_Init",
                            Util::GetBacktrace(),
                            0,
                            world_size);
  // Create a new communicator that reorders processes based on the
  // ranks assigned by the arbiter. This step is done to guarantee the
  // mapping from dbug thread IDs to MPI ranks does not change between
  // different execution of the same MPI program.
  boost::scoped_array<int> ranks(new int[world_size]);
  for (int i = 0; i < world_size; i++) { ranks[arbiter_->MPIRank(i)] = i; }
  Original::PMPI_Comm_group(MPI_COMM_WORLD, &mpi_group_world_);
  Original::PMPI_Group_incl(mpi_group_world_,
                            world_size,
                            ranks.get(),
                            &dbug_group_world_);
  Original::PMPI_Comm_create(MPI_COMM_WORLD,
                             dbug_group_world_,
                             &dbug_comm_world_);
  Original::PMPI_Comm_dup(dbug_comm_world_, &dbug_comm_internal_);
  System::MutexLock(&mpi_comms_map_mutex_);
  mpi_comms_map_[MPI_COMM_WORLD] = 0;
  System::MutexUnlock(&mpi_comms_map_mutex_);
  return ret;
}
#endif

#ifdef DBUG_HAVE_PMPI_MPROBE
int Handler::PMPI_Mprobe(int source,
                         int tag,
                         MPI_Comm comm,
                         MPI_Message *message,
                         MPI_Status *status) {
  int origin;
  MPI_Comm_rank(comm, &origin);
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  std::string message_id =
      Util::CreateResourceID("mpi_message",
                             arbiter_->ProcessID(),
                             reinterpret_cast<System::UInt64>(message));
  System::MutexLock(&mpi_comms_map_mutex_);
  int dbug_comm = mpi_comms_map_[comm];
  System::MutexUnlock(&mpi_comms_map_mutex_);
  dbug::thrift::MPIMatchingProbeReturn result =
      arbiter_->MPIMatchingProbe(arbiter_->ThreadID(),
                                 "MPI_Mprobe",
                                 Util::GetBacktrace(),
                                 origin,
                                 source,
                                 tag,
                                 dbug_comm,
                                 message_id,
                                 world_rank);
  status->MPI_SOURCE = result.source;
  status->MPI_TAG = result.tag;
  status->count = result.nbytes;
  MPIMessage mpi_message;
  mpi_message.source = result.source;
  mpi_message.tag = result.tag;
  mpi_message.comm = comm;
  System::MutexLock(&mpi_message_map_mutex_);
  mpi_message_map_[message] = mpi_message;
  System::MutexUnlock(&mpi_message_map_mutex_);
  return MPI_SUCCESS;
}
#endif

#ifdef DBUG_HAVE_PMPI_IMPROBE
int Handler::PMPI_Improbe(int source,
                          int tag,
                          MPI_Comm comm,
                          int *flag,
                          MPI_Message *message,
                          MPI_Status *status) {
  int origin;
  MPI_Comm_rank(comm, &origin);
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  std::string message_id =
      Util::CreateResourceID("mpi_message",
                             arbiter_->ProcessID(),
                             reinterpret_cast<System::UInt64>(message));
  System::MutexLock(&mpi_comms_map_mutex_);
  int dbug_comm = mpi_comms_map_[comm];
  System::MutexUnlock(&mpi_comms_map_mutex_);
  dbug::thrift::MPIImmediateMatchingProbeReturn result =
      arbiter_->MPIImmediateMatchingProbe(arbiter_->ThreadID(),
                                          "MPI_Improbe",
                                          Util::GetBacktrace(),
                                          origin,
                                          source,
                                          tag,
                                          dbug_comm,
                                          message_id,
                                          world_rank);
  if (result.flag) {
    *flag = 1;
    status->MPI_SOURCE = result.source;
    status->MPI_TAG = result.tag;
    status->count = result.nbytes;
    MPIMessage mpi_message;
    mpi_message.source = result.source;
    mpi_message.tag = result.tag;
    mpi_message.comm = comm;
    System::MutexLock(&mpi_message_map_mutex_);
    mpi_message_map_[message] = mpi_message;
    System::MutexUnlock(&mpi_message_map_mutex_);
  } else {
    *flag = 0;
  }
  return MPI_SUCCESS;
}
#endif

#ifdef DBUG_HAVE_PMPI_MRECV
int Handler::PMPI_Mrecv(void *buf,
                        int count,
                        MPI_Datatype datatype,
                        MPI_Message *message,
                        MPI_Status *status) {
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  std::string message_id =
      Util::CreateResourceID("mpi_message",
                             arbiter_->ProcessID(),
                             reinterpret_cast<System::UInt64>(message));
  arbiter_->MPIMatchingReceive(arbiter_->ThreadID(),
                               "MPI_Mrecv",
                               Util::GetBacktrace(),
                               message_id,
                               world_rank);
  System::MutexLock(&mpi_message_map_mutex_);
  MPIMessage mpi_message = mpi_message_map_[message];
  mpi_message_map_.erase(message);
  System::MutexUnlock(&mpi_message_map_mutex_);
  *message = MPI_MESSAGE_NULL;
  int ret = Original::PMPI_Recv(buf,
                                count,
                                datatype,
                                mpi_message.source,
                                mpi_message.tag,
                                (mpi_message.comm == MPI_COMM_WORLD ?
                                 dbug_comm_world_ : mpi_message.comm),
                                status);
  return ret;
}
#endif

#ifdef DBUG_HAVE_PMPI_IMRECV
void *Handler::IMRecvHelper(void *args) {
  Handler::DisableFork();
  IMRecvArguments *recv_args = reinterpret_cast<IMRecvArguments *>(args);
  arbiter_->CompleteThreadRegistration(recv_args->pid,
                                       recv_args->tid);
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  std::string message_id =
      Util::CreateResourceID("mpi_message",
                             arbiter_->ProcessID(),
                             reinterpret_cast<System::UInt64>
                             (recv_args->message));
  arbiter_->MPIMatchingReceive(recv_args->tid,
                               "MPI_Imrecv",
                               recv_args->backtrace,
                               message_id,
                               world_rank);
  MPI_Status status;
  Original::PMPI_Recv(recv_args->buffer,
                      recv_args->count,
                      recv_args->datatype,
                      recv_args->source,
                      recv_args->tag,
                      recv_args->communicator,
                      &status);
  std::string request_id =
      Util::CreateResourceID("mpi_request",
                             arbiter_->ProcessID(),
                             reinterpret_cast<System::UInt64>
                             (recv_args->request));
  arbiter_->MPIRequestComplete(arbiter_->ThreadID(),
                               request_id);
  arbiter_->UnregisterThread();
  if (System::MutexTryLock(&mpi_threads_mutex_) == 0) {
    mpi_threads_.erase(System::ThreadSelf());
    System::ThreadDetach(System::ThreadSelf());
    System::MutexUnlock(&mpi_threads_mutex_);
  }
  delete recv_args;
  Handler::EnableFork();
  return NULL;
}

void *IMrecvHelperWrapper(void *args) { return Handler::IMRecvHelper(args); }

int Handler::PMPI_Imrecv(void *buf,
                         int count,
                         MPI_Datatype datatype,
                         MPI_Message *message,
                         MPI_Request *request) {
  int type_size;
  MPI_Type_size(datatype, &type_size);
  System::MutexLock(&mpi_message_map_mutex_);
  MPIMessage mpi_message = mpi_message_map_[message];
  mpi_message_map_.erase(message);
  System::MutexUnlock(&mpi_message_map_mutex_);
  System::MutexLock(&mpi_comms_map_mutex_);
  int dbug_comm = mpi_comms_map_[mpi_message.comm];
  System::MutexUnlock(&mpi_comms_map_mutex_);
  if (mpi_message.comm == MPI_COMM_WORLD) {
    mpi_message.comm = dbug_comm_world_;
  }
  IMRecvArguments *args = new IMRecvArguments();
  args->buffer = buf;
  args->count = count;
  args->datatype = datatype;
  args->source = mpi_message.source;
  args->tag = mpi_message.tag;
  args->communicator = mpi_message.comm;
  args->dbug_communicator = dbug_comm;
  args->message = message;
  args->request = request;
  arbiter_->PostThreadRegistration(arbiter_->ThreadID(),
                                   "MPI_Imrecv",
                                   Util::GetBacktrace(),
                                   false,
                                   &args->pid,
                                   &args->tid);
  args->backtrace = Util::GetBacktrace();
  System::Thread tid;
  System::ThreadCreate(&tid, NULL, IMRecvHelper, args);
  System::MutexLock(&mpi_threads_mutex_);
  mpi_threads_.insert(tid);
  System::MutexUnlock(&mpi_threads_mutex_);
  arbiter_->AdvanceClock(arbiter_->ThreadID(), args->tid);
  *message = MPI_MESSAGE_NULL;
  return MPI_SUCCESS;
}
#endif

#ifdef DBUG_HAVE_PMPI_PROBE
int Handler::PMPI_Probe(int source,
                        int tag,
                        MPI_Comm comm,
                        MPI_Status *status) {
  int origin;
  MPI_Comm_rank(comm, &origin);
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  System::MutexLock(&mpi_comms_map_mutex_);
  int dbug_comm = mpi_comms_map_[comm];
  System::MutexUnlock(&mpi_comms_map_mutex_);
  dbug::thrift::MPIProbeReturn result =
      arbiter_->MPIProbe(arbiter_->ThreadID(),
                         "MPI_Probe",
                         Util::GetBacktrace(),
                         origin,
                         source,
                         tag,
                         dbug_comm,
                         world_rank);
  status->MPI_SOURCE = result.source;
  status->MPI_TAG = result.tag;
  status->count = result.nbytes;
  return MPI_SUCCESS;
}
#endif

#ifdef DBUG_HAVE_PMPI_IPROBE
int Handler::PMPI_Iprobe(int source,
                         int tag,
                         MPI_Comm comm,
                         int *flag,
                         MPI_Status *status) {
  int origin;
  MPI_Comm_rank(comm, &origin);
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  System::MutexLock(&mpi_comms_map_mutex_);
  int dbug_comm = mpi_comms_map_[comm];
  System::MutexUnlock(&mpi_comms_map_mutex_);
  dbug::thrift::MPIImmediateProbeReturn result =
      arbiter_->MPIImmediateProbe(arbiter_->ThreadID(),
                                  "MPI_Iprobe",
                                  Util::GetBacktrace(),
                                  origin,
                                  source,
                                  tag,
                                  dbug_comm,
                                  world_rank);
  if (result.flag) {
    *flag = 1;
    status->MPI_SOURCE = result.source;
    status->MPI_TAG = result.tag;
    status->count = result.nbytes;
  } else {
    *flag = 0;
  }
  return MPI_SUCCESS;
}
#endif

#ifdef DBUG_HAVE_PMPI_RECV
int Handler::PMPI_Recv(void *buf,
                       int count,
                       MPI_Datatype datatype,
                       int source,
                       int tag,
                       MPI_Comm comm,
                       MPI_Status *status) {
  int origin;
  MPI_Comm_rank(comm, &origin);
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  System::MutexLock(&mpi_comms_map_mutex_);
  int dbug_comm = mpi_comms_map_[comm];
  System::MutexUnlock(&mpi_comms_map_mutex_);
  dbug::thrift::MPIReceiveReturn recv_return =
      arbiter_->MPIReceive(arbiter_->ThreadID(),
                           "MPI_Recv",
                           Util::GetBacktrace(),
                           origin,
                           source,
                           tag,
                           dbug_comm,
                           world_rank);
  if (comm == MPI_COMM_WORLD) { comm = dbug_comm_world_; }
  int ret = Original::PMPI_Recv(buf,
                                count,
                                datatype,
                                recv_return.source,
                                recv_return.tag,
                                comm,
                                status);
  return ret;
}
#endif

#ifdef DBUG_HAVE_PMPI_IRECV
void *Handler::IRecvHelper(void *args) {
  Handler::DisableFork();
  IRecvArguments *recv_args = reinterpret_cast<IRecvArguments *>(args);
  arbiter_->CompleteThreadRegistration(recv_args->pid,
                                       recv_args->tid);
  int origin;
  MPI_Comm_rank(recv_args->communicator, &origin);
  int type_size;
  MPI_Type_size(recv_args->datatype, &type_size);
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  dbug::thrift::MPIReceiveReturn recv_return =
      arbiter_->MPIReceive(recv_args->tid,
                           "MPI_Irecv",
                           recv_args->backtrace,
                           origin,
                           recv_args->source,
                           recv_args->tag,
                           recv_args->dbug_communicator,
                           world_rank);
  MPI_Status status;
  Original::PMPI_Recv(recv_args->buffer,
                      recv_args->count,
                      recv_args->datatype,
                      recv_return.source,
                      recv_return.tag,
                      recv_args->communicator,
                      &status);
  std::string request_id =
      Util::CreateResourceID("mpi_request",
                             arbiter_->ProcessID(),
                             reinterpret_cast<System::UInt64>
                             (recv_args->request));
  arbiter_->MPIRequestComplete(arbiter_->ThreadID(),
                               request_id);
  arbiter_->UnregisterThread();
  if (System::MutexTryLock(&mpi_threads_mutex_) == 0) {
    mpi_threads_.erase(System::ThreadSelf());
    System::ThreadDetach(System::ThreadSelf());
    System::MutexUnlock(&mpi_threads_mutex_);
  }
  delete recv_args;
  Handler::EnableFork();
  return NULL;
}

void *IRecvHelperWrapper(void *args) { return Handler::IRecvHelper(args); }

int Handler::PMPI_Irecv(void *buf,
                        int count,
                        MPI_Datatype datatype,
                        int source,
                        int tag,
                        MPI_Comm comm,
                        MPI_Request *request) {
  int type_size;
  MPI_Type_size(datatype, &type_size);
  System::MutexLock(&mpi_comms_map_mutex_);
  int dbug_comm = mpi_comms_map_[comm];
  System::MutexUnlock(&mpi_comms_map_mutex_);
  if (comm == MPI_COMM_WORLD) { comm = dbug_comm_world_; }
  IRecvArguments *args = new IRecvArguments();
  args->buffer = buf;
  args->count = count;
  args->datatype = datatype;
  args->source = source;
  args->tag = tag;
  args->communicator = comm;
  args->dbug_communicator = dbug_comm;
  args->request = request;
  arbiter_->PostThreadRegistration(arbiter_->ThreadID(),
                                   "MPI_Irecv",
                                   Util::GetBacktrace(),
                                   false,
                                   &args->pid,
                                   &args->tid);
  args->backtrace = Util::GetBacktrace();
  System::Thread tid;
  System::ThreadCreate(&tid, NULL, IRecvHelper, args);
  System::MutexLock(&mpi_threads_mutex_);
  mpi_threads_.insert(tid);
  System::MutexUnlock(&mpi_threads_mutex_);
  arbiter_->AdvanceClock(arbiter_->ThreadID(), args->tid);
  return MPI_SUCCESS;
}
#endif

#ifdef DBUG_HAVE_PMPI_REDUCE
int Handler::PMPI_Reduce(const void *sendbuf,
                         void *recvbuf,
                         int count,
                         MPI_Datatype datatype,
                         MPI_Op op,
                         int root,
                         MPI_Comm comm) {
  int origin;
  MPI_Comm_rank(comm, &origin);
  int comm_size;
  MPI_Comm_size(comm, &comm_size);
  int type_size;
  MPI_Type_size(datatype, &type_size);
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  System::MutexLock(&mpi_comms_map_mutex_);
  int dbug_comm = mpi_comms_map_[comm];
  System::MutexUnlock(&mpi_comms_map_mutex_);
  arbiter_->MPIAllToOne(arbiter_->ThreadID(),
                        "MPI_Reduce",
                        Util::GetBacktrace(),
                        origin,
                        comm_size,
                        root,
                        dbug::arbiter::mpi::AllToOne::REDUCE,
                        dbug_comm,
                        world_rank);
  if (comm == MPI_COMM_WORLD) { comm = dbug_comm_world_; }
  boost::scoped_array<int> internal_ranks(new int[comm_size]);
  GenerateInternalRanks(comm, dbug_comm_internal_, internal_ranks.get());
  char *tmp_buffer = new char[count * type_size];
  System::CopyBytes(sendbuf, tmp_buffer, count * type_size);
  SendArguments *args = new SendArguments();
  args->buffer = tmp_buffer;
  args->count = count;
  args->datatype = datatype;
  args->destination = internal_ranks[root];
  args->tag = 0;
  args->communicator = dbug_comm_internal_;
  System::Thread tid;
  System::ThreadCreate(&tid, NULL, SendHelperWrapper, args);
  System::MutexLock(&mpi_threads_mutex_);
  mpi_threads_.insert(tid);
  System::MutexUnlock(&mpi_threads_mutex_);
  if (origin == root) {
    for (int i = 0; i < comm_size; i++) {
      MPI_Status status;
      boost::scoped_array<char> tmp_buffer(new char[count * type_size]);
      Original::PMPI_Recv((i == 0) ? recvbuf : tmp_buffer.get(),
                          count,
                          datatype,
                          internal_ranks[i],
                          0,
                          dbug_comm_internal_,
                          &status);
      if (i > 0) {
        MPI_Reduce_local(tmp_buffer.get(), recvbuf, count, datatype, op);
      }
    }
  }
  return MPI_SUCCESS;
}
#endif

#ifdef DBUG_HAVE_PMPI_IREDUCE
int Handler::PMPI_Ireduce(const void *sendbuf,
                          void *recvbuf,
                          int count,
                          MPI_Datatype datatype,
                          MPI_Op op,
                          int root,
                          MPI_Comm comm,
                          MPI_Request *request) {
  WARNING("dbug currently does not support non-blocking collectives.");
  return Original::PMPI_Ireduce(sendbuf,
                                recvbuf,
                                count,
                                datatype,
                                op,
                                root,
                                comm,
                                request);
}
#endif

#ifdef DBUG_HAVE_PMPI_REDUCE_SCATTER
int Handler::PMPI_Reduce_scatter(const void *sendbuf,
                                 void *recvbuf,
                                 const int *recvcounts,
                                 MPI_Datatype datatype,
                                 MPI_Op op,
                                 MPI_Comm comm) {
  int origin;
  MPI_Comm_rank(comm, &origin);
  int comm_size;
  MPI_Comm_size(comm, &comm_size);
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  System::MutexLock(&mpi_comms_map_mutex_);
  int dbug_comm = mpi_comms_map_[comm];
  System::MutexUnlock(&mpi_comms_map_mutex_);
  arbiter_->MPIAllToAll(arbiter_->ThreadID(),
                        "MPI_Reduce_scatter",
                        Util::GetBacktrace(),
                        origin,
                        comm_size,
                        dbug::arbiter::mpi::AllToAll::REDUCE_SCATTER,
                        dbug_comm,
                        world_rank);
  if (comm == MPI_COMM_WORLD) { comm = dbug_comm_world_; }
  return Original::PMPI_Reduce_scatter(sendbuf,
                                       recvbuf,
                                       recvcounts,
                                       datatype,
                                       op,
                                       comm);
}
#endif

#ifdef DBUG_HAVE_PMPI_IREDUCE_SCATTER
int Handler::PMPI_Ireduce_scatter(const void *sendbuf,
                                  void *recvbuf,
                                  const int *recvcounts,
                                  MPI_Datatype datatype,
                                  MPI_Op op,
                                  MPI_Comm comm,
                                  MPI_Request *request) {
  WARNING("dbug currently does not support non-blocking collectives.");
  return Original::PMPI_Ireduce_scatter(sendbuf,
                                        recvbuf,
                                        recvcounts,
                                        datatype,
                                        op,
                                        comm,
                                        request);
}
#endif

#ifdef DBUG_HAVE_PMPI_REDUCE_SCATTER_BLOCK
int Handler::PMPI_Reduce_scatter_block(const void *sendbuf,
                                       void *recvbuf,
                                       int recvcount,
                                       MPI_Datatype datatype,
                                       MPI_Op op,
                                       MPI_Comm comm) {
  int origin;
  MPI_Comm_rank(comm, &origin);
  int comm_size;
  MPI_Comm_size(comm, &comm_size);
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  System::MutexLock(&mpi_comms_map_mutex_);
  int dbug_comm = mpi_comms_map_[comm];
  System::MutexUnlock(&mpi_comms_map_mutex_);
  arbiter_->MPIAllToAll(arbiter_->ThreadID(),
                        "MPI_Reduce_scatter_block",
                        Util::GetBacktrace(),
                        origin,
                        comm_size,
                        dbug::arbiter::mpi::AllToAll::REDUCE_SCATTER_BLOCK,
                        dbug_comm,
                        world_rank);
  if (comm == MPI_COMM_WORLD) { comm = dbug_comm_world_; }
  return Original::PMPI_Reduce_scatter_block(sendbuf,
                                             recvbuf,
                                             recvcount,
                                             datatype,
                                             op,
                                             comm);
}
#endif

#ifdef DBUG_HAVE_PMPI_IREDUCE_SCATTER_BLOCK
int Handler::PMPI_Ireduce_scatter_block(const void *sendbuf,
                                        void *recvbuf,
                                        int recvcount,
                                        MPI_Datatype datatype,
                                        MPI_Op op,
                                        MPI_Comm comm,
                                        MPI_Request *request) {
  WARNING("dbug currently does not support non-blocking collectives.");
  return Original::PMPI_Ireduce_scatter_block(sendbuf,
                                              recvbuf,
                                              recvcount,
                                              datatype,
                                              op,
                                              comm,
                                              request);
}
#endif

#ifdef DBUG_HAVE_PMPI_REQUEST_FREE
int Handler::PMPI_Request_free(MPI_Request *request) {
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  if (*request != MPI_REQUEST_NULL) {
    std::string request_id =
        Util::CreateResourceID("mpi_request",
                               arbiter_->ProcessID(),
                               reinterpret_cast<System::UInt64>(request));
    arbiter_->MPIRequestFree(arbiter_->ThreadID(),
                             "MPI_Request_free",
                             Util::GetBacktrace(),
                             request_id,
                             world_rank);
  }
  return Original::PMPI_Request_free(request);
}
#endif

#ifdef DBUG_HAVE_PMPI_RSEND
int Handler::PMPI_Rsend(const void *buf,
                        int count,
                        MPI_Datatype datatype,
                        int dest,
                        int tag,
                        MPI_Comm comm) {
  int origin;
  MPI_Comm_rank(comm, &origin);
  int type_size;
  MPI_Type_size(datatype, &type_size);
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  System::MutexLock(&mpi_comms_map_mutex_);
  int dbug_comm = mpi_comms_map_[comm];
  System::MutexUnlock(&mpi_comms_map_mutex_);
  arbiter_->MPIReadySend(arbiter_->ThreadID(),
                         "MPI_Rsend",
                         Util::GetBacktrace(),
                         origin,
                         type_size * count,
                         dest,
                         tag,
                         dbug_comm,
                         world_rank);
  if (comm == MPI_COMM_WORLD) { comm = dbug_comm_world_; }
  return Original::PMPI_Rsend(buf,
                              count,
                              datatype,
                              dest,
                              tag,
                              comm);
}
#endif

#ifdef DBUG_HAVE_PMPI_IRSEND
void *Handler::IRSendHelper(void *args) {
  Handler::DisableFork();
  ISendArguments *send_args = reinterpret_cast<ISendArguments *>(args);
  arbiter_->CompleteThreadRegistration(send_args->pid,
                                       send_args->tid);
  int origin;
  MPI_Comm_rank(send_args->communicator, &origin);
  int type_size;
  MPI_Type_size(send_args->datatype, &type_size);
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  arbiter_->MPIReadySend(send_args->tid,
                         "MPI_Irsend",
                         send_args->backtrace,
                         origin,
                         type_size * send_args->count,
                         send_args->destination,
                         send_args->tag,
                         send_args->dbug_communicator,
                         world_rank);
  Original::PMPI_Rsend(send_args->buffer,
                       send_args->count,
                       send_args->datatype,
                       send_args->destination,
                       send_args->tag,
                       send_args->communicator);
  std::string request_id =
      Util::CreateResourceID("mpi_request",
                             arbiter_->ProcessID(),
                             reinterpret_cast<System::UInt64>
                             (send_args->request));
  arbiter_->MPIRequestComplete(arbiter_->ThreadID(),
                               request_id);
  arbiter_->UnregisterThread();
  if (System::MutexTryLock(&mpi_threads_mutex_) == 0) {
    mpi_threads_.erase(System::ThreadSelf());
    System::ThreadDetach(System::ThreadSelf());
    System::MutexUnlock(&mpi_threads_mutex_);
  }
  delete send_args;
  Handler::EnableFork();
  return NULL;
}

void *IRSendHelperWrapper(void *args) { return Handler::IRSendHelper(args); }

int Handler::PMPI_Irsend(const void *buf,
                         int count,
                         MPI_Datatype datatype,
                         int dest,
                         int tag,
                         MPI_Comm comm,
                         MPI_Request *request) {
  int type_size;
  MPI_Type_size(datatype, &type_size);
  System::MutexLock(&mpi_comms_map_mutex_);
  int dbug_comm = mpi_comms_map_[comm];
  System::MutexUnlock(&mpi_comms_map_mutex_);
  if (comm == MPI_COMM_WORLD) { comm = dbug_comm_world_; }
  ISendArguments *args = new ISendArguments();
  args->buffer = buf;
  args->count = count;
  args->datatype = datatype;
  args->destination = dest;
  args->tag = tag;
  args->communicator = comm;
  args->dbug_communicator = dbug_comm;
  args->request = request;
  arbiter_->PostThreadRegistration(arbiter_->ThreadID(),
                                   "MPI_Irsend",
                                   Util::GetBacktrace(),
                                   false,
                                   &args->pid,
                                   &args->tid);
  args->backtrace = Util::GetBacktrace();
  System::Thread tid;
  System::ThreadCreate(&tid, NULL, IRSendHelperWrapper, args);
  System::MutexLock(&mpi_threads_mutex_);
  mpi_threads_.insert(tid);
  System::MutexUnlock(&mpi_threads_mutex_);
  arbiter_->AdvanceClock(arbiter_->ThreadID(), args->tid);
  return MPI_SUCCESS;
}
#endif

#ifdef DBUG_HAVE_PMPI_SCAN
int Handler::PMPI_Scan(const void *sendbuf,
                       void *recvbuf,
                       int count,
                       MPI_Datatype datatype,
                       MPI_Op op,
                       MPI_Comm comm) {
  int origin;
  MPI_Comm_rank(comm, &origin);
  int comm_size;
  MPI_Comm_size(comm, &comm_size);
  int type_size;
  MPI_Type_size(datatype, &type_size);
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  System::MutexLock(&mpi_comms_map_mutex_);
  int dbug_comm = mpi_comms_map_[comm];
  System::MutexUnlock(&mpi_comms_map_mutex_);
  arbiter_->MPIScan(arbiter_->ThreadID(),
                    "MPI_Scan",
                    Util::GetBacktrace(),
                    origin,
                    comm_size,
                    dbug::arbiter::mpi::Scan::INCLUSIVE,
                    dbug_comm,
                    world_rank);
  if (comm == MPI_COMM_WORLD) { comm = dbug_comm_world_; }
  boost::scoped_array<int> internal_ranks(new int[comm_size]);
  GenerateInternalRanks(comm, dbug_comm_internal_, internal_ranks.get());
  if (origin == 0) {
    System::CopyBytes(sendbuf, recvbuf, count * type_size);
  } else {
    MPI_Status status;
    Original::PMPI_Recv(recvbuf,
                        count,
                        datatype,
                        internal_ranks[origin - 1],
                        0,
                        dbug_comm_internal_,
                        &status);
    MPI_Reduce_local(sendbuf, recvbuf, count, datatype, op);
  }
  if (origin < comm_size - 1) {
    char *tmp_buffer = new char[count * type_size];
    System::CopyBytes(recvbuf, tmp_buffer, count * type_size);
    SendArguments *args = new SendArguments();
    args->buffer = tmp_buffer;
    args->count = count;
    args->datatype = datatype;
    args->destination = internal_ranks[origin + 1];
    args->tag = 0;
    args->communicator = dbug_comm_internal_;
    System::Thread tid;
    System::ThreadCreate(&tid, NULL, SendHelperWrapper, args);
    System::MutexLock(&mpi_threads_mutex_);
    mpi_threads_.insert(tid);
    System::MutexUnlock(&mpi_threads_mutex_);
  }
  return MPI_SUCCESS;
}
#endif

#ifdef DBUG_HAVE_PMPI_ISCAN
int Handler::PMPI_Iscan(const void *sendbuf,
                        void *recvbuf,
                        int count,
                        MPI_Datatype datatype,
                        MPI_Op op,
                        MPI_Comm comm,
                        MPI_Request *request) {
  WARNING("dbug currently does not support non-blocking collectives.");
  return Original::PMPI_Iscan(sendbuf,
                              recvbuf,
                              count,
                              datatype,
                              op,
                              comm,
                              request);
}
#endif

#ifdef DBUG_HAVE_PMPI_SCATTER
int Handler::PMPI_Scatter(const void *sendbuf,
                          int sendcount,
                          MPI_Datatype sendtype,
                          void *recvbuf,
                          int recvcount,
                          MPI_Datatype recvtype,
                          int root,
                          MPI_Comm comm) {
  int origin;
  MPI_Comm_rank(comm, &origin);
  int comm_size;
  MPI_Comm_size(comm, &comm_size);
  int sendtype_size;
  MPI_Type_size(sendtype, &sendtype_size);
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  System::MutexLock(&mpi_comms_map_mutex_);
  int dbug_comm = mpi_comms_map_[comm];
  System::MutexUnlock(&mpi_comms_map_mutex_);
  arbiter_->MPIOneToAll(arbiter_->ThreadID(),
                        "MPI_Scatter",
                        Util::GetBacktrace(),
                        origin,
                        comm_size,
                        root,
                        dbug::arbiter::mpi::OneToAll::SCATTER,
                        dbug_comm,
                        world_rank);
  if (comm == MPI_COMM_WORLD) { comm = dbug_comm_world_; }
  boost::scoped_array<int> internal_ranks(new int[comm_size]);
  GenerateInternalRanks(comm, dbug_comm_internal_, internal_ranks.get());
  if (origin == root) {
    for (int i = 0; i < comm_size; i++) {
      char *tmp_buffer = new char[sendcount * sendtype_size];
      System::CopyBytes(reinterpret_cast<const char *>(sendbuf) +
                        i * sendcount * sendtype_size,
                        tmp_buffer,
                        sendcount * sendtype_size);
      SendArguments *args = new SendArguments();
      args->buffer = tmp_buffer;
      args->count = sendcount;
      args->datatype = sendtype;
      args->destination = internal_ranks[i];
      args->tag = 0;
      args->communicator = dbug_comm_internal_;
      System::Thread tid;
      System::ThreadCreate(&tid, NULL, SendHelperWrapper, args);
      System::MutexLock(&mpi_threads_mutex_);
      mpi_threads_.insert(tid);
      System::MutexUnlock(&mpi_threads_mutex_);
    }
  }
  MPI_Status status;
  return Original::PMPI_Recv(recvbuf,
                             recvcount,
                             recvtype,
                             internal_ranks[root],
                             0,
                             dbug_comm_internal_,
                             &status);
}
#endif

#ifdef DBUG_HAVE_PMPI_ISCATTER
int Handler::PMPI_Iscatter(const void *sendbuf,
                           int sendcount,
                           MPI_Datatype sendtype,
                           void *recvbuf,
                           int recvcount,
                           MPI_Datatype recvtype,
                           int root,
                           MPI_Comm comm,
                           MPI_Request *request) {
  WARNING("dbug currently does not support non-blocking collectives.");
  return Original::PMPI_Iscatter(sendbuf,
                                 sendcount,
                                 sendtype,
                                 recvbuf,
                                 recvcount,
                                 recvtype,
                                 root,
                                 comm,
                                 request);
}
#endif

#ifdef DBUG_HAVE_PMPI_SCATTERV
int Handler::PMPI_Scatterv(const void *sendbuf,
                           const int *sendcounts,
                           const int *displs,
                           MPI_Datatype sendtype,
                           void *recvbuf,
                           int recvcount,
                           MPI_Datatype recvtype,
                           int root,
                           MPI_Comm comm) {
  int origin;
  MPI_Comm_rank(comm, &origin);
  int comm_size;
  MPI_Comm_size(comm, &comm_size);
  int sendtype_size;
  MPI_Type_size(sendtype, &sendtype_size);
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  System::MutexLock(&mpi_comms_map_mutex_);
  int dbug_comm = mpi_comms_map_[comm];
  System::MutexUnlock(&mpi_comms_map_mutex_);
  arbiter_->MPIOneToAll(arbiter_->ThreadID(),
                        "MPI_Scatterv",
                        Util::GetBacktrace(),
                        origin,
                        comm_size,
                        root,
                        dbug::arbiter::mpi::OneToAll::SCATTERV,
                        dbug_comm,
                        world_rank);
  if (comm == MPI_COMM_WORLD) { comm = dbug_comm_world_; }
  boost::scoped_array<int> internal_ranks(new int[comm_size]);
  GenerateInternalRanks(comm, dbug_comm_internal_, internal_ranks.get());
  if (origin == root) {
    for (int i = 0; i < comm_size; i++) {
      System::Thread tid;
      char *tmp_buffer = new char[sendcounts[i] * sendtype_size];
      System::CopyBytes(reinterpret_cast<const char *>(sendbuf) +
                        displs[i] * sendtype_size,
                        tmp_buffer,
                        sendcounts[i] * sendtype_size);
      SendArguments *args = new SendArguments();
      args->buffer = tmp_buffer;
      args->count = sendcounts[i];
      args->datatype = sendtype;
      args->destination = internal_ranks[i];
      args->tag = 0;
      args->communicator = dbug_comm_internal_;
      System::ThreadCreate(&tid, NULL, SendHelperWrapper, args);
      System::MutexLock(&mpi_threads_mutex_);
      mpi_threads_.insert(tid);
      System::MutexUnlock(&mpi_threads_mutex_);
    }
  }
  MPI_Status status;
  return Original::PMPI_Recv(recvbuf,
                             recvcount,
                             recvtype,
                             internal_ranks[root],
                             0,
                             dbug_comm_internal_,
                             &status);
}
#endif

#ifdef DBUG_HAVE_PMPI_ISCATTERV
int Handler::PMPI_Iscatterv(const void *sendbuf,
                            const int *sendcounts,
                            const int *displs,
                            MPI_Datatype sendtype,
                            void *recvbuf,
                            int recvcount,
                            MPI_Datatype recvtype,
                            int root,
                            MPI_Comm comm,
                            MPI_Request *request) {
  WARNING("dbug currently does not support non-blocking collectives.");
  return Original::PMPI_Iscatterv(sendbuf,
                                  sendcounts,
                                  displs,
                                  sendtype,
                                  recvbuf,
                                  recvcount,
                                  recvtype,
                                  root,
                                  comm,
                                  request);
}
#endif

#ifdef DBUG_HAVE_PMPI_SEND
int Handler::PMPI_Send(const void *buf,
                       int count,
                       MPI_Datatype datatype,
                       int dest,
                       int tag,
                       MPI_Comm comm) {
  int origin;
  MPI_Comm_rank(comm, &origin);
  int type_size;
  MPI_Type_size(datatype, &type_size);
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  System::MutexLock(&mpi_comms_map_mutex_);
  int dbug_comm = mpi_comms_map_[comm];
  System::MutexUnlock(&mpi_comms_map_mutex_);
  arbiter_->MPISynchronousSend(arbiter_->ThreadID(),
                               "MPI_Send",
                               Util::GetBacktrace(),
                               origin,
                               type_size * count,
                               dest,
                               tag,
                               dbug_comm,
                               world_rank);
  if (comm == MPI_COMM_WORLD) { comm = dbug_comm_world_; }
  return Original::PMPI_Send(buf,
                             count,
                             datatype,
                             dest,
                             tag,
                             comm);
}
#endif

#ifdef DBUG_HAVE_PMPI_ISEND
void *Handler::ISendHelper(void *args) {
  Handler::DisableFork();
  ISendArguments *send_args = reinterpret_cast<ISendArguments *>(args);
  arbiter_->CompleteThreadRegistration(send_args->pid,
                                       send_args->tid);
  int origin;
  MPI_Comm_rank(send_args->communicator, &origin);
  int type_size;
  MPI_Type_size(send_args->datatype, &type_size);
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  arbiter_->MPISynchronousSend(send_args->tid,
                               "MPI_Isend",
                               send_args->backtrace,
                               origin,
                               type_size * send_args->count,
                               send_args->destination,
                               send_args->tag,
                               send_args->dbug_communicator,
                               world_rank);
  Original::PMPI_Send(send_args->buffer,
                      send_args->count,
                      send_args->datatype,
                      send_args->destination,
                      send_args->tag,
                      send_args->communicator);
  std::string request_id =
      Util::CreateResourceID("mpi_request",
                             arbiter_->ProcessID(),
                             reinterpret_cast<System::UInt64>
                             (send_args->request));
  arbiter_->MPIRequestComplete(arbiter_->ThreadID(),
                               request_id);
  arbiter_->UnregisterThread();
  if (System::MutexTryLock(&mpi_threads_mutex_) == 0) {
    mpi_threads_.erase(System::ThreadSelf());
    System::ThreadDetach(System::ThreadSelf());
    System::MutexUnlock(&mpi_threads_mutex_);
  }
  delete send_args;
  Handler::EnableFork();
  return NULL;
}

void *ISendHelperWrapper(void *args) { return Handler::ISendHelper(args); }

int Handler::PMPI_Isend(const void *buf,
                        int count,
                        MPI_Datatype datatype,
                        int dest,
                        int tag,
                        MPI_Comm comm,
                        MPI_Request *request) {
  int type_size;
  MPI_Type_size(datatype, &type_size);
  System::MutexLock(&mpi_comms_map_mutex_);
  int dbug_comm = mpi_comms_map_[comm];
  System::MutexUnlock(&mpi_comms_map_mutex_);
  if (comm == MPI_COMM_WORLD) { comm = dbug_comm_world_; }
  ISendArguments *args = new ISendArguments();
  args->buffer = buf;
  args->count = count;
  args->datatype = datatype;
  args->destination = dest;
  args->tag = tag;
  args->communicator = comm;
  args->dbug_communicator = dbug_comm;
  args->request = request;
  arbiter_->PostThreadRegistration(arbiter_->ThreadID(),
                                   "MPI_Isend",
                                   Util::GetBacktrace(),
                                   false,
                                   &args->pid,
                                   &args->tid);
  args->backtrace = Util::GetBacktrace();
  System::Thread tid;
  System::ThreadCreate(&tid, NULL, ISendHelperWrapper, args);
  System::MutexLock(&mpi_threads_mutex_);
  mpi_threads_.insert(tid);
  System::MutexUnlock(&mpi_threads_mutex_);
  arbiter_->AdvanceClock(arbiter_->ThreadID(), args->tid);
  return MPI_SUCCESS;
}
#endif

#ifdef DBUG_HAVE_PMPI_SENDRECV
int Handler::PMPI_Sendrecv(const void *sendbuf,
                           int sendcount,
                           MPI_Datatype sendtype,
                           int dest,
                           int sendtag,
                           void *recvbuf,
                           int recvcount,
                           MPI_Datatype recvtype,
                           int source,
                           int recvtag,
                           MPI_Comm comm,
                           MPI_Status *status) {
  int origin;
  MPI_Comm_rank(comm, &origin);
  int type_size;
  MPI_Type_size(sendtype, &type_size);
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  System::MutexLock(&mpi_comms_map_mutex_);
  int dbug_comm = mpi_comms_map_[comm];
  System::MutexUnlock(&mpi_comms_map_mutex_);
  dbug::thrift::MPISendReceiveReturn sendrecv_return =
      arbiter_->MPISendReceive(arbiter_->ThreadID(),
                               "MPI_Sendrecv",
                               Util::GetBacktrace(),
                               origin,
                               type_size * sendcount,
                               dest,
                               sendtag,
                               source,
                               recvtag,
                               dbug_comm,
                               world_rank);
  if (comm == MPI_COMM_WORLD) { comm = dbug_comm_world_; }
  int ret = Original::PMPI_Sendrecv(sendbuf,
                                    sendcount,
                                    sendtype,
                                    dest,
                                    sendtag,
                                    recvbuf,
                                    recvcount,
                                    recvtype,
                                    sendrecv_return.source,
                                    sendrecv_return.tag,
                                    comm,
                                    status);
  return ret;
}
#endif

#ifdef DBUG_HAVE_PMPI_SENDRECV_REPLACE
int Handler::PMPI_Sendrecv_replace(void *buf,
                                   int count,
                                   MPI_Datatype datatype,
                                   int dest,
                                   int sendtag,
                                   int source,
                                   int recvtag,
                                   MPI_Comm comm,
                                   MPI_Status *status) {
  int origin;
  MPI_Comm_rank(comm, &origin);
  int type_size;
  MPI_Type_size(datatype, &type_size);
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  System::MutexLock(&mpi_comms_map_mutex_);
  int dbug_comm = mpi_comms_map_[comm];
  System::MutexUnlock(&mpi_comms_map_mutex_);
  dbug::thrift::MPISendReceiveReturn sendrecv_return =
      arbiter_->MPISendReceive(arbiter_->ThreadID(),
                               "MPI_Sendrecv_replace",
                               Util::GetBacktrace(),
                               origin,
                               type_size * count,
                               dest,
                               sendtag,
                               source,
                               recvtag,
                               dbug_comm,
                               world_rank);
  if (comm == MPI_COMM_WORLD) { comm = dbug_comm_world_; }
  int ret = Original::PMPI_Sendrecv_replace(buf,
                                            count,
                                            datatype,
                                            dest,
                                            sendtag,
                                            sendrecv_return.source,
                                            sendrecv_return.tag,
                                            comm,
                                            status);
  return ret;
}
#endif

#ifdef DBUG_HAVE_PMPI_SSEND
int Handler::PMPI_Ssend(const void *buf,
                        int count,
                        MPI_Datatype datatype,
                        int dest,
                        int tag,
                        MPI_Comm comm) {
  int origin;
  MPI_Comm_rank(comm, &origin);
  int type_size;
  MPI_Type_size(datatype, &type_size);
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  System::MutexLock(&mpi_comms_map_mutex_);
  int dbug_comm = mpi_comms_map_[comm];
  System::MutexUnlock(&mpi_comms_map_mutex_);
  arbiter_->MPISynchronousSend(arbiter_->ThreadID(),
                               "MPI_Ssend",
                               Util::GetBacktrace(),
                               origin,
                               type_size * count,
                               dest,
                               tag,
                               dbug_comm,
                               world_rank);
  if (comm == MPI_COMM_WORLD) { comm = dbug_comm_world_; }
  return Original::PMPI_Ssend(buf,
                              count,
                              datatype,
                              dest,
                              tag,
                              comm);
}
#endif

#ifdef DBUG_HAVE_PMPI_ISSEND
void *Handler::ISSendHelper(void *args) {
  Handler::DisableFork();
  ISendArguments *send_args = reinterpret_cast<ISendArguments *>(args);
  arbiter_->CompleteThreadRegistration(send_args->pid,
                                       send_args->tid);
  int origin;
  MPI_Comm_rank(send_args->communicator, &origin);
  int type_size;
  MPI_Type_size(send_args->datatype, &type_size);
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  arbiter_->MPISynchronousSend(send_args->tid,
                               "MPI_Issend",
                               send_args->backtrace,
                               origin,
                               type_size * send_args->count,
                               send_args->destination,
                               send_args->tag,
                               send_args->dbug_communicator,
                               world_rank);
  Original::PMPI_Ssend(send_args->buffer,
                       send_args->count,
                       send_args->datatype,
                       send_args->destination,
                       send_args->tag,
                       send_args->communicator);
  std::string request_id =
      Util::CreateResourceID("mpi_request",
                             arbiter_->ProcessID(),
                             reinterpret_cast<System::UInt64>
                             (send_args->request));
  arbiter_->MPIRequestComplete(arbiter_->ThreadID(),
                               request_id);
  arbiter_->UnregisterThread();
  if (System::MutexTryLock(&mpi_threads_mutex_) == 0) {
    mpi_threads_.erase(System::ThreadSelf());
    System::ThreadDetach(System::ThreadSelf());
    System::MutexUnlock(&mpi_threads_mutex_);
  }
  delete send_args;
  Handler::EnableFork();
  return NULL;
}

void *ISSendHelperWrapper(void *args) { return Handler::ISSendHelper(args); }

int Handler::PMPI_Issend(const void *buf,
                         int count,
                         MPI_Datatype datatype,
                         int dest,
                         int tag,
                         MPI_Comm comm,
                         MPI_Request *request) {
  int type_size;
  MPI_Type_size(datatype, &type_size);
  System::MutexLock(&mpi_comms_map_mutex_);
  int dbug_comm = mpi_comms_map_[comm];
  System::MutexUnlock(&mpi_comms_map_mutex_);
  if (comm == MPI_COMM_WORLD) { comm = dbug_comm_world_; }
  ISendArguments *args = new ISendArguments();
  args->buffer = buf;
  args->count = count;
  args->datatype = datatype;
  args->destination = dest;
  args->tag = tag;
  args->communicator = comm;
  args->dbug_communicator = dbug_comm;
  args->request = request;
  arbiter_->PostThreadRegistration(arbiter_->ThreadID(),
                                   "MPI_Issend",
                                   Util::GetBacktrace(),
                                   false,
                                   &args->pid,
                                   &args->tid);
  args->backtrace = Util::GetBacktrace();
  System::Thread tid;
  System::ThreadCreate(&tid, NULL, ISSendHelperWrapper, args);
  System::MutexLock(&mpi_threads_mutex_);
  mpi_threads_.insert(tid);
  System::MutexUnlock(&mpi_threads_mutex_);
  arbiter_->AdvanceClock(arbiter_->ThreadID(), args->tid);
  return MPI_SUCCESS;
}
#endif

#ifdef DBUG_HAVE_PMPI_TEST
int Handler::PMPI_Test(MPI_Request *request,
                       int *flag,
                       MPI_Status *status) {
  if (*request == MPI_REQUEST_NULL) {
    return Original::PMPI_Test(request, flag, status);
  }
  std::string request_id =
      Util::CreateResourceID("mpi_request",
                             arbiter_->ProcessID(),
                             reinterpret_cast<System::UInt64>(request));
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  *flag = arbiter_->MPITest(arbiter_->ThreadID(),
                            "MPI_Test",
                            Util::GetBacktrace(),
                            request_id,
                            world_rank);
  return MPI_SUCCESS;
}
#endif

#ifdef DBUG_HAVE_PMPI_TESTALL
int Handler::PMPI_Testall(int count,
                          MPI_Request *requests,
                          int *flag,
                          MPI_Status *statuses) {
  std::vector<std::string> arbiter_requests;
  // Gather all non-null requests into <arbiter_requests> and record
  // mapping from <arbiter_requests> to <requests> in <index_map> such
  // that for all i, arbiter_requests[i] matches requests[index_map[i]].
  for (int i = 0; i < count; i++) {
    if (requests[i] != MPI_REQUEST_NULL) {
      std::string request_id =
          Util::CreateResourceID("mpi_request",
                                 arbiter_->ProcessID(),
                                 reinterpret_cast<System::UInt64>
                                 (requests + i));
      arbiter_requests.push_back(request_id);
    }
  }
  if (arbiter_requests.size() == 0) {
    return Original::PMPI_Testall(count, requests, flag, statuses);
  }
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  *flag = arbiter_->MPITestAll(arbiter_->ThreadID(),
                               "MPI_Testall",
                               Util::GetBacktrace(),
                               arbiter_requests,
                               world_rank);
  return MPI_SUCCESS;
}
#endif

#ifdef DBUG_HAVE_PMPI_TESTANY
int Handler::PMPI_Testany(int count,
                          MPI_Request *requests,
                          int *index,
                          int *flag,
                          MPI_Status *status) {
  std::vector<int> index_map(count);
  std::vector<std::string> arbiter_requests;
  int n = 0;
  // Gather all non-null requests into <arbiter_requests> and record
  // mapping from <arbiter_requests> to <requests> in <index_map> such
  // that for all i, arbiter_requests[i] matches requests[index_map[i]].
  for (int i = 0; i < count; i++) {
    if (requests[i] != MPI_REQUEST_NULL) {
      std::string request_id =
          Util::CreateResourceID("mpi_request",
                                 arbiter_->ProcessID(),
                                 reinterpret_cast<System::UInt64>
                                 (requests + i));
      arbiter_requests.push_back(request_id);
      index_map[n++] = i;
    }
  }
  if (arbiter_requests.size() == 0) {
    return Original::PMPI_Testany(count, requests, index, flag, status);
  }
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  int result = arbiter_->MPITestAny(arbiter_->ThreadID(),
                                    "MPI_Testany",
                                    Util::GetBacktrace(),
                                    arbiter_requests,
                                    world_rank);
  if (result >= 0) {
    *flag = 1;
    *index = index_map[result];
  } else {
    *flag = 0;
    *index = MPI_UNDEFINED;
  }
  return MPI_SUCCESS;
}
#endif

#ifdef DBUG_HAVE_PMPI_TESTSOME
int Handler::PMPI_Testsome(int incount,
                           MPI_Request *requests,
                           int *outcount,
                           int *indices,
                           MPI_Status *statuses) {
  std::vector<int> index_map(incount);
  std::vector<std::string> arbiter_requests;
  int n = 0;
  // Gather all non-null requests into <arbiter_requests> and record
  // mapping from <arbiter_requests> to <requests> in <index_map> such
  // that for all i, arbiter_requests[i] matches requests[index_map[i]].
  for (int i = 0; i < incount; i++) {
    if (requests[i] != MPI_REQUEST_NULL) {
      std::string request_id =
          Util::CreateResourceID("mpi_request",
                                 arbiter_->ProcessID(),
                                 reinterpret_cast<System::UInt64>
                                 (requests + i));
      arbiter_requests.push_back(request_id);
      index_map[n++] = i;
    }
  }
  if (arbiter_requests.size() == 0) {
    return Original::PMPI_Testsome(incount,
                                   requests,
                                   outcount,
                                   indices,
                                   statuses);
  }
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  std::vector<int> result = arbiter_->MPITestSome(arbiter_->ThreadID(),
                                                  "MPI_Testsome",
                                                  Util::GetBacktrace(),
                                                  arbiter_requests,
                                                  world_rank);
  *outcount = result.size();
  for (int i = 0; i < *outcount; i++) { indices[i] = index_map[result[i]]; }
  return MPI_SUCCESS;
}
#endif

#ifdef DBUG_HAVE_PMPI_WAIT
int Handler::PMPI_Wait(MPI_Request *request,
                       MPI_Status *status) {
  if (*request == MPI_REQUEST_NULL) {
    return Original::PMPI_Wait(request, status);
  }
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  std::string request_id =
      Util::CreateResourceID("mpi_request",
                             arbiter_->ProcessID(),
                             reinterpret_cast<System::UInt64>(request));
  arbiter_->MPIWait(arbiter_->ThreadID(),
                    "MPI_Wait",
                    Util::GetBacktrace(),
                    request_id,
                    world_rank);
  return MPI_SUCCESS;
}
#endif

#ifdef DBUG_HAVE_PMPI_WAITALL
int Handler::PMPI_Waitall(int count,
                          MPI_Request *requests,
                          MPI_Status *statuses) {
  std::vector<std::string> arbiter_requests;
  // Gather all non-null requests into <arbiter_requests> and record
  // mapping from <arbiter_requests> to <requests> in <index_map> such
  // that for all i, arbiter_requests[i] matches requests[index_map[i]].
  for (int i = 0; i < count; i++) {
    if (requests[i] != MPI_REQUEST_NULL) {
      std::string request_id =
          Util::CreateResourceID("mpi_request",
                                 arbiter_->ProcessID(),
                                 reinterpret_cast<System::UInt64>
                                 (requests + i));
      arbiter_requests.push_back(request_id);
    }
  }
  if (arbiter_requests.size() == 0) {
    return Original::PMPI_Waitall(count, requests, statuses);
  }
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  arbiter_->MPIWaitAll(arbiter_->ThreadID(),
                       "MPI_Waitany",
                       Util::GetBacktrace(),
                       arbiter_requests,
                       world_rank);
  return MPI_SUCCESS;
}
#endif

#ifdef DBUG_HAVE_PMPI_WAITANY
int Handler::PMPI_Waitany(int count,
                          MPI_Request *requests,
                          int *index,
                          MPI_Status *status) {
  std::vector<int> index_map(count);
  std::vector<std::string> arbiter_requests;
  int n = 0;
  // Gather all non-null requests into <arbiter_requests> and record
  // mapping from <arbiter_requests> to <requests> in <index_map> such
  // that for all i, arbiter_requests[i] matches requests[index_map[i]].
  for (int i = 0; i < count; i++) {
    if (requests[i] != MPI_REQUEST_NULL) {
      std::string request_id =
          Util::CreateResourceID("mpi_request",
                                 arbiter_->ProcessID(),
                                 reinterpret_cast<System::UInt64>
                                 (requests + i));
      arbiter_requests.push_back(request_id);
      index_map[n++] = i;
    }
  }
  if (arbiter_requests.size() == 0) {
    return Original::PMPI_Waitany(count, requests, index, status);
  }
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  *index = arbiter_->MPIWaitAny(arbiter_->ThreadID(),
                                "MPI_Waitany",
                                Util::GetBacktrace(),
                                arbiter_requests,
                                world_rank);
  return MPI_SUCCESS;
}
#endif

#ifdef DBUG_HAVE_PMPI_WAITSOME
int Handler::PMPI_Waitsome(int incount,
                           MPI_Request *requests,
                           int *outcount,
                           int *indices,
                           MPI_Status *statuses) {
  std::vector<int> index_map(incount);
  std::vector<std::string> arbiter_requests;
  int n = 0;
  // Gather all non-null requests into <arbiter_requests> and record
  // mapping from <arbiter_requests> to <requests> in <index_map> such
  // that for all i, arbiter_requests[i] matches requests[index_map[i]].
  for (int i = 0; i < incount; i++) {
    if (requests[i] != MPI_REQUEST_NULL) {
      std::string request_id =
          Util::CreateResourceID("mpi_request",
                                 arbiter_->ProcessID(),
                                 reinterpret_cast<System::UInt64>
                                 (requests + i));
      arbiter_requests.push_back(request_id);
      index_map[n++] = i;
    }
  }
  if (arbiter_requests.size() == 0) {
    return Original::PMPI_Waitsome(incount,
                                   requests,
                                   outcount,
                                   indices,
                                   statuses);
  }
  int world_rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &world_rank);
  std::vector<int> result = arbiter_->MPIWaitSome(arbiter_->ThreadID(),
                                                  "MPI_Waitsome",
                                                  Util::GetBacktrace(),
                                                  arbiter_requests,
                                                  world_rank);
  *outcount = result.size();
  for (int i = 0; i < *outcount; i++) { indices[i] = index_map[result[i]]; }
  return MPI_SUCCESS;
}
#endif

}  // namespace interposition
}  // namespace dbug
