/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <assert.h>

#include <sstream>

#include "dbug/arbiter/process.h"
#include "dbug/common/system.h"
#include "dbug/common/util.h"
#include "dbug/interposition/handler.h"
#include "dbug/interposition/logging.h"
#include "dbug/interposition/original.h"
#include "src/interposition/gen/process-mgmt.cc"

using dbug::arbiter::Process;

namespace dbug {
namespace interposition {

#if DBUG_HAVE__EXIT
void Handler::_exit(int status) {
  if (Interpose()) {
    Interposition off(__FUNCTION__);
    arbiter_->SetProcessReturnStatus(arbiter_->ThreadID(),
                                     System::GetProcessID(),
                                     status);
    Finalize();
    Original::_exit(status);
  } else {
    DBUG_REQUIRE_SYMBOL(Original::_exit, _EXIT_TYPE, "_exit");
    Original::_exit(status);
  }
  assert(0);
}
#endif

#if DBUG_HAVE__EXIT
void Handler::_Exit(int status) {
  if (Interpose()) {
    Interposition off(__FUNCTION__);
    arbiter_->SetProcessReturnStatus(arbiter_->ThreadID(),
                                     System::GetProcessID(),
                                     status);
    Finalize();
    Original::_Exit(status);
  } else {
    DBUG_REQUIRE_SYMBOL(Original::_Exit, _EXIT_TYPE, "_Exit");
    Original::_Exit(status);
  }
  assert(0);
}
#endif

#if DBUG_HAVE_EXECL
int Handler::execl(const char *path, const char *arg, va_list ap) {
  if (Interpose()) {
    Interposition off(__FUNCTION__);
    char **argv = NULL;
    Util::ParseArgumentList(&argv, NULL, const_cast<char *>(arg), ap);
    switch (arbiter_->Choose(arbiter_->ThreadID(),
                             "execl",
                             Util::GetBacktrace(),
                             3)) {
      case 1:
        // Simulate system limit on open files reached.
        errno = ENFILE;
        DEBUG_EXTERNAL("%s, %p", "%d", path, argv, ENFILE);
        return -1;
      case 2:
        // Simulate lack of memory.
        errno = ENOMEM;
        DEBUG_EXTERNAL("%s, %p", "%d", path, argv, ENOMEM);
        return -1;
      default:
        int pid, tid;
        arbiter_->PostThreadRegistration(arbiter_->ThreadID(),
                                         "execl",
                                         Util::GetBacktrace(),
                                         false,
                                         &pid,
                                         &tid);
        std::stringstream pid_str;
        pid_str << pid;
        System::SetEnvironmentVariable("DBUG_PID", pid_str.str().c_str(), 1);
        std::stringstream tid_str;
        tid_str << tid;
        System::SetEnvironmentVariable("DBUG_TID", tid_str.str().c_str(), 1);
        ExportResourceIdentifiers();
        arbiter_->UnregisterAllThreads();
        int ret = Original::execv(path, argv);
        ERROR("execl() failed with %d.", ret);
    }
  } else {
    ERROR("Internal call to execl().");
  }
  assert(0);
}
#endif

#if DBUG_HAVE_EXECLE
int Handler::execle(const char *path, const char *arg, va_list ap) {
  if (Interpose()) {
    Interposition off(__FUNCTION__);
    char **argv = NULL;
    char **envp = NULL;
    Util::ParseArgumentList(&argv, &envp, const_cast<char *>(arg), ap);
    switch (arbiter_->Choose(arbiter_->ThreadID(),
                             "execle",
                             Util::GetBacktrace(),
                             3)) {
      case 1:
        // Simulate system limit on open files reached.
        errno = ENFILE;
        DEBUG_EXTERNAL("%s, %p, %p", "%d", path, argv, envp, ENFILE);
        return -1;
      case 2:
        // Simulate lack of memory.
        errno = ENOMEM;
        DEBUG_EXTERNAL("%s, %p, %p", "%d", path, argv, envp, ENOMEM);
        return -1;
      default:
        int pid, tid;
        arbiter_->PostThreadRegistration(arbiter_->ThreadID(),
                                         "execle",
                                         Util::GetBacktrace(),
                                         false,
                                         &pid,
                                         &tid);
        Util::SetupEnvironment(const_cast<char ***>(&envp), pid, tid);
        ExportResourceIdentifiers();
        arbiter_->UnregisterAllThreads();
        int ret = Original::execve(path, argv, envp);
        ERROR("execle() failed with %d.", ret);
    }
  } else {
    ERROR("Internal call to execle().");
  }
  assert(0);
}
#endif

#if DBUG_HAVE_EXECLP
int Handler::execlp(const char *file, const char *arg, va_list ap) {
  if (Interpose()) {
    Interposition off(__FUNCTION__);
    char **argv = NULL;
    Util::ParseArgumentList(&argv, NULL, const_cast<char *>(arg), ap);
    switch (arbiter_->Choose(arbiter_->ThreadID(),
                             "execlp",
                             Util::GetBacktrace(),
                             3)) {
      case 1:
        // Simulate system limit on open files reached.
        errno = ENFILE;
        DEBUG_EXTERNAL("%s, %p", "%d", file, arg, ENFILE);
        return -1;
      case 2:
        // Simulate lack of memory.
        errno = ENOMEM;
        DEBUG_EXTERNAL("%s, %p", "%d", file, arg, ENOMEM);
        return -1;
      default:
        int pid, tid;
        arbiter_->PostThreadRegistration(arbiter_->ThreadID(),
                                         "execlp",
                                         Util::GetBacktrace(),
                                         false,
                                         &pid,
                                         &tid);
        std::stringstream pid_str;
        pid_str << pid;
        System::SetEnvironmentVariable("DBUG_PID", pid_str.str().c_str(), 1);
        std::stringstream tid_str;
        tid_str << tid;
        System::SetEnvironmentVariable("DBUG_TID", tid_str.str().c_str(), 1);
        ExportResourceIdentifiers();
        arbiter_->UnregisterAllThreads();
        int ret = Original::execvp(file, argv);
        ERROR("execlp() failed with %d.", ret);
    }
  } else {
    ERROR("Internal call to execlp().");
  }
  assert(0);
}
#endif

#if DBUG_HAVE_EXECV
int Handler::execv(const char *path, char *const *argv) {
  if (Interpose()) {
    Interposition off(__FUNCTION__);
    switch (arbiter_->Choose(arbiter_->ThreadID(),
                             "execv",
                             Util::GetBacktrace(),
                             3)) {
      case 1:
        // Simulate system limit on open files reached.
        errno = ENFILE;
        DEBUG_EXTERNAL("%s, %p", "%d", path, argv, ENFILE);
        return -1;
      case 2:
        // Simulate lack of memory.
        errno = ENOMEM;
        DEBUG_EXTERNAL("%s, %p", "%d", path, argv, ENOMEM);
        return -1;
      default:
        int pid, tid;
        arbiter_->PostThreadRegistration(arbiter_->ThreadID(),
                                         "execv",
                                         Util::GetBacktrace(),
                                         false,
                                         &pid,
                                         &tid);
        std::stringstream pid_str;
        pid_str << pid;
        System::SetEnvironmentVariable("DBUG_PID", pid_str.str().c_str(), 1);
        std::stringstream tid_str;
        tid_str << tid;
        System::SetEnvironmentVariable("DBUG_TID", tid_str.str().c_str(), 1);
        ExportResourceIdentifiers();
        arbiter_->UnregisterAllThreads();
        int ret = Original::execv(path, argv);
        ERROR("execv() failed with %d.", ret);
    }
  } else {
    ERROR("Internal call to execv().");
  }
  assert(0);
}
#endif

#if DBUG_HAVE_EXECVE
int Handler::execve(const char *path, char *const *argv, char *const *envp) {
  if (Interpose()) {
    Interposition off(__FUNCTION__);
    switch (arbiter_->Choose(arbiter_->ThreadID(),
                             "execve",
                             Util::GetBacktrace(),
                             3)) {
      case 1:
        // Simulate system limit on open files reached.
        errno = ENFILE;
        DEBUG_EXTERNAL("%s, %p, %p", "%d", path, argv, envp, ENFILE);
        return -1;
      case 2:
        // Simulate lack of memory.
        errno = ENOMEM;
        DEBUG_EXTERNAL("%s, %p, %p", "%d", path, argv, envp, ENOMEM);
        return -1;
      default:
        int pid, tid;
        arbiter_->PostThreadRegistration(arbiter_->ThreadID(),
                                         "execve",
                                         Util::GetBacktrace(),
                                         false,
                                         &pid,
                                         &tid);
        Util::SetupEnvironment(const_cast<char ***>(&envp), pid, tid);
        ExportResourceIdentifiers();
        arbiter_->UnregisterAllThreads();
        Original::execve(path, argv, envp);
        printf("execve() failed with %s.", strerror(errno));
    }
  } else {
    ERROR("Internal call to execve().");
  }
  assert(0);
}
#endif

#if DBUG_HAVE_EXECVP
int Handler::execvp(const char *file, char *const *argv) {
  if (Interpose()) {
    Interposition off(__FUNCTION__);
    switch (arbiter_->Choose(arbiter_->ThreadID(),
                             "execvp",
                             Util::GetBacktrace(),
                             3)) {
      case 1:
        // Simulate system limit on open files reached.
        errno = ENFILE;
        DEBUG_EXTERNAL("%s, %p", "%d", file, argv, ENFILE);
        return -1;
      case 2:
        // Simulate lack of memory.
        errno = ENOMEM;
        DEBUG_EXTERNAL("%s, %p", "%d", file, argv, ENOMEM);
        return -1;
      default:
        int pid, tid;
        arbiter_->PostThreadRegistration(arbiter_->ThreadID(),
                                         "execvp",
                                         Util::GetBacktrace(),
                                         false,
                                         &pid,
                                         &tid);
        std::stringstream pid_str;
        pid_str << pid;
        System::SetEnvironmentVariable("DBUG_PID", pid_str.str().c_str(), 1);
        std::stringstream tid_str;
        tid_str << tid;
        System::SetEnvironmentVariable("DBUG_TID", tid_str.str().c_str(), 1);
        ExportResourceIdentifiers();
        arbiter_->UnregisterAllThreads();
        int ret = Original::execvp(file, argv);
        ERROR("execvp() failed with %d.", ret);
    }
  } else {
    ERROR("Internal call to execvp().");
  }
  assert(0);
}
#endif

#if DBUG_HAVE_EXIT
void Handler::exit(int status) {
  if (Interpose()) {
    Interposition off(__FUNCTION__);
    arbiter_->SetProcessReturnStatus(arbiter_->ThreadID(),
                                     System::GetProcessID(),
                                     status);
    Finalize();
    Original::exit(status);
  } else {
    DBUG_REQUIRE_SYMBOL(Original::exit, EXIT_TYPE, "exit");
    Original::exit(status);
  }
  assert(0);
}
#endif

#if DBUG_HAVE_FORK
pid_t Handler::fork() {
  switch (arbiter_->Choose(arbiter_->ThreadID(),
                           "fork",
                           Util::GetBacktrace(),
                           3)) {
    case 1:
      // Simulate lack of resources.
      errno = EAGAIN;
      return -1;
    case 2:
      // Simulate lack of memory.
      errno = ENOMEM;
      return -1;
    default:
      // Start registration of the child process.
      int pid, tid;
      arbiter_->PostThreadRegistration(arbiter_->ThreadID(),
                                       "fork",
                                       Util::GetBacktrace(),
                                       true,
                                       &pid,
                                       &tid);
      if (arbiter_->nrunning_threads() != 1) {
        WARNING("Forking a multi-threaded program is dangerous.");
      }
      EnableFork();
      System::RWLockWriteLock(&fork_rwlock_);
      pid_t ret = Original::fork();
      System::RWLockUnlock(&fork_rwlock_);
      DisableFork();
      // If fork() failed, abort registration of the child process.
      if (ret == -1) { arbiter_->CancelThreadRegistration(tid); }
      // If fork() succeeded, complete registration of the child process.
      if (ret == 0) {
        Reset();
        arbiter_->CompleteThreadRegistration(pid, tid);
      }
      return ret;
  }
}
#endif

#if DBUG_HAVE_GETPID
pid_t Handler::getpid() {
  return Original::getpid();
}
#endif

#if DBUG_HAVE_POSIX_SPAWN
pid_t Handler::posix_spawn(
    pid_t * DBUG_RESTRICT pid,
    const char *DBUG_RESTRICT path,
    const posix_spawn_file_actions_t * file_actions,
    const posix_spawnattr_t * DBUG_RESTRICT attrp,
    char *const *DBUG_RESTRICT argv,
    char *const *DBUG_RESTRICT envp) {
  switch (arbiter_->Choose(arbiter_->ThreadID(),
                           "posix_spawn",
                           Util::GetBacktrace(),
                           3)) {
    case 1:
      // Simulate lack of resources.
      errno = EAGAIN;
      return EAGAIN;
    case 2:
      // Simulate lack of memory.
      errno = ENOMEM;
      return ENOMEM;
    default:
      int dbug_pid, dbug_tid;
      arbiter_->PostThreadRegistration(arbiter_->ThreadID(),
                                       "posix_spawn",
                                       Util::GetBacktrace(),
                                       false,
                                       &dbug_pid,
                                       &dbug_tid);
      Util::SetupEnvironment(const_cast<char ***>(&envp), dbug_pid, dbug_tid);
      return Original::posix_spawn(pid, path, file_actions, attrp, argv, envp);
  }
}
#endif

#if DBUG_HAVE_POSIX_SPAWNP
pid_t Handler::posix_spawnp(
    pid_t * DBUG_RESTRICT pid,
    const char *DBUG_RESTRICT file,
    const posix_spawn_file_actions_t * file_actions,
    const posix_spawnattr_t * DBUG_RESTRICT attrp,
    char *const *DBUG_RESTRICT argv,
    char *const *DBUG_RESTRICT envp) {
  switch (arbiter_->Choose(arbiter_->ThreadID(),
                           "posix_spawnp",
                           Util::GetBacktrace(),
                           3)) {
    case 1:
      // Simulate lack of resources.
      errno = EAGAIN;
      return EAGAIN;
    case 2:
      // Simulate lack of memory.
      errno = ENOMEM;
      return ENOMEM;
    default:
      int dbug_pid, dbug_tid;
      arbiter_->PostThreadRegistration(arbiter_->ThreadID(),
                                       "posix_spawnp",
                                       Util::GetBacktrace(),
                                       false,
                                       &dbug_pid,
                                       &dbug_tid);
      Util::SetupEnvironment(const_cast<char ***>(&envp), dbug_pid, dbug_tid);
      return Original::posix_spawnp(pid, file, file_actions, attrp, argv, envp);
  }
}
#endif

#if DBUG_HAVE_SETPGID
pid_t Handler::setpgid(pid_t pid, pid_t pgid) {
  pid_t ret = Original::setpgid(pid, pgid);
  if (ret == -1) { return ret; }
  // If pid or pgid equal 0, set them to the value of System::GetProcessID().
  if (pid == 0) { pid = System::GetProcessID(); }
  if (pgid == 0) { pgid = System::GetProcessID(); }
  arbiter_->SetProcessPGID(arbiter_->ThreadID(), pid, pgid);
  return ret;
}
#endif

#if DBUG_HAVE_SETPGRP
pid_t Handler::setpgrp() {
  pid_t ret = Original::setpgrp();
  arbiter_->SetProcessPGID(arbiter_->ThreadID(),
                           System::GetProcessID(),
                           System::GetProcessGroupID());
  return ret;
}
#endif

#if DBUG_HAVE_SETSID
pid_t Handler::setsid() {
  pid_t ret = Original::setsid();
  if (ret == -1) { return ret; }
  arbiter_->SetProcessPGID(arbiter_->ThreadID(),
                           System::GetProcessID(),
                           System::GetProcessGroupID());
  return ret;
}
#endif

#if DBUG_HAVE_WAIT
pid_t Handler::wait(System::WaitStatus stat_loc) {
  dbug::thrift::ProcessWaitReturn wait_return =
      arbiter_->ProcessWait(arbiter_->ThreadID(),
                            "wait",
                            Util::GetBacktrace(),
                            -1,
                            0);
  if (stat_loc != NULL) {
    switch (wait_return.state) {
      case Process::TERMINATED:
      case Process::SIGNALED:
        *reinterpret_cast<int *>(stat_loc) =
            W_EXITCODE(wait_return.return_status, wait_return.signal);
        break;
      case Process::STOPPED:
        *reinterpret_cast<int *>(stat_loc) =
            W_STOPCODE(wait_return.signal);
        break;
      case Process::CONTINUED:
        *reinterpret_cast<int *>(stat_loc) =
            wait_return.signal;
        break;
    }
  }
  return wait_return.value;
}
#endif

#if DBUG_HAVE_WAITPID
pid_t Handler::waitpid(pid_t pid,
                       int *stat_loc,
                       int options) {
  dbug::thrift::ProcessWaitReturn wait_return =
      arbiter_->ProcessWait(arbiter_->ThreadID(),
                            "waitpid",
                            Util::GetBacktrace(),
                            pid,
                            options);
  if (stat_loc != NULL) {
    switch (wait_return.state) {
      case Process::TERMINATED:
      case Process::SIGNALED:
        *stat_loc = W_EXITCODE(wait_return.return_status, wait_return.signal);
        break;
      case Process::STOPPED:
        *stat_loc = W_STOPCODE(wait_return.signal);
        break;
      case Process::CONTINUED:
        *stat_loc = wait_return.signal;
        break;
    }
  }
  return wait_return.value;
}
#endif

}  // namespace interposition
}  // namespace dbug
