/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <sstream>

#include "dbug/common/system.h"
#include "dbug/common/util.h"
#include "dbug/interposition/handler.h"
#include "dbug/interposition/logging.h"
#include "dbug/interposition/original.h"
#include "src/interposition/gen/pthread-barrier.cc"

namespace dbug {
namespace interposition {

#if DBUG_HAVE_PTHREAD_BARRIER_DESTROY
int Handler::pthread_barrier_destroy(pthread_barrier_t * barrier) {
  std::string resource_id;
  System::MutexLock(&barrier_map_mutex_);
  if (barrier_map_.find(barrier) != barrier_map_.end()) {
    resource_id = barrier_map_[barrier];
  } else {
    int id = Util::NewIdentifier();
    resource_id =
        Util::CreateResourceID("invalid-barrier", arbiter_->ProcessID(), id);
    barrier_map_[barrier] = resource_id;
  }
  System::MutexUnlock(&barrier_map_mutex_);
  return arbiter_->BarrierDelete(arbiter_->ThreadID(),
                                 "pthread_barrier_destroy",
                                 Util::GetBacktrace(),
                                 resource_id);
}
#endif

#if DBUG_HAVE_PTHREAD_BARRIER_INIT
int Handler::pthread_barrier_init(pthread_barrier_t * barrier,
                                  const pthread_barrierattr_t * /* attr */,
                                  unsigned count) {
  switch (arbiter_->Choose(arbiter_->ThreadID(),
                           "pthread_barrier_init",
                           Util::GetBacktrace(),
                           3)) {
    case 1:
      // Simulate lack of non-memory resources.
      errno = EAGAIN;
      return EAGAIN;
    case 2:
      // Simulate lack of memory.
      errno = ENOMEM;
      return ENOMEM;
    default:
      std::string resource_id;
      System::MutexLock(&barrier_map_mutex_);
      if (barrier_map_.find(barrier) != barrier_map_.end()) {
        resource_id = barrier_map_[barrier];
      } else {
        int id = Util::NewIdentifier();
        resource_id =
            Util::CreateResourceID("barrier", arbiter_->ProcessID(), id);
        barrier_map_[barrier] = resource_id;
      }
      System::MutexUnlock(&barrier_map_mutex_);
      return arbiter_->BarrierCreate(arbiter_->ThreadID(),
                                     "pthread_barrier_init",
                                     Util::GetBacktrace(),
                                     resource_id,
                                     count);
  }
}
#endif

#if DBUG_HAVE_PTHREAD_BARRIER_WAIT
int Handler::pthread_barrier_wait(pthread_barrier_t * barrier) {
  std::string resource_id;
  System::MutexLock(&barrier_map_mutex_);
  if (barrier_map_.find(barrier) != barrier_map_.end()) {
    resource_id = barrier_map_[barrier];
  } else {
    int id = Util::NewIdentifier();
    resource_id =
        Util::CreateResourceID("invalid-barrier", arbiter_->ProcessID(), id);
    barrier_map_[barrier] = resource_id;
  }
  System::MutexUnlock(&barrier_map_mutex_);
  return arbiter_->BarrierWait(arbiter_->ThreadID(),
                               "pthread_barrier_wait",
                               Util::GetBacktrace(),
                               resource_id);
}
#endif

}  // namespace interposition
}  // namespace dbug
