/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

// TODO(jsimsa): Simulate cond_wait(), cond_signal(), and
// cond_broadcast() errors.

// TODO(jsimsa): Model virtual time of cond_timedwait().

#include "dbug/common/system.h"
#include "dbug/common/util.h"
#include "dbug/interposition/handler.h"
#include "dbug/interposition/logging.h"
#include "dbug/interposition/original.h"
#include "src/interposition/gen/pthread-cvar.cc"

namespace dbug {
namespace interposition {

#if DBUG_HAVE_PTHREAD_COND_BROADCAST
int Handler::pthread_cond_broadcast(pthread_cond_t *cond) {
  std::string resource_id;
  System::MutexLock(&cvar_map_mutex_);
  if (cvar_map_.find(cond) != cvar_map_.end()) {
    resource_id = cvar_map_[cond];
  } else {
    int id = Util::NewIdentifier();
    resource_id =
        Util::CreateResourceID("invalid-cvar", arbiter_->ProcessID(), id);
    cvar_map_[cond] = resource_id;
  }
  System::MutexUnlock(&cvar_map_mutex_);
  return arbiter_->CVarBroadcast(arbiter_->ThreadID(),
                                 "pthread_cond_broadcast",
                                 Util::GetBacktrace(),
                                 resource_id);
  return 0;
}
#endif

#if DBUG_HAVE_PTHREAD_COND_DESTROY
int Handler::pthread_cond_destroy(pthread_cond_t *cond) {
  std::string resource_id;
  System::MutexLock(&cvar_map_mutex_);
  if (cvar_map_.find(cond) != cvar_map_.end()) {
    resource_id = cvar_map_[cond];
  } else {
    int id = Util::NewIdentifier();
    resource_id =
        Util::CreateResourceID("invalid-cvar", arbiter_->ProcessID(), id);
    cvar_map_[cond] = resource_id;
  }
  System::MutexUnlock(&cvar_map_mutex_);
  return arbiter_->CVarDelete(arbiter_->ThreadID(),
                              "pthread_cond_destroy",
                              Util::GetBacktrace(),
                              resource_id);
}
#endif

#if DBUG_HAVE_PTHREAD_COND_INIT
int Handler::pthread_cond_init(pthread_cond_t *cond,
                               const pthread_condattr_t * /* attr */) {
  switch (arbiter_->Choose(arbiter_->ThreadID(),
                           "pthread_cond_init",
                           Util::GetBacktrace(),
                           3)) {
    case 1:
      // Simulate lack of non-memory resources.
      errno = EAGAIN;
      return EAGAIN;
    case 2:
      // Simulate lack of memory.
      errno = ENOMEM;
      return ENOMEM;
    default:
      std::string resource_id;
      System::MutexLock(&cvar_map_mutex_);
      if (cvar_map_.find(cond) != cvar_map_.end()) {
        resource_id = cvar_map_[cond];
      } else {
        int id = Util::NewIdentifier();
        resource_id =
            Util::CreateResourceID("cvar", arbiter_->ProcessID(), id);
        cvar_map_[cond] = resource_id;
      }
      System::MutexUnlock(&cvar_map_mutex_);
      return arbiter_->CVarCreate(arbiter_->ThreadID(),
                                  "pthread_cond_init",
                                  Util::GetBacktrace(),
                                  resource_id);
  }
}
#endif

#if DBUG_HAVE_PTHREAD_COND_SIGNAL
int Handler::pthread_cond_signal(pthread_cond_t *cond) {
  std::string resource_id;
  System::MutexLock(&cvar_map_mutex_);
  if (cvar_map_.find(cond) != cvar_map_.end()) {
    resource_id = cvar_map_[cond];
  } else {
    int id = Util::NewIdentifier();
    resource_id =
        Util::CreateResourceID("invalid-cvar", arbiter_->ProcessID(), id);
    cvar_map_[cond] = resource_id;
  }
  System::MutexUnlock(&cvar_map_mutex_);
  return arbiter_->CVarSignal(arbiter_->ThreadID(),
                              "pthread_cond_signal",
                              Util::GetBacktrace(),
                              resource_id);
}
#endif

#if DBUG_HAVE_PTHREAD_COND_TIMEDWAIT
int Handler::pthread_cond_timedwait(pthread_cond_t *cond,
                                    pthread_mutex_t *mutex,
                                    const struct timespec * /* abstime */) {
  switch (arbiter_->Choose(arbiter_->ThreadID(),
                           "pthread_cond_timedwait",
                           Util::GetBacktrace(),
                           2)) {
    case 1:
      // Simulate time out.
      errno = ETIMEDOUT;
      return ETIMEDOUT;
    default:
      return Handler::pthread_cond_wait(cond, mutex);
  }
}
#endif

#if DBUG_HAVE_PTHREAD_COND_WAIT
int Handler::pthread_cond_wait(pthread_cond_t *cond, pthread_mutex_t *mutex) {
  std::string cvar_id;
  System::MutexLock(&cvar_map_mutex_);
  if (cvar_map_.find(cond) != cvar_map_.end()) {
    cvar_id = cvar_map_[cond];
  } else {
    int id = Util::NewIdentifier();
    cvar_id =
        Util::CreateResourceID("invalid-cvar", arbiter_->ProcessID(), id);
    cvar_map_[cond] = cvar_id;
  }
  System::MutexUnlock(&cvar_map_mutex_);
  std::string mutex_id;
  System::MutexLock(&mutex_map_mutex_);
  if (mutex_map_.find(mutex) != mutex_map_.end()) {
    mutex_id = mutex_map_[mutex];
  } else {
    int id = Util::NewIdentifier();
    mutex_id =
        Util::CreateResourceID("invalid-mutex", arbiter_->ProcessID(), id);
    mutex_map_[mutex] = mutex_id;
  }
  System::MutexUnlock(&mutex_map_mutex_);
  return arbiter_->CVarWait(arbiter_->ThreadID(),
                            "pthread_cond_wait",
                            Util::GetBacktrace(),
                            cvar_id,
                            mutex_id);
}
#endif

}  // namespace interposition
}  // namespace dbug
