/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "dbug/arbiter/thread.h"
#include "dbug/common/system.h"
#include "dbug/common/util.h"
#include "dbug/interposition/handler.h"
#include "dbug/interposition/logging.h"
#include "dbug/interposition/original.h"
#include "src/interposition/gen/pthread-mgmt.cc"

using dbug::arbiter::Thread;

namespace dbug {
namespace interposition {

#if DBUG_HAVE_PTHREAD_CREATE
struct ThreadCreateArgs {
  void *(*start_routine)(void *);  // NOLINT
  void *args;
  void *entry_point;
  int pid;
  int tid;
  sem_t *semaphore;
};

void *Handler::ThreadCreateHelper(void *args) {
  ThreadCreateArgs *helper_args = reinterpret_cast<ThreadCreateArgs *>(args);

  // Setup the initial state of the new thread.
  int cancel_state = PTHREAD_CANCEL_ENABLE;
  System::KeySetSpecific(interpose_key_, helper_args->entry_point);
  System::KeySetSpecific(cancel_state_key_, &cancel_state);
  DisableCancellation();
  DisableFork();
  arbiter_->CompleteThreadRegistration(helper_args->pid, helper_args->tid);
  // Once the registration is complete, the parent thread can resume
  // its execution.
  System::SemaphorePost(helper_args->semaphore);

  // Run the original start_routine.
  Enable();
  void *ret = helper_args->start_routine(helper_args->args);
  Disable(helper_args->entry_point);

  // When it finishes, clean up after yourself.
  delete helper_args;
  arbiter_->UnregisterThread();
  EnableFork();
  EnableCancellation();
  System::KeySetSpecific(interpose_key_, NULL);
  return ret;
}

int Handler::pthread_create(pthread_t *thread,
                            const pthread_attr_t *attr,
                            void * (*start_routine)(void *),  // NOLINT
                            void *args) {
  switch (arbiter_->Choose(arbiter_->ThreadID(),
                           "pthread_create",
                           Util::GetBacktrace(),
                           2)) {
    case 1:
      // Simulate lack of resources.
      errno = EAGAIN;
      return EAGAIN;
    default:
      // Setup the argument for the child thread. The allocated dynamic memory
      // is freed in dbug_pthread_create or below in case that the
      // pthread_create call fails.
      sem_t semaphore;
      System::SemaphoreInit(&semaphore, 0, 0);
      ThreadCreateArgs *helper_args = new ThreadCreateArgs();
      helper_args->start_routine = start_routine;
      helper_args->args = args;
      helper_args->entry_point = System::KeyGetSpecific(interpose_key_);
      arbiter_->PostThreadRegistration(arbiter_->ThreadID(),
                                       "pthread_create",
                                       Util::GetBacktrace(),
                                       false,
                                       &helper_args->pid,
                                       &helper_args->tid);
      helper_args->semaphore = &semaphore;
      int ret = Original::pthread_create(thread,
                                         attr,
                                         ThreadCreateHelper,
                                         helper_args);
      if (ret != 0) {
        arbiter_->CancelThreadRegistration(helper_args->tid);
        delete helper_args;
      } else {
        // Wait for the child thread to complete its registration.
        System::SemaphoreWait(&semaphore);
      }
      System::SemaphoreDestroy(&semaphore);
      return ret;
  }
}
#endif

#if DBUG_HAVE_PTHREAD_DETACH
int Handler::pthread_detach(pthread_t thread) {
  int ret = Original::pthread_detach(thread);
  arbiter_->SetThreadState(arbiter_->ThreadID(),
                           arbiter_->ThreadID(thread),
                           Thread::DETACHED);
  return ret;
}
#endif

#if DBUG_HAVE_PTHREAD_EXIT
void Handler::pthread_exit(void *value_ptr) {
  if (Interpose()) {
    Disable(__FUNCTION__);
    arbiter_->UnregisterThread();
    EnableFork();
    EnableCancellation();
    System::KeySetSpecific(interpose_key_, NULL);
    Original::pthread_exit(value_ptr);
  } else {
    DBUG_REQUIRE_SYMBOL(Original::pthread_exit,
                        PTHREAD_EXIT_TYPE,
                        "pthread_exit");
    Original::pthread_exit(value_ptr);
    DEBUG_INTERNAL("%p", "void", value_ptr);
  }
  assert(0);
}
#endif

#if DBUG_HAVE_PTHREAD_JOIN
int Handler::pthread_join(pthread_t thread, void **value_ptr) {
  arbiter_->ThreadJoin(arbiter_->ThreadID(),
                       "pthread_join",
                       Util::GetBacktrace(),
                       arbiter_->ThreadID(thread));
  int ret = Original::pthread_join(thread, value_ptr);
  return ret;
}
#endif

#if DBUG_HAVE_PTHREAD_SELF
pthread_t Handler::pthread_self() {
  pthread_t ret;
  if (Interpose()) {
    Interposition off(__FUNCTION__);
    ret = Original::pthread_self();
    DEBUG_EXTERNAL("void", "%u", ret);
    return ret;
  } else {
    // If the Original::pthread_self pointer has not been initialized,
    // return 0 to avoid a segmentation fault.
    ret = (Original::pthread_self ? Original::pthread_self() : 0);
    DEBUG_INTERNAL("void", "%u", ret);
    return ret;
  }
}
#endif

}  // namespace interposition
}  // namespace dbug
