/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "dbug/common/system.h"
#include "dbug/common/util.h"
#include "dbug/interposition/handler.h"
#include "dbug/interposition/logging.h"
#include "dbug/interposition/original.h"
#include "src/interposition/gen/pthread-spin.cc"

namespace dbug {
namespace interposition {

#if DBUG_HAVE_PTHREAD_SPIN_DESTROY
int Handler::pthread_spin_destroy(pthread_spinlock_t *spinlock) {
  std::string resource_id;
  System::MutexLock(&spin_map_mutex_);
  if (spin_map_.find(spinlock) != spin_map_.end()) {
    resource_id = spin_map_[spinlock];
  } else {
    int id = Util::NewIdentifier();
    resource_id =
        Util::CreateResourceID("invalid-spinlock", arbiter_->ProcessID(), id);
    spin_map_[spinlock] = resource_id;
  }
  System::MutexUnlock(&spin_map_mutex_);
  return arbiter_->SpinDelete(arbiter_->ThreadID(),
                              "pthread_spin_destroy",
                              Util::GetBacktrace(),
                              resource_id);
}
#endif

#if DBUG_HAVE_PTHREAD_SPIN_INIT
int Handler::pthread_spin_init(pthread_spinlock_t *spinlock,
                               int /* pshared */) {
  switch (arbiter_->Choose(arbiter_->ThreadID(),
                           "pthread_spin_init",
                           Util::GetBacktrace(),
                           3)) {
    case 1:
      // Simulate lack of non-memory resources.
      errno = EAGAIN;
      return EAGAIN;
    case 2:
      // Simulate lack of memory.
      errno = ENOMEM;
      return ENOMEM;
    default:
      std::string resource_id;
      System::MutexLock(&spin_map_mutex_);
      if (spin_map_.find(spinlock) != spin_map_.end()) {
        resource_id = spin_map_[spinlock];
      } else {
        int id = Util::NewIdentifier();
        resource_id =
            Util::CreateResourceID("spinlock", arbiter_->ProcessID(), id);
        spin_map_[spinlock] = resource_id;
      }
      System::MutexUnlock(&spin_map_mutex_);
      return arbiter_->SpinCreate(arbiter_->ThreadID(),
                                  "pthread_spin_init",
                                  Util::GetBacktrace(),
                                  resource_id);
  }
}
#endif

#if DBUG_HAVE_PTHREAD_SPIN_LOCK
int Handler::pthread_spin_lock(pthread_spinlock_t *spinlock) {
  std::string resource_id;
  System::MutexLock(&spin_map_mutex_);
  if (spin_map_.find(spinlock) != spin_map_.end()) {
    resource_id = spin_map_[spinlock];
  } else {
    int id = Util::NewIdentifier();
    resource_id =
        Util::CreateResourceID("invalid-spinlock", arbiter_->ProcessID(), id);
    spin_map_[spinlock] = resource_id;
  }
  System::MutexUnlock(&spin_map_mutex_);
  return arbiter_->SpinLock(arbiter_->ThreadID(),
                            "pthread_spin_lock",
                            Util::GetBacktrace(),
                            resource_id);
}
#endif

#if DBUG_HAVE_PTHREAD_SPIN_TRYLOCK
int Handler::pthread_spin_trylock(pthread_spinlock_t *spinlock) {
  std::string resource_id;
  System::MutexLock(&spin_map_mutex_);
  if (spin_map_.find(spinlock) != spin_map_.end()) {
    resource_id = spin_map_[spinlock];
  } else {
    int id = Util::NewIdentifier();
    resource_id =
        Util::CreateResourceID("invalid-spinlock", arbiter_->ProcessID(), id);
    spin_map_[spinlock] = resource_id;
  }
  System::MutexUnlock(&spin_map_mutex_);
  return arbiter_->SpinTryLock(arbiter_->ThreadID(),
                               "pthread_spin_trylock",
                               Util::GetBacktrace(),
                               resource_id);
}
#endif

#if DBUG_HAVE_PTHREAD_SPIN_UNLOCK
int Handler::pthread_spin_unlock(pthread_spinlock_t *spinlock) {
  std::string resource_id;
  System::MutexLock(&spin_map_mutex_);
  if (spin_map_.find(spinlock) != spin_map_.end()) {
    resource_id = spin_map_[spinlock];
  } else {
    int id = Util::NewIdentifier();
    resource_id =
        Util::CreateResourceID("invalid-spinlock", arbiter_->ProcessID(), id);
    spin_map_[spinlock] = resource_id;
  }
  System::MutexUnlock(&spin_map_mutex_);
  return arbiter_->SpinUnlock(arbiter_->ThreadID(),
                              "pthread_spin_unlock",
                              Util::GetBacktrace(),
                              resource_id);
}
#endif

}  // namespace interposition
}  // namespace dbug
