/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include "dbug/interposition/remote-arbiter.h"

#include <thrift/Thrift.h>

#include <boost/scoped_ptr.hpp>
#include <list>
#include <string>
#include <utility>

#include "dbug/arbiter/process.h"
#include "dbug/common/system.h"
#include "dbug/interposition/arbiter-connection.h"
#include "dbug/interposition/logging.h"
#include "src/thrift/arbiter/arbiter_types.h"

using dbug::arbiter::Process;

using dbug::thrift::Location;
using dbug::thrift::ThreadCreateReturn;

namespace dbug {
namespace interposition {

RemoteArbiter::RemoteArbiter(std::string hostname, int port)
    : DefaultRemoteArbiter(hostname, port) {}

void RemoteArbiter::CancelThreadRegistration(const int tid) {
  DefaultRemoteArbiter::UnregisterThread(tid);
}

int RemoteArbiter::Choose(const int tid,
                          const std::string &func,
                          const std::vector<Location> &backtrace,
                          const int n) {
  int ret = 0;
  if (System::CompareStrings(func.c_str(), "dbug_choose") == 0) {
    ret = DefaultRemoteArbiter::Choose(tid, func, backtrace, n);
  }
  return ret;
}

void RemoteArbiter::CompleteThreadRegistration(const int pid,
                                               const int tid) {
  DefaultRemoteArbiter::SetProcessInfo(
      tid, System::GetProcessID(), System::GetParentProcessID(),
      System::GetProcessGroupID(), Process::RUNNING, 0, 0);
  System::MutexLock(&processes_mutex_);
  System::MutexLock(&threads_mutex_);
  processes_[System::ThreadSelf()] = pid;
  ThreadInfo *new_thread = new ThreadInfo();
  new_thread->id = tid;
  new_thread->running = true;
  threads_[System::ThreadSelf()] = new_thread;
  nrunning_threads_++;
  System::MutexUnlock(&threads_mutex_);
  System::MutexUnlock(&processes_mutex_);
}

void RemoteArbiter::PostThreadRegistration(
    const int tid,
    const std::string &func,
    const std::vector<Location> &backtrace,
    const bool create_process,
    int *new_pid,
    int *new_tid) {
  ThreadCreateReturn ret =
      DefaultRemoteArbiter::ThreadCreate(tid, func, backtrace, create_process);
  *new_pid = ret.pid;
  *new_tid = ret.tid;
}

void RemoteArbiter::ProcessSignal(const int tid,
                                  const std::string &func,
                                  const std::vector<Location> &backtrace,
                                  const int pid,
                                  const int signal) {
  System::MutexLock(&threads_mutex_);
  bool terminate = false;
  for (std::pair<System::Thread, ThreadInfo *> pair : threads_) {
    ThreadInfo *thread = pair.second;
    if (thread->id == tid && thread->running) { terminate = true; }
  }
  System::MutexUnlock(&threads_mutex_);
  if (!terminate) { return; }
  bool success = false;
  try {
    boost::scoped_ptr<ArbiterConnection>
        connection(new ArbiterConnection(DefaultRemoteArbiter::hostname_,
                                         DefaultRemoteArbiter::port_));
    success =
        connection->client()->ProcessSignal(tid, func, backtrace, pid, signal);
    MESSAGE("Thread %d called ProcessSignal().", ThreadID());
  } catch (const apache::thrift::TException &tx) {
    WARNING("Connection with the arbiter failed: %s", tx.what());
  }
  if (!success) { DieQuietly(); }
}

void RemoteArbiter::ProcessTerminate(const int tid,
                                     const std::string &func,
                                     const std::vector<Location> &backtrace,
                                     const int pid) {
  System::MutexLock(&threads_mutex_);
  bool terminate = false;
  for (std::pair<System::Thread, ThreadInfo *> pair : threads_) {
    ThreadInfo *thread = pair.second;
    if (thread->id == tid && thread->running) { terminate = true; }
  }
  System::MutexUnlock(&threads_mutex_);
  if (!terminate) { return; }
  bool success = false;
  try {
    boost::scoped_ptr<ArbiterConnection>
        connection(new ArbiterConnection(DefaultRemoteArbiter::hostname_,
                                         DefaultRemoteArbiter::port_));
    success = connection->client()->ProcessTerminate(tid, func, backtrace, pid);
    MESSAGE("Thread %d called ProcessTerminate().", ThreadID());
  } catch (const apache::thrift::TException &tx) {
    WARNING("Connection with the arbiter failed: %s", tx.what());
  }
  if (!success) { DieQuietly(); }
}

void RemoteArbiter::RegisterThread() {
  System::MutexLock(&threads_mutex_);
  if (threads_.find(System::ThreadSelf()) == threads_.end()) {
    System::MutexUnlock(&threads_mutex_);
    dbug::thrift::RegisterThreadReturn ret =
        DefaultRemoteArbiter::RegisterThread(0, true);
    DefaultRemoteArbiter::SetProcessInfo(
        ret.tid,
        System::GetProcessID(),
        System::GetParentProcessID(),
        System::GetProcessGroupID(),
        Process::RUNNING,
        0,
        0);
    System::MutexLock(&processes_mutex_);
    System::MutexLock(&threads_mutex_);
    processes_[System::ThreadSelf()] = ret.pid;
    ThreadInfo *new_thread = new ThreadInfo();
    new_thread->id = ret.tid;
    new_thread->running = true;
    threads_[System::ThreadSelf()] = new_thread;
    nrunning_threads_++;
    System::MutexUnlock(&threads_mutex_);
    System::MutexUnlock(&processes_mutex_);
  } else {
    System::MutexUnlock(&threads_mutex_);
    WARNING("Trying to register an existing thread.");
  }
}

void RemoteArbiter::SetProcessReturnStatus(const int pid,
                                           const int return_status) {
  try {
    boost::scoped_ptr<ArbiterConnection>
        connection(new ArbiterConnection(DefaultRemoteArbiter::hostname_,
                                         DefaultRemoteArbiter::port_));
    connection->client()->SetProcessReturnStatus(ThreadID(),
                                                 pid,
                                                 return_status);
    MESSAGE("Thread %u called UpdateProcessReturnStatus().", ThreadID());
  } catch (const apache::thrift::TException &tx) {
    WARNING("Connection with the arbiter failed: %s", tx.what());
  }
}

void RemoteArbiter::UnregisterAllThreads() {
  std::list<int> unregister_queue;
  System::MutexLock(&threads_mutex_);
  for (std::pair<System::Thread, ThreadInfo *> pair : threads_) {
    ThreadInfo *thread = pair.second;
    if (thread->running) {
      thread->running = true;
      nrunning_threads_--;
      unregister_queue.push_back(thread->id);
    }
  }
  assert(nrunning_threads_ == 0);
  System::MutexUnlock(&threads_mutex_);
  try {
    boost::scoped_ptr<ArbiterConnection>
        connection(new ArbiterConnection(DefaultRemoteArbiter::hostname_,
                                         DefaultRemoteArbiter::port_));
    for (int tid : unregister_queue) {
      connection->client()->UnregisterThread(tid);
      MESSAGE("Thread %d called UnregisterThread(%d).", ThreadID(), tid);
    }
  } catch (const apache::thrift::TException &tx) {
    WARNING("Connection with the arbiter failed: %s", tx.what());
  }
}

void RemoteArbiter::UnregisterThread() {
  System::MutexLock(&threads_mutex_);
  threads_[System::ThreadSelf()]->running = false;
  nrunning_threads_--;
  System::MutexUnlock(&threads_mutex_);
  DefaultRemoteArbiter::UnregisterThread(ThreadID());
}

}  // namespace interposition
}  // namespace dbug
