/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

// TODO(jsimsa): Correctly return EBUSY for sem_trywait().

// TODO(jsimsa): Check that the semaphore being unlinked was
// previously created.

#include <sstream>
#include <string>

#include "dbug/common/system.h"
#include "dbug/common/util.h"
#include "dbug/interposition/handler.h"
#include "dbug/interposition/logging.h"
#include "dbug/interposition/original.h"
#include "src/interposition/gen/semaphore.cc"

namespace dbug {
namespace interposition {

#if DBUG_HAVE_SEM_CLOSE
int Handler::sem_close(sem_t *sem) {
  System::MutexLock(&nsem_map_mutex_);
  if (nsem_map_.find(sem) == nsem_map_.end()) {
    System::MutexUnlock(&nsem_map_mutex_);
    WARNING("Closing a non-existing semaphore.");
    return Original::sem_close(sem);
  }
  std::string resource_id = nsem_map_[sem];
  System::MutexUnlock(&nsem_map_mutex_);
  arbiter_->SemaphoreClose(arbiter_->ThreadID(),
                           "sem_close",
                           Util::GetBacktrace(),
                           resource_id);
  int ret = Original::sem_close(sem);
  if (ret == 0) {
    System::MutexLock(&nsem_map_mutex_);
    nsem_map_.erase(sem);
    System::MutexUnlock(&nsem_map_mutex_);
  }
  return ret;
}
#endif

#if DBUG_HAVE_SEM_DESTROY
int Handler::sem_destroy(sem_t *sem) {
  System::MutexLock(&usem_map_mutex_);
  if (usem_map_.find(sem) == usem_map_.end()) {
    System::MutexUnlock(&usem_map_mutex_);
    WARNING("Destroying an uninitialized semaphore.");
    return Original::sem_destroy(sem);
  }
  std::string resource_id =
      Util::CreateResourceID("semaphore",
                             arbiter_->ProcessID(),
                             usem_map_[sem]);
  System::MutexUnlock(&usem_map_mutex_);
  arbiter_->SemaphoreDelete(arbiter_->ThreadID(),
                            "sem_destroy",
                            Util::GetBacktrace(),
                            resource_id);
  int ret = Original::sem_destroy(sem);
  if (ret == 0) {
    System::MutexLock(&usem_map_mutex_);
    usem_map_.erase(sem);
    System::MutexUnlock(&usem_map_mutex_);
  }
  return ret;
}
#endif

#if DBUG_HAVE_SEM_INIT
int Handler::sem_init(sem_t *sem, int pshared, unsigned int value) {
  switch (arbiter_->Choose(arbiter_->ThreadID(),
                           "sem_init",
                           Util::GetBacktrace(),
                           3)) {
    case 1:
      // Simulate lack of non-memory resources.
      errno = EAGAIN;
      return EAGAIN;
    case 2:
      // Simulate lack of memory.
      errno = ENOMEM;
      return ENOMEM;
    default:
      System::MutexLock(&usem_map_mutex_);
      if (usem_map_.find(sem) != usem_map_.end()) {
        System::MutexUnlock(&usem_map_mutex_);
        WARNING("Initializing an already initialized semaphore.");
        return 0;
      }
      System::MutexUnlock(&usem_map_mutex_);
      int ret = Original::sem_init(sem, pshared, value);
      if (ret == 0) {
        int id = Util::NewIdentifier();
        std::string resource_id =
            Util::CreateResourceID("semaphore",
                                   arbiter_->ProcessID(),
                                   id);
        arbiter_->SemaphoreCreate(arbiter_->ThreadID(),
                                  "sem_init",
                                  Util::GetBacktrace(),
                                  resource_id,
                                  false,
                                  value);
        System::MutexLock(&usem_map_mutex_);
        usem_map_[sem] = id;
        System::MutexUnlock(&usem_map_mutex_);
      }
      return ret;
  }
}
#endif

#if DBUG_HAVE_SEM_OPEN
sem_t *Handler::sem_open(const char *name, int oflags, va_list ap) {
  if (Interpose()) {
    Interposition off(__FUNCTION__);
    sem_t *ret = NULL;
    mode_t mode = va_arg(ap, mode_t);
    int value = 0;
    if (mode) { value = va_arg(ap, int); }
    switch (arbiter_->Choose(arbiter_->ThreadID(),
                             "sem_open",
                             Util::GetBacktrace(),
                             3)) {
      case 1:
        // Simulate lack of non-memory resources.
        errno = EAGAIN;
        break;
      case 2:
        // Simulate lack of memory.
        errno = ENOMEM;
        break;
      default:
        bool found = false;
        System::MutexLock(&nsem_map_mutex_);
        for (std::pair<sem_t *, std::string> sem : nsem_map_) {
          if (sem.second == name) {
            found = true;
            break;
          }
        }
        if (!found) {
          System::MutexUnlock(&nsem_map_mutex_);
          WARNING("Opening an already opened semaphore.");
          return 0;
        }
        System::MutexUnlock(&nsem_map_mutex_);
        if (mode) {
          ret = Original::sem_open(name, oflags, mode, value);
        } else {
          ret = Original::sem_open(name, oflags);
        }
        if (ret != NULL) {
          arbiter_->SemaphoreCreate(arbiter_->ThreadID(),
                                    "sem_open",
                                    Util::GetBacktrace(),
                                    name,
                                    true,
                                    value);
          System::MutexLock(&nsem_map_mutex_);
          nsem_map_[ret] = name;
          System::MutexUnlock(&nsem_map_mutex_);
        }
    }
    if (mode) {
      DEBUG_EXTERNAL("%p, %d, %d, %d", "%p", name, oflags, mode, value, ret);
    } else {
      DEBUG_EXTERNAL("%p, %d", "%p", name, oflags, ret);
    }
    return ret;
  } else {
    DBUG_REQUIRE_SYMBOL(Original::sem_open, SEM_OPEN_TYPE, "sem_open");
    mode_t mode = va_arg(ap, mode_t);
    sem_t *ret;
    if (mode) {
      int value = va_arg(ap, int);
      ret = Original::sem_open(name, oflags, mode, value);
      DEBUG_INTERNAL("%s, %d, %lu, %d", "%p", name, oflags, mode, value, ret);
    } else {
      ret = Original::sem_open(name, oflags);
      DEBUG_INTERNAL("%s, %d", "%p", name, oflags, ret);
    }
    return ret;
  }
}
#endif

#if DBUG_HAVE_SEM_POST
int Handler::sem_post(sem_t *sem) {
  System::MutexLock(&usem_map_mutex_);
  System::MutexLock(&nsem_map_mutex_);
  if (usem_map_.find(sem) == usem_map_.end() &&
      nsem_map_.find(sem) == nsem_map_.end()) {
    System::MutexUnlock(&nsem_map_mutex_);
    System::MutexUnlock(&usem_map_mutex_);
    WARNING("Posting on an uninitialized semaphore.");
    return Original::sem_post(sem);
  }
  if (usem_map_.find(sem) != usem_map_.end()) {
    std::string resource_id =
        Util::CreateResourceID("semaphore",
                               arbiter_->ProcessID(),
                               usem_map_[sem]);
    System::MutexUnlock(&nsem_map_mutex_);
    System::MutexUnlock(&usem_map_mutex_);
    arbiter_->SemaphorePost(arbiter_->ThreadID(),
                            "sem_post",
                            Util::GetBacktrace(),
                            resource_id);
    return 0;
  } else {
    assert(nsem_map_.find(sem) != nsem_map_.end());
    std::string resource_id = nsem_map_[sem];
    System::MutexUnlock(&nsem_map_mutex_);
    System::MutexUnlock(&usem_map_mutex_);
    arbiter_->SemaphorePost(arbiter_->ThreadID(),
                            "sem_post",
                            Util::GetBacktrace(),
                            resource_id);
    return 0;
  }
}
#endif

#if DBUG_HAVE_SEM_TRYWAIT
int Handler::sem_trywait(sem_t *sem) {
  System::MutexLock(&usem_map_mutex_);
  System::MutexLock(&nsem_map_mutex_);
  if (usem_map_.find(sem) == usem_map_.end() &&
      nsem_map_.find(sem) == nsem_map_.end()) {
    System::MutexUnlock(&nsem_map_mutex_);
    System::MutexUnlock(&usem_map_mutex_);
    WARNING("Waiting on an uninitialized semaphore.");
    return Original::sem_trywait(sem);
  }
  if (usem_map_.find(sem) != usem_map_.end()) {
    std::string resource_id =
        Util::CreateResourceID("semaphore",
                               arbiter_->ProcessID(),
                               usem_map_[sem]);
    System::MutexUnlock(&nsem_map_mutex_);
    System::MutexUnlock(&usem_map_mutex_);
    arbiter_->SemaphoreTryWait(arbiter_->ThreadID(),
                               "sem_trywait",
                               Util::GetBacktrace(),
                               resource_id);
    return 0;
  } else {
    assert(nsem_map_.find(sem) != nsem_map_.end());
    std::string resource_id = nsem_map_[sem];
    System::MutexUnlock(&nsem_map_mutex_);
    System::MutexUnlock(&usem_map_mutex_);
    arbiter_->SemaphoreTryWait(arbiter_->ThreadID(),
                               "sem_trywait",
                               Util::GetBacktrace(),
                               resource_id);
    return 0;
  }
}
#endif

#if DBUG_HAVE_SEM_UNLINK
int Handler::sem_unlink(const char *name) {
  arbiter_->SemaphoreUnlink(arbiter_->ThreadID(),
                            "sem_unlink",
                            Util::GetBacktrace(),
                            name);
  return Original::sem_unlink(name);
}
#endif

#if DBUG_HAVE_SEM_WAIT
int Handler::sem_wait(sem_t *sem) {
  System::MutexLock(&usem_map_mutex_);
  System::MutexLock(&nsem_map_mutex_);
  if (usem_map_.find(sem) == usem_map_.end() &&
      nsem_map_.find(sem) == nsem_map_.end()) {
    System::MutexUnlock(&nsem_map_mutex_);
    System::MutexUnlock(&usem_map_mutex_);
    WARNING("Waiting on an uninitialized semaphore.");
    return Original::sem_wait(sem);
  }
  if (usem_map_.find(sem) != usem_map_.end()) {
    std::string resource_id =
        Util::CreateResourceID("semaphore",
                               arbiter_->ProcessID(),
                               usem_map_[sem]);
    System::MutexUnlock(&nsem_map_mutex_);
    System::MutexUnlock(&usem_map_mutex_);
    arbiter_->SemaphoreWait(arbiter_->ThreadID(),
                            "sem_wait",
                            Util::GetBacktrace(),
                            resource_id);
    return 0;
  } else {
    assert(nsem_map_.find(sem) != nsem_map_.end());
    std::string resource_id = nsem_map_[sem];
    System::MutexUnlock(&nsem_map_mutex_);
    System::MutexUnlock(&usem_map_mutex_);
    arbiter_->SemaphoreWait(arbiter_->ThreadID(),
                            "sem_wait",
                            Util::GetBacktrace(),
                            resource_id);
    return 0;
  }
}
#endif

}  // namespace interposition
}  // namespace dbug
