/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <mpi.h>
#include <stdio.h>

#include <boost/scoped_array.hpp>
#include <sstream>

int main(int argc, char *argv[]) {
  int i;
  double d;
  int int_size = sizeof(i);
  double double_size = sizeof(d);
  MPI_Init(&argc, &argv);
  int myrank;
  MPI_Comm_rank(MPI_COMM_WORLD, &myrank);
  int size;
  MPI_Comm_size(MPI_COMM_WORLD, &size);
  boost::scoped_array<int> sdispls(new int[size]);
  boost::scoped_array<int> send_counts(new int[size]);
  for (int i = 0; i < size; i++) {
    sdispls[i] = 0;
    send_counts[i] = i + 1;
  }
  boost::scoped_array<MPI_Datatype> send_types(new MPI_Datatype[size]);
  for (int i = 0; i < size; i++) {
    if (myrank % 2 == 0) {
      send_types[i] = MPI_INT;
    } else {
      send_types[i] = MPI_DOUBLE;
    }
  }
  int ssize;
  if (myrank % 2 == 0)  {
    ssize = int_size * size;
  } else {
    ssize = double_size * size;
  }
  boost::scoped_array<char> send_buffer(new char[ssize]);
  for (int i = 0; i < size; i++) {
    if (myrank % 2 == 0) {
      int *ptr = reinterpret_cast<int *>(send_buffer.get());
      ptr[i] = myrank;
    } else {
      double *ptr = reinterpret_cast<double *>(send_buffer.get());
      ptr[i] = myrank;
    }
  }
  std::stringstream send_stream;
  send_stream << "Process " << myrank << " sends:";
  if (myrank % 2 == 0) {
    int *ptr = reinterpret_cast<int *>(send_buffer.get());
    for (int i = 0; i < size; i++) {
      send_stream << " " << ptr[i];
    }
  } else {
    double *ptr = reinterpret_cast<double *>(send_buffer.get());
    for (int i = 0; i < size; i++) {
      send_stream << " " << ptr[i];
    }
  }
  printf("%s.\n", send_stream.str().c_str());
  boost::scoped_array<int> rdispls(new int[size]);
  boost::scoped_array<int> recv_counts(new int[size]);
  for (int i = 0; i < size; i++) {
    rdispls[i] = ((i / 2) + (i % 2)) * int_size * (myrank + 1) +
                 (i / 2) * double_size * (myrank + 1);
    recv_counts[i] = myrank + 1;
  }
  boost::scoped_array<MPI_Datatype> recv_types(new MPI_Datatype[size]);
  for (int i = 0; i < size; i++) {
    if (i % 2 == 0) {
      recv_types[i] = MPI_INT;
    } else {
      recv_types[i] = MPI_DOUBLE;
    }
  }
  int rsize = (myrank + 1) *
              ((size / 2) * double_size +
               (((size / 2) + (size % 2)) * int_size));
  boost::scoped_array<char> recv_buffer(new char[rsize]);
  for (int i = 0; i < rsize; i++) { recv_buffer[i] = 0; }
  MPI_Alltoallw(send_buffer.get(),
                send_counts.get(),
                sdispls.get(),
                send_types.get(),
                recv_buffer.get(),
                recv_counts.get(),
                rdispls.get(),
                recv_types.get(),
                MPI_COMM_WORLD);
  std::stringstream stream;
  stream << "Process " << myrank << " received:";
  int offset = 0;
  for (int i = 0; i < size; i++) {
    for (int j = 0; j < myrank + 1; j++) {
      if (i % 2 == 0) {
        int *ptr = reinterpret_cast<int *>(recv_buffer.get() + offset);
        stream << " " << *ptr;
        offset += int_size;
      } else {
        double *ptr = reinterpret_cast<double *>(recv_buffer.get() + offset);
        stream << " " << *ptr;
        offset += double_size;
      }
    }
  }
  printf("%s.\n", stream.str().c_str());
  MPI_Finalize();
  return 0;
}
