/*
 * Authors: Jiri Simsa
 * Copyright (c) 2013, Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY
 * WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#include <sstream>
#include <string>

#include "dbug/common/command.h"
#include "dbug/common/system.h"
#include "dbug/common/util.h"
#include "gtest/gtest.h"

namespace dbug {
namespace {

void CreateXMLConfigFile(int num_processes,
                         int timeout,
                         bool use_dpor,
                         std::string test_filename) {
  System::File *file = System::FileOpen("config.xml", "w");
  std::stringstream stream;
  stream << "<config>\n"
         << "  <arbiter arguments=\"--start_num_threads "
         << num_processes << "\"/>\n"
         << "  <explorer timeout=\"" << timeout << "\""
         << " dpor=\"" << (use_dpor ? "true" : "false") << "\"/>\n"
         << "  <program command=\"" << BUILDDIR << "/"
         << test_filename << "\"/>\n"
         << "</config>\n";
  System::FilePrint(file, stream.str().c_str());
  System::FileClose(file);
}

TEST(PThreadTest, Barrier) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(1, 10, false, "pthread-barrier");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(PThreadTest, CVar) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(1, 10, false, "pthread-cvar");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(PThreadTest, Mutex) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(1, 10, false, "pthread-mutex");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(PThreadTest, RWLock) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(1, 10, false, "pthread-rwlock");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

TEST(PThreadTest, Spin) {
  System::Process child = System::Fork();
  if (child == 0) {
    CreateXMLConfigFile(1, 10, false, "pthread-spin");
    Command command;
    std::map<std::string, std::string> empty_environment;
    command.set_program(BINDIR "/dbug-explorer");
    command.add_arg("config.xml");
    Util::StartCommand(&command, empty_environment);
  } else {
    int status;
    System::Wait(child, &status, 0);
  }
}

}  // anonymous namespace
}  // namespace dbug

int main(int argc, char **argv) {
  testing::InitGoogleTest(&argc, argv);
  return RUN_ALL_TESTS();
}
